@@ -1,24 +1,43 @@
 """Base forest-related models using LightGBM as the gradient booster."""
 from abc import abstractmethod
 import os
-from typing import Any, Dict, Optional, Sequence, Tuple, Union, cast
+from typing import Any, ClassVar, Dict, Optional, Sequence, Tuple, Union, cast
 
 import lightgbm as lgb
 from marshmallow import fields
 import numpy as np
 
-from bitfount.data.databunch import _BitfountDataBunch
-from bitfount.data.dataloader import _BitfountDataLoader
-from bitfount.data.datasource import DataSource
+from bitfount.data.databunch import BitfountDataBunch
+from bitfount.data.dataloaders import BitfountDataLoader
+from bitfount.data.datasources.base_source import BaseSource
 from bitfount.models.base_models import ClassifierMixIn, _BaseModel
-from bitfount.types import _DataFrameType, _JSONDict
+from bitfount.types import T_FIELDS_DICT
 
 # TODO: [BIT-1041] Introduce non-implementation base model
+from bitfount.utils import delegates
 
 
+@delegates()
 class BaseLGBMRandomForest(_BaseModel):
     """Implements an (optionally Gradient Boosted) Random Forest from LightGBM."""
 
+    fields_dict: ClassVar[T_FIELDS_DICT] = {
+        "gradient_boosting": fields.Boolean(),
+        "num_leaves": fields.Integer(allow_none=True),
+        "max_depth": fields.Integer(allow_none=True),
+        "subsample_for_bin": fields.Integer(allow_none=True),
+        "num_iterations": fields.Integer(allow_none=True),
+        "learning_rate": fields.Float(allow_none=True),
+        "reg_alpha": fields.Float(allow_none=True),
+        "reg_lambda": fields.Float(allow_none=True),
+        "bagging_freq": fields.Integer(allow_none=True),
+        "bagging_fraction": fields.Float(allow_none=True),
+        "feature_fraction": fields.Float(allow_none=True),
+        "early_stopping_rounds": fields.Integer(allow_none=True),
+        "verbose": fields.Integer(allow_none=True),
+        "min_split_gain": fields.Float(allow_none=True),
+    }
+
     def __init__(
         self,
         gradient_boosting: bool = False,
@@ -61,27 +80,6 @@ class BaseLGBMRandomForest(_BaseModel):
         self.objective: Optional[str] = None
         self.training_metrics: Optional[Sequence[str]] = None
 
-    class _Schema(_BaseModel._Schema):
-        gradient_boosting = fields.Boolean()
-        num_leaves = fields.Integer(allow_none=True)
-        max_depth = fields.Integer(allow_none=True)
-        subsample_for_bin = fields.Integer(allow_none=True)
-        num_iterations = fields.Integer(allow_none=True)
-        learning_rate = fields.Float(allow_none=True)
-        reg_alpha = fields.Float(allow_none=True)
-        reg_lambda = fields.Float(allow_none=True)
-        bagging_freq = fields.Integer(allow_none=True)
-        bagging_fraction = fields.Float(allow_none=True)
-        feature_fraction = fields.Float(allow_none=True)
-        early_stopping_rounds = fields.Integer(allow_none=True)
-        verbose = fields.Integer(allow_none=True)
-        min_split_gain = fields.Float(allow_none=True)
-
-        @abstractmethod
-        def recreate_model(self, data: _JSONDict, **kwargs: Any) -> _BaseModel:
-            """Recreates model using Schema."""
-            raise NotImplementedError
-
     @abstractmethod
     def _set_objective(self) -> None:
         """Set training objective."""
@@ -101,16 +99,22 @@ class BaseLGBMRandomForest(_BaseModel):
         self._model = lgb.Booster(model_file=filename)
 
     def evaluate(
-        self, test_dl: Optional[_BitfountDataLoader] = None, *args: Any, **kwargs: Any
+        self, test_dl: Optional[BitfountDataLoader] = None, *args: Any, **kwargs: Any
     ) -> Tuple[np.ndarray, np.ndarray]:
         """Perform inference on test set and save dictionary of metrics."""
         if test_dl is None:
-            if isinstance(self.test_dl, _BitfountDataLoader):
+            if isinstance(self.test_dl, BitfountDataLoader):
                 test_dl = self.test_dl
             else:
                 raise ValueError("There is no test data to evaluate the model on.")
 
-        test_df: _DataFrameType = test_dl.get_x_dataframe()
+        test_df = test_dl.get_x_dataframe()
+        if isinstance(test_df, tuple):
+            raise ValueError(
+                "Multiple dataframes retrieved unexpectedly; "
+                "this model does not support combination tabular and image data."
+            )
+
         test_preds = np.zeros(test_df.shape[0])
         best_iteration = self._model.best_iteration
         test_preds += self._model.predict(test_df, num_iteration=best_iteration)
@@ -142,7 +146,13 @@ class BaseLGBMRandomForest(_BaseModel):
 
     def _create_dataset(self) -> Tuple[lgb.Dataset, Optional[lgb.Dataset]]:
         """Create LightGBM Dataset object for better memory management."""
-        x_train: _DataFrameType = self.train_dl.get_x_dataframe()
+        x_train = self.train_dl.get_x_dataframe()
+        if isinstance(x_train, tuple):
+            raise ValueError(
+                "Multiple dataframes retrieved unexpectedly; "
+                "this model does not support combination tabular and image data."
+            )
+
         y_train = self.train_dl.get_y_dataframe().astype("int64")
         train_df = lgb.Dataset(
             data=x_train,
@@ -153,8 +163,14 @@ class BaseLGBMRandomForest(_BaseModel):
         weights_col = self.datastructure.loss_weights_col
         if weights_col is not None:
             train_df = train_df.set_weight(weight=list(x_train[weights_col]))
-        if isinstance(self.validation_dl, _BitfountDataLoader):
-            x_valid: _DataFrameType = self.validation_dl.get_x_dataframe()
+        if isinstance(self.validation_dl, BitfountDataLoader):
+            x_valid = self.validation_dl.get_x_dataframe()
+            if isinstance(x_valid, tuple):
+                raise ValueError(
+                    "Multiple dataframes retrieved unexpectedly; "
+                    "this model does not support combination tabular and image data."
+                )
+
             y_valid = self.validation_dl.get_y_dataframe().astype("int64")
             # For LightGBM DataFrame.dtypes for label must be int, float or bool
 
@@ -173,12 +189,12 @@ class BaseLGBMRandomForest(_BaseModel):
 
         return train_df, validation_df
 
-    def fit(self, data: Optional[DataSource] = None, *args: Any, **kwargs: Any) -> None:
-        """Trains a model using the training set provided by the DataSource object."""
+    def fit(self, data: Optional[BaseSource] = None, *args: Any, **kwargs: Any) -> None:
+        """Trains a model using the training set provided by the BaseSource object."""
         if data:
             if self.datastructure.query:
                 table_schema = self.datastructure._override_schema()
-                self._databunch = _BitfountDataBunch(
+                self.databunch = BitfountDataBunch(
                     data_structure=self.datastructure,
                     schema=table_schema,
                     datasource=data,
@@ -208,19 +224,21 @@ class BaseLGBMRandomForest(_BaseModel):
 
         params = self.get_params()
         lgb_train, lgb_valid = self._create_dataset()
+        if lgb_valid:
+            valid_sets = [lgb_valid, lgb_train]
+            valid_names = ["valid", "train"]
+        else:
+            valid_sets = [lgb_train]
+            valid_names = ["train"]
+
         model = lgb.train(
             params=params,
             num_boost_round=self.num_iterations,
             train_set=lgb_train,
-            valid_sets=[lgb_valid, lgb_train],
-            valid_names=["valid", "train"],
+            valid_sets=valid_sets,
+            valid_names=valid_names,
             verbose_eval=self.verbose,
             early_stopping_rounds=self.early_stopping_rounds,
             init_model=self._model,
         )
-        self._model: lgb = model
-
-    @property
-    def training_needed(self) -> bool:
-        """Dictates whether the model needs training."""
-        return self.num_iterations > 0
+        self._model: lgb.Booster = model
