@@ -3,36 +3,42 @@ from abc import abstractmethod
 from collections import defaultdict
 import logging
 import os
+import re
 from typing import Any, Dict, List, MutableMapping, Optional, Tuple, Union, cast
 
 import numpy as np
 import pytorch_lightning as pl
+from pytorch_lightning.callbacks import Callback
+from pytorch_lightning.callbacks.progress import TQDMProgressBar
 import torch
 from torch import nn as nn
 from torch.optim.lr_scheduler import _LRScheduler
 from torch.utils.data import DataLoader as PyTorchDataLoader
 
-from bitfount.backends.pytorch.data.dataloaders import _PyTorchBitfountDataLoader
-from bitfount.backends.pytorch.federated.models import _PyTorchDistributedModelMixIn
+from bitfount.backends.pytorch.data.dataloaders import _BasePyTorchBitfountDataLoader
+from bitfount.backends.pytorch.federated.mixins import _PyTorchDistributedModelMixIn
 from bitfount.backends.pytorch.models.base_models import (
     _STEP_OUTPUT,
     _TORCH_DTYPES,
     _OptimizerType,
 )
 from bitfount.backends.pytorch.utils import _autodetect_gpu
-from bitfount.data.databunch import _BitfountDataBunch
-from bitfount.data.dataloader import _BitfountDataLoader
-from bitfount.data.datasource import DataSource
+from bitfount.data.databunch import BitfountDataBunch
+from bitfount.data.dataloaders import BitfountDataLoader
+from bitfount.data.datasources.base_source import BaseSource
 from bitfount.metrics import Metric
 from bitfount.models.base_models import ClassifierMixIn, ModelContext
 from bitfount.models.bitfount_model import BitfountModel
-from bitfount.types import T_DTYPE
+from bitfount.types import T_DTYPE, _StrAnyDict
+from bitfount.utils import _merge_list_of_dicts, delegates
+from bitfount.utils.logging_utils import filter_stderr
 
 logger = logging.getLogger(__name__)
 
 
+@delegates()
 class PyTorchBitfountModel(
-    _PyTorchDistributedModelMixIn, pl.LightningModule, BitfountModel
+    _PyTorchDistributedModelMixIn, BitfountModel, pl.LightningModule
 ):
     """Blueprint for a pytorch custom model in the lightning format.
 
@@ -48,6 +54,14 @@ class PyTorchBitfountModel(
 
     https://pytorch-lightning.readthedocs.io/en/stable/common/lightning_module.html
 
+    :::caution
+
+    Ensure you set `self.metrics` in the `__init__` method of your subclass to ensure
+    they pertain appropriately to your model. If not, Bitfount will attempt to set
+    these appropriately for you but there is no guarantee it will get it right.
+
+    :::
+
     Args:
         batch_size: The batch size to use for training. Defaults to 32.
         epochs: The number of epochs to train for.
@@ -59,14 +73,17 @@ class PyTorchBitfountModel(
         epochs: The number of epochs to train for.
         steps: The number of steps to train for.
         preds: The predictions from the most recent test run.
-        targs: The targets from the most recent test run.
+        target: The targets from the most recent test run.
         val_stats: Metrics from the validation set during training.
 
     Raises:
         ValueError: If both `epochs` and `steps` are specified.
     """
 
-    train_dl: _PyTorchBitfountDataLoader
+    train_dl: _BasePyTorchBitfountDataLoader
+    # Test attributes
+    _test_preds: Optional[List[np.ndarray]]
+    _test_targets: Optional[List[np.ndarray]]
 
     def __init__(
         self,
@@ -92,22 +109,33 @@ class PyTorchBitfountModel(
         self._model: Optional[nn.Module] = None
         self._pl_trainer: pl.Trainer = self.trainer_init()  # cannot be "self.trainer"
         self.preds: List[float] = []
-        self.targs: List[float] = []
+        self.target: List[float] = []
         self.val_stats: List[Dict[str, float]] = []
         self._trained_on_previous_batch: bool = False
         self._total_num_batches_trained: int = 0
 
+    @staticmethod
+    def _get_import_statements() -> List[str]:
+        """Returns a list of import statements likely to be required for the model.
+
+        Returns:
+            A list of import statements.
+        """
+        return [
+            "import os",
+            "import torch",
+            "from torch import nn as nn",
+            "from torch.nn import functional as F",
+            "from bitfount import *",
+            "import bitfount",
+        ]
+
     @abstractmethod
     def forward(self, x: Any) -> Any:  # type:ignore[override] # Reason: see below
         """Forward method of the model - just like a regular `torch.nn.Module` class.
 
-        Args:
-            x: Input to the model.
-
-        Returns:
-            Output of the model.
-
         :::tip
+
         This will depend on your model but could be as simple as:
 
         ```python
@@ -115,6 +143,12 @@ class PyTorchBitfountModel(
         ```
 
         :::
+
+        Args:
+            x: Input to the model.
+
+        Returns:
+            Output of the model.
         """
         # pl.LightningModule incompatible with our forward function.
         raise NotImplementedError
@@ -144,15 +178,6 @@ class PyTorchBitfountModel(
     def training_step(self, batch: Any, batch_idx: int) -> _STEP_OUTPUT:  # type: ignore[override] # Reason: see below  # noqa: B950
         """Training step.
 
-        Args:
-            batch: The batch to be trained on.
-            batch_idx: The index of the batch to be trained on from the train
-                dataloader.
-
-        Returns:
-            The loss from this batch as a `torch.Tensor`. Or a dictionary which includes
-            the key `loss` and the loss as a `torch.Tensor`.
-
         :::caution
 
         If iterations have been specified in terms of steps, the default behaviour of
@@ -175,6 +200,15 @@ class PyTorchBitfountModel(
         ```
 
         :::
+
+        Args:
+            batch: The batch to be trained on.
+            batch_idx: The index of the batch to be trained on from the train
+                dataloader.
+
+        Returns:
+            The loss from this batch as a `torch.Tensor`. Or a dictionary which includes
+            the key `loss` and the loss as a `torch.Tensor`.
         """
         # Override the pl.lightning method, as it requires *args
         # and **kwargs as arguments whereas we will always have
@@ -182,7 +216,7 @@ class PyTorchBitfountModel(
         raise NotImplementedError
 
     @abstractmethod
-    def validation_step(self, batch: Any, batch_idx: int) -> _STEP_OUTPUT:  # type: ignore[override] # Reason: see below # noqa: B950
+    def validation_step(self, batch: Any, batch_idx: int) -> _StrAnyDict:  # type: ignore[override] # Reason: see below # noqa: B950
         """Validation step.
 
         Args:
@@ -201,10 +235,8 @@ class PyTorchBitfountModel(
         raise NotImplementedError
 
     @abstractmethod
-    def test_step(self, batch: Any, batch_idx: int) -> Optional[_STEP_OUTPUT]:  # type: ignore[override] # Reason: see below  # noqa: B950
-        """Performs test step and must set `self.preds` and `self.targs`.
-
-        They will be returned by the `evaluate` method.
+    def test_step(self, batch: Any, batch_idx: int) -> _STEP_OUTPUT:  # type: ignore[override] # Reason: see below  # noqa: B950
+        """Operates on a single batch of data from the test set.
 
         Args:
             batch: The batch to be evaluated.
@@ -212,9 +244,9 @@ class PyTorchBitfountModel(
                 dataloader.
 
         Returns:
-            Any object or value of interest. These will be passed to the
-            `test_epoch_end` method. If returning None, testing will skip to the next
-            batch.
+            A dictionary of predictions and targets, with the dictionary
+            keys being "predictions" and "targets" for each of them, respectively.
+            These will be passed to the `test_epoch_end` method.
         """
         # Override the pl.lightning method, as it requires
         # *args and **kwargs as arguments whereas we will
@@ -224,20 +256,20 @@ class PyTorchBitfountModel(
     def tensor_precision(self) -> T_DTYPE:
         """Returns tensor dtype used by Pytorch Lightning Trainer.
 
-        Returns:
-            Pytorch tensor dtype.
-
         :::note
 
         Currently only 32-bit training is supported.
 
         :::
+
+        Returns:
+            Pytorch tensor dtype.
         """
         # TODO: [BIT-727] support non-32 bit training
         return cast(T_DTYPE, _TORCH_DTYPES[cast(int, self._pl_trainer.precision)])
 
     def initialise_model(
-        self, data: Optional[DataSource] = None, context: Optional[ModelContext] = None
+        self, data: Optional[BaseSource] = None, context: Optional[ModelContext] = None
     ) -> None:
         """Any initialisation of models/dataloaders to be done here.
 
@@ -252,26 +284,58 @@ class PyTorchBitfountModel(
         """
         self._context = context
         self._initialised = True
-        table_schema = self.datastructure.get_table_schema(
-            schema=self.schema, pod_identifier=self.pod_identifier, datasource=data
-        )
+        if self._context == ModelContext.MODELLER:
+            # In a distributed setting, the Modeller needs to first initialise its
+            # own model before it can be used. The pod identifier needs to be set
+            # before the model is initialised so the the relevant details can be
+            # retrieved from the schema. For this we just use the first pod
+            # identifier specified in the datastructure as it is assumed the the
+            # schemas for all the Pods are the same.
+            pod_identifiers = self.datastructure.get_pod_identifiers()
+            if pod_identifiers:
+                self.set_pod_identifier(pod_identifiers[0])
         if data is not None:
-            if context:
-                self._databunch = _BitfountDataBunch(
+            if self.datastructure.query:
+                table_schema = self.datastructure._override_schema(
+                    pod_identifier=self.pod_identifier
+                )
+                self.databunch = BitfountDataBunch(
                     data_structure=self.datastructure,
                     schema=table_schema,
                     datasource=data,
                 )
-            else:
-                # For local training, we can add the datasource to the schema.
-                if not data._data_is_loaded:
-                    data.load_data()
-                self._add_datasource_to_schema(data)
+            elif self.datastructure.table:
+                if context:
+                    table_schema = self.schema.get_table_schema(
+                        self.datastructure.get_table_name(self.pod_identifier)
+                    )
+                    self.databunch = BitfountDataBunch(
+                        data_structure=self.datastructure,
+                        schema=table_schema,
+                        datasource=data,
+                    )
+                else:
+                    # For local training, we can add the datasource to the schema.
+                    data.load_data(table_name=self.datastructure.table)
+                    self._add_datasource_to_schema(data)
+                    table_schema = self.schema.get_table_schema(
+                        self.datastructure.get_table_name(self.pod_identifier)
+                    )
+
             if self._context != ModelContext.MODELLER:
                 self._set_dataloaders(self.batch_size)
+        else:
+            if self.datastructure.query:
+                table_schema = self.datastructure._override_schema(
+                    pod_identifier=self.pod_identifier
+                )
+            else:
+                table_schema = self.schema.get_table_schema(
+                    self.datastructure.get_table_name(self.pod_identifier)
+                )
         self.datastructure.set_training_input_size(table_schema)
 
-        if self._objective == "classification":
+        if hasattr(self, "_objective") and self._objective == "classification":
             # The casts here are to assuage mypy because it (incorrectly) asserts
             # that a subclass of both ClassifierMixIn and BitfountModel cannot exist.
             # We utilise a subclass of both in the tests to assure ourselves.
@@ -290,42 +354,52 @@ class PyTorchBitfountModel(
         Documentation for pytorch-lightning trainer can be found here:
         https://pytorch-lightning.readthedocs.io/en/stable/common/trainer.html
 
-        Returns:
-            The pytorch lightning trainer.
-
         :::tip
 
         Override this method to choose your own `Trainer` arguments.
 
         :::
-        """
-        trainer = pl.Trainer(
-            max_epochs=self.epochs,
-            max_steps=self.steps,
-            progress_bar_refresh_rate=1,
-            deterministic=True,
-            auto_lr_find=True,
-            gpus=_autodetect_gpu(),  # will be 1 (first GPU) or 0 (CPU)
-        )
-        return trainer
 
-    def train_dataloader(self) -> _PyTorchBitfountDataLoader:  # type: ignore[override] # Reason: see below # noqa: B950
+        Returns:
+            The pytorch lightning trainer.
+        """
+        callbacks: List[Callback] = [TQDMProgressBar(refresh_rate=1)]
+
+        # torch emits warnings to stderr that are not relevant for us, so we need
+        # to filter them out
+        with filter_stderr(
+            re.escape(
+                "[W Context.cpp:70] Warning:"
+                " torch.use_deterministic_algorithms is in beta"
+            )
+        ):
+            trainer = pl.Trainer(
+                max_epochs=self.epochs or -1,
+                max_steps=self.steps or -1,
+                deterministic=True,
+                auto_lr_find=True,
+                gpus=_autodetect_gpu(),  # will be 1 (first GPU) or 0 (CPU)
+                callbacks=callbacks,
+            )
+            return trainer
+
+    def train_dataloader(self) -> _BasePyTorchBitfountDataLoader:  # type: ignore[override] # Reason: see below # noqa: B950
         """Returns training dataloader."""
         # We override the dataloader return annotation as the LightningModule
         # expects a pytorch DataLoader, and we return out PyTorchBitfountDataLoader
         return self.train_dl
 
-    def val_dataloader(self) -> _PyTorchBitfountDataLoader:  # type: ignore[override] # Reason: see below # noqa: B950
+    def val_dataloader(self) -> _BasePyTorchBitfountDataLoader:  # type: ignore[override] # Reason: see below # noqa: B950
         """Returns validation dataloader."""
         # We override the dataloader return annotation as the LightningModule
         # expects a pytorch DataLoader, and we return out PyTorchBitfountDataLoader
-        return cast(_PyTorchBitfountDataLoader, self.validation_dl)
+        return cast(_BasePyTorchBitfountDataLoader, self.validation_dl)
 
-    def test_dataloader(self) -> _PyTorchBitfountDataLoader:  # type: ignore[override] # Reason: see below # noqa: B950
+    def test_dataloader(self) -> _BasePyTorchBitfountDataLoader:  # type: ignore[override] # Reason: see below # noqa: B950
         """Returns test dataloader."""
         # We override the dataloader return annotation as the LightningModule
         # expects a pytorch DataLoader, and we return out PyTorchBitfountDataLoader
-        return cast(_PyTorchBitfountDataLoader, self.test_dl)
+        return cast(_BasePyTorchBitfountDataLoader, self.test_dl)
 
     def serialize(self, filename: Union[str, os.PathLike]) -> None:
         """Serialize model to file with provided `filename`.
@@ -337,28 +411,29 @@ class PyTorchBitfountModel(
             logger.info("Model not yet initialised. Auto-initialising model.")
             self.initialise_model()
         # Model has been initialised, assuring mypy of this
-        assert self._model is not None  # nosec
+        assert self._model is not None  # nosec[assert_used]
         torch.save(self._model.state_dict(), filename)
 
-    def deserialize(self, filename: Union[str, os.PathLike]) -> None:
+    def deserialize(self, filename: Union[str, os.PathLike], **kwargs: Any) -> None:
         """Deserialize model.
 
-        Args:
-            filename: Path to file containing serialized model.
-
         :::danger
 
         This should not be used on a model file that has been received across a
         trust boundary due to underlying use of `pickle` by `torch`.
 
         :::
+
+        Args:
+            filename: Path to file containing serialized model.
+            **kwargs: Keyword arguments provided to `torch.load` under the hood.
         """
         if not self._initialised:
             logger.info("Model not yet initialised. Auto-initialising model.")
             self.initialise_model()
         # Model has been initialised, assuring mypy of this
-        assert self._model is not None  # nosec
-        self._model.load_state_dict(torch.load(filename))
+        assert self._model is not None  # nosec[assert_used]
+        self._model.load_state_dict(torch.load(filename, **kwargs))
 
     def skip_training_batch(self, batch_idx: int) -> bool:
         """Checks if the current batch from the training set should be skipped.
@@ -381,8 +456,7 @@ class PyTorchBitfountModel(
             # If we have trained on the previous batch, we can avoid the checks because
             # it means we have already reached the target start batch.
             if not self._trained_on_previous_batch:
-                # mypy cannot see that the trainer has this attribute due to how it is set in pytorch lightning # noqa: B950
-                if (self.steps != self._pl_trainer.max_steps) and (  # type: ignore[attr-defined] # Reason: see above # noqa: B950
+                if (self.steps != self._pl_trainer.max_steps) and (
                     batch_idx < (self._total_num_batches_trained % len(self.train_dl))
                 ):
                     return True
@@ -391,8 +465,7 @@ class PyTorchBitfountModel(
 
             # `_total_num_batches_trained` hasn't been incremented yet so we need to add
             # 1 here to get the correct batch number.
-            # mypy cannot see that the trainer has this attribute due to how it is set in pytorch lightning # noqa: B950
-            if self._total_num_batches_trained + 1 == self._pl_trainer.max_steps:  # type: ignore[attr-defined] # Reason: see above # noqa: B950
+            if self._total_num_batches_trained + 1 == self._pl_trainer.max_steps:
                 self._trained_on_previous_batch = False
 
         if not self._pl_trainer.sanity_checking:
@@ -401,7 +474,7 @@ class PyTorchBitfountModel(
         return False
 
     def validation_epoch_end(  # type: ignore[override] # Reason: see below
-        self, outputs: List[Dict[str, Any]]
+        self, outputs: List[_StrAnyDict]
     ) -> None:
         """Called at the end of the validation epoch with all validation step outputs.
 
@@ -411,8 +484,8 @@ class PyTorchBitfountModel(
         Args:
             outputs: List of outputs from each validation step.
         """
-        # Override the pl.lightning method, as it outputs can be
-        # List[Union[Tensor, Dict[str, Any]]],
+        # Override the pl.lightning method, as its outputs can be
+        # List[Union[Tensor, _StrAnyDict]],
         # whereas we force outputs to be a Dict
         avgs = self._compute_metric_averages(outputs)
         self.val_stats.append(avgs)
@@ -421,8 +494,19 @@ class PyTorchBitfountModel(
         for k, v in avgs.items():
             self.log(f"avg_{k}", v)
 
+    def test_epoch_end(self, outputs: List[Dict[str, torch.Tensor]]) -> None:  # type: ignore[override] # Reason: see below # noqa: B950
+        """Aggregates the predictions and targets from the test set.
+
+        Args:
+            outputs: List of outputs from each test step.
+        """
+        # Override the pl.lightning method, as it requires a different type for outputs.
+        merged_outputs: Dict[str, List[torch.Tensor]] = _merge_list_of_dicts(outputs)
+        self._test_preds = [i.cpu().numpy() for i in merged_outputs["predictions"]]
+        self._test_targets = [i.cpu().numpy() for i in merged_outputs["targets"]]
+
     @staticmethod
-    def _compute_metric_averages(outputs: List[Dict[str, Any]]) -> Dict[str, float]:
+    def _compute_metric_averages(outputs: List[_StrAnyDict]) -> Dict[str, float]:
         """Compute the average metrics from a list of outputs."""
         # Stack up shared dict keys into lists of entries
         stacked_dict = defaultdict(list)
@@ -437,7 +521,7 @@ class PyTorchBitfountModel(
         return avgs
 
     def _evaluate_local(
-        self, test_dl: Optional[_BitfountDataLoader] = None, **kwargs: Any
+        self, test_dl: Optional[BitfountDataLoader] = None, **kwargs: Any
     ) -> Tuple[np.ndarray, np.ndarray]:
         """This method runs inference on the test dataloader.
 
@@ -451,15 +535,22 @@ class PyTorchBitfountModel(
         Returns:
             A tuple of predictions and targets as numpy arrays.
         """
+        # Reset test attributes to None
+        self._reset_test_attrs()
         if test_dl is None:
-            if isinstance(self.test_dl, _PyTorchBitfountDataLoader):
+            if isinstance(self.test_dl, _BasePyTorchBitfountDataLoader):
                 test_dl = self.test_dl
             else:
                 raise ValueError("No test data to evaluate the model on.")
-        self._pl_trainer.test(test_dataloaders=cast(PyTorchDataLoader, test_dl))
-        return np.asarray(self.preds), np.asarray(self.targs)
+        self._pl_trainer.test(model=self, dataloaders=cast(PyTorchDataLoader, test_dl))
+        return np.asarray(self._test_preds), np.asarray(self._test_targets)
 
-    def predict(self, data: DataSource, **kwargs: Any) -> np.ndarray:
+    def _reset_test_attrs(self) -> None:
+        """Resets test attributes to None."""
+        self._test_preds = None
+        self._test_targets = None
+
+    def predict(self, data: BaseSource, **kwargs: Any) -> np.ndarray:
         """This method runs inference on the test data, returns predictions.
 
         This is done by calling `test_step` under the hood. Customise this method as you
@@ -474,24 +565,22 @@ class PyTorchBitfountModel(
         """
         data.load_data()
         self._add_datasource_to_schema(datasource=data)
-        test_dl = self._databunch.get_test_dataloader(self.batch_size)
-        if isinstance(test_dl, _BitfountDataLoader):
+        test_dl = self.databunch.get_test_dataloader(self.batch_size)
+        if isinstance(test_dl, BitfountDataLoader):
             logger.info(
-                f"Using test portion of dataset for inference - this has"
-                f"{len(test_dl)} records."
+                f"Using test portion of dataset for inference - this has "
+                f"{len(test_dl)} record(s)."
             )
         else:
             raise ValueError("No data provided to run inference for the model on.")
 
-        self._pl_trainer.test(
-            model=self, test_dataloaders=cast(PyTorchDataLoader, test_dl)
-        )
-        return cast(np.ndarray, self.preds)
+        self._pl_trainer.test(model=self, dataloaders=cast(PyTorchDataLoader, test_dl))
+        return cast(np.ndarray, self._test_preds)
 
     def _fit_local(
         self,
-        data: DataSource,
-        metrics: Optional[MutableMapping[str, Metric]] = None,
+        data: BaseSource,
+        metrics: Optional[Union[str, List[str], MutableMapping[str, Metric]]] = None,
         **kwargs: Any,
     ) -> Dict[str, str]:
         """Trains the model on local data.
@@ -507,12 +596,3 @@ class PyTorchBitfountModel(
 
         # Return the validation stats to be sent back
         return {k: ("%.4f" % v) for k, v in self.val_stats[-1].items()}
-
-    @property
-    def training_needed(self) -> bool:
-        """Dictates whether the model needs training.
-
-        Returns:
-            True if one of either `self.epochs` or `self.steps` are > 0.
-        """
-        return sum(filter(None, (self.epochs, self.steps))) > 0
