@@ -2,16 +2,19 @@
 from __future__ import annotations
 
 import logging
-from typing import Any, List, Optional, Tuple, Type, Union, cast
+from typing import Any, ClassVar, List, Optional, Tuple, Union, cast
 
-from marshmallow import fields, post_load
+from marshmallow import fields
 import pandas as pd
 from pytorch_tabnet.tab_model import TabNetClassifier as TabNetClassifier_
 import torch
 from torch import nn as nn
 from torch.nn import Module
 
-from bitfount.backends.pytorch.data.dataloaders import _PyTorchBitfountDataLoader
+from bitfount.backends.pytorch.data.dataloaders import (
+    PyTorchBitfountDataLoader,
+    _BasePyTorchBitfountDataLoader,
+)
 from bitfount.backends.pytorch.models.base_models import (
     _OPTIMIZERS,
     _SCHEDULERS,
@@ -33,6 +36,7 @@ from bitfount.backends.pytorch.types import (
     ImgXorTabDataSplit,
     TabDataReturnType,
 )
+from bitfount.data.datastructure import DataStructure
 from bitfount.data.types import SemanticType
 from bitfount.models.base_models import (
     CNNModelStructure,
@@ -41,32 +45,33 @@ from bitfount.models.base_models import (
     NeuralNetworkPredefinedModel,
     Scheduler,
 )
-from bitfount.types import _JSONDict
-from bitfount.utils import DEFAULT_SEED
+from bitfount.types import T_FIELDS_DICT
+from bitfount.utils import DEFAULT_SEED, delegates
 
 logger = logging.getLogger(__name__)
 
 
+@delegates()
 class PyTorchImageClassifier(PyTorchClassifierMixIn, BasePyTorchModel):
     """A Pytorch model designed specifically for image classification problems.
 
     The model can handle binary, multiclass and multilabel classification problems.
 
-    Raises:
-        ValueError: If the model structure is not an instance of `CNNModelStructure` or
-            `NeuralNetworkPredefinedModel`.
-        ValueError: If the loss function is not `nn.BCEWithLogitsLoss` or
-            `nn.CrossEntropyLoss`.
-
     :::info
 
     Currently images are scaled to 224 x 224.
 
     :::
+
+    Raises:
+        ValueError: If the model structure is not an instance of `CNNModelStructure` or
+            `NeuralNetworkPredefinedModel`.
+        ValueError: If the loss function is not `nn.BCEWithLogitsLoss` or
+            `nn.CrossEntropyLoss`.
     """
 
-    train_dl: _PyTorchBitfountDataLoader
-
+    train_dl: _BasePyTorchBitfountDataLoader
+    datastructure: DataStructure
     input_dim: int = 224  # default input images are 224 x 224
 
     def __init__(self, **kwargs: Any):
@@ -77,7 +82,6 @@ class PyTorchImageClassifier(PyTorchClassifierMixIn, BasePyTorchModel):
             raise ValueError("Model structure does not match model")
 
         super().__init__(model_structure=model_structure, **kwargs)
-
         if self.loss_func is None:
             if self.multilabel:
                 self.loss_func = nn.BCEWithLogitsLoss
@@ -183,9 +187,9 @@ class PyTorchImageClassifier(PyTorchClassifierMixIn, BasePyTorchModel):
                 self.model_structure.ff_layers,
                 self._get_convolution_final_output_dimension(),
             )
+            logger.debug(f"Creating model with {self.model_structure.num_heads} heads")
             # self.model_structure.dropout_probs is set in post_init,
             # so we can just cast below
-            logger.debug(f"Creating model with {self.model_structure.num_heads} heads")
             model = _PyTorchConvNeuralNet(
                 layer_sizes,
                 cast(List[float], self.model_structure.dropout_probs),
@@ -200,20 +204,8 @@ class PyTorchImageClassifier(PyTorchClassifierMixIn, BasePyTorchModel):
             )
         return model
 
-    class _Schema(BasePyTorchModel._Schema, PyTorchClassifierMixIn._Schema):
-        @post_load
-        def recreate_model(
-            self, data: _JSONDict, **kwargs: Any
-        ) -> PyTorchImageClassifier:
-            """Recreates a PyTorchImageClassifier."""
-            return PyTorchImageClassifier(**data)
-
-    @classmethod
-    def get_schema(cls) -> Type[PyTorchImageClassifier._Schema]:
-        """Returns the Schema for the model."""
-        return cls._Schema
-
 
+@delegates()
 class PyTorchTabularClassifier(PyTorchClassifierMixIn, BasePyTorchModel):
     """A Pytorch model designed specifically for tabular classification problems.
 
@@ -226,7 +218,7 @@ class PyTorchTabularClassifier(PyTorchClassifierMixIn, BasePyTorchModel):
             `nn.CrossEntropyLoss`.
     """
 
-    train_dl: _PyTorchBitfountDataLoader
+    train_dl: _BasePyTorchBitfountDataLoader
 
     def __init__(self, **kwargs: Any):
         model_structure = kwargs.pop("model_structure", FeedForwardModelStructure())
@@ -241,6 +233,7 @@ class PyTorchTabularClassifier(PyTorchClassifierMixIn, BasePyTorchModel):
             model_structure=cast(NeuralNetworkModelStructure, model_structure),
             **kwargs,
         )
+
         if self.loss_func is None:
             if hasattr(self, "multilabel") and (self.multilabel is not False):
                 self.loss_func = nn.BCEWithLogitsLoss
@@ -255,13 +248,14 @@ class PyTorchTabularClassifier(PyTorchClassifierMixIn, BasePyTorchModel):
         Takes number of continuous features and number of heads. Creates and
         returns model.
         """
-        ignore_cols_for_training = self.datastructure.get_columns_ignored_for_training()
-
         table_schema = self.datastructure.get_table_schema(
             schema=self.schema, pod_identifier=self.pod_identifier
         )
-        self.set_number_of_classes(table_schema)
+        ignore_cols_for_training = self.datastructure.get_columns_ignored_for_training(
+            table_schema
+        )
 
+        self.set_number_of_classes(table_schema)
         num_continuous = len(
             [
                 col
@@ -295,18 +289,15 @@ class PyTorchTabularClassifier(PyTorchClassifierMixIn, BasePyTorchModel):
             )
             for _ in range(num_heads)
         ]
-        logger.debug(f"Creating model with {num_heads} heads")
-        # self.model_structure.dropout probs is set in post_init,
+        # self.model_structure.dropout_probs is set in post_init,
         # so we can just cast below
+        logger.debug(f"Creating model with {num_heads} heads")
         model = _PyTorchFeedForwardNeuralNet(
             embedding_sizes,
             self.model_structure.embedding_dropout,
             num_continuous,
             layer_sizes,
-            cast(
-                List[float],
-                self.model_structure.dropout_probs,
-            ),
+            cast(List[float], self.model_structure.dropout_probs),
             self.model_structure.mish_activation_function,
             head_sizes,
         )
@@ -326,11 +317,14 @@ class PyTorchTabularClassifier(PyTorchClassifierMixIn, BasePyTorchModel):
         NB: `ignore_classes` is never returned
         """
         tab, sup = cast(ImgXorTabDataSplit, data)
-        ignore_cols_for_training = self.datastructure.get_columns_ignored_for_training()
 
         table_schema = self.datastructure.get_table_schema(
             schema=self.schema, pod_identifier=self.pod_identifier
         )
+        ignore_cols_for_training = self.datastructure.get_columns_ignored_for_training(
+            table_schema
+        )
+
         n_cat = len(
             _calculate_embedding_sizes(
                 table_schema.get_categorical_feature_sizes(ignore_cols_for_training)
@@ -359,29 +353,37 @@ class PyTorchTabularClassifier(PyTorchClassifierMixIn, BasePyTorchModel):
             category = None
         return (x_1.t(), x_2), weights, category
 
-    class _Schema(BasePyTorchModel._Schema, PyTorchClassifierMixIn._Schema):
-        @post_load
-        def recreate_model(
-            self, data: _JSONDict, **kwargs: Any
-        ) -> PyTorchTabularClassifier:
-            """Recreates a PyTorchTabularClassifier."""
-            return PyTorchTabularClassifier(**data)
-
-    @classmethod
-    def get_schema(cls) -> Type[PyTorchTabularClassifier._Schema]:
-        """Returns the Schema for the model."""
-        return cls._Schema
-
 
+@delegates()
 class PyTorchLogisticRegressionClassifier(PyTorchClassifierMixIn, BasePyTorchModel):
     """A Logistic Regression classifier implemented in PyTorch.
 
     Utilises softmax regression to allow extension to more than 2 classes.
-
     The input and output dimensions are calculated from the data automatically.
+    A softmax regressor is used to ensure this can work on multi-class
+        problems.
+
+    Args:
+        bias: Whether the underlying linear layer should learn an additive
+            bias. Default True.
+        l1_regularization_weight: The weight of L1 regularization to apply, if any.
+        l2_regularization_weight: The weight of L2 regularization to apply, if any.
+        embed_categorical: Whether to use categorical embedding layers to handle
+            categorical variables, or to treat them as inherently label-encoded.
+            Default True.
+        embed_categorical_dropout: The dropout probability to apply to categorical
+            variables if using embed_categorical.
+        **kwargs: Other keyword arguments for the model.
     """
 
-    train_dl: _PyTorchBitfountDataLoader
+    train_dl: PyTorchBitfountDataLoader
+    fields_dict: ClassVar[T_FIELDS_DICT] = {
+        "bias": fields.Boolean(),
+        "l1_regularization_weight": fields.Float(allow_none=True),
+        "l2_regularization_weight": fields.Float(allow_none=True),
+        "embed_categorical": fields.Boolean(),
+        "embed_categorical_dropout": fields.Float(),
+    }
 
     def __init__(
         self,
@@ -392,23 +394,7 @@ class PyTorchLogisticRegressionClassifier(PyTorchClassifierMixIn, BasePyTorchMod
         embed_categorical_dropout: float = 0.04,
         **kwargs: Any,
     ) -> None:
-        """Create a logistic regression classifier using PyTorch.
-
-        A softmax regressor is used to ensure this can work on multi-class
-        problems.
-
-        Args:
-            bias: Whether the underlying linear layer should learn an additive
-                bias. Default True.
-            l1_regularization_weight: The weight of L1 regularization to apply, if any.
-            l2_regularization_weight: The weight of L2 regularization to apply, if any.
-            embed_categorical: Whether to use categorical embedding layers to handle
-                categorical variables, or to treat them as inherently label-encoded.
-                Default True.
-            embed_categorical_dropout: The dropout probability to apply to categorical
-                variables if using embed_categorical.
-            **kwargs: Other keyword arguments for the model.
-        """
+        """Create a logistic regression classifier using PyTorch."""
         # Override model_structure; model_structure is required by the
         # NeuralNetworkMixin, but is not used in this class, so we override it
         # with an empty one and log out that this has happened
@@ -463,7 +449,9 @@ class PyTorchLogisticRegressionClassifier(PyTorchClassifierMixIn, BasePyTorchMod
         self.set_number_of_classes(table_schema)
 
         # Get details on continuous and categorical features
-        ignore_cols_for_training = self.datastructure.get_columns_ignored_for_training()
+        ignore_cols_for_training = self.datastructure.get_columns_ignored_for_training(
+            table_schema
+        )
 
         if self.embed_categorical:
             # If categorical embedding requested, calculate embedding sizes
@@ -524,10 +512,12 @@ class PyTorchLogisticRegressionClassifier(PyTorchClassifierMixIn, BasePyTorchMod
         tab, sup = cast(ImgXorTabDataSplit, data)
 
         # Gather schema-related information
-        ignore_cols_for_training = self.datastructure.get_columns_ignored_for_training()
         table_schema = self.datastructure.get_table_schema(
             schema=self.schema, pod_identifier=self.pod_identifier
         )
+        ignore_cols_for_training = self.datastructure.get_columns_ignored_for_training(
+            table_schema
+        )
 
         # Calculate the number of continuous and categorical columns
         n_categorical = table_schema.get_num_categorical(
@@ -625,44 +615,28 @@ class PyTorchLogisticRegressionClassifier(PyTorchClassifierMixIn, BasePyTorchMod
 
         return loss
 
-    class _Schema(BasePyTorchModel._Schema, PyTorchClassifierMixIn._Schema):
-        bias = fields.Boolean()
-        l1_regularization_weight = fields.Float(allow_none=True)
-        l2_regularization_weight = fields.Float(allow_none=True)
-        embed_categorical = fields.Boolean()
-        embed_categorical_dropout = fields.Float()
-
-        @post_load
-        def recreate_model(
-            self, data: _JSONDict, **kwargs: Any
-        ) -> PyTorchLogisticRegressionClassifier:
-            """Recreates a PyTorchImageClassifier."""
-            return PyTorchLogisticRegressionClassifier(**data)
-
-    @classmethod
-    def get_schema(cls) -> Type[PyTorchLogisticRegressionClassifier._Schema]:
-        """Returns the Schema for the model."""
-        return cls._Schema
-
 
+@delegates()
 class TabNetClassifier(PyTorchClassifierMixIn, BaseTabNetModel):
     """TabNet Classifier for binary and multiclass tabular classification problems.
 
     See base class for more information.
     """
 
-    train_dl: _PyTorchBitfountDataLoader
+    train_dl: PyTorchBitfountDataLoader
 
     def __init__(self, **kwargs: Any) -> None:
         super().__init__(**kwargs)
 
     def _create_model(self) -> TabNetClassifier_:
         """Create model for Binary or Multiclass classification."""
-        ignore_cols_for_training = self.datastructure.get_columns_ignored_for_training()
-
         table_schema = self.datastructure.get_table_schema(
             schema=self.schema, pod_identifier=self.pod_identifier
         )
+        ignore_cols_for_training = self.datastructure.get_columns_ignored_for_training(
+            table_schema
+        )
+
         self.set_number_of_classes(table_schema)
 
         # Only consider the tabular part
@@ -701,14 +675,3 @@ class TabNetClassifier(PyTorchClassifierMixIn, BaseTabNetModel):
             mask_type=self.mask_type,
             seed=self.seed or DEFAULT_SEED,
         )
-
-    class _Schema(BaseTabNetModel._Schema, PyTorchClassifierMixIn._Schema):
-        @post_load
-        def recreate_model(self, data: _JSONDict, **kwargs: Any) -> TabNetClassifier:
-            """Recreate TabNetClassifier model."""
-            return TabNetClassifier(**data)
-
-    @classmethod
-    def get_schema(cls) -> Type[TabNetClassifier._Schema]:
-        """Returns the Schema for the model."""
-        return cls._Schema
