@@ -1,51 +1,60 @@
 """Classes concerning data structures.
 
-DataStructures provide information about the columns of a DataSource for a specific
+DataStructures provide information about the columns of a BaseSource for a specific
 Modelling Job.
 """
 from __future__ import annotations
 
 from dataclasses import dataclass
+import inspect
 import logging
-from os import PathLike
+from types import MappingProxyType
 from typing import (
     TYPE_CHECKING,
     Any,
+    ClassVar,
     Dict,
     List,
     Mapping,
     MutableMapping,
     Optional,
+    Type,
     Union,
     cast,
 )
 
 import desert
-from marshmallow import Schema as MarshmallowSchema
-from marshmallow import fields, post_load
-import yaml
+from marshmallow import fields
 
-from bitfount.data.datasource import DataFrameSource
-from bitfount.data.datasplitters import _DatasetSplitter
+from bitfount.data.datasources.database_source import DatabaseSource
+from bitfount.data.datasplitters import DatasetSplitter
 from bitfount.data.exceptions import DataStructureError
 from bitfount.data.schema import BitfountSchema, TableSchema
 from bitfount.data.types import (
     CategoricalRecord,
     ContinuousRecord,
+    DataSplit,
     ImageRecord,
     SchemaOverrideMapping,
     SemanticType,
     StrDictField,
     TextRecord,
+    _SemanticTypeRecord,
     _SemanticTypeValue,
 )
+from bitfount.transformations.base_transformation import TRANSFORMATION_REGISTRY
 from bitfount.transformations.batch_operations import BatchTimeOperation
 from bitfount.transformations.parser import TransformationsParser
-from bitfount.types import _JSONDict
+from bitfount.types import (
+    T_FIELDS_DICT,
+    T_NESTED_FIELDS,
+    _BaseSerializableObjectMixIn,
+    _JSONDict,
+)
 from bitfount.utils import _add_this_to_list
 
 if TYPE_CHECKING:
-    from bitfount.data.datasource import DataSource
+    from bitfount.data.datasources.base_source import BaseSource
     from bitfount.runners.config_schemas import (
         DataStructureAssignConfig,
         DataStructureSelectConfig,
@@ -61,14 +70,35 @@ DEFAULT_IMAGE_TRANSFORMATIONS: List[Union[str, _JSONDict]] = [
     "ToTensorV2",
 ]
 
+_registry: Dict[str, Type[BaseDataStructure]] = {}
+registry: Mapping[str, Type[BaseDataStructure]] = MappingProxyType(_registry)
+
 
 @dataclass
-class DataStructure:
-    """Information about the columns of a DataSource.
+class BaseDataStructure:
+    """Base DataStructure class."""
+
+    @classmethod
+    def __init_subclass__(cls, **kwargs: Any):
+        if not inspect.isabstract(cls):
+            logger.debug(f"Adding {cls.__name__}: {cls} to registry")
+            _registry[cls.__name__] = cls
+
+
+@dataclass
+class DataStructure(BaseDataStructure, _BaseSerializableObjectMixIn):
+    """Information about the columns of a BaseSource.
 
     This component provides the desired structure of data
     to be used by discriminative machine learning models.
 
+    :::note
+
+    If the datastructure includes image columns, batch transformations will be applied
+    to them.
+
+    :::
+
     Args:
         table: The table in the Pod schema to be used for local data. If executing a
             remote task, this should be a mapping of Pod names to table names.
@@ -80,12 +110,12 @@ class DataStructure:
             will be returned after the sql query is executed. For a local training
             task it will be a mapping of column names to their types, for a remote
             task it will be a mapping of the Pod name to the new columns and types.
-            If a column is defined as `categorical`, the mapping should include a
+            If a column is defined as "categorical", the mapping should include a
             mapping to the categories. Required if a sql query is provided.
-            E.g. {'Pod_id': {'categorical': [{'col1': {'value_1':0, 'value_2': 1}}],
-            'continuous': ['col2']} for remote training or {'categorical':
-            [{'col1': {'value_1':0, 'value_2': 1}}],'continuous': ['col2']}
-            for local training. Defaults to None.
+            E.g. {'Pod_id': {'categorical': [{'col1': {'value_1':0, 'value_2': 1
+            }}], "continuous": ['col2']} for remote training
+            or {'categorical':[{ "col1" : {'value_1':0, 'value_2': 1}}],'continuous':
+            ['col2']} for local training. Defaults to None.
         target: The training target column or list of columns.
         ignore_cols: A list of columns to ignore when getting the
             data. Defaults to None.
@@ -112,14 +142,46 @@ class DataStructure:
             Defaults to None.
         dataset_transforms: A dictionary of transformations to apply to
             the whole dataset. Defaults to None.
+        auto_convert_grayscale_images: Whether or not to automatically convert grayscale
+            images to RGB. Defaults to True.
 
     Raises:
         DataStructureError: If 'sql_query' is provided as well as either `selected_cols`
             or `ignore_cols`.
         DataStructureError: If both `ignore_cols` and `selected_cols` are provided.
         DataStructureError: If the `multihead_col` is provided without `multihead_size`.
+        ValueError: If a batch transformation name is not recognised.
+
     """
 
+    fields_dict = {
+        "table": StrDictField(allow_none=True),
+        "query": StrDictField(allow_none=True),
+        "schema_types_override": fields.Dict(allow_none=True),
+        "target": fields.Raw(allow_none=True),
+        "ignore_cols": fields.List(fields.Str(), allow_none=True),
+        "selected_cols": fields.List(fields.Str(), allow_none=True),
+        "loss_weights_col": fields.Str(allow_none=True),
+        "multihead_col": fields.Str(allow_none=True),
+        "multihead_size": fields.Int(allow_none=True),
+        "ignore_classes_col": fields.Str(allow_none=True),
+        "batch_transforms": fields.List(
+            fields.Dict(
+                keys=fields.Str(),
+                values=fields.Dict(keys=fields.Str()),
+            ),
+            allow_none=True,
+        ),
+        "dataset_transforms": fields.List(
+            fields.Dict(
+                keys=fields.Str(),
+                values=fields.Dict(keys=fields.Str()),
+            ),
+            allow_none=True,
+        ),
+        "auto_convert_grayscale_images": fields.Boolean(),
+    }  # type: ClassVar[T_FIELDS_DICT]
+    nested_fields = {}  # type: ClassVar[T_NESTED_FIELDS]
     table: Optional[Union[str, Mapping[str, str]]] = None
     query: Optional[Union[str, Mapping[str, str]]] = None
     schema_types_override: Optional[
@@ -133,7 +195,7 @@ class DataStructure:
     selected_cols: List[str] = desert.field(  # type: ignore[assignment] # Reason: above
         fields.List(fields.String()), default_factory=list
     )
-    data_splitter: Optional[_DatasetSplitter] = None
+    data_splitter: Optional[DatasetSplitter] = None
     loss_weights_col: Optional[str] = None
     multihead_col: Optional[str] = None
     multihead_size: Optional[int] = None
@@ -141,8 +203,10 @@ class DataStructure:
     image_cols: Optional[List[str]] = None
     batch_transforms: Optional[List[Dict[str, _JSONDict]]] = None
     dataset_transforms: Optional[List[Dict[str, _JSONDict]]] = None
+    auto_convert_grayscale_images: bool = True
 
     def __post_init__(self) -> None:
+        self.class_name = type(self).__name__
         if self.table and self.query:
             raise DataStructureError(
                 "Invalid parameter specification. "
@@ -185,19 +249,37 @@ class DataStructure:
         if self.batch_transforms is None and self.image_cols:
             default_image_transformations = []
             for col in self.image_cols:
-                for step in ["train", "validation"]:
+                for step in DataSplit:
                     default_image_transformations.append(
                         {
-                            "Image": {
+                            "albumentations": {
                                 "arg": col,
                                 "output": True,
                                 "transformations": DEFAULT_IMAGE_TRANSFORMATIONS,
-                                "step": step,
+                                "step": step.value,
                             }
                         }
                     )
             self.batch_transforms = default_image_transformations
 
+        # Ensure specified batch transformations are all valid transformations
+        if self.batch_transforms is not None:
+            invalid_batch_transforms = []
+            for _dict in self.batch_transforms:
+                for tfm in _dict:
+                    if tfm not in TRANSFORMATION_REGISTRY:
+                        invalid_batch_transforms.append(tfm)
+            if invalid_batch_transforms:
+                raise ValueError(
+                    f"The following batch transformations are not recognised: "
+                    f"{', '.join(sorted(invalid_batch_transforms))}."
+                )
+
+        # Create mapping of all feature names used in training together with the
+        # corresponding semantic type. This is the final mapping that will be used
+        # to decide which features will be actually be used.
+        self.selected_cols_w_types: Dict[_SemanticTypeValue, List[str]] = {}
+
     @classmethod
     def create_datastructure(
         cls,
@@ -209,8 +291,9 @@ class DataStructure:
         """Creates a datastructure based on the yaml config.
 
         Args:
-            table: The table in the Pod schema to be used for local data. If executing a
-                remote task, this should a mapping of Pod names to table names.
+            table_config: The table in the Pod schema to be used for local data.
+                If executing a remote task, this should a mapping of Pod names
+                to table names.
             select: The configuration for columns to be included/excluded
                 from the `DataStructure`.
             transform: The configuration for dataset and batch transformations
@@ -242,21 +325,6 @@ class DataStructure:
             dataset_transforms=transform.dataset,
         )
 
-    @classmethod
-    def load_from_file(cls, file_path: Union[str, PathLike]) -> DataStructure:
-        """Loads DataStructure from yaml file.
-
-        Args:
-            file: A yaml file with the `DataStructure` configuration.
-
-        Returns:
-            The loaded `DataStructure`.
-        """
-        with open(file_path, "r") as f:
-            datastructure_yaml = yaml.safe_load(f)
-        datastructure: DataStructure = cls._Schema().load(datastructure_yaml)
-        return datastructure
-
     def get_table_name(self, pod_identifier: Optional[str] = None) -> str:
         """Returns the relevant table name of the `DataStructure`.
 
@@ -284,23 +352,32 @@ class DataStructure:
             if isinstance(self.table, str):
                 return None
             else:
-                pod_identifiers = list(self.table.keys())
+                pod_identifiers = list(self.table)
         elif self.query is not None:
             if isinstance(self.query, str):
                 return None
             else:
-                pod_identifiers = list(self.query.keys())
+                pod_identifiers = list(self.query)
         else:
             return None
         return pod_identifiers
 
-    def get_columns_ignored_for_training(self) -> List[str]:
+    def get_columns_ignored_for_training(self, table_schema: TableSchema) -> List[str]:
         """Adds all the extra columns that will not be used in model training.
 
+        Args:
+            table_schema: The schema of the table.
+
         Returns:
             ignore_cols_aux: A list of columns that will be ignored when
                 training a model.
         """
+        if self.selected_cols:
+            self.ignore_cols = [
+                feature
+                for feature in table_schema.get_feature_names()
+                if feature not in self.selected_cols
+            ]
         ignore_cols_aux = self.ignore_cols[:]
         ignore_cols_aux = _add_this_to_list(self.target, ignore_cols_aux)
         ignore_cols_aux = _add_this_to_list(self.loss_weights_col, ignore_cols_aux)
@@ -318,7 +395,7 @@ class DataStructure:
             [
                 col
                 for col in schema.get_feature_names()
-                if col not in self.get_columns_ignored_for_training()
+                if col not in self.get_columns_ignored_for_training(schema)
                 and col not in schema.get_feature_names(SemanticType.TEXT)
             ]
         )
@@ -343,23 +420,35 @@ class DataStructure:
                 for feature in schema.get_feature_names()
                 if feature not in self.selected_cols
             ]
-
+        else:
+            # Make sure we set self.selected_cols
+            self.selected_cols = [
+                feature
+                for feature in schema.get_feature_names()
+                if feature not in self.ignore_cols
+            ]
+        if self.target and self.target not in self.selected_cols:
+            self.selected_cols = _add_this_to_list(self.target, self.selected_cols)
         # Get the list of all columns ignored for training
-        ignore_cols_aux = self.get_columns_ignored_for_training()
+        ignore_cols_aux = self.get_columns_ignored_for_training(schema)
 
-        # Create mapping of all feature names used in training
+        # Populate mapping of all feature names used in training
         # together with the corresponding semantic type
-        self.selected_cols_w_types: Dict[_SemanticTypeValue, List[str]] = {}
         for stype, features in schema.features.items():
-            columns_stype_list: List[_SemanticTypeValue] = list(features.keys())  # type: ignore[attr-defined] # Reason: the features will always be a FeaturesDict # noqa: B950
+            columns_stype_list = list(cast(Dict[str, _SemanticTypeRecord], features))
+
+            # Iterating over `self.selected_cols` ensures we preserve the order that the
+            # user specified the columns
             self.selected_cols_w_types[cast(_SemanticTypeValue, stype)] = [
-                col for col in columns_stype_list if col not in ignore_cols_aux
+                col
+                for col in self.selected_cols
+                if (col in columns_stype_list and col not in ignore_cols_aux)
             ]
         # Add mapping to empty list for all stypes not present
         # in the current datastructure
         all_stypes = [stype.value for stype in SemanticType]
         for stype in all_stypes:
-            if stype not in self.selected_cols_w_types.keys():
+            if stype not in self.selected_cols_w_types:
                 self.selected_cols_w_types[cast(_SemanticTypeValue, stype)] = []
 
         # Get the number of images present in the datastructure.
@@ -393,17 +482,21 @@ class DataStructure:
                         col for col in value["col"] if col not in original_cols_to_keep
                     ]
                     new_columns = [f"{col}_{key}" for col in value["col"]]
-                    self.selected_cols.extend(new_columns)
-                    self.ignore_cols.extend(discard_columns)
+                    # Error raised in the pods if we set both ignore_cols
+                    # and selected_cols here.
+                    if self.selected_cols:
+                        self.selected_cols.extend(new_columns)
+                    else:
+                        self.ignore_cols.extend(discard_columns)
                     self.selected_cols = [
                         col for col in self.selected_cols if col not in discard_columns
                     ]
 
-    def apply_dataset_transformations(self, datasource: DataSource) -> DataSource:
+    def apply_dataset_transformations(self, datasource: BaseSource) -> BaseSource:
         """Applies transformations to whole dataset.
 
         Args:
-            datasource: The `DataSource` object to be transformed.
+            datasource: The `BaseSource` object to be transformed.
 
         Returns:
             datasource: The transformed datasource.
@@ -439,7 +532,7 @@ class DataStructure:
 
         if isinstance(self.query, str):
             # That means that the datastructure is intended for local training.
-            if "categorical" in self.schema_types_override.keys():
+            if "categorical" in self.schema_types_override:
                 # For local training, categorical features should be
                 # defined inside a dictionary.
                 for item in self.schema_types_override["categorical"]:
@@ -450,8 +543,8 @@ class DataStructure:
                             "Please specify encodings."
                         )
         elif isinstance(self.query, dict):
-            for pod_id in self.query.keys():
-                if pod_id not in self.schema_types_override.keys():
+            for pod_id in self.query:
+                if pod_id not in self.schema_types_override:
                     raise DataStructureError(
                         f"The pod id:{pod_id} given to the query was "
                         f"not found in the schema override. Please "
@@ -460,7 +553,7 @@ class DataStructure:
                     )
             for overrides in self.schema_types_override.values():
                 overrides = cast(SchemaOverrideMapping, overrides)
-                if "categorical" in overrides.keys():
+                if "categorical" in overrides:
                     for item in overrides["categorical"]:
                         if not isinstance(item, dict):
                             raise DataStructureError(
@@ -471,7 +564,7 @@ class DataStructure:
 
     def _override_schema(
         self,
-        datasource: Optional[DataSource] = None,
+        datasource: Optional[BaseSource] = None,
         pod_identifier: Optional[str] = None,
     ) -> TableSchema:
         """Method to override the pod schema when a sql query is given."""
@@ -485,13 +578,17 @@ class DataStructure:
             feature_overrides = cast(
                 SchemaOverrideMapping, self.schema_types_override[pod_identifier]  # type: ignore[index] # Reason: An error will be caught by the _validate_schema_features. # noqa: B950
             )
+        else:
+            raise ValueError(
+                "No query or dictionary of pod_identifiers to queries was given."
+            )
+
         table_name = "data"
         table_dtypes = None
         if datasource and datasource._data_is_loaded:
             # Data is loaded into a dataframe by the time we get here,
             # so we use the DataFrameSource to get the dtypes.
-            datasource_loader = DataFrameSource(data=datasource.data)
-            table_dtypes = datasource_loader.get_dtypes()
+            table_dtypes = datasource.get_dtypes()
 
         self.selected_cols = []
         schema = TableSchema(name=table_name)
@@ -574,7 +671,7 @@ class DataStructure:
         self,
         schema: BitfountSchema,
         pod_identifier: Optional[str] = None,
-        datasource: Optional[DataSource] = None,
+        datasource: Optional[BaseSource] = None,
     ) -> TableSchema:
         """Returns the table schema based on the datastructure arguments.
 
@@ -592,6 +689,11 @@ class DataStructure:
         """
         if self.query:
             # If the datastructure is given a query, then we use the schema override.
+            if datasource and isinstance(datasource, DatabaseSource):
+                if isinstance(self.query, dict) and pod_identifier:
+                    datasource.datastructure_query = self.query[pod_identifier]
+                elif isinstance(self.query, str):
+                    datasource.datastructure_query = self.query
             table_schema = self._override_schema(
                 pod_identifier=pod_identifier, datasource=datasource
             )
@@ -623,34 +725,3 @@ class DataStructure:
                     )
                 ),
             )
-
-    class _Schema(MarshmallowSchema):
-        table = StrDictField(allow_none=True)
-        query = StrDictField(allow_none=True)
-        schema_types_override = fields.Dict(allow_none=True)
-        target = fields.Raw(required=True)
-        ignore_cols = fields.List(fields.Str(), allow_none=True)
-        loss_weights_col = fields.Str(allow_none=True)
-        multihead_col = fields.Str(allow_none=True)
-        ignore_classes_col = fields.Str(allow_none=True)
-        batch_transforms = fields.List(
-            fields.Dict(
-                keys=fields.Str(),
-                values=fields.Dict(keys=fields.Str()),
-            ),
-            allow_none=True,
-        )
-        dataset_transforms = fields.List(
-            fields.Dict(
-                keys=fields.Str(),
-                values=fields.Dict(keys=fields.Str()),
-            ),
-            allow_none=True,
-        )
-
-        @post_load
-        def recreate_datastructure(
-            self, data: _JSONDict, **_kwargs: Any
-        ) -> DataStructure:
-            """Recreates DataStructure."""
-            return DataStructure(**data)
