@@ -20,7 +20,6 @@ from typing import (
 )
 
 from PIL import Image
-import databricks.koalas as ks
 from marshmallow import fields, post_dump, post_load, pre_dump, pre_load
 import numpy as np
 import pandas as pd
@@ -29,8 +28,9 @@ from pandas.core.dtypes.common import is_numeric_dtype
 import yaml
 
 import bitfount
-from bitfount.data.datasource import DataSource
-from bitfount.data.exceptions import BitfountSchemaError
+from bitfount.data import BaseSource
+from bitfount.data.datasources.base_source import MultiTableSource
+from bitfount.data.exceptions import BitfountSchemaError, DataSourceError
 from bitfount.data.types import (
     CategoricalRecord,
     ContinuousRecord,
@@ -42,9 +42,9 @@ from bitfount.data.types import (
     _SemanticTypeRecord,
     _SemanticTypeValue,
 )
-from bitfount.data.utils import DatabaseConnection, _hash_str
-from bitfount.types import _DataFrameType, _Dtypes, _JSONDict
-from bitfount.utils import _add_this_to_list
+from bitfount.data.utils import _hash_str
+from bitfount.types import _Dtypes, _JSONDict
+from bitfount.utils import _add_this_to_list, delegates
 
 logger = logging.getLogger(__name__)
 
@@ -108,7 +108,7 @@ class _TableSchemaMarshmallowMixIn:
                 for semantic_type in result:
                     data[f"{semantic_type}Features"] = result[semantic_type]
                 return data
-            elif any([key for key in data.keys() if "Features" in key]):
+            elif any([key for key in data if "Features" in key]):
                 # Data has been already preprocessed
                 return data
             else:
@@ -131,6 +131,7 @@ class _TableSchemaMarshmallowMixIn:
             return new_schema
 
 
+@delegates()
 class TableSchema(_TableSchemaMarshmallowMixIn):
     """A schema that defines the features of a dataframe.
 
@@ -157,14 +158,17 @@ class TableSchema(_TableSchemaMarshmallowMixIn):
     def _dtype_based_stype_split(
         data: _Dtypes, ignore_cols: Optional[Sequence[str]] = None
     ) -> Dict[SemanticType, List[str]]:
-        """Returns dictionary of Semantic types and the corresponding columns in `data`.
+        """Returns dictionary of Semantic types and corresponding columns in `data`.
 
-        We first call `convert_dtypes` on the data to convert older pandas dtypes to
-        newer ones. This call is idempotent and allows us to identify the data type of
-        columns with the `object` dtype.
+        This method determines which data types correspond to which semantic types.
         """
         converted_data = data.copy()
         if ignore_cols:
+            missing_cols = [i for i in ignore_cols if i not in data]
+            if missing_cols:
+                logger.warning(
+                    f"Could not find ignored columns: {', '.join(missing_cols)}"
+                )
             converted_data = {
                 k: v for k, v in converted_data.items() if k not in ignore_cols
             }
@@ -193,7 +197,7 @@ class TableSchema(_TableSchemaMarshmallowMixIn):
     def _add_categorical_feature(
         self,
         name: str,
-        values: Union[np.ndarray, pd.Series, ks.Series],
+        values: Union[np.ndarray, pd.Series],
         dtype: Optional[Union[Dtype, np.dtype]] = None,
         description: Optional[str] = None,
     ) -> None:
@@ -222,7 +226,7 @@ class TableSchema(_TableSchemaMarshmallowMixIn):
             try:
                 stype = SemanticType(new_stype)
 
-                if stype not in existing_stypes.keys():
+                if stype not in existing_stypes:
                     existing_stypes[stype] = []
                 existing_stypes[stype] = _add_this_to_list(
                     feature_list, existing_stypes[stype]
@@ -236,17 +240,17 @@ class TableSchema(_TableSchemaMarshmallowMixIn):
                     k: v for k, v in table_dtypes.items() if k in feature_list
                 }
                 dtype_features = self._dtype_based_stype_split(feature_dtypes, [])
-                stype = list(dtype_features.keys())[0]
-                if stype not in existing_stypes.keys():
+                stype = list(dtype_features)[0]
+                if stype not in existing_stypes:
                     existing_stypes[stype] = []
                 existing_stypes[stype] = _add_this_to_list(
                     feature_list, existing_stypes[stype]
                 )
         return existing_stypes
 
-    def _add_dataframe_features(
+    def _add_datasource_features(
         self,
-        datasource: DataSource,
+        datasource: BaseSource,
         ignore_cols: List[str],
         force_stype: MutableMapping[_SemanticTypeValue, List[str]],
         descriptions: Mapping[str, str],
@@ -256,7 +260,7 @@ class TableSchema(_TableSchemaMarshmallowMixIn):
         Adds all the features in the dataframe to the schema, using the dtype to decide
         the semantic type of the feature.
         """
-        table_dtypes = datasource.get_dtypes(self.name)
+        table_dtypes = datasource.get_dtypes(table_name=self.name)
         for item in force_stype.values():
             ignore_cols = _add_this_to_list(item, ignore_cols)
         inferred_semantic_types = self._dtype_based_stype_split(
@@ -265,9 +269,9 @@ class TableSchema(_TableSchemaMarshmallowMixIn):
         semantic_types = self._combine_existing_stypes_with_forced_stypes(
             inferred_semantic_types, force_stype, table_dtypes
         )
-        if SemanticType.CATEGORICAL in semantic_types.keys():
+        if SemanticType.CATEGORICAL in semantic_types:
             values = datasource.get_values(
-                semantic_types[SemanticType.CATEGORICAL], self.name
+                semantic_types[SemanticType.CATEGORICAL], table_name=self.name
             )
 
         for stype, features in semantic_types.items():
@@ -295,10 +299,20 @@ class TableSchema(_TableSchemaMarshmallowMixIn):
                         )
 
                 elif stype == SemanticType.CATEGORICAL:
+                    # Convert arbitrary iterable to the right form
+                    feature_values: Union[np.ndarray, pd.Series]
+                    if not isinstance(
+                        (raw_feature_values := values[feature_name]),
+                        (np.ndarray, pd.Series),
+                    ):
+                        feature_values = np.asarray(raw_feature_values)
+                    else:
+                        feature_values = raw_feature_values
+
                     self._add_categorical_feature(
                         name=feature_name,
                         dtype=dtype,
-                        values=values[feature_name],
+                        values=feature_values,
                         description=description,
                     )
                 elif stype == SemanticType.IMAGE:
@@ -313,23 +327,25 @@ class TableSchema(_TableSchemaMarshmallowMixIn):
                             description=descriptions.get(feature_name),
                         )
                     record = self.features["image"][feature_name]
-                    for img in datasource.get_column(feature_name, self.name):
+                    for img in datasource.get_column(
+                        feature_name, table_name=self.name
+                    ):
                         im = Image.open(img)
                         record.dimensions[im.size] += 1
                         record.modes[im.mode] += 1
                         record.formats[im.format] += 1
 
-    def _expand_dataframe(self, dataframe: _DataFrameType) -> _DataFrameType:
+    def _expand_dataframe(self, dataframe: pd.DataFrame) -> pd.DataFrame:
         """Expands dataframe to include missing columns specified in the schema.
 
         Simply adds columns populated with default values: 'nan' for categorical
         and text columns and '0' for continuous columns.
 
         Args:
-            dataframe (DataFrameType): dataframe without all the required columns
+            dataframe: dataframe without all the required columns
 
         Returns:
-            DataFrameType: dataframe that includes all the required columns
+            Dataframe that includes all the required columns
 
         Raises:
             BitfountSchemaError: if there is a missing image column as this cannot be
@@ -339,7 +355,6 @@ class TableSchema(_TableSchemaMarshmallowMixIn):
         missing_text_value = "nan"
         missing_continuous_value = 0
         columns = list(dataframe.columns)
-
         for stype in self.features:
             # Iterate through semantic types
             for feature_name in self.features[cast(_SemanticTypeValue, stype)]:
@@ -370,37 +385,59 @@ class TableSchema(_TableSchemaMarshmallowMixIn):
         return dataframe
 
     def _reduce_dataframe(
-        self, dataframe: _DataFrameType, keep_cols: Optional[List[str]] = None
-    ) -> _DataFrameType:
+        self, dataframe: pd.DataFrame, keep_cols: Optional[List[str]] = None
+    ) -> pd.DataFrame:
         """Drops any columns that are not part of the schema.
 
         Args:
-            dataframe (DataFrameType): dataframe which includes extra columns
-            keep_cols (Optional[List[str]]): optional list of columns to keep even if
+            dataframe: dataframe which includes extra columns
+            keep_cols: optional list of columns to keep even if
                 they are not part of the schema. Defaults to None.
 
         Returns:
-            DataFrameType: dataframe with extra columns removed
+            Dataframe with extra columns removed
         """
         cols_to_keep = self.get_feature_names()
         cols_to_keep = _add_this_to_list(keep_cols, cols_to_keep)
         return dataframe[cols_to_keep]
 
-    def _apply_types(self, dataframe: _DataFrameType) -> _DataFrameType:
+    def _apply_types(
+        self, dataframe: pd.DataFrame, selected_cols: Optional[List[str]] = None
+    ) -> pd.DataFrame:
         """Applies the prescribed feature types to the dataframe.
 
         Args:
-            dataframe (DataFrameType): dataframe with varied types
+            dataframe: dataframe with varied types
+            selected_cols: optional list of columns selected for
+                training to which the types will be applied.
 
         Returns:
-            DataFrameType: dataframe with types that are specified in the schema
+            Dataframe with types that are specified in the schema
         """
+        if selected_cols:
+            selected = [
+                feature_name
+                for stype in self.features
+                for feature_name, record in self.features[
+                    cast(_SemanticTypeValue, stype)
+                ].items()
+                if feature_name in selected_cols
+            ]
+        else:
+            selected = [
+                feature_name
+                for stype in self.features
+                for feature_name, record in self.features[
+                    cast(_SemanticTypeValue, stype)
+                ].items()
+            ]
         types: Dict[str, Union[Dtype, np.dtype]] = {
             feature_name: record.dtype
             for stype in self.features
             for feature_name, record in self.features[
                 cast(_SemanticTypeValue, stype)
             ].items()
+            if feature_name in selected
         }
 
         if "categorical" in self.features:
@@ -408,22 +445,23 @@ class TableSchema(_TableSchemaMarshmallowMixIn):
                 {
                     feature_name: record.encoder.dtype
                     for feature_name, record in self.features["categorical"].items()
+                    if feature_name in selected
                 }
             )
 
         return dataframe.astype(types)
 
-    def _encode_dataframe(self, data: _DataFrameType) -> _DataFrameType:
+    def _encode_dataframe(self, data: pd.DataFrame) -> pd.DataFrame:
         """Encodes the dataframe categorical columns according to the schema.
 
         Args:
-            data (DataFrameType): the dataframe to be encoded
+            data: the dataframe to be encoded
 
         Raises:
             ValueError: if the encoder fails to encode a particular column
 
         Returns:
-            DataFrameType: the dataframe with the categorical columns encoded
+            The dataframe with the categorical columns encoded
         """
         if "categorical" in self.features:
             for feature_name, record in self.features["categorical"].items():
@@ -556,9 +594,9 @@ class TableSchema(_TableSchemaMarshmallowMixIn):
             ]
         )
 
-    def add_dataframe_features(
+    def add_datasource_features(
         self,
-        datasource: DataSource,
+        datasource: BaseSource,
         ignore_cols: Optional[Sequence[str]] = None,
         force_stype: Optional[MutableMapping[_SemanticTypeValue, List[str]]] = None,
         descriptions: Optional[Mapping[str, str]] = None,
@@ -566,8 +604,8 @@ class TableSchema(_TableSchemaMarshmallowMixIn):
         """Adds datasource features to schema.
 
         Args:
-            dataframe: The dataframe whose features this method adds.
-            ignore_cols: Columns to ignore from the `DataSource`. Defaults to None.
+            datasource: The datasource whose features this method adds.
+            ignore_cols: Columns to ignore from the `BaseSource`. Defaults to None.
             force_stype: Columns for which to change the semantic type.
                 Format: semantictype: [columnnames]. Defaults to None.
                 Example: {'categorical': ['target_column'],
@@ -585,7 +623,7 @@ class TableSchema(_TableSchemaMarshmallowMixIn):
         if force_stype is None:
             force_stype = {}
 
-        self._add_dataframe_features(
+        self._add_datasource_features(
             datasource=datasource,
             ignore_cols=ignore_cols_aux,
             force_stype=force_stype,
@@ -593,8 +631,8 @@ class TableSchema(_TableSchemaMarshmallowMixIn):
         )
 
     def apply(
-        self, dataframe: _DataFrameType, keep_cols: Optional[List[str]] = None
-    ) -> _DataFrameType:
+        self, dataframe: pd.DataFrame, keep_cols: Optional[List[str]] = None
+    ) -> pd.DataFrame:
         """Applies the schema to a dataframe and returns the transformed dataframe.
 
         Sequentially adds missing columns to the dataframe, removes superfluous columns
@@ -612,9 +650,8 @@ class TableSchema(_TableSchemaMarshmallowMixIn):
         """
         dataframe = self._expand_dataframe(dataframe)
         dataframe = self._reduce_dataframe(dataframe, keep_cols=keep_cols)
-        dataframe = self._apply_types(dataframe)
+        dataframe = self._apply_types(dataframe, selected_cols=keep_cols)
         dataframe = self._encode_dataframe(dataframe)
-
         return dataframe
 
     def __eq__(self, other: Any) -> bool:
@@ -767,13 +804,19 @@ class _BitfountSchemaMarshmallowMixIn:
             return new_schema
 
 
+@delegates()
 class BitfountSchema(_BitfountSchemaMarshmallowMixIn):
-    """A schema that defines the tables of a `DataSource`.
+    """A schema that defines the tables of a `BaseSource`.
 
-    It lists all the tables found in the `DataSource` and the features in those tables.
+    It lists all the tables found in the `BaseSource` and the features in those tables.
+
+    Args:
+        datasource: An optional `BaseSource` object. This can be provided as a utility
+            to avoid calling `add_datasource_tables`.
+        **kwargs: Optional keyword arguments to be provided to `add_datasource_tables`.
     """
 
-    def __init__(self, datasource: Optional[DataSource] = None, **kwargs: Any):
+    def __init__(self, datasource: Optional[BaseSource] = None, **kwargs: Any):
         self.tables: List[TableSchema] = []
         # self._orig_hash is used to store the hash when loading a previously
         # generated schema.
@@ -795,13 +838,13 @@ class BitfountSchema(_BitfountSchemaMarshmallowMixIn):
     def hash(self) -> str:
         """The hash of this schema.
 
-        This relates to the DataSource(s) that were used in the generation of this
+        This relates to the BaseSource(s) that were used in the generation of this
         schema to assure that this schema is used against compatible data sources.
 
         Returns:
             A sha256 hash of the `_datasource_hashes`.
         """
-        # Must be sorted to ensure ordering of DataSources being added doesn't
+        # Must be sorted to ensure ordering of BaseSources being added doesn't
         # change things.
         frozen_hashes: str = str(sorted(self._datasource_hashes))
         return _hash_str(frozen_hashes)
@@ -826,7 +869,7 @@ class BitfountSchema(_BitfountSchemaMarshmallowMixIn):
 
     def add_datasource_tables(
         self,
-        datasource: DataSource,
+        datasource: BaseSource,
         table_name: Optional[str] = None,
         table_descriptions: Optional[Mapping[str, str]] = None,
         column_descriptions: Optional[Mapping[str, Mapping[str, str]]] = None,
@@ -835,12 +878,12 @@ class BitfountSchema(_BitfountSchemaMarshmallowMixIn):
             Mapping[str, MutableMapping[_SemanticTypeValue, List[str]]]
         ] = None,
     ) -> None:
-        """Adds the tables from a `DataSource` to the schema.
+        """Adds the tables from a `BaseSource` to the schema.
 
         Args:
-            datasource: The `DataSource` to add the tables from.
+            datasource: The `BaseSource` to add the tables from.
             table_name: The name of the table if there is only one table in the
-                `DataSource`.
+                `BaseSource`.
             table_descriptions: A mapping of table names to descriptions.
             column_descriptions: A mapping of table names to a mapping of column names
                 to descriptions.
@@ -850,16 +893,26 @@ class BitfountSchema(_BitfountSchemaMarshmallowMixIn):
 
         Raises:
             BitfountSchemaError: If the schema is already frozen.
+            DataSourceError: If the `BaseSource` for the provided datasource has
+                not been initialised properly. This can be done by calling
+                `super().__init__(**kwargs)` in the `__init__` of the DataSource.
             ValueError: If a table name hasn't been provided for a single table
-                `DataSource`.
+                `BaseSource`.
         """
+        # Raise error if BaseSource has not been initialised
+        if not datasource.is_initialised:
+            raise DataSourceError(
+                "The datasource provided has not initialised the BaseSource "
+                "parent class. Please make sure that you call "
+                "`super().__init__(**kwargs)` in your child method."
+            )
         if not self._frozen:
 
             if datasource.multi_table:
-                assert isinstance(datasource.data_ref, DatabaseConnection)  # nosec
-                assert datasource.data_ref.table_names is not None  # nosec
+                assert isinstance(datasource, MultiTableSource)  # nosec[assert_used]
+                assert datasource.table_names is not None  # nosec[assert_used]
 
-                table_names = datasource.data_ref.table_names
+                table_names = datasource.table_names
                 if table_name:
                     logger.warning(
                         "Ignoring table_name argument for multi-table datasource."
@@ -887,7 +940,7 @@ class BitfountSchema(_BitfountSchemaMarshmallowMixIn):
                     self.tables.append(TableSchema(t, table_description))
 
                 table_schema = self.get_table_schema(t)
-                table_schema.add_dataframe_features(
+                table_schema.add_datasource_features(
                     datasource=datasource,
                     descriptions=column_descriptions_,
                     ignore_cols=ignore_cols_,
@@ -964,8 +1017,8 @@ class BitfountSchema(_BitfountSchemaMarshmallowMixIn):
         return table_schema.get_categorical_feature_sizes(ignore_cols)
 
     def apply(
-        self, dataframe: _DataFrameType, keep_cols: Optional[List[str]] = None
-    ) -> _DataFrameType:
+        self, dataframe: pd.DataFrame, keep_cols: Optional[List[str]] = None
+    ) -> pd.DataFrame:
         """Applies the schema to a dataframe and returns the transformed dataframe.
 
         Sequentially adds missing columns to the dataframe, removes superfluous columns
