@@ -5,43 +5,25 @@ import asyncio
 from dataclasses import dataclass
 from datetime import datetime, timezone
 from enum import Enum
-from typing import (
-    TYPE_CHECKING,
-    Callable,
-    Final,
-    List,
-    Mapping,
-    Protocol,
-    Tuple,
-    Type,
-    TypeVar,
-    Union,
-)
+from typing import Callable, Final, List, Mapping, Protocol, Tuple, Type, TypeVar, Union
 
 from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey, RSAPublicKey
 import requests
 from requests import HTTPError, RequestException, Response
 from requests.exceptions import InvalidJSONError
 
-from bitfount.federated.algorithms.model_algorithms.base import (
-    _BaseModelAlgorithmFactory,
-)
 from bitfount.federated.encryption import _RSAEncryption
 from bitfount.federated.logging import _get_federated_logger
-from bitfount.federated.model_reference import BitfountModelReference
 from bitfount.federated.pod_response_message import _PodResponseMessage
 from bitfount.federated.task_requests import (
     _EncryptedTaskRequest,
-    _ProtocolDetails,
     _SignedEncryptedTaskRequest,
     _TaskRequest,
     _TaskRequestMessage,
 )
 from bitfount.federated.transport.worker_transport import _WorkerMailbox
 from bitfount.federated.types import (
-    AggregatorType,
-    AlgorithmType,
-    ProtocolType,
+    SerializedProtocol,
     _PodResponseType,
     _TaskRequestMessageGenerator,
 )
@@ -58,9 +40,7 @@ from bitfount.hub.types import (
     _PKCEAccessTokenRequestDict,
     _PKCEAccessTokenResponseJSON,
 )
-
-if TYPE_CHECKING:
-    from bitfount.federated.protocols.base import _BaseProtocolFactory
+from bitfount.utils import web_utils
 
 logger = _get_federated_logger(__name__)
 
@@ -72,8 +52,7 @@ requests.packages.urllib3.util.connection.HAS_IPV6 = False  # type: ignore[attr-
 class IdentityVerificationMethod(Enum):
     """Allowed types of identity verification."""
 
-    # TODO: [BIT-1291] Re-enable KEYS
-    # KEYS = "key-based"
+    KEYS = "key-based"
     LOCAL = "local"
     OIDC_ACF_PKCE = "oidc-auth-code"
     OIDC_DEVICE_CODE = "oidc-device-code"
@@ -108,12 +87,11 @@ _M_co = TypeVar("_M_co", covariant=True)
 class _AuthorisationChecker(Protocol[_M_co]):
     """Protocol for checking authorisation.
 
-    This is a simple protocol for checking authorisation.
-    It's used to inform a decision on whether a Pod accepts
-    or rejects a job.
+    This is a simple protocol for checking authorisation. It's used to inform a decision
+    on whether a Pod accepts or rejects a job.
     """
 
-    task_protocol_details: _ProtocolDetails
+    serialized_protocol: SerializedProtocol
 
     @classmethod
     def create_task_request_message_generator(
@@ -152,97 +130,19 @@ class _AuthorisationChecker(Protocol[_M_co]):
 
     async def check_authorisation(self) -> _PodResponseMessage:
         """Check if the task should be executed or not."""
-        ...  # Empty method body (explicit '...')
-
-    def verify_protocol(self, protocol: _BaseProtocolFactory) -> bool:
-        """Verify the protocol that was received is the same one that is authorised.
-
-        This includes the algorithm and aggregator components of the protocol. If there
-        is any kind of exception, the method defaults to returning False, i.e.
-        rejecting the protocol.
-
-        Args:
-            protocol (BaseProtocolFactory): the protocol actually received from the
-                Modeller
-
-        Returns:
-            bool: True if the protocol is the same as the authorised one
-        """
-        try:
-            if ProtocolType[protocol.name].value != self.task_protocol_details.protocol:
-                logger.error(
-                    f"Protocol {protocol.name} does not match "
-                    f"{self.task_protocol_details.protocol}"
-                )
-                return False
-
-            if (
-                AlgorithmType[protocol.algorithm.name].value
-                != self.task_protocol_details.algorithm
-            ):
-                logger.error(
-                    f"Algorithm {protocol.algorithm.name} does not match "
-                    f"{self.task_protocol_details.algorithm}"
-                )
-                return False
-
-            if (
-                hasattr(protocol, "aggregator")
-                # hasattr check above means we can ignore the attr-defined mypy error
-                and AggregatorType[protocol.aggregator.name].value != self.task_protocol_details.aggregator  # type: ignore[attr-defined] # Reason: hasattr check # noqa: B950
-            ):
-                logger.error(
-                    f"Aggregator {protocol.aggregator.name} does not match "  # type: ignore[attr-defined] # Reason: hasattr check # noqa: B950
-                    f"{self.task_protocol_details.aggregator}"
-                )
-                return False
-
-            if isinstance(protocol.algorithm, _BaseModelAlgorithmFactory):
-                # If the algorithm has a model, it is either a `BitfountModelReference`
-                # or a built-in model.
-                if isinstance(protocol.algorithm.model, BitfountModelReference):
-                    # If the model is a `BitfountModelReference`, we need to check the
-                    # the username and model_ref combo match what was authorised in the
-                    # task request.
-                    if (
-                        f"{protocol.algorithm.model.username}.{protocol.algorithm.model.model_ref}"  # noqa: B950
-                        != self.task_protocol_details.model
-                    ):
-                        logger.error(
-                            f"Model {protocol.algorithm.model.username}.{protocol.algorithm.model.model_ref} "  # noqa: B950
-                            f"does not match {self.task_protocol_details.model}"
-                        )
-                        return False
-                else:
-                    # If the model is a built-in model, it is prefixed with `bitfount.`
-                    # in the task request.
-                    if (
-                        f"bitfount.{protocol.algorithm.model.name}"
-                        != self.task_protocol_details.model
-                    ):
-                        logger.error(
-                            f"Model {protocol.algorithm.model.name} does not match "
-                            f"{self.task_protocol_details.model}"
-                        )
-                        return False
-
-        except Exception as e:
-            logger.exception(e)
-            return False
-
-        return True
+        ...
 
 
 def _make_encrypted_task_request_message(
     auth_type: IdentityVerificationMethod,
-    protocol_details: _ProtocolDetails,
+    serialized_protocol: SerializedProtocol,
     pod_identifiers: List[str],
     aes_key: bytes,
     pod_public_key: RSAPublicKey,
 ) -> bytes:
     """Common function for generating messages with _EncryptedTaskRequest contents."""
     # Create task request
-    tr = _TaskRequest(protocol_details, pod_identifiers, aes_key)
+    tr = _TaskRequest(serialized_protocol, pod_identifiers, aes_key)
 
     # Encrypt and pack in class
     packed_tr = _EncryptedTaskRequest(
@@ -251,7 +151,7 @@ def _make_encrypted_task_request_message(
 
     # Create outer message
     message = _TaskRequestMessage(
-        protocol_details,
+        serialized_protocol,
         auth_type.value,
         packed_tr.serialize(),
     )
@@ -314,19 +214,19 @@ class _LocalAuthorisation(_AuthorisationChecker[_EncryptedTaskRequest]):
     """
 
     pod_response_message: _PodResponseMessage
-    task_protocol_details: _ProtocolDetails
+    serialized_protocol: SerializedProtocol
 
     @classmethod
     def _generate_task_request_message(
         cls,
-        protocol_details: _ProtocolDetails,
+        serialized_protocol: SerializedProtocol,
         pod_identifiers: List[str],
         aes_key: bytes,
         pod_public_key: RSAPublicKey,
     ) -> bytes:
         return _make_encrypted_task_request_message(
             IdentityVerificationMethod.LOCAL,
-            protocol_details,
+            serialized_protocol,
             pod_identifiers,
             aes_key,
             pod_public_key,
@@ -363,17 +263,14 @@ class _SignatureBasedAuthorisation(_AuthorisationChecker[_SignedEncryptedTaskReq
 
     Checks whether the Modeller provided signature of the task
     can be verified against the modeller's public key.
-
-    Currently, this is not incorporated into the access manager.
-    The signature verification takes places in the pod.
-    In the future we will likely move this verification to the access manager,
-    so this will become a simpler HTTP call.
     """
 
     pod_response_message: _PodResponseMessage
+    access_manager: BitfountAM
+    modeller_name: str
     encrypted_task_request: bytes
     signature: bytes
-    task_protocol_details: _ProtocolDetails
+    serialized_protocol: SerializedProtocol
 
     @dataclass
     class _SignedMessageMaker:
@@ -381,20 +278,20 @@ class _SignatureBasedAuthorisation(_AuthorisationChecker[_SignedEncryptedTaskReq
 
         def generate_task_request_message(
             self,
-            protocol_details: _ProtocolDetails,
+            serialized_protocol: SerializedProtocol,
             pod_identifiers: List[str],
             aes_key: bytes,
             pod_public_key: RSAPublicKey,
         ) -> bytes:
-            tr = _TaskRequest(protocol_details, pod_identifiers, aes_key)
+            tr = _TaskRequest(serialized_protocol, pod_identifiers, aes_key)
 
             encrypted_tr = _RSAEncryption.encrypt(tr.serialize(), pod_public_key)
             signature = _RSAEncryption.sign_message(self.private_key, encrypted_tr)
             packed_tr = _SignedEncryptedTaskRequest(encrypted_tr, signature)
 
             message = _TaskRequestMessage(
-                protocol_details,
-                IdentityVerificationMethod.KEYS.value,  # type: ignore[attr-defined] # Reason: Disabled until [BIT-1291] resolved # noqa: B950
+                serialized_protocol,
+                IdentityVerificationMethod.KEYS.value,
                 packed_tr.serialize(),
             )
             return message.serialize()
@@ -438,9 +335,15 @@ class _SignatureBasedAuthorisation(_AuthorisationChecker[_SignedEncryptedTaskReq
 
     def _check_access(self) -> _PodResponseType:
         """Check that access is allowed for the authorised user and request."""
-        # TODO: [BIT-1291] Implement this method once signature-based access is
-        #       implemented in Access Manager.
-        raise NotImplementedError
+        return self.access_manager.check_signature_based_access_request(
+            unsigned_task=self.encrypted_task_request,
+            signed_task=_SignedEncryptedTaskRequest(
+                self.encrypted_task_request, self.signature
+            ).serialize(),
+            pod_identifier=self.pod_response_message.pod_identifier,
+            serialized_protocol=self.serialized_protocol,
+            modeller_name=self.modeller_name,
+        )
 
     async def check_authorisation(self) -> _PodResponseMessage:
         """Check signature matches modeller public key."""
@@ -462,19 +365,19 @@ class _SAMLAuthorisation(_AuthorisationChecker[_EncryptedTaskRequest]):
     pod_response_message: _PodResponseMessage
     access_manager: BitfountAM
     mailbox: _WorkerMailbox
-    task_protocol_details: _ProtocolDetails
+    serialized_protocol: SerializedProtocol
 
     @classmethod
     def _generate_task_request_message(
         cls,
-        protocol_details: _ProtocolDetails,
+        serialized_protocol: SerializedProtocol,
         pod_identifiers: List[str],
         aes_key: bytes,
         pod_public_key: RSAPublicKey,
     ) -> bytes:
         return _make_encrypted_task_request_message(
             IdentityVerificationMethod.SAML,
-            protocol_details,
+            serialized_protocol,
             pod_identifiers,
             aes_key,
             pod_public_key,
@@ -532,7 +435,7 @@ class _SAMLAuthorisation(_AuthorisationChecker[_EncryptedTaskRequest]):
                 saml_request_id,
                 self.pod_response_message.pod_identifier,
                 self.mailbox.modeller_name,
-                self.task_protocol_details,
+                self.serialized_protocol,
             )
         )
         logger.info(
@@ -553,7 +456,7 @@ class _OIDCAuthorisationCode(_AuthorisationChecker[_EncryptedTaskRequest]):
     pod_response_message: _PodResponseMessage
     access_manager: BitfountAM
     mailbox: _WorkerMailbox
-    task_protocol_details: _ProtocolDetails
+    serialized_protocol: SerializedProtocol
 
     # Auth-related params
     _auth_domain: str = None  # type: ignore[assignment] # Reason: filled in __post_init__ if not provided # noqa: B950
@@ -572,14 +475,14 @@ class _OIDCAuthorisationCode(_AuthorisationChecker[_EncryptedTaskRequest]):
     @classmethod
     def _generate_task_request_message(
         cls,
-        protocol_details: _ProtocolDetails,
+        serialized_protocol: SerializedProtocol,
         pod_identifiers: List[str],
         aes_key: bytes,
         pod_public_key: RSAPublicKey,
     ) -> bytes:
         return _make_encrypted_task_request_message(
             IdentityVerificationMethod.OIDC_ACF_PKCE,
-            protocol_details,
+            serialized_protocol,
             pod_identifiers,
             aes_key,
             pod_public_key,
@@ -660,7 +563,7 @@ class _OIDCAuthorisationCode(_AuthorisationChecker[_EncryptedTaskRequest]):
             "code_verifier": code_verifier,
             "redirect_uri": redirect_uri,
         }
-        response: Response = requests.post(
+        response: Response = web_utils.post(
             self._token_endpoint,
             data=request_data,
             timeout=20,
@@ -675,7 +578,7 @@ class _OIDCAuthorisationCode(_AuthorisationChecker[_EncryptedTaskRequest]):
         """Check that access is allowed for the authorised user and request."""
         return self.access_manager.check_oidc_access_request(
             self.pod_response_message.pod_identifier,
-            self.task_protocol_details,
+            self.serialized_protocol,
             self.mailbox.modeller_name,
             modeller_access_token,
         )
@@ -691,7 +594,7 @@ class _OIDCDeviceCode(_AuthorisationChecker[_EncryptedTaskRequest]):
     pod_response_message: _PodResponseMessage
     access_manager: BitfountAM
     mailbox: _WorkerMailbox
-    task_protocol_details: _ProtocolDetails
+    serialized_protocol: SerializedProtocol
 
     # Auth-related params
     _auth_domain: str = None  # type: ignore[assignment] # Reason: filled in __post_init__ if not provided # noqa: B950
@@ -710,14 +613,14 @@ class _OIDCDeviceCode(_AuthorisationChecker[_EncryptedTaskRequest]):
     @classmethod
     def _generate_task_request_message(
         cls,
-        protocol_details: _ProtocolDetails,
+        serialized_protocol: SerializedProtocol,
         pod_identifiers: List[str],
         aes_key: bytes,
         pod_public_key: RSAPublicKey,
     ) -> bytes:
         return _make_encrypted_task_request_message(
             IdentityVerificationMethod.OIDC_DEVICE_CODE,
-            protocol_details,
+            serialized_protocol,
             pod_identifiers,
             aes_key,
             pod_public_key,
@@ -798,7 +701,7 @@ class _OIDCDeviceCode(_AuthorisationChecker[_EncryptedTaskRequest]):
                 "device_code": device_code,
                 "grant_type": "urn:ietf:params:oauth:grant-type:device_code",
             }
-            response = requests.post(
+            response = web_utils.post(
                 self._token_endpoint,
                 data=request_data,
                 timeout=10,
@@ -870,7 +773,7 @@ class _OIDCDeviceCode(_AuthorisationChecker[_EncryptedTaskRequest]):
         """Check that access is allowed for the authorised user and request."""
         return self.access_manager.check_oidc_access_request(
             self.pod_response_message.pod_identifier,
-            self.task_protocol_details,
+            self.serialized_protocol,
             self.mailbox.modeller_name,
             modeller_access_token,
         )
@@ -879,8 +782,7 @@ class _OIDCDeviceCode(_AuthorisationChecker[_EncryptedTaskRequest]):
 _IDENTITY_VERIFICATION_METHODS_MAP: Final[
     Mapping[IdentityVerificationMethod, Type[_AuthorisationChecker]]
 ] = {
-    # TODO: [BIT-1291] Re-enable KEYS
-    # IdentityVerificationMethod.KEYS: _SignatureBasedAuthorisation,
+    IdentityVerificationMethod.KEYS: _SignatureBasedAuthorisation,
     IdentityVerificationMethod.LOCAL: _LocalAuthorisation,
     IdentityVerificationMethod.OIDC_ACF_PKCE: _OIDCAuthorisationCode,
     IdentityVerificationMethod.OIDC_DEVICE_CODE: _OIDCDeviceCode,
