@@ -2,7 +2,6 @@
 from __future__ import annotations
 
 import asyncio
-from functools import partial
 import os
 from pathlib import Path
 from typing import (
@@ -13,20 +12,25 @@ from typing import (
     Literal,
     Mapping,
     Optional,
+    Type,
     Union,
     cast,
 )
 
 from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey, RSAPublicKey
 
+from bitfount.federated.aggregators.secure import SecureAggregator
 from bitfount.federated.algorithms.model_algorithms.base import (
     _BaseModelAlgorithmFactory,
 )
 from bitfount.federated.authorisation_checkers import (
     _IDENTITY_VERIFICATION_METHODS_MAP,
     IdentityVerificationMethod,
+    _SignatureBasedAuthorisation,
     check_identity_verification_method,
 )
+from bitfount.federated.encryption import _RSAEncryption
+from bitfount.federated.exceptions import PodResponseError
 from bitfount.federated.helper import (
     _check_and_update_pod_ids,
     _create_message_service,
@@ -34,7 +38,7 @@ from bitfount.federated.helper import (
 )
 from bitfount.federated.logging import _get_federated_logger
 from bitfount.federated.model_reference import BitfountModelReference
-from bitfount.federated.task_requests import _ProtocolDetails
+from bitfount.federated.monitoring import task_monitor_context
 from bitfount.federated.transport.base_transport import _run_func_and_listen_to_mailbox
 from bitfount.federated.transport.identity_verification.oidc import (
     _OIDCAuthFlowChallengeHandler,
@@ -52,14 +56,10 @@ from bitfount.federated.transport.message_service import (
     _MessageService,
 )
 from bitfount.federated.transport.modeller_transport import _ModellerMailbox
-from bitfount.federated.types import (
-    AggregatorType,
-    AlgorithmType,
-    ProtocolType,
-    _TaskRequestMessageGenerator,
-)
+from bitfount.federated.types import _TaskRequestMessageGenerator
 from bitfount.hub.api import BitfountHub
 from bitfount.hub.authentication_flow import _get_auth_environment
+from bitfount.hub.exceptions import PodDoesNotExistError
 from bitfount.hub.helper import _default_bitfounthub, _get_pod_public_keys
 from bitfount.types import DistributedModelProtocol
 
@@ -70,7 +70,7 @@ if TYPE_CHECKING:
 logger = _get_federated_logger(__name__)
 
 
-class Modeller:
+class _Modeller:
     """Dispatches tasks to pods and runs the modeller side of the provided protocol.
 
     ```python title="Example usage:"
@@ -91,8 +91,6 @@ class Modeller:
         pod_public_key_paths: Optional. Mapping of pod identifiers to public
             key files for existing pod public keys. Expired or non-existent
             keys will be downloaded from the Hub.
-        pretrained_file: Path to a file containing pretrained model parameters.
-            Defaults to None.
         identity_verification_method: The identity verification method to use.
             Defaults to OIDC_DEVICE_CODE.
         private_key: This modeller's private key either as an `RSAPrivateKey`
@@ -115,7 +113,6 @@ class Modeller:
         message_service: Optional[_MessageService] = None,
         bitfounthub: Optional[BitfountHub] = None,
         pod_public_key_paths: Optional[Mapping[str, Path]] = None,
-        pretrained_file: Optional[Union[str, os.PathLike]] = None,
         identity_verification_method: Union[
             str, IdentityVerificationMethod
         ] = IdentityVerificationMethod.DEFAULT,
@@ -131,41 +128,15 @@ class Modeller:
         )
         self._pod_public_key_paths: Mapping[str, Path] = pod_public_key_paths or dict()
 
-        # TODO: [BIT-1252] pretrained file should not be passed here
-        self._pretrained_file = pretrained_file
-
         self._identity_verification_method: IdentityVerificationMethod = (
             check_identity_verification_method(identity_verification_method)
         )
 
         # Load private key if needed
-        self._private_key: Optional[RSAPrivateKey] = None
-        # TODO: [BIT-1291] Uncomment when key-based validation enabled again
-        # if self._identity_verification_method == IdentityVerificationMethod.KEYS:
-        #     if not private_key:
-        #         raise ValueError(
-        #             "Key-based identity verification selected but "
-        #             "no private key provided."
-        #         )
-        #     elif isinstance(private_key, Path):
-        #         self._private_key = _RSAEncryption.load_private_key(private_key)
-        #     else:
-        #         self._private_key = private_key
-        # else:
-        #     if private_key:
-        #         logger.warning(
-        #             f"Private key provided but identity verification method "
-        #             f'"{self._identity_verification_method.value}" was chosen. '
-        #             f"Private key will be ignored."
-        #         )
+        if isinstance(private_key, Path):
+            private_key = _RSAEncryption.load_private_key(private_key)
+        self._private_key = private_key
 
-        # TODO: [BIT-1291] Remove this when key-based validation enabled again
-        if private_key:
-            logger.warning(
-                f"Private key provided but identity verification method "
-                f'"{self._identity_verification_method.value}" was chosen. '
-                f"Private key will be ignored."
-            )
         self._idp_url: Optional[str]
         if self._identity_verification_method == IdentityVerificationMethod.SAML:
             # We also need to set the idp_url based on the environment if using SAML
@@ -173,6 +144,31 @@ class Modeller:
         else:
             self._idp_url = None
 
+    @property
+    def _private_key(self) -> Optional[RSAPrivateKey]:
+        """Private key attribute."""
+        return self.__private_key
+
+    @_private_key.setter
+    def _private_key(self, private_key: Optional[RSAPrivateKey]) -> None:
+        """Set _private_key attribute."""
+        self.__private_key: Optional[RSAPrivateKey] = None
+
+        if self._identity_verification_method == IdentityVerificationMethod.KEYS:
+            if not private_key:
+                raise ValueError(
+                    "Key-based identity verification selected but "
+                    "no private key provided."
+                )
+            self.__private_key = private_key
+        else:
+            if private_key:
+                logger.warning(
+                    f"Private key provided but identity verification method "
+                    f'"{self._identity_verification_method.value}" was chosen. '
+                    f"Private key will be ignored."
+                )
+
     async def _send_task_requests(
         self, pod_identifiers: Iterable[str]
     ) -> _ModellerMailbox:
@@ -191,51 +187,13 @@ class Modeller:
             pod_public_key_paths=self._pod_public_key_paths,
         )
 
-        # The bare minimum task request has as just a protocol and an algorithm.
-        partial_protocol_details = partial(
-            _ProtocolDetails,
-            protocol=ProtocolType[self.protocol.name].value,
-            algorithm=AlgorithmType[self.protocol.algorithm.name].value,
-        )
-
-        if isinstance(self.protocol.algorithm, _BaseModelAlgorithmFactory):
-            # If the algorithm is an instance of `_BaseModelAlgorithmFactory`, this
-            # means it has a model. This model is either a `BitfountModelReference`
-            # or a built-in model.
-            if isinstance(self.protocol.algorithm.model, BitfountModelReference):
-                # If the model is a `BitfountModelReference`, we send the `model_ref`
-                # prefixed with the `username` of the model.
-                model = self.protocol.algorithm.model
-                model_name = f"{model.username}.{model.model_ref}"
-            else:
-                # If the model is a built-in model, we send the model name prefixed
-                # with `bitfount.`
-                model_name = f"bitfount.{self.protocol.algorithm.model.name}"
-
-            # Currently there can only be an aggregator if the algorithm has a model so
-            # we can safely add the aggregator to the protocol details here.
-            if hasattr(self.protocol, "aggregator"):
-                protocol_details = partial_protocol_details(
-                    model=model_name,
-                    aggregator=AggregatorType[self.protocol.aggregator.name].value,  # type: ignore[attr-defined] # Reason: hasattr check # noqa: B950
-                )
-            # If there is no aggregator, we can just add the model to the protocol
-            # details.
-            else:
-                protocol_details = partial_protocol_details(model=model_name)
-        else:
-            # If the algorithm is not an instance of `_BaseModelAlgorithmFactory`, this
-            # means it has no model (and therefore no aggregator either).
-            protocol_details = partial_protocol_details()
-
         task_request_msg_gen = self._get_task_request_msg_gen()
-
         # Send task requests to chosen pods. We haven't attached a handler for the
         # responses yet but this is handled below. There is no issue due to the
         # fact that the listening task hasn't been started yet, and the inherent
         # handler backoff.
         modeller_mailbox: _ModellerMailbox = await _ModellerMailbox.send_task_requests(
-            protocol_details=protocol_details,
+            serialized_protocol=self.protocol.dump(),
             pod_public_keys=loaded_pod_public_keys,
             task_request_msg_gen=task_request_msg_gen,
             message_service=self._message_service,
@@ -249,28 +207,22 @@ class Modeller:
             self._identity_verification_method
         ]
 
-        # TODO: [BIT-1291] Uncomment when key-based validation enabled again
-        # if self._identity_verification_method == IdentityVerificationMethod.KEYS:
-        #     authorization_checker_cls = cast(
-        #         Type[_SignatureBasedAuthorisation], authorization_checker_cls
-        #     )
-        #     if not self._private_key:
-        #         raise ValueError(
-        #             "Signature-based identification selected but no "
-        #             "private key provided."
-        #         )
-        #     return authorization_checker_cls.create_task_request_message_generator(
-        #         self._private_key
-        #     )
-        # else:
-        #     return authorization_checker_cls.create_task_request_message_generator()
-
-        # TODO: [BIT-1291] Remove this when key-based validation enabled again
-        return authorization_checker_cls.create_task_request_message_generator()
+        if self._identity_verification_method == IdentityVerificationMethod.KEYS:
+            authorization_checker_cls = cast(
+                Type[_SignatureBasedAuthorisation], authorization_checker_cls
+            )
+            assert isinstance(self._private_key, RSAPrivateKey)  # nosec[assert_used]
+            return authorization_checker_cls.create_task_request_message_generator(
+                self._private_key
+            )
+        else:
+            return authorization_checker_cls.create_task_request_message_generator()
 
     async def _modeller_run(
         self,
         modeller_mailbox: _ModellerMailbox,
+        pod_identifiers: Iterable[str],
+        require_all_pods: bool = False,
         response_handler: Optional[_ResponseHandler] = None,
     ) -> Union[Literal[False], Optional[Any]]:
         """Waits for pod responses and handles any who don't respond in time.
@@ -284,7 +236,7 @@ class Modeller:
         # Process job accept/reject
         await modeller_mailbox.process_task_request_responses()
 
-        # Fail-fast if no pod accepted.
+        # Fail-fast if no pods accepted the task request.
         if not modeller_mailbox.accepted_worker_mailboxes:
             logger.error(
                 "No workers with which to train. "
@@ -293,20 +245,43 @@ class Modeller:
             )
             return False
 
+        # Otherwise run the protocol
+        unaccepted_worker_ids = set(pod_identifiers).difference(
+            set(modeller_mailbox.accepted_worker_mailboxes)
+        )
+        if unaccepted_worker_ids:
+            unaccepted_worker_msg = (
+                f"Pods {', '.join(unaccepted_worker_ids)} "
+                "rejected task request or failed to respond. "
+            )
+            if require_all_pods:
+                raise PodResponseError(
+                    unaccepted_worker_msg
+                    + "Task requires all pods accept the task request."
+                )
+            else:
+                logger.warning(
+                    unaccepted_worker_msg + "Continuing task without these pods ..."
+                )
+
+        await modeller_mailbox.send_task_start_message()
+
         return await self._run_modeller_protocol(modeller_mailbox)
 
     async def _run_modeller_protocol(
-        self, modeller_mailbox: _ModellerMailbox
+        self,
+        modeller_mailbox: _ModellerMailbox,
     ) -> Optional[Any]:
         """Runs the modeller-side of the protocol."""
-        # Run the task
-        await modeller_mailbox.send_task_details(self.protocol.dump())
         protocol = self.protocol.modeller(mailbox=modeller_mailbox)
-        result = await protocol.run(pretrained_file=self._pretrained_file)
+        # Run the task
+        result = await protocol.run()
         return result
 
     async def run_async(
-        self, pod_identifiers: Iterable[str]
+        self,
+        pod_identifiers: Iterable[str],
+        require_all_pods: bool,
     ) -> Union[Literal[False], Optional[Any]]:
         """Runs the modeller's task with a set of pods.
 
@@ -318,6 +293,10 @@ class Modeller:
 
         Returns:
             Whatever the protocol's return value is.
+
+        Raises:
+            PodResponseError: If require_all_pods is True and at least one
+                pod_identifier rejects or fails to respond to the task request.
         """
         try:
             # We do this early to give any handler components time to spin up
@@ -328,18 +307,30 @@ class Modeller:
                 pod_identifiers
             )
 
-            run_result: Union[
-                Literal[False], Optional[Any]
-            ] = await _run_func_and_listen_to_mailbox(
-                self._modeller_run(
-                    modeller_mailbox=modeller_mailbox, response_handler=response_handler
-                ),
-                modeller_mailbox,
-            )
-            modeller_mailbox.delete_handler(_BitfountMessageType.LOG_MESSAGE)
+            with task_monitor_context(
+                hub=self._hub,
+                task_id=modeller_mailbox.task_id,
+                sender_id=modeller_mailbox.mailbox_id,
+            ):
+                run_result: Union[
+                    Literal[False], Optional[Any]
+                ] = await _run_func_and_listen_to_mailbox(
+                    self._modeller_run(
+                        modeller_mailbox=modeller_mailbox,
+                        pod_identifiers=pod_identifiers,
+                        require_all_pods=require_all_pods,
+                        response_handler=response_handler,
+                    ),
+                    modeller_mailbox,
+                )
+                modeller_mailbox.delete_all_handlers(_BitfountMessageType.LOG_MESSAGE)
 
-            logger.info("Task complete")
-            return run_result
+                logger.info("Task complete")
+                return run_result
+        except PodDoesNotExistError as e:
+            logger.exception(e)
+            logger.error("Aborted task request.")
+            return False
         finally:
             # Stop any response handler web servers that may be running still
             try:
@@ -396,7 +387,7 @@ class Modeller:
 
         return response_handler
 
-    def _serialize(self, model_out: Path) -> None:
+    def _serialize(self, model_out: Union[str, os.PathLike]) -> None:
         """Save model to file.
 
         Save the model to the file specified.
@@ -404,10 +395,10 @@ class Modeller:
         Args:
             model_out: The path to save the model out to.
         """
-        if self.protocol.name != "ResultsOnly":
+        if self.protocol.class_name != "ResultsOnly":
             # If protocol is not ResultsOnly, then the algorithm has a model
             # Assertion to reassure mypy of this fact
-            assert isinstance(  # nosec
+            assert isinstance(  # nosec[assert_used]
                 self.protocol.algorithm, _BaseModelAlgorithmFactory
             )
             self.protocol.algorithm.model = cast(
@@ -421,7 +412,10 @@ class Modeller:
             )
 
     def run(
-        self, pod_identifiers: Iterable[str], model_out: Optional[Path] = None
+        self,
+        pod_identifiers: Iterable[str],
+        require_all_pods: bool = False,
+        model_out: Optional[Union[Path, str]] = None,
     ) -> Union[Literal[False], Optional[Any]]:
         """Runs the modeller's task with a set of pods.
 
@@ -434,9 +428,49 @@ class Modeller:
 
         Returns:
             Whatever the protocol's return value is.
+
+        Raises:
+            PodResponseError: If require_all_pods is True and at least one
+                pod_identifier rejects or fails to respond to the task request.
         """
         pod_identifiers = _check_and_update_pod_ids(pod_identifiers, self._hub)
-        result = asyncio.run(self.run_async(pod_identifiers))
-        if model_out:
+
+        # We need to pass the `SecureShare` parameters to the
+        # model when SecureAggregation is in use.
+        # First we check the protocol's model and aggregators.
+        # if steps_between_parameter_updates is 1, then we do
+        # the clipping at the 'SecureShare' level.
+        if (
+            (
+                hasattr(self.protocol, "aggregator")
+                and isinstance(self.protocol.aggregator, SecureAggregator)  # type: ignore[attr-defined] # Reason: see below # noqa: B950
+            )
+            and (
+                hasattr(self.protocol, "steps_between_parameter_updates")
+                and self.protocol.steps_between_parameter_updates != 1  # type: ignore[attr-defined] # Reason: see below # noqa: B950
+            )
+            and hasattr(self.protocol.algorithm, "model")
+        ):
+            logger.warning(
+                "SecureAggregation in use. "
+                "We recommend normalization of continuous features prior to training."
+            )
+            # Show warning to user if using custom models that
+            # they might have to implement clipping in their custom model.
+
+            if isinstance(self.protocol.algorithm.model, BitfountModelReference):  # type: ignore[attr-defined] # Reason: see below # noqa: B950
+                logger.warning(
+                    "You are using a custom model with Secure Aggregation."
+                    "We recommend clipping the model parameters."
+                )
+            # Mypy: hasattr check on the protocol aggregator, model and steps.
+            # Pass the `SecureShare` parameters to the model for clipping
+            self.protocol.algorithm.model.param_clipping = {  # type: ignore[attr-defined] # Reason: see above # noqa: B950
+                "prime_q": self.protocol.aggregator._secure_share.prime_q,  # type: ignore[attr-defined] # Reason: see above # noqa: B950
+                "precision": self.protocol.aggregator._secure_share.precision,  # type: ignore[attr-defined] # Reason: see above # noqa: B950
+                "num_workers": len(pod_identifiers),
+            }
+        result = asyncio.run(self.run_async(pod_identifiers, require_all_pods))
+        if model_out and result is not False:
             self._serialize(model_out)
         return result
