@@ -1,10 +1,16 @@
 """Pods for responding to tasks."""
 import asyncio
+from contextlib import nullcontext
+from dataclasses import asdict
 import os
+from pathlib import Path
 import threading
 from typing import (
+    Any,
     Callable,
+    ContextManager,
     Coroutine,
+    Dict,
     Iterable,
     List,
     MutableSequence,
@@ -15,13 +21,13 @@ from typing import (
 )
 
 from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey, RSAPublicKey
-from pydantic import AnyUrl
 from requests import HTTPError, RequestException
+import yaml
 
 from bitfount.config import BITFOUNT_STORAGE_PATH
-from bitfount.data.datasource import DataSource
+from bitfount.data.datasources.base_source import BaseSource
+from bitfount.data.exceptions import DataSourceError
 from bitfount.data.schema import BitfountSchema
-from bitfount.data.utils import DatabaseConnection
 from bitfount.federated.aggregators.secure import _is_secure_share_task_request
 from bitfount.federated.authorisation_checkers import (
     _IDENTITY_VERIFICATION_METHODS_MAP,
@@ -30,6 +36,7 @@ from bitfount.federated.authorisation_checkers import (
     _OIDCAuthorisationCode,
     _OIDCDeviceCode,
     _SAMLAuthorisation,
+    _SignatureBasedAuthorisation,
     check_identity_verification_method,
 )
 from bitfount.federated.exceptions import PodNameError, PodRegistrationError
@@ -38,11 +45,15 @@ from bitfount.federated.helper import (
     _create_and_connect_pod_mailbox,
 )
 from bitfount.federated.logging import _get_federated_logger
+from bitfount.federated.monitoring import task_monitor_context
 from bitfount.federated.pod_keys_setup import PodKeys, _get_pod_keys
 from bitfount.federated.pod_response_message import _PodResponseMessage
 from bitfount.federated.pod_vitals import _PodVitals, _PodVitalsHandler
 from bitfount.federated.privacy.differential import DPPodConfig
-from bitfount.federated.task_requests import _ProtocolDetails, _TaskRequestMessage
+from bitfount.federated.task_requests import (
+    _SignedEncryptedTaskRequest,
+    _TaskRequestMessage,
+)
 from bitfount.federated.transport.base_transport import _run_func_and_listen_to_mailbox
 from bitfount.federated.transport.config import MessageServiceConfig
 from bitfount.federated.transport.message_service import (
@@ -54,13 +65,16 @@ from bitfount.federated.transport.worker_transport import (
     _InterPodWorkerMailbox,
     _WorkerMailbox,
 )
-from bitfount.federated.types import AggregatorType, _PodResponseType
+from bitfount.federated.types import (
+    AggregatorType,
+    SerializedProtocol,
+    _PodResponseType,
+)
+from bitfount.federated.utils import _keys_exist
 from bitfount.federated.worker import _Worker
 from bitfount.hub.api import BitfountAM, BitfountHub, PodPublicMetadata
-from bitfount.hub.authentication_flow import (
-    _DEFAULT_USER_DIRECTORY,
-    _get_auth_environment,
-)
+from bitfount.hub.authentication_flow import _DEFAULT_USERNAME, _get_auth_environment
+from bitfount.hub.exceptions import SchemaUploadError
 from bitfount.hub.helper import (
     _create_access_manager,
     _create_bitfounthub,
@@ -68,6 +82,8 @@ from bitfount.hub.helper import (
 )
 from bitfount.runners.config_schemas import (
     POD_NAME_REGEX,
+    AccessManagerConfig,
+    PodConfig,
     PodDataConfig,
     PodDetailsConfig,
 )
@@ -76,12 +92,35 @@ from bitfount.transformations.dataset_operations import (
     NormalizeDataTransformation,
 )
 from bitfount.transformations.processor import TransformationProcessor
-from bitfount.types import _DataFrameType
-from bitfount.utils import _is_notebook
+from bitfount.utils import _handle_fatal_error, _is_notebook
 
 logger = _get_federated_logger(__name__)
 
 
+class _StoppableThead(threading.Thread):
+    """Stoppable thread by using a stop `threading.Event`.
+
+    Args:
+        stop_event: This is a `threading.Event` which when set, should stop the thread.
+            The function that is being executed in the thread is required to regularly
+            check the status of this event.
+        **kwargs: Keyword arguments passed to parent constructor.
+    """
+
+    def __init__(self, stop_event: threading.Event, **kwargs: Any):
+        self._stop_event = stop_event
+        super().__init__(**kwargs)
+
+    @property
+    def stopped(self) -> bool:
+        """Returns whether or not the stop event has been set."""
+        return self._stop_event.is_set()
+
+    def stop(self) -> None:
+        """Sets the stop event."""
+        self._stop_event.set()
+
+
 class Pod:
     """Makes data and computation available remotely and responds to tasks.
 
@@ -100,11 +139,17 @@ class Pod:
     pod.start()
     ```
 
+    :::tip
+
+    Once you start a `Pod`, you can just leave it running in the background. It will
+    automatically respond to any tasks without any intervention required.
+
+    :::
+
     Args:
         name: Name of the pod. This will appear on `Bitfount Hub` and `Bitfount AM`.
-            This is also used for the name of the table in a single-table `DataSource`.
-        data: Data to be made available to the pod. This can be a Path, URL, Database
-            connection or a DataFrame.
+            This is also used for the name of the table in a single-table `BaseSource`.
+        datasource: A concrete instance of the `BaseSource` object.
         username: Username of the user who is registering the pod. Defaults to None.
         data_config: Configuration for the data. Defaults to None.
         schema: Schema for the data. This can be a `BitfountSchema` object or a Path
@@ -119,6 +164,8 @@ class Pod:
             to share a training task with. Required if the protocol uses the
             `SecureAggregator` aggregator.
         pod_dp: Differential privacy configuration for the pod. Defaults to None.
+        update_schema: Whether the schema needs to be re-generated even if provided.
+            Defaults to False.
 
     Attributes:
         data: Data that the pod encapsulates.
@@ -131,46 +178,59 @@ class Pod:
 
     Raises:
         PodRegistrationError: If the pod could not be registered for any reason.
-
-    :::tip
-
-    Once you start a `Pod`, you can just leave it running in the background. It will
-    automatically respond to any tasks without any intervention required.
-
-    :::
+        DataSourceError: If the `BaseSource` for the provided datasource has
+            not been initialised properly. This can be done by calling
+            `super().__init__(**kwargs)` in the `__init__` of the DataSource.
     """
 
     def __init__(
         self,
         name: str,
-        data: Union[
-            os.PathLike,
-            AnyUrl,
-            DatabaseConnection,
-            _DataFrameType,
-        ],
+        datasource: BaseSource,
         username: Optional[str] = None,
         data_config: Optional[PodDataConfig] = None,
         schema: Optional[Union[str, os.PathLike, BitfountSchema]] = None,
         pod_details_config: Optional[PodDetailsConfig] = None,
-        bitfounthub: Optional[BitfountHub] = None,
-        ms_config: Optional[MessageServiceConfig] = None,
+        hub: Optional[BitfountHub] = None,
+        message_service: Optional[MessageServiceConfig] = None,
         access_manager: Optional[BitfountAM] = None,
         pod_keys: Optional[PodKeys] = None,
         approved_pods: Optional[List[str]] = None,
-        pod_dp: Optional[DPPodConfig] = None,
+        differential_privacy: Optional[DPPodConfig] = None,
+        update_schema: bool = False,
     ):
         self.name = name
+
         if not data_config:
             data_config = PodDataConfig()
-        # Create datasource
-        self.datasource = DataSource(
-            data_ref=data,
-            koalas=data_config.koalas,
-            ignore_cols=data_config.ignore_cols,
-            modifiers=data_config.modifiers,
-            **data_config.datasource_args,
+
+        pod_details_config = (
+            pod_details_config
+            if pod_details_config is not None
+            else self._get_default_pod_details_config()
         )
+        self._pod_config = PodConfig(
+            datasource=datasource.__class__.__name__,
+            data_config=data_config,
+            pod_details_config=pod_details_config,
+            name=name,
+            schema=schema if isinstance(schema, Path) else None,
+            access_manager=AccessManagerConfig(),
+            message_service=message_service
+            if message_service
+            else MessageServiceConfig(),
+            approved_pods=approved_pods,
+            update_schema=update_schema,
+        )
+        # Fast-fail if BaseSource has not been initialised
+        if not datasource.is_initialised:
+            raise DataSourceError(
+                "The datasource provided has not initialised the BaseSource "
+                "parent class. Please make sure that you call "
+                "`super().__init__(**kwargs)` in your child method."
+            )
+        # Create datasource
+        self.datasource = datasource
 
         # Load schema if necessary
         if schema and not isinstance(schema, BitfountSchema):
@@ -178,22 +238,43 @@ class Pod:
 
         # If the schema was a path, it should be loaded by here, so we can cast:
         schema = cast(Optional[BitfountSchema], schema)
-        self.schema = self._setup_schema(schema=schema, data_config=data_config)
 
-        pod_details_config = (
-            pod_details_config
-            if pod_details_config is not None
-            else self._get_default_pod_details_config()
-        )
+        if self.datasource.multi_table and schema is not None:
+            # check if all table names are in the schema
+            schema_tables = [table.name for table in schema.tables]
+            if any(table not in schema_tables for table in self.datasource.table_names):  # type: ignore[attr-defined] # reason: see below: # noqa: B950
+                # mypy ignore - we only check the table names
+                # for multi-table the datasources
+                logger.info(
+                    "Datasource has additional tables to the "
+                    "schema provided, re-generating schema."
+                )
+                update_schema = True
+        elif data_config.auto_tidy is True and schema is not None:
+            # Auto-tidy is not applied to multi-table datasources,
+            # so we only cover the single table case.
+            logger.info("Auto-tidying the datasource, schema will be re-generated.")
+            update_schema = True
+        elif schema is not None and self.name not in [
+            table.name for table in schema.tables
+        ]:
+            logger.info(
+                "Provided schema table name does not "
+                "match to pod name, re-generating schema."
+            )
+            update_schema = True
+
+        if update_schema is True or schema is None:
+            self.schema = self._setup_schema(schema=schema, data_config=data_config)
+        else:
+            logger.info("Using user provided schema.")
+            self.schema = schema
+
         self.public_metadata = self._get_public_metadata(
             pod_details_config, self.schema
         )
 
-        self._hub = (
-            bitfounthub
-            if bitfounthub is not None
-            else _create_bitfounthub(username=username)
-        )
+        self._hub = hub if hub is not None else _create_bitfounthub(username=username)
         self._session = self._hub.session
 
         self._access_manager = (
@@ -208,17 +289,22 @@ class Pod:
         self.pod_identifier = f"{self._session.username}/{self.name}"
         if approved_pods is None:
             approved_pods = []
-        _check_and_update_pod_ids([*approved_pods, self.pod_identifier], self._hub)
+        approved_pods = _check_and_update_pod_ids(
+            [*approved_pods, self.pod_identifier], self._hub
+        )
         self.approved_pods = approved_pods
-
-        self._pod_dp = pod_dp
+        self._pod_dp = differential_privacy
         self._pod_vitals = _PodVitals()
 
         # Connecting the pod to the message service must happen AFTER registering
         # it on the hub as the message service uses hub information to verify that
         # the relevant message queue is available.
-        self._register_pod()
-        self._ms_config: Optional[MessageServiceConfig] = ms_config
+        try:
+            self._register_pod()
+        except PodRegistrationError as pre:
+            _handle_fatal_error(pre, logger=logger)
+
+        self._ms_config: Optional[MessageServiceConfig] = message_service
         self._mailbox: Optional[_PodMailbox] = None
 
         # Marker for when initialization is complete
@@ -240,6 +326,19 @@ class Pod:
                 f"Pod names must match: {POD_NAME_REGEX.pattern}"
             )
 
+    def to_yaml(self, output_file: Optional[str] = None) -> Optional[Dict]:
+        """Convert the pod specification to yaml.
+
+        If output_file argument is given, write yaml to that locaton.
+        """
+        output = asdict(self._pod_config)
+        if output_file:
+            with open(output_file, "w") as file:
+                yaml.dump(output, file)
+            return None
+        else:
+            return output
+
     def _setup_schema(
         self,
         data_config: PodDataConfig,
@@ -252,10 +351,11 @@ class Pod:
         if not schema:
             schema = BitfountSchema()
 
-        # Add DataSource to schema
+        # Add BaseSource to schema
         schema.add_datasource_tables(
             datasource=self.datasource,
             table_name=self.name,
+            ignore_cols=data_config.ignore_cols,
             force_stypes=data_config.force_stypes,
         )
 
@@ -264,19 +364,20 @@ class Pod:
             logger.warning("Can't autotidy multi-table data.")
         elif data_config.auto_tidy:
             clean_data = CleanDataTransformation()
-            # TODO: [BIT-962] data normalizing should not be done here as this
-            # converts all integer columns to floats
+            # Normalization is applied to all float columns if `auto-tidy` is true
             normalize = NormalizeDataTransformation()
             processor = TransformationProcessor(
-                [clean_data, normalize], schema.get_table_schema(self.name)
+                [clean_data, normalize],
+                schema.get_table_schema(self.name),
             )
             self.datasource.data = processor.transform(self.datasource.data)
 
-            # Add DataSource to schema again because features will have changed by
+            # Add BaseSource to schema again because features will have changed by
             # auto-tidying
             schema.add_datasource_tables(
                 datasource=self.datasource,
                 table_name=self.name,
+                ignore_cols=data_config.ignore_cols,
                 force_stypes=data_config.force_stypes,
             )
 
@@ -285,24 +386,26 @@ class Pod:
         return schema
 
     def _get_default_pod_details_config(self) -> PodDetailsConfig:
+        """Get default pod details config."""
         return PodDetailsConfig(display_name=self.name, description=self.name)
 
     def _get_public_metadata(
         self, pod_details_config: PodDetailsConfig, schema: BitfountSchema
     ) -> PodPublicMetadata:
+        """Get PodPublicMetadata."""
         return PodPublicMetadata(
             self.name,
             pod_details_config.display_name,
             pod_details_config.description,
             schema.to_json(),
-            pod_details_config.is_public,
         )
 
     def _get_default_pod_keys(
         self, pod_keys: Optional[PodKeys]
     ) -> Tuple[RSAPrivateKey, RSAPublicKey]:
+        """Get default pod keys."""
         if pod_keys is None:
-            user_storage_path = BITFOUNT_STORAGE_PATH / _DEFAULT_USER_DIRECTORY
+            user_storage_path = BITFOUNT_STORAGE_PATH / _DEFAULT_USERNAME
             pod_directory = user_storage_path / "pods" / self.name
             pod_keys = _get_pod_keys(pod_directory)
         return pod_keys.private, pod_keys.public
@@ -322,7 +425,7 @@ class Pod:
                 self.pod_public_key,
                 self._access_manager_public_key,
             )
-        except HTTPError as ex:
+        except (HTTPError, SchemaUploadError) as ex:
             logger.critical(f"Failed to register with hub: {ex}")
             raise PodRegistrationError("Failed to register with hub") from ex
         except RequestException as ex:
@@ -378,7 +481,7 @@ class Pod:
     def _check_for_unapproved_pods(
         self,
         pods_involved_in_task: Iterable[str],
-        task_protocol_details: _ProtocolDetails,
+        serialized_protocol: SerializedProtocol,
     ) -> Optional[List[str]]:
         """Returns the pods that we're not happy to work with.
 
@@ -389,16 +492,14 @@ class Pod:
         around security with other pods.
 
         Args:
-            pods_involved_in_task (List[str]): A list of other pods
-                that have been contacted by the modeller for this task
-            task_protocol_details (ProtocolDetails): The decrypted protocol details
-                portion of the task request.
+            pods_involved_in_task: A list of other pods that have been contacted by
+                the modeller for this task.
+            serialized_protocol: The decrypted serialized protocol portion of the task
+                request.
 
         Returns:
-            Optional[List[str]]
-                - List of unapproved pods
-                    or None if all are approved
-                    or None if secure aggregation not in use
+            Either a list of unapproved pods or `None` if all are approved or if secure
+            aggregation not in use.
         """
         unapproved_workers = None
 
@@ -411,7 +512,11 @@ class Pod:
         except ValueError:  # if not in list to remove
             pass
 
-        if task_protocol_details.aggregator == AggregatorType.SecureAggregator.value:
+        aggregator = serialized_protocol.get("aggregator")
+        if (
+            aggregator
+            and aggregator["class_name"] == AggregatorType.SecureAggregator.value
+        ):
             logger.info(
                 "Secure aggregation is in use, checking responses from other pods."
             )
@@ -426,20 +531,19 @@ class Pod:
         logger.info(f"Training task request received from '{message.sender}'")
         try:
             await self._create_and_run_worker(message)
-        except TimeoutError as te:
-            logger.warning("Training timed out, modeller was too slow!")
-            logger.warning(te)
-            # TODO: [BIT-841][BIT-788] Send modeller a message?
+        except asyncio.TimeoutError:
+            logger.info("Ready for next task...")
             return
 
     async def _create_and_run_worker(self, message: _BitfountMessage) -> None:
         """Creates and runs a worker instance."""
         # `_initialise` is always called before this method, so we can assume
         # that the mailbox is initialised. Reassuring mypy that this is True.
-        assert isinstance(self._mailbox, _PodMailbox)  # nosec
+        assert isinstance(self._mailbox, _PodMailbox)  # nosec[assert_used]
 
         # Unpack task details from received message
         logger.info("Unpacking task details from message...")
+        task_id = message.task_id
         task_request_message: _TaskRequestMessage = _TaskRequestMessage.deserialize(
             message.body
         )
@@ -455,11 +559,11 @@ class Pod:
         # we are not, `unapproved_workers` will be `None`.
         other_pods = [
             pod_id
-            for pod_id in message.pod_mailbox_ids.keys()
+            for pod_id in message.pod_mailbox_ids
             if pod_id != self.pod_identifier
         ]
         unapproved_workers = self._check_for_unapproved_pods(
-            other_pods, task_request.protocol_details
+            other_pods, task_request.serialized_protocol
         )
 
         # If we are dealing with secure aggregation (and hence need inter-pod
@@ -476,14 +580,15 @@ class Pod:
             other_pod_public_keys = _get_pod_public_keys(other_pods, self._hub)
 
             worker_mailbox = _InterPodWorkerMailbox(
+                pod_public_keys=other_pod_public_keys,
+                private_key=self.private_key,
                 pod_identifier=self.pod_identifier,
                 modeller_mailbox_id=message.sender_mailbox_id,
                 modeller_name=message.sender,
                 aes_encryption_key=task_request.aes_key,
                 message_service=self._mailbox.message_service,
                 pod_mailbox_ids=message.pod_mailbox_ids,
-                pod_public_keys=other_pod_public_keys,
-                private_key=self.private_key,
+                task_id=task_id,
             )
         else:
             logger.debug("Creating modeller<->worker-only mailbox.")
@@ -494,6 +599,7 @@ class Pod:
                 aes_encryption_key=task_request.aes_key,
                 message_service=self._mailbox.message_service,
                 pod_mailbox_ids=message.pod_mailbox_ids,
+                task_id=task_id,
             )
 
         # TODO: [BIT-1045] Move the secure aggregation allowed check to the access
@@ -506,7 +612,7 @@ class Pod:
                 message.sender, self.pod_identifier
             )
             authorisation_errors.add(
-                _PodResponseType.SECURE_AGGREGATION_WORKERS_NOT_AUTHORISED,
+                _PodResponseType.NO_ACCESS,
                 unapproved_workers,
             )
             await worker_mailbox.reject_task(authorisation_errors.messages)
@@ -528,9 +634,36 @@ class Pod:
             pod_vitals=self._pod_vitals,
             pod_dp=self._pod_dp,
             pod_identifier=self.pod_identifier,
+            serialized_protocol=task_request.serialized_protocol,
         )
-        # Run the worker and the mailbox listening simultaneously
-        await _run_func_and_listen_to_mailbox(worker.run(), worker_mailbox)
+
+        # If interacting with an older modeller version then task_id won't be supplied
+        task_monitor_cm: ContextManager
+        if worker_mailbox.task_id:
+            task_monitor_cm = task_monitor_context(
+                hub=self._hub,
+                task_id=worker_mailbox.task_id,
+                sender_id=worker_mailbox.mailbox_id,
+            )
+        else:
+            task_monitor_cm = nullcontext()
+
+        with task_monitor_cm:
+            # Run the worker and the mailbox listening simultaneously
+            try:
+                await _run_func_and_listen_to_mailbox(worker.run(), worker_mailbox)
+            except Exception as e:
+                logger.federated_error(e)
+                logger.exception(e)
+
+                if worker_mailbox.task_id:
+                    logger.error(
+                        f"Exception whilst running task {worker_mailbox.task_id}."
+                    )
+                else:
+                    logger.error("Exception whilst running task.")
+
+        logger.info("Ready for next task...")
 
     def _create_authorisation_checker(
         self,
@@ -558,7 +691,10 @@ class Pod:
         task_request = authorisation_checker_cls.unpack_task_request(
             task_request_message, self.private_key
         )
-        task_protocol_details = task_request.protocol_details
+        serialized_protocol = task_request.serialized_protocol
+        # remove schema since it is the largest task element.
+        if _keys_exist(serialized_protocol, "algorithm", "model", "schema"):  # type: ignore[arg-type] # Reason: a typed dict is still a dict. # noqa: B950
+            serialized_protocol["algorithm"]["model"].pop("schema")  # type: ignore[typeddict-item] # Reason: all models have a schema associated with them # noqa: B950
 
         pod_response_message = _PodResponseMessage(
             modeller_name=sender,
@@ -567,30 +703,30 @@ class Pod:
 
         authorisation_checker: _AuthorisationChecker
 
-        # TODO: [BIT-1291] Uncomment when key-based validation enabled again
-        # if auth_type == IdentityVerificationMethod.KEYS:
-        #     # Public Key Signature authorisation
-        #     packed_request = (
-        #         authorisation_checker_cls.extract_from_task_request_message(
-        #             task_request_message
-        #         )
-        #     )
-        #
-        #     authorisation_checker = _SignatureBasedAuthorisation(
-        #         pod_response_message=pod_response_message,
-        #         encrypted_task_request=packed_request.encrypted_request,
-        #         signature=packed_request.signature,
-        #         task_protocol_details=task_protocol_details,
-        #     )
-        # elif auth_type == IdentityVerificationMethod.OIDC_ACF_PKCE:
-        if auth_type == IdentityVerificationMethod.OIDC_ACF_PKCE:
+        if auth_type == IdentityVerificationMethod.KEYS:
+            # Public Key Signature authorisation
+            packed_request: _SignedEncryptedTaskRequest = (
+                authorisation_checker_cls.extract_from_task_request_message(
+                    task_request_message
+                )
+            )
+
+            authorisation_checker = _SignatureBasedAuthorisation(
+                pod_response_message=pod_response_message,
+                access_manager=self._access_manager,
+                modeller_name=worker_mailbox.modeller_name,
+                encrypted_task_request=packed_request.encrypted_request,
+                signature=packed_request.signature,
+                serialized_protocol=serialized_protocol,
+            )
+        elif auth_type == IdentityVerificationMethod.OIDC_ACF_PKCE:
             # OIDC Authorization Code Flow
             auth_env = _get_auth_environment()
             authorisation_checker = _OIDCAuthorisationCode(
                 pod_response_message=pod_response_message,
                 access_manager=self._access_manager,
                 mailbox=worker_mailbox,
-                task_protocol_details=task_protocol_details,
+                serialized_protocol=serialized_protocol,
                 _auth_domain=auth_env.auth_domain,
                 _client_id=auth_env.client_id,
             )
@@ -601,7 +737,7 @@ class Pod:
                 pod_response_message=pod_response_message,
                 access_manager=self._access_manager,
                 mailbox=worker_mailbox,
-                task_protocol_details=task_protocol_details,
+                serialized_protocol=serialized_protocol,
                 _auth_domain=auth_env.auth_domain,
                 _client_id=auth_env.client_id,
             )
@@ -611,12 +747,14 @@ class Pod:
                 pod_response_message=pod_response_message,
                 access_manager=self._access_manager,
                 mailbox=worker_mailbox,
-                task_protocol_details=task_protocol_details,
+                serialized_protocol=serialized_protocol,
             )
         return authorisation_checker
 
     @staticmethod
-    async def _repeat(interval: int, func: Callable[..., Coroutine]) -> None:
+    async def _repeat(
+        stop_event: threading.Event, interval: int, func: Callable[..., Coroutine]
+    ) -> None:
         """Run coroutine func every interval seconds.
 
         If func has not finished before *interval*, will run again
@@ -626,7 +764,10 @@ class Pod:
             interval: run interval in seconds
             func: function to call which returns a coroutine to await
         """
-        while True:
+        while not stop_event.is_set():
+            # Don't need to worry about gather tasks cancellation as func() (in
+            # this case _pod_heartbeat()) is short running, so if one of the tasks
+            # raises an exception the other won't be left running long.
             await asyncio.gather(func(), asyncio.sleep(interval))
 
     async def _pod_heartbeat(self) -> None:
@@ -638,13 +779,25 @@ class Pod:
         except RequestException as ex:
             logger.warning(f"Could not connect to hub for status: {ex}")
 
-    def _run_pod_heartbeat_task(self) -> None:
+    def _run_pod_heartbeat_task(self, stop_event: threading.Event) -> None:
         """Makes 10-second interval pod heartbeats to the hub."""
         if _is_notebook():
             # We need to create a new event loop here for jupyter
             # As it's run in a new thread and can't be patched by nest_asyncio
             asyncio.set_event_loop(asyncio.new_event_loop())
-        asyncio.run(self._repeat(10, self._pod_heartbeat))
+        asyncio.run(self._repeat(stop_event, 10, self._pod_heartbeat))
+
+    def _get_pod_heartbeat_thread(self) -> _StoppableThead:
+        """Returns pod heartbeat thread."""
+        logger.info(f"Starting pod {self.name}...")
+        thread_stop_event = threading.Event()
+        pod_heartbeat = _StoppableThead(
+            stop_event=thread_stop_event,
+            target=self._run_pod_heartbeat_task,
+            args=(thread_stop_event,),
+            name="pod_heartbeat",
+        )
+        return pod_heartbeat
 
     def _pod_vitals_server(self, vitals_handler: _PodVitalsHandler) -> None:
         """Run _PodVitals webserver."""
@@ -667,7 +820,10 @@ class Pod:
             vitals_handler = _PodVitalsHandler(self._pod_vitals)
             logger.debug("Starting Pod Vitals interface...")
             threading.Thread(
-                daemon=True, target=self._pod_vitals_server, args=(vitals_handler,)
+                daemon=True,
+                target=self._pod_vitals_server,
+                args=(vitals_handler,),
+                name="pod_vitals_interface",
             ).start()
             return vitals_handler
         else:
@@ -685,11 +841,11 @@ class Pod:
 
         # `_initialise` has just been called which sets the mailbox so we can assume
         # that the mailbox is initialised. Reassuring mypy that this is True.
-        assert isinstance(self._mailbox, _PodMailbox)  # nosec
+        assert isinstance(self._mailbox, _PodMailbox)  # nosec[assert_used]
 
         # Setup heartbeat to hub
-        logger.info(f"Starting pod {self.name}...")
-        threading.Thread(daemon=True, target=self._run_pod_heartbeat_task).start()
+        pod_heartbeat = self._get_pod_heartbeat_thread()
+        pod_heartbeat.start()
 
         # Start pod vitals webserver
         vitals_handler = self._run_pod_vitals_server()
@@ -701,13 +857,24 @@ class Pod:
 
         # Start pod listening for messages
         logger.info("Pod started... press Ctrl+C to stop")
-        await self._mailbox.listen_for_messages()
+        try:
+            await self._mailbox.listen_indefinitely()
+        finally:
+            logger.info(f"Pod {self.name} stopped.")
 
-        logger.info(f"Pod {self.name} stopped.")
+            # Shutdown pod heartbeat thread
+            pod_heartbeat.stop()
+            logger.debug("Waiting up to 15 seconds for pod heartbeat thread to stop")
+            pod_heartbeat.join(15)
+            if pod_heartbeat.stopped:
+                logger.debug("Shut down pod heartbeat thread")
+            else:
+                logger.error("Unable to shut down pod heartbeat thread")
 
-        # Shutdown pod vitals webserver
-        if vitals_handler:
-            vitals_handler.runner.cleanup()
+            # Shutdown pod vitals webserver
+            if vitals_handler:
+                await vitals_handler.runner.cleanup()
+                logger.debug("Shut down vitals handler thread")
 
     def start(self) -> None:
         """Starts a pod instance, listening for tasks.
