@@ -1,12 +1,14 @@
 """Handles Modeller sending training requests and receiving pod responses."""
 from __future__ import annotations
 
+from dataclasses import dataclass
 import logging
 from typing import (
     Any,
     Callable,
     Dict,
     Final,
+    Iterable,
     List,
     Mapping,
     MutableMapping,
@@ -14,22 +16,22 @@ from typing import (
     Optional,
     Set,
     Tuple,
+    Union,
     cast,
 )
 
 from cryptography.hazmat.primitives.asymmetric.rsa import RSAPublicKey
 from grpc import RpcError
 
+from bitfount import config
 from bitfount.federated.encryption import _AESEncryption
 from bitfount.federated.exceptions import BitfountTaskStartError
 from bitfount.federated.logging import _federate_logger, _get_federated_logger
-from bitfount.federated.task_requests import _ProtocolDetails
 from bitfount.federated.transport.base_transport import (
     Handler,
     MessageRetrievalError,
     _BaseMailbox,
     _send_aes_encrypted_message,
-    _WorkerMailboxDetails,
 )
 from bitfount.federated.transport.handlers import _AsyncMultipleResponsesHandler
 from bitfount.federated.transport.message_service import (
@@ -46,15 +48,36 @@ from bitfount.federated.transport.types import (
 from bitfount.federated.transport.utils import _average_training_metrics
 from bitfount.federated.types import (
     _RESPONSE_MESSAGES,
+    SerializedProtocol,
     _PodResponseType,
     _TaskRequestMessageGenerator,
 )
-from bitfount.types import _JSONDict, _SAMLResponse, _SerializedWeights
+from bitfount.types import _JSONDict, _SAMLResponse, _SerializedWeights, _StrAnyDict
 
 logger = _get_federated_logger(__name__)
 
-_DEFAULT_TASK_RESPONSE_TIMEOUT: Final[int] = 5 * 60
-_DEFAULT_OIDC_CLIENT_IDS_TIMEOUT: Final[int] = 5 * 60
+_DEFAULT_TASK_RESPONSE_TIMEOUT: Final[int] = 5 * 60  # 5 minutes
+_DEFAULT_OIDC_CLIENT_IDS_TIMEOUT: Final[int] = 5 * 60  # 5 minutes
+_SOFT_LIMIT_MESSAGE_TIMEOUT: Final[int] = config.BITFOUNT_ONLINE_CHECK_SOFT_LIMIT
+
+
+@dataclass
+class _WorkerMailboxDetails:
+    """Mailbox details for a specific task/worker on a pod.
+
+    Used by the modeller to encapsulate details of worker mailboxes.
+
+    Attributes:
+        pod_identifier: The parent pod's identifier.
+        public_key: The parent pod's public key.
+        mailbox_id: The mailbox ID for this specific task/worker.
+        aes_encryption_key: The encryption key to use for this specific task/worker.
+    """
+
+    pod_identifier: str
+    public_key: RSAPublicKey
+    mailbox_id: str
+    aes_encryption_key: bytes
 
 
 class _ModellerMailbox(_BaseMailbox):
@@ -64,8 +87,11 @@ class _ModellerMailbox(_BaseMailbox):
         self,
         mailbox_id: str,
         worker_mailboxes: Mapping[str, _WorkerMailboxDetails],
+        task_id: str,
         message_service: _MessageService,
-        handlers: Optional[Mapping[_BitfountMessageType, Handler]] = None,
+        handlers: Optional[
+            Mapping[_BitfountMessageType, Union[Handler, Iterable[Handler]]]
+        ] = None,
     ):
         """Creates a new ModellerMailbox.
 
@@ -77,6 +103,7 @@ class _ModellerMailbox(_BaseMailbox):
             mailbox_id: The mailbox ID for this modeller mailbox.
             worker_mailboxes: A mapping of pod identifiers to worker mailbox details
                               for the pods/workers that will be involved in this task.
+            task_id: The ID for the task this mailbox belongs to.
             message_service: The underlying message service.
             handlers: Optional. A set of handlers to initialise with.
         """
@@ -85,10 +112,17 @@ class _ModellerMailbox(_BaseMailbox):
         )
         self.worker_mailboxes: Dict[str, _WorkerMailboxDetails] = dict(worker_mailboxes)
         self._pod_identifiers: Set[str] = set(worker_mailboxes.keys())
+        self._task_id = task_id
 
         self.accepted_worker_mailboxes: Dict[str, _WorkerMailboxDetails] = {}
 
         self._setup_federated_logging()
+        self._setup_online_status_handler()
+
+    @property
+    def task_id(self) -> str:
+        """The task ID of the task associated with this mailbox."""
+        return self._task_id
 
     ############################
     # Task Setup Phase Methods #
@@ -96,7 +130,7 @@ class _ModellerMailbox(_BaseMailbox):
     @classmethod
     async def send_task_requests(
         cls,
-        protocol_details: _ProtocolDetails,
+        serialized_protocol: SerializedProtocol,
         pod_public_keys: Mapping[str, RSAPublicKey],
         task_request_msg_gen: _TaskRequestMessageGenerator,
         message_service: _MessageService,
@@ -107,22 +141,18 @@ class _ModellerMailbox(_BaseMailbox):
         for the pods.
 
         Args:
-            protocol_details:
-                A tuple of protocol name, algorithm name and optionally model and
-                aggregator names.
-            pod_public_keys:
-                A mapping of pod identifiers to their public keys for all the pods
-                involved in this task request.
-            task_request_msg_gen:
-                A callable which will generate a task request message appropriate
-                to the chosen verification method.
+            serialized_protocol: The serialized protocol to use for the task.
+            pod_public_keys: A mapping of pod identifiers to their public keys for all
+                the pods involved in this task request.
+            task_request_msg_gen: A callable which will generate a task request message
+                appropriate to the chosen verification method.
             message_service: The underlying message service to use.
 
         Returns:
             The created modeller mailbox for this task.
         """
-        modeller_mailbox_id, worker_mailboxes = await cls._send_task_requests(
-            protocol_details,
+        modeller_mailbox_id, worker_mailboxes, task_id = await cls._send_task_requests(
+            serialized_protocol,
             pod_public_keys,
             task_request_msg_gen=task_request_msg_gen,
             message_service=message_service,
@@ -130,6 +160,7 @@ class _ModellerMailbox(_BaseMailbox):
         modeller_mailbox = cls(
             mailbox_id=modeller_mailbox_id,
             worker_mailboxes=worker_mailboxes,
+            task_id=task_id,
             message_service=message_service,
         )
         return modeller_mailbox
@@ -137,32 +168,29 @@ class _ModellerMailbox(_BaseMailbox):
     @classmethod
     async def _send_task_requests(
         cls,
-        protocol_details: _ProtocolDetails,
+        serialized_protocol: SerializedProtocol,
         pod_public_keys: Mapping[str, RSAPublicKey],
         task_request_msg_gen: _TaskRequestMessageGenerator,
         message_service: _MessageService,
-    ) -> Tuple[str, Dict[str, _WorkerMailboxDetails]]:
+    ) -> Tuple[str, Dict[str, _WorkerMailboxDetails], str]:
         """Manage sending of task requests, such as training requests, to pods.
 
         Args:
-            protocol_details:
-                A tuple of protocol name, algorithm name and optionally model and
-                aggregator names.
-            pod_public_keys:
-                A mapping of pod identifiers to their public keys for all the pods
-                involved in this task request.
-            task_request_msg_gen:
-                A callable which will generate a task request message appropriate
-                to the chosen verification method.
+            serialized_protocol: The serialized protocol to use for the task.
+            pod_public_keys: A mapping of pod identifiers to their public keys for all
+                the pods involved in this task request.
+            task_request_msg_gen: A callable which will generate a task request message
+                appropriate to the chosen verification method.
             message_service: The underlying message service to use.
 
         Returns:
             Tuple of:
                 - (str) modeller mailbox ID
                 - (dict) of pod identifier to worker mailbox details
+                - (str) the task ID
         """
         # Shorthand the pod identifiers for ease of use
-        pod_identifiers: List[str] = list(pod_public_keys.keys())
+        pod_identifiers: List[str] = list(pod_public_keys)
 
         # Generate encryption keys and then task request messages for each pod.
         aes_key_per_pod = {
@@ -171,7 +199,7 @@ class _ModellerMailbox(_BaseMailbox):
         }
         task_request_per_pod: Dict[str, bytes] = {
             pod_identifier: task_request_msg_gen(
-                protocol_details,
+                serialized_protocol,
                 pod_identifiers,
                 aes_key_per_pod[pod_identifier],
                 pod_public_key,
@@ -185,6 +213,7 @@ class _ModellerMailbox(_BaseMailbox):
             (
                 modeller_mailbox_id,
                 worker_mailbox_ids,
+                task_id,
             ) = await message_service.setup_communication_with_pods(
                 task_request_per_pod
             )
@@ -208,6 +237,7 @@ class _ModellerMailbox(_BaseMailbox):
                 )
                 for pod_identifier, worker_mailbox_id in worker_mailbox_ids.items()
             },
+            task_id,
         )
 
     async def get_oidc_client_ids(
@@ -272,6 +302,7 @@ class _ModellerMailbox(_BaseMailbox):
                 recipient_mailbox_id=pod_mailbox.mailbox_id,
                 sender=self.message_service.username,
                 sender_mailbox_id=self.mailbox_id,
+                task_id=self._task_id,
             )
 
     async def send_oidc_device_code_responses(
@@ -289,6 +320,7 @@ class _ModellerMailbox(_BaseMailbox):
                 recipient_mailbox_id=pod_mailbox.mailbox_id,
                 sender=self.message_service.username,
                 sender_mailbox_id=self.mailbox_id,
+                task_id=self._task_id,
             )
 
     async def get_saml_challenges(
@@ -300,7 +332,8 @@ class _ModellerMailbox(_BaseMailbox):
         is reached. Responses are expected from all pods assigned at mailbox init.
 
         Returns:
-            A list of received SAML challenges.
+            A list of received SAML challenges. Note that there is no notion of
+            "pod order" in this list, the elements are in the order they are received.
         """
         saml_challenges = []
 
@@ -425,21 +458,10 @@ class _ModellerMailbox(_BaseMailbox):
 
                     # Process different forms of rejection
                     for response_type in response.body:
-                        if (
-                            response_type
-                            == _PodResponseType.SECURE_AGGREGATION_WORKERS_NOT_AUTHORISED.name  # noqa: B950
-                        ):
-                            unapproved_pods = ", ".join(response.body[response_type])
-                            logger.error(
-                                f"Received rejection from {pod_identifier}. "
-                                f"{_RESPONSE_MESSAGES[_PodResponseType[response_type]]}: "  # noqa: B950
-                                f"{unapproved_pods}"
-                            )
-                        else:
-                            logger.error(
-                                f"Received rejection from {pod_identifier}. "
-                                f"{_RESPONSE_MESSAGES[_PodResponseType[response_type]]}"  # noqa: B950
-                            )
+                        logger.error(
+                            f"Received rejection from {pod_identifier}. "
+                            f"{_RESPONSE_MESSAGES[_PodResponseType[response_type]]}"  # noqa: B950
+                        )
 
             # Handle cases where response didn't arrive
             else:
@@ -467,6 +489,7 @@ class _ModellerMailbox(_BaseMailbox):
             recipient_mailbox_id=pod_mailbox.mailbox_id,
             sender=self.message_service.username,
             sender_mailbox_id=self.mailbox_id,
+            task_id=self._task_id,
         )
 
     ################################
@@ -495,31 +518,31 @@ class _ModellerMailbox(_BaseMailbox):
                 recipient_mailbox_id=mailbox.mailbox_id,
                 sender_mailbox_id=self.mailbox_id,
                 sender=self.message_service.username,
+                task_id=self._task_id,
             )
 
-    async def send_task_details(self, task_details: _JSONDict) -> None:
-        """Sends the task details (protocol, algorithm, etc) to the workers.
-
-        Args:
-            task_details: The serialized task details.
-        """
-        await self._send_to_all_pods_aes_encrypt(
-            task_details, _BitfountMessageType.ALGORITHM_EXCHANGE
-        )
-
     async def send_training_iteration_complete_update(
         self, training_complete: bool
     ) -> None:
         """Sends whether training is complete or not to the workers."""
+        logger.debug(f"Sending TRAINING_COMPLETE from {self.mailbox_id}")
         await self._send_to_all_pods_aes_encrypt(
             training_complete, _BitfountMessageType.TRAINING_COMPLETE
         )
 
+    async def send_task_start_message(self) -> None:
+        """Sends task start message to the workers.
+
+        Note: The message is not important here, the message type is.
+        """
+        await self._send_to_all_pods_aes_encrypt(None, _BitfountMessageType.TASK_START)
+
     async def send_task_complete_message(self) -> None:
         """Sends task complete message to the workers.
 
         Note: The message is not important here, the message type is.
         """
+        logger.debug(f"Sending TASK_COMPLETE message to workers from {self.mailbox_id}")
         await self._send_to_all_pods_aes_encrypt(
             None, _BitfountMessageType.TASK_COMPLETE
         )
@@ -537,15 +560,16 @@ class _ModellerMailbox(_BaseMailbox):
 
     async def get_evaluation_results_from_workers(
         self, timeout: Optional[int] = None
-    ) -> List[Any]:
+    ) -> _StrAnyDict:
         """Get evaluation results from workers."""
-        all_eval_results: List[Any] = []
+        logger.info("Waiting to receive results from Pods...")
+        all_eval_results: _StrAnyDict = {}
 
         # Create light-weight handler to append to shared list
         def evaluation_results_handler(message: _BitfountMessage) -> None:
             logger.debug(f"Receiving evaluation results from worker {message.sender}")
             eval_results = self._decrypt_message(message).body
-            all_eval_results.append(eval_results)
+            all_eval_results[message.sender] = eval_results
 
         # We use `self` rather than `self.modeller_mailbox` as the mailbox below
         # because this is ensures things are correctly delegated.
@@ -570,10 +594,11 @@ class _ModellerMailbox(_BaseMailbox):
         _federate_logger(self)
 
         def log_message_handler(message: _BitfountMessage) -> None:
+            """Locally logs the log message that has been received from a pod."""
             log_message_wrapper: _DecryptedBitfountMessage = self._decrypt_message(
                 message
             )
-            log_message: Dict[str, Any] = log_message_wrapper.body
+            log_message: _JSONDict = log_message_wrapper.body
             # We prepend the name of the pod to the log message
             log_message["msg"] = f"{message.sender} Pod: {log_message['msg']}"
             # We remove the `federated` key to avoid recursively sending a federated
@@ -583,6 +608,35 @@ class _ModellerMailbox(_BaseMailbox):
 
         self.register_handler(_BitfountMessageType.LOG_MESSAGE, log_message_handler)
 
+    def _setup_online_status_handler(self) -> None:
+        """Respond to online status requests from Pods."""
+
+        async def status_request_handler(message: _BitfountMessage) -> None:
+            """Responds to an ONLINE_CHECK request with an ONLINE_RESPONSE."""
+            logger.info(f"Informing {message.sender} that we are still online.")
+
+            # We use the message service sending directly as we don't want to
+            # re-encrypt the already encrypted body, we just want to send it back
+            # to the worker.
+            await self.message_service.send_message(
+                _BitfountMessage(
+                    message_type=_BitfountMessageType.ONLINE_RESPONSE,
+                    body=message.body,
+                    recipient=message.sender,
+                    recipient_mailbox_id=message.sender_mailbox_id,
+                    sender=self.message_service.username,
+                    sender_mailbox_id=self.mailbox_id,
+                    task_id=self._task_id,
+                ),
+                already_packed=True,
+            )
+
+        self.register_handler(
+            _BitfountMessageType.ONLINE_CHECK,
+            status_request_handler,
+            high_priority=True,
+        )
+
     ##################################
     # End Task Running Phase Methods #
     ##################################
@@ -600,6 +654,109 @@ async def _send_model_parameters(
     )
 
 
+async def _send_psi_dataset_modeller(
+    dataset: List[str], modeller_mailbox: _ModellerMailbox
+) -> None:
+    """Sends psi datasets to the worker."""
+    await modeller_mailbox._send_to_all_pods_aes_encrypt(
+        dataset, _BitfountMessageType.PSI_DATASET
+    )
+
+
+def _psi_dataset_handler(
+    modeller_mailbox: _ModellerMailbox,
+    psi_datasets_str: List[Tuple[List[str], List[str]]],
+) -> Callable[[_BitfountMessage], None]:
+    """Handle receiving of PSI datasets from workers.
+
+    Appends them to the `psi_datasets_str` list.
+
+    Note that there is no notion of "worker order" in this list, the elements are
+    in the order they are received.
+    """
+    # Create light-weight handler to append to shared list
+    def psi_dataset_handler(message: _BitfountMessage) -> None:
+        logger.debug(f"Receiving psi dataset from worker {message.sender}")
+        # Deliberate access to private method here as that method shouldn't be used in
+        # any other context than transport layer access.
+        # noinspection PyProtectedMember
+        psi_results = modeller_mailbox._decrypt_message(message).body
+        psi_datasets_str.append(psi_results)
+
+    return psi_dataset_handler
+
+
+async def _get_psi_datasets_from_workers(
+    modeller_mailbox: _ModellerMailbox, timeout: Optional[int] = None
+) -> List[Tuple[List[str], List[str]]]:
+    """Get psi datasets from workers.
+
+    Note that there is no notion of "worker order" in this list, the elements are
+    in the order they are received.
+    """
+    psi_datasets_str: List[Tuple[List[str], List[str]]] = []
+    psi_dataset_handler = _psi_dataset_handler(modeller_mailbox, psi_datasets_str)
+    # We use `self` rather than `self.modeller_mailbox` as the mailbox below
+    # because this is ensures things are correctly delegated.
+    with _AsyncMultipleResponsesHandler(
+        handler=psi_dataset_handler,
+        message_types=_BitfountMessageType.PSI_DATASET,
+        mailbox=modeller_mailbox,
+        responders=modeller_mailbox.accepted_worker_mailboxes.keys(),
+    ) as response_handler:
+        await response_handler.wait_for_responses(timeout=timeout)
+    return psi_datasets_str
+
+
+def _public_key_handler(
+    modeller_mailbox: _ModellerMailbox, public_key: List[bytes]
+) -> Callable[[_BitfountMessage], None]:
+    """Public key handler.
+
+    Will mutate the passed in sequence by appending responses to it.
+    Should only be used for the PrivateSetIntersection protocol.
+
+    Note that there is no notion of "worker order" in this list, the elements are
+    in the order they are received.
+    """
+
+    def public_key_handler(message: _BitfountMessage) -> None:
+        logger.debug(f"Receiving the public key from worker {message.sender}")
+        # Deliberate access to private method here as that method shouldn't be used in
+        # any other context than transport layer access.
+        # noinspection PyProtectedMember
+        public_key_ = modeller_mailbox._decrypt_message(message).body
+        public_key.append(public_key_)
+
+    return public_key_handler
+
+
+async def _get_public_key(
+    modeller_mailbox: _ModellerMailbox, timeout: Optional[int] = None
+) -> List[bytes]:
+    """Get public key from worker.
+
+    Used for the PrivateSetIntersection protocol.
+
+    Note that there is no notion of "pod order" in this list, the elements are in
+    the order they are received.
+    """
+    public_key_lst: List[bytes] = []
+    public_key_handler = _public_key_handler(modeller_mailbox, public_key_lst)
+    with _AsyncMultipleResponsesHandler(
+        handler=public_key_handler,
+        message_types=_BitfountMessageType.KEY_EXCHANGE,
+        mailbox=modeller_mailbox,
+        responders=modeller_mailbox.accepted_worker_mailboxes.keys(),
+    ) as response_handler:
+        non_responders = await response_handler.wait_for_responses(timeout=timeout)
+        if non_responders:
+            logger.info(
+                f"The following did not send training metrics in time: {non_responders}"
+            )
+    return public_key_lst
+
+
 def _training_metrics_handler(
     modeller_mailbox: _ModellerMailbox,
     training_metrics: MutableSequence[Mapping[str, str]],
@@ -607,6 +764,9 @@ def _training_metrics_handler(
     """Training metrics handler.
 
     Will mutate the passed in sequence by appending responses to it.
+
+    Note that there is no notion of "worker order" in this list, the elements are
+    in the order they are received.
     """
 
     def training_metrics_handler(message: _BitfountMessage) -> None:
@@ -614,7 +774,9 @@ def _training_metrics_handler(
         # Deliberate access to private method here as that method shouldn't be used in
         # any other context than transport layer access.
         # noinspection PyProtectedMember
-        single_training_metrics = modeller_mailbox._decrypt_message(message).body
+        single_training_metrics: Mapping[str, str] = modeller_mailbox._decrypt_message(
+            message
+        ).body
         training_metrics.append(single_training_metrics)
 
     return training_metrics_handler
@@ -624,7 +786,7 @@ async def _get_training_metrics_from_workers(
     modeller_mailbox: _ModellerMailbox,
     timeout: Optional[int] = None,
 ) -> Dict[str, float]:
-    """Get training metrics from workers."""
+    """Get average training metrics from workers."""
     training_metrics: List[Mapping[str, str]] = []
     training_metrics_handler = _training_metrics_handler(
         modeller_mailbox, training_metrics
@@ -652,7 +814,7 @@ def _parameter_updates_handler(
 ) -> Callable[[_BitfountMessage], None]:
     """Parameter update handler.
 
-    Will mutate the passed in sequence by appending responses to it.
+    Will mutate the passed in mapping by appending responses to it.
     """
 
     def parameter_update_handler(message: _BitfountMessage) -> None:
@@ -670,7 +832,15 @@ def _parameter_updates_handler(
 async def _get_parameter_updates_from_workers(
     modeller_mailbox: _ModellerMailbox, timeout: Optional[int] = None
 ) -> Dict[str, _SerializedWeights]:
-    """Get model parameter updates from workers."""
+    """Get model parameter updates from workers.
+
+    Args:
+        modeller_mailbox: The modeller mailbox.
+        timeout: The timeout for the request.
+
+    Returns:
+        A dictionary of the form {worker_name: weight_update}.
+    """
     weight_updates: Dict[str, _SerializedWeights] = {}
     parameter_updates_handler = _parameter_updates_handler(
         modeller_mailbox, weight_updates
