@@ -1,10 +1,25 @@
 """Handles messages for a Pods and Modellers at the protocol level during a task."""
+import asyncio
 import logging
-from typing import Any, Callable, Dict, Final, Iterable, List, Mapping, Optional
+from typing import (
+    Any,
+    Callable,
+    Dict,
+    Final,
+    Iterable,
+    List,
+    Mapping,
+    Optional,
+    Tuple,
+    Union,
+)
+import warnings
 
+from cryptography.hazmat.primitives import serialization
 from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey, RSAPublicKey
 import msgpack
 
+from bitfount import config
 from bitfount.federated.encryption import _RSAEncryption
 from bitfount.federated.logging import _federate_logger, _get_federated_logger
 from bitfount.federated.transport.base_transport import (
@@ -15,6 +30,7 @@ from bitfount.federated.transport.base_transport import (
 from bitfount.federated.transport.handlers import (
     _AsyncMultipleResponsesHandler,
     _get_message_awaitable,
+    _OnlineResponseHandling,
 )
 from bitfount.federated.transport.message_service import (
     _BitfountMessage,
@@ -32,7 +48,15 @@ from bitfount.types import _JSONDict, _SAMLResponse, _SerializedWeights
 
 logger = _get_federated_logger(__name__)
 
-_DEFAULT_OIDC_MODELLER_RESPONSE_TIMEOUT: Final[int] = 5 * 60
+# How long to wait for Modeller authentication response
+_DEFAULT_AUTHENTICATION_MODELLER_RESPONSE_TIMEOUT: Final[int] = 5 * 60  # 5 minutes
+# How long in seconds to wait for a response from the modeller before checking if
+# the modeller is still online.
+_SOFT_LIMIT_MESSAGE_TIMEOUT: Final[int] = config.BITFOUNT_ONLINE_CHECK_SOFT_LIMIT
+# How long in seconds to wait for a response from the modeller regarding their
+# online status before aborting the task. The hard limit can only be reached after
+# the soft limit has already been reached.
+_HARD_LIMIT_MESSAGE_TIMEOUT: Final[int] = config.BITFOUNT_ONLINE_CHECK_HARD_LIMIT
 
 
 class _WorkerMailbox(_BaseMailbox):
@@ -46,7 +70,10 @@ class _WorkerMailbox(_BaseMailbox):
         aes_encryption_key: bytes,
         message_service: _MessageService,
         pod_mailbox_ids: Mapping[str, str],
-        handlers: Optional[Mapping[_BitfountMessageType, Handler]] = None,
+        task_id: Optional[str] = None,
+        handlers: Optional[
+            Mapping[_BitfountMessageType, Union[Handler, Iterable[Handler]]]
+        ] = None,
     ):
         """Create new worker mailbox for a specific task.
 
@@ -58,6 +85,7 @@ class _WorkerMailbox(_BaseMailbox):
             message_service: the underlying message service.
             pod_mailbox_ids: mapping of pod_identifier to worker mailbox IDs for
                              all pods involved in the task.
+            task_id: The ID of the task that this mailbox is associated with.
             handlers: an optional mapping of message types to handlers to initialise
                       with.
         """
@@ -72,18 +100,47 @@ class _WorkerMailbox(_BaseMailbox):
         self.modeller_name = modeller_name
         self.aes_encryption_key = aes_encryption_key
         self.pod_mailbox_ids: Dict[str, str] = dict(pod_mailbox_ids)
+        self._task_id = task_id
+        if self._task_id is None:
+            warnings.warn(
+                "Worker mailboxes should be instantiated with task IDs;"
+                " ask the Modeller to upgrade their bitfount version if possible.",
+                FutureWarning,
+            )
+            logger.warning(
+                "Worker mailboxes should be instantiated with task IDs;"
+                " ask the Modeller to upgrade their bitfount version if possible."
+            )
+
+        # Create modeller online checker and set it to monitor all messages and
+        # specifically ONLINE_RESPONSE messages
+        self._online_response_handler = _OnlineResponseHandling(
+            self.modeller_name, self.aes_encryption_key
+        )
+        self.register_universal_handler(
+            self._online_response_handler.response_handler, high_priority=True
+        )
+        self.register_handler(
+            _BitfountMessageType.ONLINE_RESPONSE,
+            self._online_response_handler.response_handler,
+            high_priority=True,
+        )
 
         # Gather list of other pods. Iteration order is important here (to ensure
         # consistency between share generation order), so we use List rather than
         # another collection.
         self.other_pods: List[str] = [
             pod_identifier
-            for pod_identifier in pod_mailbox_ids.keys()
+            for pod_identifier in pod_mailbox_ids
             if pod_identifier != self.pod_identifier
         ]
-
         self._setup_federated_logging()
 
+    @property
+    def task_id(self) -> Optional[str]:
+        """The task ID of the task associated with this mailbox."""
+        return self._task_id
+
     async def accept_task(self) -> None:
         """Sends an acceptance message to the modeller."""
         await self._send_aes_encrypted_message(
@@ -123,7 +180,9 @@ class _WorkerMailbox(_BaseMailbox):
             saml_request, _BitfountMessageType.SAML_REQUEST
         )
 
-    async def get_saml_response(self, timeout: Optional[int] = None) -> _SAMLResponse:
+    async def get_saml_response(
+        self, timeout: Optional[int] = _DEFAULT_AUTHENTICATION_MODELLER_RESPONSE_TIMEOUT
+    ) -> _SAMLResponse:
         """Awaits for the SAML Response from the modeller."""
         saml_response: _SAMLResponse = await self._get_message_and_decrypt(
             _BitfountMessageType.SAML_RESPONSE, timeout
@@ -139,16 +198,15 @@ class _WorkerMailbox(_BaseMailbox):
         )
 
     async def get_oidc_auth_flow_response(
-        self, timeout: Optional[int] = _DEFAULT_OIDC_MODELLER_RESPONSE_TIMEOUT
+        self, timeout: Optional[int] = _DEFAULT_AUTHENTICATION_MODELLER_RESPONSE_TIMEOUT
     ) -> _OIDCAuthFlowResponse:
         """Get OIDC Auth Code Flow response from the modeller.
 
         Response will contain an authorization code, code verifier, and redirect URI.
 
         Args:
-            timeout:
-                How long to wait for response, in seconds. If `None`, will wait
-                indefinitely.
+            timeout: How long to wait for response, in seconds. If `None`, will wait
+                indefinitely. Defaults to 5 minutes.
 
         Returns:
             Tuple of authorization code, code verifier, and redirect URI.
@@ -176,22 +234,21 @@ class _WorkerMailbox(_BaseMailbox):
                 )
 
     async def get_oidc_device_code_response(
-        self, timeout: Optional[int] = _DEFAULT_OIDC_MODELLER_RESPONSE_TIMEOUT
+        self, timeout: Optional[int] = _DEFAULT_AUTHENTICATION_MODELLER_RESPONSE_TIMEOUT
     ) -> _PodDeviceCodeDetails:
         """Get OIDC Device Code Flow response from the modeller.
 
         Response will contain a device code, and polling details.
 
         Args:
-            timeout:
-                How long to wait for response, in seconds. If `None`, will wait
-                indefinitely.
+            timeout: How long to wait for response, in seconds. If `None`, will wait
+                indefinitely. Defaults to 5 minutes.
 
         Returns:
             Tuple device code, when the code expires, and the polling interval to use.
         """
         logger.info(f"Awaiting OIDC device response from '{self.modeller_name}'")
-        decrypted: Dict[str, Any] = await self._get_message_and_decrypt(
+        decrypted: _JSONDict = await self._get_message_and_decrypt(
             _BitfountMessageType.OIDC_DEVICE_CODE_RESPONSE, timeout
         )
 
@@ -230,16 +287,18 @@ class _WorkerMailbox(_BaseMailbox):
             recipient_mailbox_id=self.modeller_mailbox_id,
             sender=self.pod_identifier,
             sender_mailbox_id=self.mailbox_id,
+            task_id=self._task_id,
         )
 
     async def send_evaluation_results(self, eval_results: Mapping[str, float]) -> None:
         """Sends evaluation results to the modeller."""
+        logger.info("Sending results to modeller...")
         await self._send_aes_encrypted_message(
             eval_results, _BitfountMessageType.EVALUATION_RESULTS
         )
 
     async def _get_message(
-        self, message_type: _BitfountMessageType, timeout: Optional[int] = None
+        self, message_type: _BitfountMessageType, timeout: Optional[int]
     ) -> _BitfountMessage:
         """Generic handler for single message retrieval.
 
@@ -249,10 +308,83 @@ class _WorkerMailbox(_BaseMailbox):
 
         Returns:
             The message.
+
+        Raises:
+            asyncio.TimeoutError: If the message is not received within the timeout.
         """
+        logger.debug(
+            f"Waiting for message ({message_type}) retrieval in"
+            f" {self.__class__.__name__}._get_message() from mailbox {self.mailbox_id}"
+        )
+
+        # Registers the handler for the expected message type
         async_awaitable = _get_message_awaitable()
         self.register_temp_handler(message_type, async_awaitable)
-        message = await async_awaitable.result(timeout)
+
+        # Waits for the message until a `TimeoutError` occurs. If a message is received
+        # before the timeout, it is simply returned. Otherwise, a message is sent to the
+        # modeller to check that they are still online. If they are online, the timeout
+        # resets and we wait for the original message again. If they are not online,
+        # this will raise another `TimeoutError` waiting for the Modeller's response. On
+        # the second timeout, the task is cancelled by sending a `TASK_ABORT` message to
+        # the modeller and re-raising the original `TimeoutError`.
+        while True:
+            try:
+                message = await async_awaitable.result(timeout)
+                break
+            except asyncio.TimeoutError as te:
+                logger.debug(
+                    f"Soft limit timeout hit waiting for message {message_type.name}."
+                )
+
+                logger.info("Checking if Modeller is still online...")
+                online_check_uuid = self._online_response_handler.get_online_check_id()
+                await self.check_modeller_online(online_check_uuid)
+
+                # Want to wait for _either_ the original message to come through or
+                # for the modeller to indicate it is online.
+                message_result_task = asyncio.create_task(async_awaitable.result())
+                online_response_task = asyncio.create_task(
+                    self._online_response_handler.wait_for_response(online_check_uuid)
+                )
+                joint_wait = asyncio.wait(
+                    (message_result_task, online_response_task),
+                    return_when=asyncio.FIRST_COMPLETED,
+                )
+
+                try:
+                    done, pending = await asyncio.wait_for(
+                        joint_wait, timeout=_HARD_LIMIT_MESSAGE_TIMEOUT
+                    )
+                except asyncio.TimeoutError:
+                    logger.warning("Modeller is offline. Aborting task.")
+                    await self.send_task_abort_message()
+
+                    # Double-check these tasks are cancelled
+                    message_result_task.cancel()
+                    online_response_task.cancel()
+
+                    raise te
+                else:
+                    if message_result_task in done:
+                        # We have our message, cancel the response waiter and return
+                        logger.info(
+                            "Modeller is online, responded with expected message."
+                        )
+                        online_response_task.cancel()
+                        self._online_response_handler.remove_waiter(online_check_uuid)
+                        return message_result_task.result()
+                    else:  # online_response_task in done
+                        logger.info("Modeller is online, continuing to wait.")
+
+                        # One final check to see if it's there
+                        if not message_result_task.done():
+                            # Cancel message result task as we're going to start the
+                            # loop again and await on it there
+                            message_result_task.cancel()
+                        else:
+                            return message_result_task.result()
+
         return message
 
     def _aes_decrypt(self, message: _BitfountMessage) -> Any:
@@ -267,7 +399,9 @@ class _WorkerMailbox(_BaseMailbox):
         return message.decrypt(self.aes_encryption_key).body
 
     async def _get_message_and_decrypt(
-        self, message_type: _BitfountMessageType, timeout: Optional[int] = None
+        self,
+        message_type: _BitfountMessageType,
+        timeout: Optional[int] = _SOFT_LIMIT_MESSAGE_TIMEOUT,
     ) -> Any:
         """Generic handler for single encrypted message retrieval.
 
@@ -281,15 +415,8 @@ class _WorkerMailbox(_BaseMailbox):
         message = await self._get_message(message_type, timeout)
         return self._aes_decrypt(message)
 
-    async def get_task_details(self, timeout: Optional[int] = None) -> _JSONDict:
-        """Awaits serialized task details (protocol, etc.) from the modeller."""
-        task_details: _JSONDict = await self._get_message_and_decrypt(
-            _BitfountMessageType.ALGORITHM_EXCHANGE, timeout=timeout
-        )
-        return task_details
-
     async def get_training_iteration_complete_update(
-        self, timeout: Optional[int] = None
+        self, timeout: Optional[int] = _SOFT_LIMIT_MESSAGE_TIMEOUT
     ) -> bool:
         """Awaits for the next training complete update from the modeller."""
         training_complete: bool = await self._get_message_and_decrypt(
@@ -297,12 +424,37 @@ class _WorkerMailbox(_BaseMailbox):
         )
         return training_complete
 
-    async def get_task_complete_update(self, timeout: Optional[int] = None) -> None:
+    async def get_task_start_update(
+        self, timeout: Optional[int] = _SOFT_LIMIT_MESSAGE_TIMEOUT
+    ) -> None:
+        """Awaits for the task start message from the modeller."""
+        await self._get_message_and_decrypt(
+            _BitfountMessageType.TASK_START, timeout=timeout
+        )
+
+    async def get_task_complete_update(
+        self, timeout: Optional[int] = _SOFT_LIMIT_MESSAGE_TIMEOUT
+    ) -> None:
         """Awaits for the task complete message from the modeller."""
+        logger.debug(
+            f"Awaiting TASK_COMPLETE message"
+            f" from {self.modeller_name}"
+            f" in mailbox {self.mailbox_id}"
+        )
         await self._get_message_and_decrypt(
             _BitfountMessageType.TASK_COMPLETE, timeout=timeout
         )
 
+    async def check_modeller_online(self, online_check_uuid: str) -> None:
+        """Send message to Modeller to check if the Modeller is online."""
+        await self._send_aes_encrypted_message(
+            online_check_uuid, _BitfountMessageType.ONLINE_CHECK
+        )
+
+    async def send_task_abort_message(self) -> None:
+        """Send task abort message to Modeller."""
+        await self._send_aes_encrypted_message(None, _BitfountMessageType.TASK_ABORT)
+
     async def log(self, message: Mapping[str, object]) -> None:
         """Log message to Modeller."""
         await self._send_aes_encrypted_message(
@@ -314,7 +466,7 @@ class _WorkerMailbox(_BaseMailbox):
         _federate_logger(self)
 
         def log_message_handler(message: _BitfountMessage) -> None:
-            log_message: Dict[str, Any] = self._aes_decrypt(message)
+            log_message: _JSONDict = self._aes_decrypt(message)
             # We prepend the log message to show it's come from the Modeller
             log_message["msg"] = f"Modeller: {log_message['msg']}"
             # We remove the `federated` key to avoid recursively sending a federated
@@ -355,20 +507,51 @@ async def _send_parameter_update(
     # Deliberate access to private method here as that method shouldn't be used in
     # any other context than transport layer access.
     # noinspection PyProtectedMember
+    logger.debug(f"Sending TRAINING_UPDATE from {worker_mailbox.mailbox_id}")
     await worker_mailbox._send_aes_encrypted_message(
         parameter_update, _BitfountMessageType.TRAINING_UPDATE
     )
 
 
+async def _send_public_key(
+    public_key: RSAPublicKey, worker_mailbox: _WorkerMailbox
+) -> None:
+    """Sends the public key to the modeller.
+
+    Used for the PSI protocol.
+
+    Args:
+        public_key: The public key to use in the PSI.
+        worker_mailbox: The worker mailbox to use to send the parameter update.
+    """
+    # Serialize the public key first
+    serialized_key = public_key.public_bytes(
+        encoding=serialization.Encoding.PEM,
+        format=serialization.PublicFormat.SubjectPublicKeyInfo,
+    )
+    await worker_mailbox._send_aes_encrypted_message(
+        serialized_key, _BitfountMessageType.KEY_EXCHANGE
+    )
+
+
+async def _send_psi_dataset_worker(
+    dataset: Tuple[List[str], List[str]], worker_mailbox: _WorkerMailbox
+) -> None:
+    """Sends psi datasets to the modeller."""
+    await worker_mailbox._send_aes_encrypted_message(
+        dataset, _BitfountMessageType.PSI_DATASET
+    )
+
+
 async def _get_model_parameters(
-    worker_mailbox: _WorkerMailbox, timeout: Optional[int] = None
+    worker_mailbox: _WorkerMailbox, timeout: Optional[int] = _SOFT_LIMIT_MESSAGE_TIMEOUT
 ) -> _SerializedWeights:
     """Awaits for the next model parameter update from the modeller.
 
     Args:
         worker_mailbox: The worker mailbox that the message will be sent to.
         timeout: Optional. The time to wait in seconds for the next model
-                 parameter update.
+            parameter update.
 
     Returns:
         The received parameter update.
@@ -384,6 +567,25 @@ async def _get_model_parameters(
     return weights
 
 
+async def _get_psi_dataset(
+    worker_mailbox: _WorkerMailbox, timeout: Optional[int] = _SOFT_LIMIT_MESSAGE_TIMEOUT
+) -> List[str]:
+    """Awaits for the psi dataset from the modeller.
+
+    Args:
+        worker_mailbox: The worker mailbox that the message will be sent to.
+        timeout: Optional. The time to wait in seconds for the next model
+            parameter update.
+
+    Returns:
+        The psi dataset.
+    """
+    psi_dataset: List[str] = await worker_mailbox._get_message_and_decrypt(
+        _BitfountMessageType.PSI_DATASET, timeout=timeout
+    )
+    return psi_dataset
+
+
 class _InterPodWorkerMailbox(_WorkerMailbox):
     """A worker mailbox that also handles pod-to-pod communication.
 
@@ -457,6 +659,7 @@ class _InterPodWorkerMailbox(_WorkerMailbox):
                 recipient_mailbox_id=recipient_mailbox_id,
                 sender=self.pod_identifier,
                 sender_mailbox_id=self.mailbox_id,
+                task_id=self._task_id,
             ),
             already_packed=True,
         )
@@ -510,7 +713,8 @@ async def _get_worker_secure_shares(
         timeout: Optional. The number of seconds to wait for secure shares to arrive.
 
     Returns:
-        The list of received shares.
+        The list of received shares. Note that there is no notion of "worker order"
+        in this list, the shares will be in the order they were received.
     """
     shares: List[int] = []
 
