@@ -1,11 +1,11 @@
 """Typed wrappers around the Bitfount REST apis."""
 from __future__ import annotations
 
-import dataclasses
 from dataclasses import dataclass
+import hashlib
 import logging
 from pathlib import Path
-from typing import TYPE_CHECKING, Any, Final, List, Mapping, Optional, Tuple, Type, cast
+from typing import Any, Final, List, Mapping, Optional, Tuple, Type, cast
 import warnings
 
 from cryptography.exceptions import InvalidSignature
@@ -25,13 +25,24 @@ from bitfount.config import (
 )
 from bitfount.data.schema import BitfountSchema
 from bitfount.federated.encryption import _RSAEncryption
-from bitfount.federated.types import _PodResponseType
+from bitfount.federated.types import SerializedProtocol, _PodResponseType
 from bitfount.hub.authentication_flow import BitfountSession
+from bitfount.hub.exceptions import (
+    ModelTooLargeError,
+    ModelUploadError,
+    ModelValidationError,
+    SchemaUploadError,
+)
 from bitfount.hub.types import (
+    _DEV_AM_URL,
+    _STAGING_AM_URL,
+    PRODUCTION_AM_URL,
+    PRODUCTION_HUB_URL,
     _AccessManagerKeyResponseJSON,
     _AMAccessCheckResponseJSON,
     _ModelDetailsResponseJSON,
     _ModelUploadResponseJSON,
+    _MonitorPostJSON,
     _MultiModelDetailsResponseJSON,
     _MultiPodDetailsResponseJSON,
     _OIDCAccessCheckPostJSON,
@@ -39,6 +50,7 @@ from bitfount.hub.types import (
     _PodRegistrationResponseJSON,
     _SAMLAdditionalInfoPOSTJSON,
     _SAMLChallengeResponseJSON,
+    _SignatureBasedAccessCheckPostJSON,
 )
 from bitfount.models.bitfount_model import BitfountModel
 from bitfount.storage import (
@@ -55,10 +67,11 @@ from bitfount.types import (
     _S3PresignedURL,
     _SAMLResponse,
 )
-from bitfount.utils import _get_mb_from_bytes, _get_non_abstract_classes_from_module
-
-if TYPE_CHECKING:
-    from bitfount.federated.task_requests import _ProtocolDetails
+from bitfount.utils import (
+    _get_mb_from_bytes,
+    _get_non_abstract_classes_from_module,
+    web_utils,
+)
 
 logger = logging.getLogger(__name__)
 
@@ -66,25 +79,6 @@ logger = logging.getLogger(__name__)
 # TODO: [BIT-1443] Remove this once Hub/AM support IPv6
 requests.packages.urllib3.util.connection.HAS_IPV6 = False  # type: ignore[attr-defined] # Reason: see above # noqa: B950
 
-PRODUCTION_HUB_URL: Final[str] = "https://hub.bitfount.com"
-PRODUCTION_AM_URL: Final[str] = "https://am.bitfount.com"
-_STAGING_HUB_URL: Final[str] = "https://hub.staging.bitfount.com"
-_STAGING_AM_URL: Final[str] = "https://am.staging.bitfount.com"
-_DEV_HUB_URL: Final[str] = "http://localhost:3000"
-_DEV_AM_URL: Final[str] = "http://localhost:3001"
-
-_PRODUCTION_IDP_URL: Final[str] = (
-    "https://prod-bitfount.eu.auth0.com/"
-    "samlp/XJAxDKePxvyTKYdwPZMGjTwUAkX7Hts0?SAMLRequest="
-)
-_STAGING_IDP_URL: Final[str] = (
-    "https://dev-bitfount.eu.auth0.com/"
-    "samlp/kLEmmLCZB8frJsimz6A30DZFpvAUbrDt?SAMLRequest="
-)
-_DEV_IDP_URL: Final[str] = (
-    "https://dev-bitfount.eu.auth0.com/"
-    "samlp/FUo7nqEpUlARg11KE9WbYA0trvPJJ5ms?SAMLRequest="
-)
 
 # This should match the corresponding value in
 # bitfount/bitfount-web:bf-packages/bitfount-hub/pages/api/pods.api.js
@@ -108,7 +102,17 @@ class PodPublicMetadata:
     display_name: str
     description: str
     schema: _JSONDict  # This should be set to schema.to_json()
-    is_public: bool = False
+
+
+def hash_file_contents(file_path: Path) -> str:
+    """Hash file contents using sha256."""
+    sha256_hash = hashlib.sha256()
+    with open(file_path, "rb") as f:
+        # Read and update hash string value in blocks of 4K
+        for byte_block in iter(lambda: f.read(4096), b""):
+            sha256_hash.update(byte_block)
+
+    return sha256_hash.hexdigest()
 
 
 def _verify_signature(
@@ -380,7 +384,7 @@ class BitfountHub:
 
         return pod_key
 
-    def send_model(self, model_code_path: Path) -> bool:
+    def send_model(self, model_code_path: Path) -> None:
         """Sends the provided `model_code` to Hub, associated to session username.
 
         The name of the model is taken from the name of the file.
@@ -388,24 +392,26 @@ class BitfountHub:
         Args:
             model_code_path: The path to the file containing the model code.
 
-        Returns:
-            True if model was uploaded successfully, False otherwise.
+        Raises:
+            ModelUploadError: If the model upload fails. Specific subclasses of
+                ModelUploadError are used to indicate different failure conditions.
         """
         # Check model is correctly formed
         try:
             self._verify_bitfount_model_format(model_code_path)
         except (ValueError, TypeError) as e:
             logger.error(f"Model incorrectly structured. Error: {e}")
-            return False
+            raise ModelValidationError("Model incorrectly structured.") from e
         except ImportError as ie:
             logger.error(ie)
-            return False
+            raise ModelValidationError("Unable to import model.") from ie
 
         model_name: str = model_code_path.stem
         model_code_text: bytes = model_code_path.read_text().encode("utf-8")
         model_size = len(model_code_text)
+        model_hash = hash_file_contents(model_code_path)
         if model_size > _MAX_CUSTOM_MODEL_SIZE_BYTES:
-            raise ValueError(
+            raise ModelTooLargeError(
                 f"Model is too large to upload: "
                 f"expected max {_MAX_CUSTOM_MODEL_SIZE_MEGABYTES} megabytes, "
                 f"got {_get_mb_from_bytes(model_size).fractional} megabytes."
@@ -419,6 +425,7 @@ class BitfountHub:
                 json={
                     "modelName": model_name,
                     "modelSize": model_size,
+                    "modelHash": model_hash,
                 },
             )
             response.raise_for_status()
@@ -430,7 +437,9 @@ class BitfountHub:
                 # response variable wasn't created yet
                 err_response = None
             self._handle_hub_request_exception(err, err_response)
-            return False
+            raise ModelUploadError(
+                "Request exception occurred when uploading model details to hub."
+            ) from err
 
         # Extract response
         response_json: _ModelUploadResponseJSON = _extract_response_json(response)
@@ -439,7 +448,10 @@ class BitfountHub:
                 f"Could not send model to Bitfount Hub. "
                 f"Failed with message: {response_json['errorMessage']}"
             )
-            return False
+            raise ModelUploadError(
+                f"Failed to upload model details to hub: "
+                f"{response_json['errorMessage']}"
+            )
         logger.info(
             "Model successfully registered to Bitfount Hub, "
             "uploading to object storage."
@@ -448,21 +460,37 @@ class BitfountHub:
         # Extract URL to upload model to and upload it
         upload_url: _S3PresignedPOSTURL = response_json["uploadUrl"]
         upload_fields: _S3PresignedPOSTFields = response_json["uploadFields"]
-        _upload_file_to_s3(
-            upload_url,
-            upload_fields,
-            file_contents=model_code_text,
-            file_name=model_code_path.name,
-        )
-        logger.info("Model successfully uploaded to object storage.")
+        try:
+            _upload_file_to_s3(
+                upload_url,
+                upload_fields,
+                file_contents=model_code_text,
+                file_name=model_code_path.name,
+            )
+        except RequestException as re:
+            logger.error(f"Failed to upload model to S3: {re}")
+            raise ModelUploadError("Failed to upload model to S3") from re
 
-        return True
+        logger.info("Model successfully uploaded to object storage.")
+        return None
 
     @staticmethod
     def _verify_bitfount_model_format(path_to_model: Path) -> Type[BitfountModel]:
-        """Verifies that the the model file is correctly formatted.
+        """Verifies that the model file is correctly formatted.
+
+        Args:
+            path_to_model: Path to the model file.
+
+        Returns:
+            The parsed model class.
+
+        Raises:
+            ImportError: If the file could not be loaded.
+            TypeError: If there is no BitfountModel subclass in the file, or it
+                is still abstract.
+            ValueError: If multiple model classes are contained in the file.
+            ValueError: If the name of the class and the name of the file differ.
 
-        Returns the model class
         """
         models = _get_non_abstract_classes_from_module(path_to_model)
         models = {
@@ -493,25 +521,29 @@ class BitfountHub:
         return model[1]
 
     def get_model(
-        self, username: str, model_name: str
+        self, username: str, model_name: str, model_version: Optional[int] = None
     ) -> Optional[Type[BitfountModel]]:
         """Gets model code corresponding to `model_name` from user `username`.
 
         Args:
             username: The model's owner.
             model_name: The name of the model.
+            model_version: (Optional) The version of the model. Defaults to latest.
 
         Returns:
             The loaded Bitfount model class or None if unable to retrieve it.
         """
         # Retrieve model details from API endpoint
         try:
+            params = {
+                "modellerName": username,
+                "modelName": model_name,
+            }
+            if model_version:
+                params.update({"modelVersion": str(model_version)})
             response = self.session.get(
                 f"{self.url}/api/models",
-                params={
-                    "modellerName": username,
-                    "modelName": model_name,
-                },
+                params=params,
             )
             response.raise_for_status()
         except RequestException as err:
@@ -546,11 +578,17 @@ class BitfountHub:
         # easier to import and verify.
         # This will be a str because we provide an encoding arg.
         model_code: str = cast(str, _download_file_from_s3(model_url, encoding="utf-8"))
-        # TODO: [BIT-1073] Reintroduce model hash verification at this point in
-        #  the download process.
+
         model_file = self._write_model_code_to_file(
             model_code, self.user_storage_path, model_name
         )
+
+        file_hash = hash_file_contents(model_file)
+        if file_hash != response_json["modelHash"]:
+            raise ModelValidationError(
+                "Stored hash does not match hashed model file contents. "
+                "Model must be re-uploaded to hub."
+            )
         try:
             model_class = self._verify_bitfount_model_format(model_file)
         except (ValueError, TypeError, ImportError) as e:
@@ -594,12 +632,12 @@ class BitfountHub:
 
         Raises:
             RequestException: If unable to connect to the hub.
-            HTTPError: If the response if not successful.
+            HTTPError: If the response is not successful.
         """
         # Need to determine how large it will be once packed and uploaded to S3
         schema_size = _get_packed_data_object_size(public_metadata.schema)
         if schema_size > _MAX_SCHEMA_SIZE_BYTES:
-            raise ValueError(
+            raise SchemaUploadError(
                 f"Schema is too large to upload: "
                 f"expected max {_MAX_SCHEMA_SIZE_MEGABYTES} megabytes, "
                 f"got {_get_mb_from_bytes(schema_size).fractional} megabytes."
@@ -615,7 +653,6 @@ class BitfountHub:
                 Encoding.OpenSSH, PublicFormat.OpenSSH
             ).decode("utf-8"),
             "description": public_metadata.description,
-            "isPublic": public_metadata.is_public,
             "schemaSize": schema_size,
         }
 
@@ -731,6 +768,32 @@ class BitfountHub:
         schema_json: _JSONDict = _download_data_from_s3(schema_url)
         return BitfountSchema.load(schema_json)
 
+    def send_monitor_update(self, update: _MonitorPostJSON) -> None:
+        """Send an update to the monitor service.
+
+        Args:
+            update: The monitor service update to send as a JSON dict.
+
+        Raises:
+            RequestException: If unable to connect to the hub.
+            HTTPError: If the response is not successful.
+        """
+        try:
+            response = self.session.post(
+                f"{self.url}/api/ingest", json=update, timeout=2
+            )
+            response.raise_for_status()
+        except RequestException as err:
+            try:
+                # noinspection PyUnboundLocalVariable
+                err_response: Optional[Response] = response
+            except NameError:
+                # response variable wasn't created yet
+                err_response = None
+            err_msg = self._handle_hub_request_exception(err, err_response)
+            # Raise the same type of exception that was caught
+            raise type(err)(err_msg) from err
+
 
 class BitfountAM:
     """Typed API for communicating with the Bitfount Access Manager.
@@ -792,7 +855,7 @@ class BitfountAM:
                 f"Retrieving access manager public key from "
                 f"{self.access_manager_url}/api/access-manager-key"
             )
-            response = requests.get(
+            response = web_utils.get(
                 f"{self.access_manager_url}/api/access-manager-key",
                 timeout=5,
             )
@@ -837,7 +900,7 @@ class BitfountAM:
         """
         try:
             response = self.session.get(
-                f"{self.access_manager_url}/api/saml", timeout=10
+                f"{self.access_manager_url}/api/saml?idp=cli", timeout=10
             )
             response.raise_for_status()
         except RequestException as err:
@@ -869,7 +932,7 @@ class BitfountAM:
         saml_request_id: str,
         pod_identifier: str,
         modeller_name: str,
-        task_protocol_details: _ProtocolDetails,
+        serialized_protocol: SerializedProtocol,
     ) -> _PodResponseType:
         """Check if modeller task should be accepted.
 
@@ -884,8 +947,8 @@ class BitfountAM:
                 stored this.
             pod_identifier: The pod identifier of the pod that the task is for
             modeller_name: The username of the modeller that has sent the task
-            task_protocol_details: The protocol details portion of the task request from
-                the Modeller
+            serialized_protocol: The serialized protocol to use for the task received
+                from the modeller.
 
         Returns:
             Response code from the Access manager.
@@ -900,19 +963,19 @@ class BitfountAM:
             originalSAMLRequestID=saml_request_id,
             podIdentifier=pod_identifier,
             modellerName=modeller_name,
-            modellerRequest=dataclasses.astuple(task_protocol_details),
+            modellerProtocolRequest=serialized_protocol,
             identityProvider="SAML",
         )
         post_json: _JSONDict = dict(saml_response, **additional_info)
 
         return self._do_authorization_check(
-            f"{self.access_manager_url}/api/access", post_json
+            f"{self.access_manager_url}/api/access?idp=cli", post_json
         )
 
     def check_oidc_access_request(
         self,
         pod_identifier: str,
-        task_protocol_details: _ProtocolDetails,
+        serialized_protocol: SerializedProtocol,
         modeller_name: str,
         modeller_access_token: str,
     ) -> _PodResponseType:
@@ -920,7 +983,7 @@ class BitfountAM:
 
         Args:
             pod_identifier: The pod identifier for the request.
-            task_protocol_details: The details of the request.
+            serialized_protocol: The protocol received from the modeller.
             modeller_name: The name of the modeller.
             modeller_access_token: OAuth access token for the modeller.
 
@@ -933,13 +996,56 @@ class BitfountAM:
             HTTPError: If it is unable to extract the JSON from the response.
             InvalidJSONError: If JSON does not match expected format.
         """
-        post_json: _OIDCAccessCheckPostJSON = {
-            "podIdentifier": pod_identifier,
-            "modellerRequest": dataclasses.astuple(task_protocol_details),
-            "modellerName": modeller_name,
-            "modellerToken": modeller_access_token,
-            "identityProvider": "OIDC",
-        }
+        post_json = _OIDCAccessCheckPostJSON(
+            podIdentifier=pod_identifier,
+            modellerProtocolRequest=serialized_protocol,
+            modellerName=modeller_name,
+            modellerToken=modeller_access_token,
+            identityProvider="OIDC",
+        )
+
+        return self._do_authorization_check(
+            f"{self.access_manager_url}/api/access", post_json
+        )
+
+    def check_signature_based_access_request(
+        self,
+        unsigned_task: bytes,
+        signed_task: bytes,
+        pod_identifier: str,
+        serialized_protocol: SerializedProtocol,
+        modeller_name: str,
+    ) -> _PodResponseType:
+        """Check access by verifying the signed task against the unsigned task.
+
+        Check that when modeller_name signs the unsigned task it matches the signed
+        version of the task.
+
+        Args:
+            unsigned_task: Unsigned task.
+            signed_task: Signed task.
+            pod_identifier: The pod identifier for the request.
+            serialized_protocol: The protocol received from the modeller.
+            modeller_name: The name of the modeller.
+            modeller_access_token: OAuth access token for the modeller.
+
+        Returns:
+            Response code from the Access manager.
+
+        Raises:
+            RequestException: If there is a problem communicating with the
+                access manager.
+            HTTPError: If it is unable to extract the JSON from the response.
+            InvalidJSONError: If JSON does not match expected format.
+        """
+        post_json = _SignatureBasedAccessCheckPostJSON(
+            podIdentifier=pod_identifier,
+            modellerRequest=serialized_protocol,
+            modellerName=modeller_name,
+            unsignedTask=unsigned_task,
+            signedTask=signed_task,
+            identityProvider="SIGNATURE",
+        )
 
         return self._do_authorization_check(
             f"{self.access_manager_url}/api/access", post_json
@@ -989,7 +1095,7 @@ class BitfountAM:
             try:
                 return _PodResponseType[response_json["code"]]
             except KeyError:
-                return _PodResponseType.ERROR_IN_VERIFICATION
+                return _PodResponseType.NO_ACCESS
             except TypeError:
                 raise AttributeError()
         except AttributeError:
