@@ -8,48 +8,70 @@ from __future__ import annotations
 from abc import ABC, abstractmethod
 from dataclasses import dataclass, field
 from enum import Enum, auto
+import inspect
 import logging
 import os
 from pathlib import Path
+from types import MappingProxyType
 from typing import (
     TYPE_CHECKING,
     Any,
     Callable,
+    ClassVar,
     Dict,
     Generic,
     List,
+    Mapping,
     MutableMapping,
     Optional,
     Tuple,
     Type,
     TypeVar,
     Union,
-    cast,
 )
 
 import desert
-from marshmallow import Schema as MarshmallowSchema
-from marshmallow import fields, post_load
+from marshmallow import fields
 import numpy as np
 
 from bitfount.config import BITFOUNT_LOGS_DIR
-from bitfount.data.databunch import _BitfountDataBunch
-from bitfount.data.dataloader import _BitfountDataLoader
-from bitfount.data.datasource import DataSource
+from bitfount.data.databunch import BitfountDataBunch
+from bitfount.data.dataloaders import BitfountDataLoader
+from bitfount.data.datasources.base_source import BaseSource
 from bitfount.data.datastructure import DataStructure
+from bitfount.data.datastructure import registry as datastructure_registry
 from bitfount.data.schema import BitfountSchema, TableSchema
-from bitfount.types import DistributedModelProtocol, _JSONDict
-from bitfount.utils import _add_this_to_list, seed_all
+from bitfount.types import (
+    T_FIELDS_DICT,
+    T_NESTED_FIELDS,
+    T_OPTIONAL_NESTED,
+    DistributedModelProtocol,
+    _BaseSerializableObjectMixIn,
+    _StrAnyDict,
+)
+from bitfount.utils import _add_this_to_list, delegates, seed_all
 
 if TYPE_CHECKING:
     from bitfount.metrics import Metric
 
+
 # Valid pooling functions for Convolutional Neural Networks.
 POOLING_FUNCTIONS = ["max", "avg"]
 
 
 logger = logging.getLogger(__name__)
 
+# Main Model registry that is used to collect all model classes.
+# Used for serialization and deserialization of model classes.
+# Includes both abstract and non-abstract classes.
+MAIN_MODEL_REGISTRY: _StrAnyDict = {}
+
+# MYPY ISSUES
+# Due to the use of desert.field to create non-standard fields in these dataclasses,
+# mypy complains about incompatible type assignment. This is a known bug in desert
+# (https://github.com/python-desert/desert/issues/101) and so we can ignore it
+# until the issue is resolved.
+
 
 @dataclass
 class LoggerConfig:
@@ -67,9 +89,6 @@ class LoggerConfig:
             empty dictionary
     """
 
-    # Desert bug causing incompatability with mypy regarding type assignment
-    # (https://github.com/python-desert/desert/issues/101)
-
     #: same as argument
     name: str
     #: same as argument
@@ -78,11 +97,11 @@ class LoggerConfig:
             deserialize=lambda path: path if path is None else Path(path).expanduser()
         ),
         default=BITFOUNT_LOGS_DIR,
-    )  # type: ignore[assignment] # Reason: see above
+    )  # type: ignore[assignment] # Reason: see top of file
     #: same as argument
-    params: Optional[_JSONDict] = desert.field(
+    params: Optional[_StrAnyDict] = desert.field(
         fields.Dict(keys=fields.Str), default_factory=dict
-    )  # type: ignore[assignment] # Reason: see above
+    )  # type: ignore[assignment] # Reason: see top of file
 
 
 class ModelContext(Enum):
@@ -97,9 +116,31 @@ class ModelContext(Enum):
 
 
 ModelType = TypeVar("ModelType", bound="_BaseModel")
+_model_structure_registry: Dict[str, Type[_BaseModelStructureMixIn]] = {}
+model_structure_registry: Mapping[
+    str, Type[_BaseModelStructureMixIn]
+] = MappingProxyType(_model_structure_registry)
 
 
-class _BaseModel(ABC, Generic[ModelType]):
+@dataclass
+class _BaseModelStructureMixIn:
+    @classmethod
+    def __init_subclass__(cls, **kwargs: Any):
+        if not inspect.isabstract(cls):
+            logger.debug(f"Adding {cls.__name__}: {cls} to registry")
+            _model_structure_registry[cls.__name__] = cls
+
+
+class _BaseModelRegistryMixIn:
+    @classmethod
+    def __init_subclass__(cls, **kwargs: Any):
+        if cls.__name__ not in MAIN_MODEL_REGISTRY:
+            MAIN_MODEL_REGISTRY[cls.__name__] = cls
+
+
+class _BaseModel(
+    _BaseModelRegistryMixIn, _BaseSerializableObjectMixIn, ABC, Generic[ModelType]
+):
     """Abstract Base Model from which all other models must inherit.
 
     This class is designed to be at the very bottom of the inheritance hierarchy.
@@ -107,31 +148,58 @@ class _BaseModel(ABC, Generic[ModelType]):
     other classes defined in other libraries. It also takes kwargs so that we do not
     throw an error if there are unexpected keyword arguments. These unexpected keyword
     arguments will end up in this constructor where they will simply be ignored.
+
+    Args:
+        datastructure: `DataStructure` to be passed to the model when initialised
+        schema: The `BitfountSchema` object associated with the datasource
+            on which the model will be trained on.
+        seed: Random number seed. Used for setting random seed for all libraries.
+            Defaults to None.
+        param_clipping: Arguments for clipping for BatchNorm parameters.
+            Used for federated models with secure aggregation.
+            It should contain the SecureShare variables and the
+            number of workers in a dictionary,
+            e.g. {"prime_q":13, "precision": 10**3,"num_workers":2}.
+            Defaults to None.
+
     """
 
+    fields_dict: ClassVar[T_FIELDS_DICT] = {
+        "seed": fields.Integer(allow_none=True),
+        "schema": fields.Nested(BitfountSchema._Schema),
+        "param_clipping": fields.Dict(
+            keys=fields.String(), values=fields.Integer(), allow_none=True
+        ),
+    }
+    # TODO: [BIT-1623] BitfountSchema here temporarily as a nested field,
+    #  should be removed following the RFC
+    nested_fields: ClassVar[T_NESTED_FIELDS] = {"datastructure": datastructure_registry}
+
     def __init__(
         self,
         datastructure: DataStructure,
         schema: BitfountSchema,
         seed: Optional[int] = None,
+        param_clipping: Optional[Dict[str, int]] = None,
         **kwargs: Any,
     ):
-        self.name = type(self).__name__
+        self.class_name = f"bitfount.{type(self).__name__}"
         self._context: Optional[ModelContext] = None
         self.metrics: Optional[MutableMapping[str, Metric]] = None
         self._model: Optional[ModelType] = None
         self._initialised: bool = False
         self.seed = seed
+        self.param_clipping = param_clipping
         seed_all(self.seed)
         self.datastructure = datastructure
-        self._databunch: _BitfountDataBunch
+        self.databunch: BitfountDataBunch
         self.schema = schema
         self._objective: str
 
         # Placeholders for dataloaders
-        self.train_dl: _BitfountDataLoader
-        self.validation_dl: Optional[_BitfountDataLoader] = None
-        self.test_dl: Optional[_BitfountDataLoader] = None
+        self.train_dl: BitfountDataLoader
+        self.validation_dl: Optional[BitfountDataLoader] = None
+        self.test_dl: Optional[BitfountDataLoader] = None
 
         for unexpected_kwarg in kwargs:
             logger.warning(f"Ignoring unexpected keyword argument {unexpected_kwarg}")
@@ -147,18 +215,18 @@ class _BaseModel(ABC, Generic[ModelType]):
         Args:
             batch_size: The batch size to use for the dataloaders. Defaults to None.
         """
-        if self._databunch is None:
+        if self.databunch is None:
             raise ValueError(
                 "_set_dataloaders() requires the databunch to be set "
                 "before being called."
             )
-        self.train_dl = self._databunch.get_train_dataloader(batch_size)
-        self.validation_dl = self._databunch.get_validation_dataloader(batch_size)
-        self.test_dl = self._databunch.get_test_dataloader(batch_size)
+        self.train_dl = self.databunch.get_train_dataloader(batch_size)
+        self.validation_dl = self.databunch.get_validation_dataloader(batch_size)
+        self.test_dl = self.databunch.get_test_dataloader(batch_size)
 
     def _add_datasource_to_schema(
         self,
-        datasource: DataSource,
+        datasource: BaseSource,
     ) -> None:
         """Adds the datasource to the schema.
 
@@ -168,10 +236,10 @@ class _BaseModel(ABC, Generic[ModelType]):
         # TODO: [BIT-1167] Once we add the transformations we can force the target
         #  to be categorical in the transformations as this can be specified directly
         if self.datastructure.image_cols:
-            if not datasource._data_is_loaded:
-                datasource.load_data()
+            datasource.load_data()
+            datasource_columns = datasource.get_dtypes()
             for image in self.datastructure.image_cols:
-                if image not in datasource.data.columns:
+                if image not in datasource_columns:
                     raise ValueError(f"Could not find {image} in dataset columns")
 
         # If the model is distributed, then we need to get the table name that
@@ -179,7 +247,7 @@ class _BaseModel(ABC, Generic[ModelType]):
         pod_identifier: Optional[str] = None
         if hasattr(self, "pod_identifier"):
             # Assuring mypy that the attribute exists
-            assert isinstance(self, DistributedModelProtocol)  # nosec
+            assert isinstance(self, DistributedModelProtocol)  # nosec[assert_used]
             pod_identifier = self.pod_identifier
         table_name = self.datastructure.get_table_name(pod_identifier)
 
@@ -193,7 +261,7 @@ class _BaseModel(ABC, Generic[ModelType]):
             force_stypes={table_name: self.datastructure._force_stype},
             table_name=table_name,
         )
-        self._databunch = _BitfountDataBunch(
+        self.databunch = BitfountDataBunch(
             data_structure=self.datastructure,
             schema=self.schema.get_table_schema(table_name),
             datasource=datasource,
@@ -201,7 +269,7 @@ class _BaseModel(ABC, Generic[ModelType]):
 
     def initialise_model(
         self,
-        data: Optional[DataSource] = None,
+        data: Optional[BaseSource] = None,
         context: Optional[ModelContext] = None,
     ) -> None:
         """Can be implemented to initialise model if necessary.
@@ -238,7 +306,7 @@ class _BaseModel(ABC, Generic[ModelType]):
     @abstractmethod
     def evaluate(
         self,
-        test_dl: Optional[_BitfountDataLoader] = None,
+        test_dl: Optional[BitfountDataLoader] = None,
         pod_identifiers: Optional[List[str]] = None,
         **kwargs: Any,
     ) -> Union[Tuple[np.ndarray, np.ndarray], Dict[str, float]]:
@@ -246,7 +314,7 @@ class _BaseModel(ABC, Generic[ModelType]):
 
         Args:
             test_dl: Optional `BitfountDataLoader` object containing test data. If this
-                is not provided, the test set from the `DataSource` used to train the
+                is not provided, the test set from the `BaseSource` used to train the
                 model should be used if present.
             pod_identifiers: A list of pod identifiers to use for remote evaluation.
 
@@ -255,11 +323,11 @@ class _BaseModel(ABC, Generic[ModelType]):
         """
         raise NotImplementedError
 
-    def predict(self, data: DataSource, **kwargs: Any) -> np.ndarray:
+    def predict(self, data: BaseSource, **kwargs: Any) -> np.ndarray:
         """This method runs inference on the test data, returns predictions.
 
         Args:
-            data: `DataSource` object containing the data to run prediction on.
+            data: `BaseSource` object containing the data to run prediction on.
                 Predictions will be generated for the test subset (as defined
                 by the `DataSetSplitter`).
 
@@ -271,7 +339,7 @@ class _BaseModel(ABC, Generic[ModelType]):
     @abstractmethod
     def fit(
         self,
-        data: Optional[DataSource] = None,
+        data: Optional[BaseSource] = None,
         metrics: Optional[Dict[str, Metric]] = None,
         pod_identifiers: Optional[List[str]] = None,
         **kwargs: Any,
@@ -289,32 +357,8 @@ class _BaseModel(ABC, Generic[ModelType]):
         """
         raise NotImplementedError
 
-    @property
-    @abstractmethod
-    def training_needed(self) -> bool:
-        """Dictates whether the model needs training."""
-        raise NotImplementedError
 
-    class _Schema(MarshmallowSchema):
-        name = fields.Str()
-        datastructure = fields.Nested(DataStructure._Schema)
-        schema = fields.Nested(BitfountSchema._Schema, allow_none=True)
-        seed = fields.Integer(allow_none=True)
-
-        @abstractmethod
-        @post_load
-        def recreate_model(self, data: _JSONDict, **_kwargs: Any) -> _BaseModel:
-            """Recreates model using Schema."""
-            raise NotImplementedError
-
-    @classmethod
-    @abstractmethod
-    def get_schema(cls) -> Type[_BaseModel._Schema]:
-        """Get the model schema."""
-        raise NotImplementedError
-
-
-class ClassifierMixIn:
+class ClassifierMixIn(_BaseModelRegistryMixIn, _BaseSerializableObjectMixIn):
     """MixIn for classification problems.
 
     Classification models must have this class in their inheritance hierarchy.
@@ -322,6 +366,12 @@ class ClassifierMixIn:
     Args:
         multilabel: Whether the problem is a multi-label problem. i.e. each datapoint
             belongs to multiple classes
+        param_clipping: Arguments for clipping for BatchNorm parameters.
+            Used for federated models with secure aggregation.
+            It should contain the SecureShare variables and the
+            number of workers in a dictionary,
+            e.g. {"prime_q":13, "precision": 10**3,"num_workers":2}
+
 
     Attributes:
         multilabel: Whether the problem is a multi-label problem
@@ -332,10 +382,23 @@ class ClassifierMixIn:
     datastructure: DataStructure
     #: set in _BaseModel
     schema: BitfountSchema
+    fields_dict: ClassVar[T_FIELDS_DICT] = {
+        "multilabel": fields.Bool(),
+        "param_clipping": fields.Dict(
+            keys=fields.String(), values=fields.Integer(), allow_none=True
+        ),
+    }
+    nested_fields: ClassVar[T_NESTED_FIELDS] = {"datastructure": datastructure_registry}
 
-    def __init__(self, multilabel: bool = False, **kwargs: Any) -> None:
+    def __init__(
+        self,
+        multilabel: bool = False,
+        param_clipping: Optional[Dict[str, int]] = None,
+        **kwargs: Any,
+    ) -> None:
         super().__init__(**kwargs)
         self.multilabel = multilabel
+        self.param_clipping = param_clipping
         self.n_classes: int
         if "n_classes" in kwargs:
             self.n_classes = int(kwargs["n_classes"])
@@ -371,7 +434,7 @@ class ClassifierMixIn:
 
     def _add_datasource_to_schema(
         self,
-        datasource: DataSource,
+        datasource: BaseSource,
     ) -> None:
         """Adds the datasource to the schema.
 
@@ -386,11 +449,13 @@ class ClassifierMixIn:
             self.datastructure.multihead_col, force_categorical
         )
         if self.datastructure.image_cols:
+            datasource.load_data()
+            datasource_columns = datasource.get_dtypes()
             for image in self.datastructure.image_cols:
-                if image not in datasource.data.columns:
+                if image not in datasource_columns:
                     raise ValueError(f"Could not find {image} in dataset columns")
 
-        if "categorical" not in self.datastructure._force_stype.keys():
+        if "categorical" not in self.datastructure._force_stype:
             self.datastructure._force_stype["categorical"] = force_categorical
         else:
             _add_this_to_list(
@@ -407,17 +472,14 @@ class ClassifierMixIn:
             force_stypes={table_name: self.datastructure._force_stype},
             table_name=table_name,
         )
-        self._databunch = _BitfountDataBunch(
+        self.databunch = BitfountDataBunch(
             data_structure=self.datastructure,
             schema=self.schema.get_table_schema(table_name),
             datasource=datasource,
         )
 
-    class _Schema(MarshmallowSchema):
-        multilabel = fields.Bool()
 
-
-class RegressorMixIn:
+class RegressorMixIn(_BaseModelRegistryMixIn, _BaseSerializableObjectMixIn):
     """MixIn for regression problems.
 
     Currently, just used for tagging purposes.
@@ -427,7 +489,9 @@ class RegressorMixIn:
 
 
 @dataclass
-class NeuralNetworkModelStructure(ABC):
+class NeuralNetworkModelStructure(
+    ABC, _BaseModelStructureMixIn, _BaseSerializableObjectMixIn
+):
     """Dataclass defining the structure of a neural network model.
 
     Args:
@@ -454,10 +518,17 @@ class NeuralNetworkModelStructure(ABC):
     #: same as argument
     mish_activation_function: bool = True
 
+    fields_dict = {
+        "layers": fields.List(fields.Int()),
+        "dropout_probs": fields.List(fields.Float()),
+        "mish_activation_function": fields.Bool(),
+    }  # type: ClassVar[T_FIELDS_DICT]
+    nested_fields = {}  # type: ClassVar[T_NESTED_FIELDS]
+
     def __post_init__(self) -> None:
+        self.class_name = type(self).__name__
         if self.layers is None:
             self.layers = [1000, 500]
-
         if self.dropout_probs is None:
             self.dropout_probs = [0.01, 0.1]
         if len(self.layers) != len(self.dropout_probs):
@@ -485,14 +556,12 @@ class NeuralNetworkModelStructure(ABC):
             else 1
         )
 
-    class _Schema(MarshmallowSchema):
-        layers = fields.List(fields.Int())
-        dropout_probs = fields.List(fields.Float())
-        mish_activation_function = fields.Bool()
-
 
+@delegates()
 @dataclass
-class FeedForwardModelStructure(NeuralNetworkModelStructure):
+class FeedForwardModelStructure(
+    NeuralNetworkModelStructure, _BaseModelStructureMixIn, _BaseSerializableObjectMixIn
+):
     """Dataclass defining the structure of a feedforward neural network model.
 
     This model structure only defines linear and dropout layers.
@@ -503,20 +572,17 @@ class FeedForwardModelStructure(NeuralNetworkModelStructure):
 
     #: same as argument
     embedding_dropout: float = 0.04
-
-    class _Schema(NeuralNetworkModelStructure._Schema):
-        embedding_dropout = fields.Float()
-
-        @post_load
-        def recreate_model(
-            self, data: _JSONDict, **_kwargs: Any
-        ) -> FeedForwardModelStructure:
-            """Recreates model using Schema."""
-            return FeedForwardModelStructure(**data)
+    fields_dict = {
+        "embedding_dropout": fields.Float(default=0.04)
+    }  # type: ClassVar[T_FIELDS_DICT]
+    nested_fields = {}  # type: ClassVar[T_NESTED_FIELDS]
 
 
+@delegates()
 @dataclass
-class CNNModelStructure(NeuralNetworkModelStructure):
+class CNNModelStructure(
+    NeuralNetworkModelStructure, _BaseModelStructureMixIn, _BaseSerializableObjectMixIn
+):
     """Dataclass defining the structure of a convolutional neural network model.
 
     This model structure has multiple convolutional layers followed by multiple
@@ -550,6 +616,18 @@ class CNNModelStructure(NeuralNetworkModelStructure):
     ff_layers: List[int] = field(default_factory=list)
     #: same as argument
     ff_dropout_probs: List[float] = field(default_factory=list)
+    fields_dict = {
+        "kernel_size": fields.Int(),
+        "padding": fields.Int(),
+        "stride": fields.Int(),
+        "pooling_function": fields.Str(),
+        "layers": fields.List(fields.Int()),
+        "ff_layers": fields.List(fields.Int(), allow_none=True, default=[1000, 500]),
+        "ff_dropout_probs": fields.List(
+            fields.Float(), allow_none=True, default=[0.01, 0.1]
+        ),
+    }  # type: ClassVar[T_FIELDS_DICT]
+    nested_fields = {}  # type: ClassVar[T_NESTED_FIELDS]
 
     def __post_init__(self) -> None:
         if not isinstance(self.layers, list):
@@ -568,23 +646,11 @@ class CNNModelStructure(NeuralNetworkModelStructure):
                 "Number of neural layers must equal number of dropout layers."
             )
 
-    class _Schema(NeuralNetworkModelStructure._Schema):
-
-        kernel_size = fields.Int()
-        padding = fields.Int()
-        stride = fields.Int()
-        pooling_function = fields.Str()
-        ff_layers = fields.List(fields.Int())
-        ff_dropout_probs = fields.List(fields.Float())
-
-        @post_load
-        def recreate_model(self, data: _JSONDict, **_kwargs: Any) -> CNNModelStructure:
-            """Recreates model using Schema."""
-            return CNNModelStructure(**data)
-
 
 @dataclass
-class NeuralNetworkPredefinedModel:
+class NeuralNetworkPredefinedModel(
+    _BaseModelStructureMixIn, _BaseSerializableObjectMixIn
+):
     """This class encompasses what is required to use a predefined model e.g. ResNet.
 
     The currently supported models are:
@@ -607,22 +673,34 @@ class NeuralNetworkPredefinedModel:
     pretrained: bool = True
     #: same as argument
     kwargs: Optional[Dict[str, bool]] = None
+    fields_dict = {
+        "name": fields.String(),
+        "pretrained": fields.Bool(),
+        "kwargs": fields.Dict(keys=fields.Str(), values=fields.Bool(), allow_none=True),
+    }  # type: ClassVar[T_FIELDS_DICT]
+    nested_fields = {}  # type: ClassVar[T_NESTED_FIELDS]
 
-    class _Schema(MarshmallowSchema):
-        name = fields.Str()
-        pretrained = fields.Bool()
-        kwargs = fields.Dict(keys=fields.Str(), values=fields.Bool(), allow_none=True)
+    def __post_init__(self) -> None:
+        self.class_name = type(self).__name__
 
-        @post_load
-        def recreate_model(
-            self, data: _JSONDict, **_kwargs: Any
-        ) -> NeuralNetworkPredefinedModel:
-            """Recreates model using Schema."""
-            return NeuralNetworkPredefinedModel(**data)
+
+_optimizer_registry: Dict[str, Type[_BaseOptimizer]] = {}
+optimizer_registry: Mapping[str, Type[_BaseOptimizer]] = MappingProxyType(
+    _optimizer_registry
+)
 
 
 @dataclass
-class Optimizer:
+class _BaseOptimizer:
+    @classmethod
+    def __init_subclass__(cls, **kwargs: Any):
+        if not inspect.isabstract(cls):
+            logger.debug(f"Adding {cls.__name__}: {cls} to registry")
+            _optimizer_registry[cls.__name__] = cls
+
+
+@dataclass
+class Optimizer(_BaseOptimizer, _BaseSerializableObjectMixIn):
     """Class for specifying the optimizer for a neural network.
 
     The options for the optimizer will depend on the backend being used.
@@ -635,20 +713,34 @@ class Optimizer:
     #: same as argument
     name: str
     #: same as argument
-    params: Dict[str, Any] = field(default_factory=dict)
+    params: _StrAnyDict = field(default_factory=dict)
+    fields_dict = {
+        "name": fields.String(),
+        "params": fields.Dict(),
+    }  # type: ClassVar[T_FIELDS_DICT]
+    nested_fields = {}  # type: ClassVar[T_NESTED_FIELDS]
 
-    class _Schema(MarshmallowSchema):
-        name = fields.String()
-        params = fields.Dict()
+    def __post_init__(self) -> None:
+        self.class_name = type(self).__name__
 
-        @post_load
-        def recreate(self, data: _JSONDict, **_kwargs: Any) -> Optimizer:
-            """Recreate Optimizer instance from schema."""
-            return Optimizer(**data)
+
+_scheduler_registry: Dict[str, Type[_BaseScheduler]] = {}
+scheduler_registry: Mapping[str, Type[_BaseScheduler]] = MappingProxyType(
+    _scheduler_registry
+)
 
 
 @dataclass
-class Scheduler:
+class _BaseScheduler:
+    @classmethod
+    def __init_subclass__(cls, **kwargs: Any):
+        if not inspect.isabstract(cls):
+            logger.debug(f"Adding {cls.__name__}: {cls} to registry")
+            _scheduler_registry[cls.__name__] = cls
+
+
+@dataclass
+class Scheduler(_BaseScheduler, _BaseSerializableObjectMixIn):
     """Class for specifying the scheduler for a neural network.
 
     The options for the scheduler will depend on the backend being used.
@@ -661,20 +753,34 @@ class Scheduler:
     #: same as argument
     name: str
     #: same as argument
-    params: Dict[str, Any] = field(default_factory=dict)
+    params: _StrAnyDict = field(default_factory=dict)
+    fields_dict = {
+        "name": fields.String(),
+        "params": fields.Dict(),
+    }  # type: ClassVar[T_FIELDS_DICT]
+    nested_fields = {}  # type: ClassVar[T_NESTED_FIELDS]
 
-    class _Schema(MarshmallowSchema):
-        name = fields.String()
-        params = fields.Dict()
+    def __post_init__(self) -> None:
+        self.class_name = type(self).__name__
 
-        @post_load
-        def recreate(self, data: _JSONDict, **_kwargs: Any) -> Scheduler:
-            """Recreate Scheduler instance from schema."""
-            return Scheduler(**data)
+
+_early_stopping_registry: Dict[str, Type[_BaseEarlyStopping]] = {}
+early_stopping_registry: Mapping[str, Type[_BaseEarlyStopping]] = MappingProxyType(
+    _early_stopping_registry
+)
 
 
 @dataclass
-class EarlyStopping:
+class _BaseEarlyStopping:
+    @classmethod
+    def __init_subclass__(cls, **kwargs: Any):
+        if not inspect.isabstract(cls):
+            logger.debug(f"Adding {cls.__name__}: {cls} to registry")
+            _early_stopping_registry[cls.__name__] = cls
+
+
+@dataclass
+class EarlyStopping(_BaseEarlyStopping, _BaseSerializableObjectMixIn):
     """Class for specifying early stopping parameters.
 
     Args:
@@ -683,22 +789,28 @@ class EarlyStopping:
     """
 
     #: same as argument
-    params: Dict[str, Any] = field(default_factory=dict)
+    params: _StrAnyDict = field(default_factory=dict)
+    fields_dict = {
+        "params": fields.Dict(),
+    }  # type: ClassVar[T_FIELDS_DICT]
+    nested_fields = {}  # type: ClassVar[T_NESTED_FIELDS]
 
-    class _Schema(MarshmallowSchema):
-        params = fields.Dict()
-
-        @post_load
-        def recreate(self, data: _JSONDict, **_kwargs: Any) -> EarlyStopping:
-            """Recreate EarlyStopping instance from schema."""
-            return EarlyStopping(**data)
+    def __post_init__(self) -> None:
+        self.class_name = type(self).__name__
 
 
-class NeuralNetworkMixIn:
+class NeuralNetworkMixIn(_BaseSerializableObjectMixIn):
     """Specifies model structure and hyperparameters for neural network models.
 
     All neural network models must inherit from this class.
 
+    :::caution
+
+    `custom_loss_func` cannot be serialized and therefore cannot be used in Federated
+    Learning.
+
+    :::
+
     Args:
         model_structure: The structure of the model.
         batch_size: The number of data points in each batch. Defaults to 512.
@@ -724,17 +836,24 @@ class NeuralNetworkMixIn:
 
     Raises:
         ValueError: If both `epochs` and `steps` are provided.
-
-    :::caution
-
-    `custom_loss_func` cannot be serialized and therefore cannot be used in Federated
-    Learning.
-
-    :::
     """
 
     #: set in _BaseModel
     datastructure: DataStructure
+    fields_dict: ClassVar[T_FIELDS_DICT] = {
+        "epochs": fields.Integer(allow_none=True),
+        "steps": fields.Integer(allow_none=True),
+        "batch_size": fields.Integer(allow_none=True),
+        "custom_loss_func": fields.Raw(allow_none=True),
+    }
+    nested_fields: ClassVar[T_NESTED_FIELDS] = {
+        "datastructure": datastructure_registry,
+        "model_structure": model_structure_registry,
+        "optimizer": optimizer_registry,
+        "scheduler": scheduler_registry,
+        "early_stopping": early_stopping_registry,
+    }
+    optional_nested: ClassVar[T_OPTIONAL_NESTED] = ["scheduler", "early_stopping"]
 
     def __init__(
         self,
@@ -751,7 +870,6 @@ class NeuralNetworkMixIn:
         **kwargs: Any,
     ):
         super().__init__(**kwargs)
-
         if (steps is None and epochs is None) or (
             isinstance(steps, int) and isinstance(epochs, int)
         ):
@@ -778,6 +896,11 @@ class NeuralNetworkMixIn:
         if isinstance(self.model_structure, NeuralNetworkModelStructure):
             self.model_structure.set_num_heads(self.datastructure)
 
+    @classmethod
+    def __init_subclass__(cls, **kwargs: Any):
+        if cls.__name__ not in MAIN_MODEL_REGISTRY:
+            MAIN_MODEL_REGISTRY[cls.__name__] = cls
+
     @staticmethod
     @abstractmethod
     def _get_optimizer(optimizer: Optimizer) -> Callable:
@@ -789,72 +912,3 @@ class NeuralNetworkMixIn:
     def _get_scheduler(scheduler: Scheduler) -> Callable:
         """Returns appropriate scheduler class."""
         raise NotImplementedError
-
-    @property
-    def training_needed(self) -> bool:
-        """Dictates whether the model needs training.
-
-        Returns:
-            bool: True if one of either epochs or steps are > 0.
-        """
-        return sum(filter(None, (self.epochs, self.steps))) > 0
-
-    class _Schema(MarshmallowSchema):
-        model_structure = fields.Method(
-            serialize="_nn_model_structure_serialize",
-            deserialize="_nn_model_structure_deserialize",
-        )
-        epochs = fields.Integer(allow_none=True)
-        steps = fields.Integer(allow_none=True)
-        batch_size = fields.Integer(allow_none=True)
-        optimizer = fields.Nested(Optimizer._Schema, allow_none=True)
-        scheduler = fields.Nested(Scheduler._Schema, allow_none=True)
-        early_stopping = fields.Nested(EarlyStopping._Schema, allow_none=True)
-        custom_loss_func = fields.Raw(allow_none=True)
-
-        def _nn_model_structure_serialize(
-            self, base_object: NeuralNetworkMixIn
-        ) -> _JSONDict:
-            """Custom serialization of Neural Network structure.
-
-            Required for the NeuralNetworkMixIn to know which Schema to use
-            for serialization.
-            """
-            class_to_schema = {
-                "NeuralNetworkPredefinedModel": NeuralNetworkPredefinedModel._Schema,
-                "CNNModelStructure": CNNModelStructure._Schema,
-                "FeedForwardModelStructure": FeedForwardModelStructure._Schema,
-            }
-            try:
-                schema: MarshmallowSchema = class_to_schema[
-                    base_object.model_structure.__class__.__name__
-                ]()
-                return cast(_JSONDict, schema.dump(base_object.model_structure))
-            except KeyError as e:
-                raise KeyError(f"{e}. Schema Object not recognised")
-
-        def _nn_model_structure_deserialize(
-            self, object_dict: _JSONDict
-        ) -> Union[
-            CNNModelStructure, FeedForwardModelStructure, NeuralNetworkPredefinedModel
-        ]:
-            """Custom deserialization of Neural Network structure.
-
-            Required for the NeuralNetworkMixIn to know which Schema to use
-            for deserialization.
-            """
-            if object_dict.get("kernel_size"):
-                cnn_model_structure: CNNModelStructure = (
-                    CNNModelStructure._Schema().load(object_dict)
-                )
-                return cnn_model_structure
-            elif object_dict.get("embedding_dropout"):
-                ff_model_structure: FeedForwardModelStructure = (
-                    FeedForwardModelStructure._Schema().load(object_dict)
-                )
-                return ff_model_structure
-            else:
-                predefined_model: NeuralNetworkPredefinedModel = (
-                    NeuralNetworkPredefinedModel._Schema().load(object_dict)
-                )
-                return predefined_model
