@@ -3,6 +3,7 @@ from __future__ import annotations
 
 from dataclasses import dataclass
 import logging
+import os
 from pathlib import Path
 import re
 import typing
@@ -36,7 +37,8 @@ from bitfount.federated.transport.config import (
     _STAGING_MESSAGE_SERVICE_URL,
     MessageServiceConfig,
 )
-from bitfount.hub.api import (
+from bitfount.hub.authentication_flow import _DEFAULT_USERNAME
+from bitfount.hub.types import (
     _DEV_AM_URL,
     _DEV_HUB_URL,
     _DEV_IDP_URL,
@@ -46,8 +48,8 @@ from bitfount.hub.api import (
     _STAGING_IDP_URL,
     PRODUCTION_AM_URL,
     PRODUCTION_HUB_URL,
+    APIKeys,
 )
-from bitfount.hub.authentication_flow import _DEFAULT_USER_DIRECTORY
 from bitfount.models.base_models import (
     EarlyStopping,
     LoggerConfig,
@@ -117,11 +119,21 @@ class HubConfig:
 # MODELLER SCHEMAS
 @dataclass
 class ModellerUserConfig:
-    """Configuration for the modeller."""
+    """Configuration for the modeller.
+
+    Args:
+        username: The username of the modeller. This can be picked up automatically
+            from the session but can be overridden here.
+        identity_verification_method: The method to use for identity verification.
+            Accepts one of the values in `IDENTITY_VERIFICATION_METHODS`, i.e. one of
+            `key-based`, `saml`, `oidc-auth-code` or `oidc-device-code`.
+        private_key_file: The path to the private key file for key-based identity
+            verification.
+    """
 
     username: str = desert.field(
         fields.String(validate=validate.Regexp(USERNAME_REGEX)),
-        default=_DEFAULT_USER_DIRECTORY,
+        default=_DEFAULT_USERNAME,
     )  # type: ignore[assignment] # Reason: see top of file
 
     identity_verification_method: str = desert.field(
@@ -131,16 +143,17 @@ class ModellerUserConfig:
     private_key_file: Optional[Path] = desert.field(  # type: ignore[assignment] # Reason: see top of file  # noqa: B950
         fields.Function(deserialize=_deserialize_path), default=None
     )
-    identity_provider_url: Optional[str] = None
 
     def __post_init__(self) -> None:
         environment = _get_environment()
+        self._identity_provider_url: str
+
         if environment == _STAGING_ENVIRONMENT:
-            self.identity_provider_url = _STAGING_IDP_URL
+            self._identity_provider_url = _STAGING_IDP_URL
         elif environment == _DEVELOPMENT_ENVIRONMENT:
-            self.identity_provider_url = _DEV_IDP_URL
+            self._identity_provider_url = _DEV_IDP_URL
         else:
-            self.identity_provider_url = _PRODUCTION_IDP_URL
+            self._identity_provider_url = _PRODUCTION_IDP_URL
 
 
 @dataclass
@@ -192,6 +205,7 @@ class BitfountModelReferenceConfig:
     model_ref: Union[Path, str] = desert.field(
         fields.Function(deserialize=_deserialize_model_ref)
     )  # type: ignore[assignment] # Reason: see top of file
+    model_version: Optional[int] = None
     username: Optional[str] = None
 
 
@@ -212,9 +226,6 @@ class ModelConfig:
     )  # type: ignore[assignment] # Reason: see top of file
     logger_config: Optional[LoggerConfig] = None
     dp_config: Optional[DPModellerConfig] = None
-    pretrained_file: Optional[Path] = desert.field(
-        fields.Function(deserialize=_deserialize_path), default=None
-    )  # type: ignore[assignment] # Reason: see top of file
 
     def __post_init__(self) -> None:
         # Validate either name or bitfount_model reference provided
@@ -297,6 +308,9 @@ class AlgorithmConfig:
     arguments: _JSONDict = desert.field(
         fields.Dict(keys=fields.Str), default_factory=dict
     )  # type: ignore[assignment] # Reason: see top of file
+    pretrained_file: Optional[Path] = desert.field(
+        fields.Function(deserialize=_deserialize_path), default=None
+    )  # type: ignore[assignment] # Reason: see top of file
 
     def __post_init__(self) -> None:
         # TODO: [BIT-1300] Remove this once custom algorithms are supported
@@ -322,14 +336,19 @@ class TaskConfig:
 # POD SCHEMAS
 @dataclass
 class PodConfig:
-    """Full configuration for the pod."""
+    """Full configuration for the pod.
 
-    data: Union[DatabaseConfig, PathConfig]
+    Raises:
+        ValueError: If a username is not provided alongside API keys.
+    """
+
+    datasource: str
     data_config: PodDataConfig
-    pod_details: PodDetailsConfig
-    pod_name: str = fields.String(
-        validate=validate.Regexp(POD_NAME_REGEX)
+    name: str = desert.field(
+        fields.String(validate=validate.Regexp(POD_NAME_REGEX))
     )  # type: ignore[assignment] # Reason: see top of file
+    api_key: Optional[APIKeys] = None
+    pod_details_config: Optional[PodDetailsConfig] = None
     schema: Optional[Path] = desert.field(
         fields.Function(deserialize=_deserialize_path), default=None
     )  # type: ignore[assignment] # Reason: see top of file
@@ -345,11 +364,12 @@ class PodConfig:
         default_factory=MessageServiceConfig,
     )  # type: ignore[assignment] # Reason: see top of file
     differential_privacy: Optional[DPPodConfig] = None
-    other_pods: Optional[List[str]] = None
+    approved_pods: Optional[List[str]] = None
     username: str = desert.field(
         fields.String(validate=validate.Regexp(USERNAME_REGEX)),
-        default=_DEFAULT_USER_DIRECTORY,
+        default=_DEFAULT_USERNAME,
     )  # type: ignore[assignment] # Reason: see top of file
+    update_schema: bool = False
 
     def __post_init__(self) -> None:
         environment = _get_environment()
@@ -364,10 +384,27 @@ class PodConfig:
             self.message_service.port = _DEV_MESSAGE_SERVICE_PORT
             self.message_service.tls = _DEV_MESSAGE_SERVICE_TLS
 
+        # Use API Keys for authentication if provided
+        if self.api_key:
+            if self.username == _DEFAULT_USERNAME:
+                raise ValueError("Must specify a username when using API Keys.")
+
+            logger.info("Setting API Keys as environment variables.")
+
+            if os.environ.get("BITFOUNT_API_KEY_ID") or os.environ.get(
+                "BITFOUNT_API_KEY"
+            ):
+                logger.warning(
+                    "Existing environment variable API keys detected. Overriding with "
+                    "those provided in the pod config."
+                )
+            os.environ["BITFOUNT_API_KEY_ID"] = self.api_key.access_key_id
+            os.environ["BITFOUNT_API_KEY"] = self.api_key.access_key
+
     @property
     def pod_id(self) -> str:
         """The pod ID of the pod specified."""
-        return f"{self.username}/{self.pod_name}"
+        return f"{self.username}/{self.name}"
 
 
 @dataclass
@@ -423,18 +460,22 @@ class DataSplitConfig:
 
 @dataclass
 class PodDataConfig:
-    """Configuration for the Schema, DataSource and Pod.
+    """Configuration for the Schema, BaseSource and Pod.
 
     Args:
         force_stypes: The semantic types to force for the data. This is passed to the
-            `BitfountSchema`.
-        ignore_cols: The columns to ignore. This is passed to the `DataSource`.
+            `BitfountSchema`. This should be a mapping from pod name to a mapping from
+            the type to a list of column names (e.g. {"pod_name": {categorical:
+            ["col1", "col2"]}}, or a mappping from each table name in a multitable
+            datasource to a mapping from the type to a list of column names (e.g.
+            {"table1": {categorical: ["col1", "col2"]}, "table2":  {continuous:
+            ["col3", "col4"]}}.
+        ignore_cols: The columns to ignore. This is passed to the data source.
         modifiers: The modifiers to apply to the data. This is passed to the
-            `DataSource`.
-        datasource_args: Key-value pairs of arguments to pass to the `DataSource`
+            `BaseSource`.
+        datasource_args: Key-value pairs of arguments to pass to the data source
             constructor.
-        data_split: The data split configuration. This is passed to the `DataSource`.
-        koalas: Whether to use Koalas. This is passed to the `DataSource`.
+        data_split: The data split configuration. This is passed to the data source.
         auto_tidy: Whether to automatically tidy the data. This is used by the
             `Pod` and will result in removal of NaNs and normalisation of numeric
             values. Defaults to False.
@@ -458,8 +499,9 @@ class PodDataConfig:
         default=None,
     )  # type: ignore[assignment] # Reason: see comment and top of file
 
-    ignore_cols: Optional[List[str]] = desert.field(
-        fields.List(fields.String()), default=None
+    ignore_cols: Optional[Mapping[str, List[str]]] = desert.field(
+        fields.Dict(keys=fields.String(), values=fields.List(fields.String())),
+        default=None,
     )  # type: ignore[assignment] # Reason: see comment and top of file
 
     modifiers: Optional[Dict[str, DataPathModifiers]] = desert.field(
@@ -481,7 +523,6 @@ class PodDataConfig:
         fields.Nested(desert.schema_class(DataSplitConfig)),
         default_factory=DataSplitConfig,
     )  # type: ignore[assignment] # Reason: see top of file
-    koalas: bool = False
     auto_tidy: bool = False
 
 
@@ -492,12 +533,10 @@ class PodDetailsConfig:
     Args:
         display_name: The display name of the pod.
         description: The description of the pod.
-        is_public: Whether the pod is public. Defaults to False.
     """
 
     display_name: str
     description: str
-    is_public: bool = False
 
 
 @dataclass
