@@ -8,7 +8,7 @@ from typing import (
     Callable,
     Dict,
     List,
-    Literal,
+    Mapping,
     Optional,
     Set,
     Tuple,
@@ -18,16 +18,17 @@ from typing import (
 )
 
 import numpy as np
+import pandas as pd
 from pandas.core.dtypes.common import is_numeric_dtype
 
-from bitfount.data.types import SemanticType
+from bitfount.data.types import DataSplit, SemanticType
 from bitfount.transformations.base_transformation import (
     MultiColumnOutputTransformation,
     Transformation,
 )
 from bitfount.transformations.batch_operations import (
+    AlbumentationsImageTransformation,
     BatchTimeOperation,
-    ImageTransformation,
 )
 from bitfount.transformations.binary_operations import (
     AdditionTransformation,
@@ -39,6 +40,8 @@ from bitfount.transformations.binary_operations import (
 from bitfount.transformations.dataset_operations import (
     CleanDataTransformation,
     NormalizeDataTransformation,
+    ScalarAdditionDataTransformation,
+    ScalarMultiplicationDataTransformation,
 )
 from bitfount.transformations.exceptions import (
     InvalidBatchTransformationError,
@@ -51,8 +54,6 @@ from bitfount.transformations.unary_operations import (
     InclusionTransformation,
     OneHotEncodingTransformation,
 )
-from bitfount.types import _DataFrameLib, _DataFrameType, _SeriesType
-from bitfount.utils import _get_df_library, _get_df_library_type
 
 if TYPE_CHECKING:
     from bitfount.data.schema import TableSchema
@@ -64,6 +65,13 @@ logger = logging.getLogger(__name__)
 class TransformationProcessor:
     """Processes Transformations on a given dataframe.
 
+    :::caution
+
+    The Transformation processor does not add any of the newly created columns
+    to the Schema. This must be done separately after processing the transformations.
+
+    :::
+
     Args:
         transformations: The list of transformations to apply.
         schema: The schema of the data to be transformed.
@@ -73,13 +81,6 @@ class TransformationProcessor:
         transformations: The list of transformations to apply.
         schema: The schema of the data to be transformed.
         col_refs: The set of columns referenced in those transformations.
-
-    ::: warning
-
-    The Transformation processor does not add any of the newly created columns
-    to the Schema. This must be done separately after processing the transformations.
-
-    :::
     """
 
     def __init__(
@@ -113,10 +114,13 @@ class TransformationProcessor:
             NormalizeDataTransformation: self._do_normalize_data,
             OneHotEncodingTransformation: self._do_one_hot_encoding,
             SubtractionTransformation: self._do_subtraction,
-            ImageTransformation: self._do_image_transformation,
+            AlbumentationsImageTransformation: self._do_image_transformation,
+            ScalarMultiplicationDataTransformation: self._do_scalar_multiplication,
+            ScalarAdditionDataTransformation: self._do_scalar_addition,
         }
+        cols_to_reverse: Mapping[str, Tuple[float, float]]
 
-    def transform(self, data: _DataFrameType) -> _DataFrameType:
+    def transform(self, data: pd.DataFrame) -> pd.DataFrame:
         """Performs `self.transformations` on `data` sequentially.
 
         Arguments to an operation are extracted by first checking if they are
@@ -128,25 +132,21 @@ class TransformationProcessor:
         output.
 
         Args:
-            data: The `pandas` or `koalas` dataframe to be transformed.
+            data: The `pandas` dataframe to be transformed.
 
-        :::warning
-
-        Differing behaviour between `pandas` and `koalas` in this method.
-        Pandas **will not** create columns where there is a type mismatch for the
-        operation but koalas **will** create a column of `NaNs`/`None`s etc.
-
-        :::
+        Raises:
+            MissingColumnReferenceError: If there is a reference to a non-existing
+                column.
+            TypeError: if there are clashes between column names or if unable
+                to apply transformation.
         """
         data_columns = set(data.columns)
-
         # Check that all referenced columns are present in `data`
         missing_cols = sorted(self.col_refs.difference(data_columns))
         if missing_cols:
             raise MissingColumnReferenceError(
                 [f"Reference to non-existent column: {c}" for c in missing_cols]
             )
-
         # Loop through transformations and perform them sequentially
         application_errors = []
         cols_to_drop = []
@@ -182,7 +182,7 @@ class TransformationProcessor:
 
             # Attempt to perform transformation. If there is a type mismatch
             # pandas will throw a type error which we catch and skip the
-            # transformation. Koalas does not throw TypeErrors
+            # transformation.
             try:
                 data = operation(data, transformation)
             except TypeError as e:
@@ -201,13 +201,12 @@ class TransformationProcessor:
 
         return data
 
-    def batch_transform(
-        self, data: np.ndarray, step: Literal["train", "validation"]
-    ) -> np.ndarray:
+    def batch_transform(self, data: np.ndarray, step: DataSplit) -> np.ndarray:
         """Performs batch transformations.
 
         Args:
             data: The data to be transformed at batch time as a numpy array.
+            step: The step at which the data should be transformed.
 
         Raises:
             InvalidBatchTransformationError: If one of the specified transformations
@@ -237,8 +236,8 @@ class TransformationProcessor:
 
     @staticmethod
     def _apply_arg_conversion(
-        data: _DataFrameType, *args: Union[Any, str, Transformation]
-    ) -> Union[_SeriesType, List[_SeriesType]]:
+        data: pd.DataFrame, *args: Union[Transformation, str, Any]
+    ) -> Union[Union[pd.Series, Any], List[Union[pd.Series, Any]]]:
         """Applies argument conversion to each of the supplied *args.
 
         Check, in order, if it is a:
@@ -253,7 +252,7 @@ class TransformationProcessor:
         Returns: A single converted arg if only one was supplied, or a list of converted
             args in the same order they were supplied.
         """
-        converted = []
+        converted: List[Union[pd.Series, Any]] = []
         for arg in args:
             # See if it is a transformation
             if isinstance(arg, Transformation):
@@ -279,7 +278,7 @@ class TransformationProcessor:
 
     @staticmethod
     def _get_list_of_col_refs(
-        data: _DataFrameType, cols: Union[str, List[str]], schema_cols: List[str]
+        data: pd.DataFrame, cols: Union[str, List[str]], schema_cols: List[str]
     ) -> Tuple[List[str], bool]:
         """Gets the list of actual column references from a list of potentials.
 
@@ -304,25 +303,38 @@ class TransformationProcessor:
         """
         pre_extracted = False
         out_cols = cols
-        if out_cols == "all":
-            # Extract all target columns from schema
-            pre_extracted = True
-            out_cols = [col for col in schema_cols if col in data.columns]
+
+        if isinstance(out_cols, list):
+            # If columns are not pre-extracted, they will have start
+            # with "c:" or "col:"
+            if not any(":" in col_name for col_name in out_cols):
+                pre_extracted = True
+            out_cols = out_cols
         elif isinstance(out_cols, str):
-            # Wrap single column in iterable
-            out_cols = [out_cols]
+            if out_cols == "float":
+                pre_extracted = True
+                out_cols = cast(
+                    list, data.select_dtypes(include=["float"]).columns.to_list()
+                )
+            elif out_cols == "all":
+                # Extract all target columns from schema
+                pre_extracted = True
+                out_cols = [col for col in schema_cols if col in data.columns]
+            else:
+                # Wrap single column in iterable
+                out_cols = [out_cols]
         return out_cols, pre_extracted
 
     @staticmethod
-    def _do_addition(data: _DataFrameType, t: AdditionTransformation) -> _DataFrameType:
+    def _do_addition(data: pd.DataFrame, t: AdditionTransformation) -> pd.DataFrame:
         """Performs addition transformation on `data` and returns it."""
         arg1, arg2 = TransformationProcessor._apply_arg_conversion(data, t.arg1, t.arg2)
         data[t.name] = arg1 + arg2
         return data
 
     def _do_clean_data(
-        self, data: _DataFrameType, t: CleanDataTransformation
-    ) -> _DataFrameType:
+        self, data: pd.DataFrame, t: CleanDataTransformation
+    ) -> pd.DataFrame:
         """Cleans categorical and continuous columns in the data.
 
         Replaces infinities and NAs.
@@ -349,9 +361,7 @@ class TransformationProcessor:
         return data
 
     @staticmethod
-    def _do_comparison(
-        data: _DataFrameType, t: ComparisonTransformation
-    ) -> _DataFrameType:
+    def _do_comparison(data: pd.DataFrame, t: ComparisonTransformation) -> pd.DataFrame:
         """Performs comparison between arg1 and arg2 of comparison transformation."""
         arg1, arg2 = TransformationProcessor._apply_arg_conversion(data, t.arg1, t.arg2)
 
@@ -370,45 +380,47 @@ class TransformationProcessor:
         return data
 
     @staticmethod
-    def _do_division(data: _DataFrameType, t: DivisionTransformation) -> _DataFrameType:
+    def _do_division(data: pd.DataFrame, t: DivisionTransformation) -> pd.DataFrame:
         """Performs division transformation on `data` and returns it."""
         arg1, arg2 = TransformationProcessor._apply_arg_conversion(data, t.arg1, t.arg2)
         data[t.name] = arg1 / arg2
         return data
 
     @staticmethod
-    def _do_inclusion(
-        data: _DataFrameType, t: InclusionTransformation
-    ) -> _DataFrameType:
+    def _do_inclusion(data: pd.DataFrame, t: InclusionTransformation) -> pd.DataFrame:
         # Only arg should be a column name in this transformation
         """Performs inclusion transformation on `data` and returns it."""
         arg = TransformationProcessor._apply_arg_conversion(data, t.arg)
-        arg = cast(_SeriesType, arg)
+        arg = cast(pd.Series, arg)
         data[t.name] = arg.str.contains(t.in_str)
         return data
 
     @staticmethod
     def _do_multiplication(
-        data: _DataFrameType, t: MultiplicationTransformation
-    ) -> _DataFrameType:
+        data: pd.DataFrame, t: MultiplicationTransformation
+    ) -> pd.DataFrame:
         """Performs multiplication transformation on `data` and returns it."""
         arg1, arg2 = TransformationProcessor._apply_arg_conversion(data, t.arg1, t.arg2)
         data[t.name] = arg1 * arg2
         return data
 
     def _do_normalize_data(
-        self, data: _DataFrameType, t: NormalizeDataTransformation
-    ) -> _DataFrameType:
+        self,
+        data: pd.DataFrame,
+        t: NormalizeDataTransformation,
+    ) -> pd.DataFrame:
         """Normalizes numeric columns using their mean and stddev.
 
-        Results in mean of 0 and stddev of 1.
+        Note: The default value of the transformation.cols for this
+        transformation is "float". Results in mean of 0 and stddev of 1.
         """
         cols, pre_extracted = self._get_list_of_col_refs(
-            data, t.cols, self._schema_cont_cols
+            data,
+            t.cols,
+            self._schema_cont_cols,
         )
         if not pre_extracted:
             cols = [_extract_col_ref(col) for col in cols]
-
         for col_ref in cols:
             if not is_numeric_dtype(data[col_ref]):
                 raise TypeError(
@@ -417,29 +429,104 @@ class TransformationProcessor:
             mean = data[col_ref].mean()
             stddev = data[col_ref].std()
             data[col_ref] = (data[col_ref] - mean) / (1e-7 + stddev)
-
-        for col in self._schema_cont_cols:
-            # If self._schema_cont_cols is not empty then self.schema cannot be None
-            assert self.schema is not None  # nosec
-            self.schema.features["continuous"][col].dtype = data[col].dtype
+        if self.schema is not None:
+            for col in cols:
+                self.schema.features["continuous"][col].dtype = data[col].dtype
 
         return data
 
+    def _do_scalar_multiplication(
+        self, data: pd.DataFrame, t: ScalarMultiplicationDataTransformation
+    ) -> pd.DataFrame:
+        """Performs column multiplication with scalar.
+
+        Only applicable to numeric columns.
+
+        Raises:
+            TypeError: if one of the referenced columns is not numeric.
+            MissingColumnReferenceError: if one of the referenced columns is missing.
+        """
+        # If the scalar is a numerical value, we apply the scalar
+        # multiplication to all numerical columns.
+        if isinstance(t.scalar, (int, float)):
+            cols, pre_extracted = self._get_list_of_col_refs(
+                data, t.cols, self._schema_cont_cols
+            )
+            if not pre_extracted:
+                cols = [_extract_col_ref(col) for col in cols]
+
+            data[cols] = data[cols] * t.scalar
+            return data
+        # Otherwise, it is a dictionary mapping column names to scalar,
+        # so we multiply each column with the scalar it is mapped to.
+        else:
+            missing_cols = set(t.scalar.keys()) - set(data.columns)
+            if missing_cols:
+                raise MissingColumnReferenceError(
+                    [f"Reference to non-existent column: {c}" for c in missing_cols]
+                )
+            for col_ref, scalar in t.scalar.items():
+                if not is_numeric_dtype(data[col_ref]):
+                    raise TypeError(
+                        f'Cannot normalize column "{col_ref}" as it is not numeric.'
+                    )
+                data[col_ref] = data[col_ref] * scalar
+            return data
+
+    def _do_scalar_addition(
+        self, data: pd.DataFrame, t: ScalarAdditionDataTransformation
+    ) -> pd.DataFrame:
+        """Performs column addition with scalar.
+
+        Only applicable to numeric columns.
+
+        Raises:
+            TypeError: if one of the referenced columns is not numeric.
+            MissingColumnReferenceError: if one of the referenced columns is missing.
+        """
+        # If the scalar is a numerical value, we apply the scalar
+        # addition to all numerical columns.
+        if isinstance(t.scalar, (int, float)):
+            cols, pre_extracted = self._get_list_of_col_refs(
+                data, t.cols, self._schema_cont_cols
+            )
+            if not pre_extracted:
+                cols = [_extract_col_ref(col) for col in cols]
+
+            data[cols] = data[cols] + t.scalar
+            return data
+        # Otherwise, it is a dictionary mapping column names to scalar,
+        # so we add the scalar to the column it is mapped to.
+        else:
+            missing_cols = set(t.scalar.keys()) - set(data.columns)
+            if missing_cols:
+                raise MissingColumnReferenceError(
+                    [f"Reference to non-existent column: {c}" for c in missing_cols]
+                )
+            for col_ref, scalar in t.scalar.items():
+                if not is_numeric_dtype(data[col_ref]):
+                    raise TypeError(
+                        f'Cannot normalize column "{col_ref}" as it is not numeric.'
+                    )
+                data[col_ref] = data[col_ref] + scalar
+            return data
+
     @staticmethod
     def _do_one_hot_encoding(
-        data: _DataFrameType, t: OneHotEncodingTransformation
-    ) -> _DataFrameType:
+        data: pd.DataFrame, t: OneHotEncodingTransformation
+    ) -> pd.DataFrame:
         """Performs one hot encoding transformation on `data` and returns it."""
         # Get unencoded data and columns
-        arg: _SeriesType = TransformationProcessor._apply_arg_conversion(data, t.arg)
+        # Only one arg supplied so therefore is single series that's returned
+        arg: pd.Series = cast(
+            pd.Series, TransformationProcessor._apply_arg_conversion(data, t.arg)
+        )
         ohe_cols: List[str] = sorted(t.columns)
         n_rows: int = len(arg)
         n_cols: int = len(ohe_cols)
 
         # Create an appropriately sized dataframe, filled with zeros
-        df_lib_type = _get_df_library_type(data)
-        df_lib = _get_df_library(data)
-        ohe_df: _DataFrameType = df_lib.DataFrame(
+        ohe_df: pd.DataFrame = pd.DataFrame(
             data=np.zeros(shape=(n_rows, n_cols), dtype=np.int8), columns=ohe_cols
         )
 
@@ -447,48 +534,23 @@ class TransformationProcessor:
         # non-null values in the arg Series. These will be marked into the correct
         # column as we find matches, and if no match is found, they should be in
         # this column anyway.
-        if df_lib_type == _DataFrameLib.KOALAS:
-            # koalas indexes aren't subscriptable so must use a BooleanIndex instead
-            not_null_idxs = arg[arg.notnull()].index
-            # koalas indexes don't support __iter__ so will break the .loc
-            # setting below; instead we must convert it to a numpy array (which
-            # still won't work as it doesn't match a protocol pyspark expects) and
-            # then to a list. ks.Index.to_list() isn't implemented either.
-            #
-            # This approach is relatively slow compared to the native pandas
-            # implementation and should probably be replaced with something
-            # more performant.
-            # TODO: [BIT-1042] Increase performance, avoid koalas multi-conversion.
-            not_null_idxs = not_null_idxs.to_numpy().tolist()
-        else:
-            not_null_idxs = arg.index[arg.notnull()]
+        not_null_idxs = arg.index[arg.notnull()]
         ohe_df.loc[not_null_idxs, t.unknown_col] = 1
 
         # Iterate through the (value, target_col) pairs and set the correct column
         # to 1 for all locations the value is found. Sets the unknown column in
         # the same locations to 0.
         for val, target_col in t.values.items():
-            if df_lib_type == _DataFrameLib.KOALAS:
-                # Same issues as above
-                match_idxs = arg[arg == val].index
-                match_idxs = match_idxs.to_numpy().tolist()
-            else:
-                match_idxs = arg.index[arg == val]
+            match_idxs = arg.index[arg == val]
             ohe_df.loc[match_idxs, target_col] = 1
             ohe_df.loc[match_idxs, t.unknown_col] = 0
 
-        if df_lib_type == _DataFrameLib.KOALAS:
-            with df_lib.option_context("compute.ops_on_diff_frames", True):  # type: ignore[attr-defined] # Reason: koalas has this function # noqa: B950
-                return df_lib.concat(
-                    [data, ohe_df], axis=1, sort=True
-                )  # sort=True is needed for koalas
-        else:
-            return df_lib.concat([data, ohe_df], axis=1)
+        return pd.concat([data, ohe_df], axis=1)
 
     @staticmethod
     def _do_subtraction(
-        data: _DataFrameType, t: SubtractionTransformation
-    ) -> _DataFrameType:
+        data: pd.DataFrame, t: SubtractionTransformation
+    ) -> pd.DataFrame:
         """Performs subtraction transformation on `data` and returns it."""
         arg1, arg2 = TransformationProcessor._apply_arg_conversion(data, t.arg1, t.arg2)
         data[t.name] = arg1 - arg2
@@ -496,9 +558,8 @@ class TransformationProcessor:
 
     @staticmethod
     def _do_image_transformation(
-        data: np.ndarray, t: ImageTransformation
+        data: np.ndarray, t: AlbumentationsImageTransformation
     ) -> np.ndarray:
         """Performs image transformation on `data` and returns it."""
-        logger.debug(f"Applying image transformations: {t.transformations}")
         tfm = t.get_callable()
         return tfm(image=data)["image"]
