@@ -8,7 +8,8 @@ import numpy as np
 import pandas as pd
 import yaml
 
-from bitfount.data.datasource import DataSource
+from bitfount import config
+from bitfount.data.datasources.dataframe_source import DataFrameSource
 from bitfount.data.datastructure import DataStructure
 from bitfount.data.schema import BitfountSchema
 from bitfount.federated.utils import _MODELS
@@ -22,6 +23,8 @@ from bitfount.transformations.processor import TransformationProcessor
 
 loggers = setup_loggers([logging.getLogger("bitfount")])
 
+config._BITFOUNT_CLI_MODE = True
+
 
 def evaluate_model(
     path_to_config_yaml: str, path_to_test_csv: str, path_to_model: str
@@ -41,10 +44,7 @@ def evaluate_model(
     if len(schema.tables) > 1:
         raise ValueError("Only single-table schemas are supported.")
 
-    datasource = DataSource(
-        data=data,
-        **datasource_args,
-    )
+    datasource = DataFrameSource(data=data, **datasource_args)
     # Transform dataset
     clean_data = CleanDataTransformation()
     normalize = NormalizeDataTransformation()
@@ -58,7 +58,7 @@ def evaluate_model(
     # Load model
     neural_network = _MODELS[algorithm](datastructure=data_structure, schema=schema)
     neural_network.initialise_model(data=datasource)
-    test_data_loader = neural_network._databunch.get_train_dataloader(
+    test_data_loader = neural_network.databunch.get_train_dataloader(
         batch_size=batch_size
     )
 
