@@ -2,7 +2,7 @@
 import copy
 from pathlib import Path
 import re
-from typing import Any, Dict, List, MutableMapping, Optional, Union, cast
+from typing import Dict, List, MutableMapping, Optional, Union
 from unittest.mock import Mock, PropertyMock, create_autospec
 
 from _pytest.logging import LogCaptureFixture
@@ -18,7 +18,10 @@ import sqlalchemy
 import yaml
 
 import bitfount
-from bitfount.data.datasource import DataSource
+from bitfount.data.datasources.base_source import BaseSource
+from bitfount.data.datasources.database_source import DatabaseSource
+from bitfount.data.datasources.dataframe_source import DataFrameSource
+from bitfount.data.exceptions import DataSourceError
 from bitfount.data.schema import (
     BitfountSchema,
     BitfountSchemaError,
@@ -31,10 +34,16 @@ from bitfount.data.types import (
     SemanticType,
     _SemanticTypeValue,
 )
-from bitfount.data.utils import DatabaseConnection, _hash_str
-from bitfount.types import _Dtypes
+from bitfount.data.utils import DatabaseConnection, _generate_dtypes_hash, _hash_str
+from bitfount.types import _Dtypes, _JSONDict
 from tests.utils import PytestRequest
-from tests.utils.helper import TABLE_NAME, create_dataset, integration_test, unit_test
+from tests.utils.helper import (
+    TABLE_NAME,
+    create_dataset,
+    get_warning_logs,
+    integration_test,
+    unit_test,
+)
 
 SCHEMA_DUMP_PATH: str = "schema_dump_test.yaml"
 
@@ -42,7 +51,7 @@ SCHEMA_DUMP_PATH: str = "schema_dump_test.yaml"
 @fixture
 def data() -> pd.DataFrame:
     """Returns dataset."""
-    return cast(pd.DataFrame, create_dataset())
+    return create_dataset()
 
 
 @fixture(params=["dataframe", "db", "db_multi_table", "db_query"])
@@ -50,34 +59,35 @@ def datasource(
     request: PytestRequest,
     data: pd.DataFrame,
     db_session: sqlalchemy.engine.base.Engine,
-) -> DataSource:
+) -> BaseSource:
     """Returns datasource."""
+    datasource: BaseSource
     if request.param == "db":
         db_conn = DatabaseConnection(db_session, table_names=["dummy_data"])
-        datasource = DataSource(db_conn, seed=420)
+        datasource = DatabaseSource(db_conn, seed=420)
     elif request.param == "db_multi_table":
         db_conn = DatabaseConnection(
             db_session, table_names=["dummy_data", "dummy_data_2"]
         )
-        datasource = DataSource(db_conn, seed=420)
+        datasource = DatabaseSource(db_conn, seed=420)
     elif request.param == "db_query":
         db_conn = DatabaseConnection(db_session, query="SELECT * FROM dummy_data")
-        datasource = DataSource(db_conn, seed=420)
+        datasource = DatabaseSource(db_conn, seed=420)
     elif request.param == "dataframe":
-        datasource = DataSource(data, seed=420)
+        datasource = DataFrameSource(data, seed=420)
     else:
         raise ValueError(
-            "Supported DataSources are : dataframe, db, db_multi_table and db_query."
+            "Supported Data Sources are : dataframe, db, db_multi_table and db_query."
         )
 
     return datasource
 
 
 @fixture
-def image_datasource() -> DataSource:
+def image_datasource() -> DataFrameSource:
     """Datasource with image_col for testing."""
     data = create_dataset(image=True)
-    ds = DataSource(data, image_col=["image"])
+    ds = DataFrameSource(data, image_col=["image"])
     ds.load_data()
     return ds
 
@@ -96,26 +106,34 @@ def schema() -> BitfountSchema:
 
 @fixture
 def table_schema_with_data(
-    datasource: DataSource, table_schema: TableSchema
+    datasource: BaseSource, table_schema: TableSchema
 ) -> TableSchema:
     """Schema with datasource already set."""
-    if isinstance(datasource.data_ref, DatabaseConnection) and (
-        table_names := datasource.data_ref.table_names
+    if isinstance(datasource, DatabaseSource) and (
+        table_names := datasource.table_names
     ):
         table_schema = TableSchema(name=table_names[0])
-    table_schema.add_dataframe_features(datasource)
+    table_schema.add_datasource_features(datasource)
     return table_schema
 
 
 @fixture
-def schema_with_data(datasource: DataSource, schema: BitfountSchema) -> BitfountSchema:
+def schema_with_data(datasource: BaseSource, schema: BitfountSchema) -> BitfountSchema:
     """Schema with datasource already set."""
-    if isinstance(datasource.data_ref, DatabaseConnection) and (
-        table_names := datasource.data_ref.table_names
+    if isinstance(datasource, DatabaseSource) and (
+        table_names := datasource.table_names
     ):
-        schema.add_datasource_tables(datasource, table_name=table_names[0])
+        schema.add_datasource_tables(
+            datasource,
+            table_name=table_names[0],
+            force_stypes={table_names[0]: {"categorical": ["Date"]}},
+        )
     else:
-        schema.add_datasource_tables(datasource, table_name=TABLE_NAME)
+        schema.add_datasource_tables(
+            datasource,
+            table_name=TABLE_NAME,
+            force_stypes={TABLE_NAME: {"categorical": ["Date"]}},
+        )
     return schema
 
 
@@ -146,7 +164,7 @@ class TestSchemaSerialization:
         }
 
     def test_schema_dump_load(
-        self, datasource: DataSource, schema: BitfountSchema, tmp_path: Path
+        self, datasource: BaseSource, schema: BitfountSchema, tmp_path: Path
     ) -> None:
         """Tests dumping and loading data schema."""
         schema.add_datasource_tables(datasource, table_name=TABLE_NAME)
@@ -162,24 +180,26 @@ class TestSchemaSerialization:
     def test_multi_table_schema_dump_load(
         self,
         mock_engine: Mock,
-        mock_pandas_read_sql_table: None,
+        mock_pandas_read_sql_query: None,
+        mocker: MockerFixture,
         schema: BitfountSchema,
         tmp_path: Path,
-        mocker: MockerFixture,
     ) -> None:
         """Tests dumping and loading multi-table schema."""
         db_conn = DatabaseConnection(
             mock_engine,
             table_names=["dummy_data", "dummy_data_2"],
         )
-        ds = DataSource(db_conn, seed=420)
-        mocker.patch.object(
-            ds.loader, "get_dtypes", return_value=create_dataset().dtypes.to_dict()
-        )
+        ds = DatabaseSource(db_conn, seed=420)
+        mock_dtypes = create_dataset().dtypes.to_dict()
+        mocker.patch.object(ds, "get_dtypes", return_value=mock_dtypes)
+        ds._table_hashes.add(_generate_dtypes_hash(mock_dtypes))
         mocker.patch.object(
             ds,
             "get_values",
-            side_effect=lambda x, y: {col: create_dataset()[col].unique() for col in x},
+            side_effect=lambda x, table_name: {
+                col: create_dataset()[col].unique() for col in x
+            },
         )
         schema.add_datasource_tables(ds)
         assert schema.table_names == ["dummy_data", "dummy_data_2"]
@@ -206,12 +226,11 @@ class TestSchemaSerialization:
         }
 
     def test_load_works_without_metadata(
-        self, datasource: DataSource, schema: BitfountSchema, tmp_path: Path
+        self, datasource: BaseSource, schema: BitfountSchema, tmp_path: Path
     ) -> None:
         """Tests loading works if metadata tag not present."""
         schema.add_datasource_tables(datasource, table_name=TABLE_NAME)
         dumped = schema.dumps()
-
         # Remove metadata from dumped schema and save to file
         loaded_yaml = yaml.safe_load(dumped)
         del loaded_yaml["metadata"]
@@ -227,12 +246,11 @@ class TestSchemaSerialization:
         assert loaded_schema._orig_hash is None
 
     def test_schema_dump_load_sorts_keys_alphabetically(
-        self, datasource: DataSource, tmp_path: Path
+        self, datasource: BaseSource, tmp_path: Path
     ) -> None:
         """Tests that key ordering is made alphabetical by dump and load."""
         schema = BitfountSchema()
         schema.add_datasource_tables(datasource, table_name=TABLE_NAME)
-
         file_path = tmp_path / SCHEMA_DUMP_PATH
         schema.dump(file_path)
         loaded_schema = BitfountSchema.load_from_file(file_path)
@@ -246,6 +264,16 @@ class TestSchemaSerialization:
                 )
                 assert sorted(original_feature_names) == loaded_feature_names
 
+    def test_schema_raises_error_basesource_flag(
+        self,
+        datasource: BaseSource,
+    ) -> None:
+        """Tests schema raises error when BaseSource is not initialised."""
+        del datasource._base_source_init
+        schema = BitfountSchema()
+        with pytest.raises(DataSourceError):
+            schema.add_datasource_tables(datasource, table_name=TABLE_NAME)
+
     def test_schema_json_dump_has_correct_casing(
         self, empty_schema_metadata: Dict[str, str]
     ) -> None:
@@ -332,7 +360,7 @@ class TestSchemaSerialization:
         assert dumped_schema == expected_schema
 
     def test_schema_load_invalid_stype(
-        self, datasource: DataSource, tmp_path: Path
+        self, datasource: BaseSource, tmp_path: Path
     ) -> None:
         """Tests that semantic types must be predefined in the Enum."""
         schema = BitfountSchema()
@@ -355,7 +383,7 @@ class TestSchemaSerialization:
             BitfountSchema.load_from_file(file_path)
 
     def test_schema_load_invalid_dtype(
-        self, datasource: DataSource, tmp_path: Path
+        self, datasource: BaseSource, tmp_path: Path
     ) -> None:
         """Tests schema load fails if invalid dtype.
 
@@ -377,7 +405,7 @@ class TestSchemaSerialization:
             BitfountSchema.load_from_file(file_path)
 
     def test_schema_with_image_dump_load(
-        self, image_datasource: DataSource, schema: BitfountSchema, tmp_path: Path
+        self, image_datasource: DataFrameSource, schema: BitfountSchema, tmp_path: Path
     ) -> None:
         """Tests dumping and (re)loading a schema gives equal item."""
         file_path = tmp_path / SCHEMA_DUMP_PATH
@@ -430,17 +458,16 @@ class TestTableSchema:
     """Tests TableSchema class."""
 
     def test_schema_add_datasource(
-        self, datasource: DataSource, table_schema: TableSchema
+        self, datasource: BaseSource, table_schema: TableSchema
     ) -> None:
         """Checks ability to add a dataset."""
-        if isinstance(datasource.data_ref, DatabaseConnection):
-            if table_names := datasource.data_ref.table_names:
+        if isinstance(datasource, DatabaseSource):
+            if table_names := datasource.table_names:
                 table_schema = TableSchema(name=table_names[0])
-        table_schema.add_dataframe_features(datasource)
+        table_schema.add_datasource_features(datasource)
         datasource.load_data()
         if not datasource.multi_table:
             assert len(table_schema.get_feature_names()) == len(datasource.data.columns)
-
         assert (
             "I" in table_schema.features["text"]
         )  # I's values are letters of alphabet
@@ -458,10 +485,10 @@ class TestTableSchema:
         expect2 = pd.DataFrame(
             [["a", "b", "nan"], ["c", "d", "nan"]], columns=["B", "C", "A"]
         )
-        ds1 = DataSource(data1)
-        ds2 = DataSource(data2)
-        table_schema.add_dataframe_features(ds1)
-        table_schema.add_dataframe_features(ds2)
+        ds1 = DataFrameSource(data1)
+        ds2 = DataFrameSource(data2)
+        table_schema.add_datasource_features(ds1)
+        table_schema.add_datasource_features(ds2)
         data1_exp = table_schema._expand_dataframe(data1)
         data2_exp = table_schema._expand_dataframe(data2)
         assert_frame_equal(data1_exp, expect1)
@@ -479,10 +506,10 @@ class TestTableSchema:
             [[True, False, "nan"], [True, False, "nan"]],
             columns=["B", "C", "A"],
         )
-        ds1 = DataSource(data1)
-        ds2 = DataSource(data2)
-        table_schema.add_dataframe_features(ds1)
-        table_schema.add_dataframe_features(ds2)
+        ds1 = DataFrameSource(data1)
+        ds2 = DataFrameSource(data2)
+        table_schema.add_datasource_features(ds1)
+        table_schema.add_datasource_features(ds2)
         data1_exp = table_schema._expand_dataframe(data1)
         data2_exp = table_schema._expand_dataframe(data2)
         assert_frame_equal(data1_exp, expect1)
@@ -494,10 +521,10 @@ class TestTableSchema:
         data2 = pd.DataFrame([[1.0, 2.0], [3.0, 4.0]], columns=["B", "C"])
         expect1 = pd.DataFrame([[1.0, 2.0, 0], [3.0, 4.0, 0]], columns=["A", "B", "C"])
         expect2 = pd.DataFrame([[1.0, 2.0, 0], [3.0, 4.0, 0]], columns=["B", "C", "A"])
-        ds1 = DataSource(data1)
-        ds2 = DataSource(data2)
-        table_schema.add_dataframe_features(ds1)
-        table_schema.add_dataframe_features(ds2)
+        ds1 = DataFrameSource(data1)
+        ds2 = DataFrameSource(data2)
+        table_schema.add_datasource_features(ds1)
+        table_schema.add_datasource_features(ds2)
         data1_exp = table_schema._expand_dataframe(data1)
         data2_exp = table_schema._expand_dataframe(data2)
         assert_frame_equal(data1_exp, expect1)
@@ -509,20 +536,20 @@ class TestTableSchema:
         data2 = pd.DataFrame([[1, 2], [3, 4]], columns=["B", "C"])
         expect1 = pd.DataFrame([[1, 2, 0], [3, 4, 0]], columns=["A", "B", "C"])
         expect2 = pd.DataFrame([[1, 2, 0], [3, 4, 0]], columns=["B", "C", "A"])
-        ds1 = DataSource(data1)
-        ds2 = DataSource(data2)
-        table_schema.add_dataframe_features(ds1)
-        table_schema.add_dataframe_features(ds2)
+        ds1 = DataFrameSource(data1)
+        ds2 = DataFrameSource(data2)
+        table_schema.add_datasource_features(ds1)
+        table_schema.add_datasource_features(ds2)
         data1_exp = table_schema._expand_dataframe(data1)
         data2_exp = table_schema._expand_dataframe(data2)
         assert_frame_equal(data1_exp, expect1)
         assert_frame_equal(data2_exp, expect2)
 
     def test_schema_add_datasource_with_image(
-        self, image_datasource: DataSource, table_schema: TableSchema
+        self, image_datasource: DataFrameSource, table_schema: TableSchema
     ) -> None:
         """Tests add_dataframe_features works correctly."""
-        table_schema.add_dataframe_features(
+        table_schema.add_datasource_features(
             image_datasource, force_stype={"image": ["image"]}
         )
         assert len(table_schema.get_feature_names()) == len(
@@ -533,20 +560,19 @@ class TestTableSchema:
         image_feature = table_schema.features["image"]["image"]
         assert dict(image_feature.dimensions) == {(50, 50): 4000}
         assert "RGB" in image_feature.modes
-        assert "L" in image_feature.modes
         assert dict(image_feature.formats) == {"PNG": 4000}
 
     def test_schema_add_datasource_without_force_stype_doesnt_contain_image_col(
         self,
-        image_datasource: DataSource,
+        image_datasource: DataFrameSource,
         table_schema: TableSchema,
     ) -> None:
         """Tests that image col is not added to images."""
-        table_schema.add_dataframe_features(image_datasource)
+        table_schema.add_datasource_features(image_datasource)
         assert "image" not in table_schema.features
 
     def test_force_stypes(
-        self, image_datasource: DataSource, table_schema: TableSchema
+        self, image_datasource: DataFrameSource, table_schema: TableSchema
     ) -> None:
         """Tests that all force_stypes are added to schema."""
         force_stype: MutableMapping[_SemanticTypeValue, List[str]]
@@ -560,7 +586,7 @@ class TestTableSchema:
         # Added the 'blah' dict key to check that if a value is given a semantic type
         # which we don't support, it defaults to the semantic value based
         # on dtype.
-        table_schema.add_dataframe_features(image_datasource, force_stype=force_stype)
+        table_schema.add_datasource_features(image_datasource, force_stype=force_stype)
         assert "M" in table_schema.features["categorical"]
         assert "J" in table_schema.features["categorical"]
         assert "A" in table_schema.features["continuous"]
@@ -571,11 +597,11 @@ class TestTableSchema:
     def test_stype_images(self, table_schema: TableSchema) -> None:
         """Tests multiple images added to schema."""
         data = create_dataset(classification=False, multiimage=True, img_size=2)
-        ds = DataSource(data[["image1", "image2", "TARGET"]])
+        ds = DataFrameSource(data[["image1", "image2", "TARGET"]])
         ds.load_data()
         force_stype: MutableMapping[_SemanticTypeValue, List[str]]
         force_stype = {"image": ["image1", "image2"], "text": ["TARGET"]}
-        table_schema.add_dataframe_features(ds, force_stype=force_stype)
+        table_schema.add_datasource_features(ds, force_stype=force_stype)
         assert "image1" in table_schema.features["image"]
         assert "image2" in table_schema.features["image"]
         assert "TARGET" in table_schema.features["text"]
@@ -583,8 +609,10 @@ class TestTableSchema:
     def test_encode_dataframe(self, table_schema: TableSchema) -> None:
         """Tests schema encodes dataframe correctly."""
         data = pd.DataFrame([["a", "b"], ["c", "d"]], columns=["A", "B"])
-        ds = DataSource(data)
-        table_schema.add_dataframe_features(ds, force_stype={"categorical": ["A", "B"]})
+        ds = DataFrameSource(data)
+        table_schema.add_datasource_features(
+            ds, force_stype={"categorical": ["A", "B"]}
+        )
         expect = pd.DataFrame(
             [
                 [
@@ -604,9 +632,9 @@ class TestTableSchema:
     def test_reduce_dataframe(self, table_schema: TableSchema) -> None:
         """Tests reduce_dataframe removes appropriate columns."""
         data = pd.DataFrame([["a", "b"], ["c", "d"]], columns=["A", "B"])
-        ds = DataSource(data)
+        ds = DataFrameSource(data)
         ds.load_data()
-        table_schema.add_dataframe_features(ds)
+        table_schema.add_datasource_features(ds)
         data2 = pd.DataFrame(
             [["a", "b", "c"], ["d", "e", "f"]], columns=["A", "B", "C"]
         )
@@ -618,8 +646,8 @@ class TestTableSchema:
     def test_apply_types(self, table_schema: TableSchema) -> None:
         """Tests apply_types changes types of columns appropriately."""
         data = pd.DataFrame([[1.1, 2.2], [3.3, 4.4]], columns=["A", "B"])
-        ds = DataSource(data)
-        table_schema.add_dataframe_features(ds)
+        ds = DataFrameSource(data)
+        table_schema.add_datasource_features(ds)
         data2 = pd.DataFrame([[5, 6], [7, 8]], columns=["A", "B"])
         expect = pd.DataFrame(
             [[5.0, 6.0], [7.0, 8.0]], columns=["A", "B"]
@@ -627,13 +655,25 @@ class TestTableSchema:
         data_exp = table_schema._apply_types(data2)
         assert_frame_equal(data_exp, expect)
 
+    def test_apply_types_w_keep_cols(self, table_schema: TableSchema) -> None:
+        """Tests apply_types changes types of columns appropriately."""
+        data = pd.DataFrame([[1.1, 2.2, 5.5], [3.3, 4.4, 6.6]], columns=["A", "B", "C"])
+        ds = DataFrameSource(data)
+        table_schema.add_datasource_features(ds)
+        data2 = pd.DataFrame([[5, 6, 1], [7, 8, 9]], columns=["A", "B", "C"])
+        expect = pd.DataFrame(
+            [[5.0, 6.0, 1], [7.0, 8.0, 9]], columns=["A", "B", "C"]
+        ).convert_dtypes(convert_integer=False)
+        data_exp = table_schema._apply_types(data2, selected_cols=["A", "B"])
+        assert_frame_equal(data_exp, expect)
+
     def test_categorical_feature_sizes_with_no_ignore_cols(
         self, table_schema: TableSchema
     ) -> None:
         """Schema returns all categorical sizes if no ignore_cols provided."""
         data = pd.DataFrame([[True, False], [False, True]], columns=["A", "B"])
-        ds = DataSource(data)
-        table_schema.add_dataframe_features(ds)
+        ds = DataFrameSource(data)
+        table_schema.add_datasource_features(ds)
         assert table_schema.get_categorical_feature_sizes() == [2, 2]
 
     def test_categorical_feature_sizes_with_ignore_cols(
@@ -641,8 +681,8 @@ class TestTableSchema:
     ) -> None:
         """Schema returns correct categorical sizes when ignore_cols is categorical."""
         data = pd.DataFrame([[True, False], [False, True]], columns=["A", "B"])
-        ds = DataSource(data)
-        table_schema.add_dataframe_features(ds)
+        ds = DataFrameSource(data)
+        table_schema.add_datasource_features(ds)
         assert table_schema.get_categorical_feature_sizes(ignore_cols="A") == [2]
 
     def test_get_categorical_feature_size_raises_value_error_when_there_are_no_categorical_features(  # noqa: B950
@@ -702,9 +742,9 @@ class TestTableSchema:
         """Tests get_num_categorical for different types of ignore_cols."""
         # Create schema with categorical data columns
         data = pd.DataFrame([[True, False], [False, True]], columns=["A", "B"])
-        ds = DataSource(data)
+        ds = DataFrameSource(data)
         ds.load_data()
-        table_schema.add_dataframe_features(ds)
+        table_schema.add_datasource_features(ds)
 
         num_categorical = table_schema.get_num_categorical(ignore_cols)
 
@@ -731,9 +771,9 @@ class TestTableSchema:
         """Tests get_num_continuous for different types of ignore_cols."""
         # Create schema with continuous data columns
         data = pd.DataFrame([[1, 2], [3, 4]], columns=["A", "B"])
-        ds = DataSource(data)
+        ds = DataFrameSource(data)
         ds.load_data()
-        table_schema.add_dataframe_features(ds)
+        table_schema.add_datasource_features(ds)
 
         num_continuous = table_schema.get_num_continuous(ignore_cols)
 
@@ -838,7 +878,7 @@ class TestTableSchemaSplittingByDtype:
             {"A": int},
         ],
     )
-    def test_int_dtype(self, data: Dict[str, Any]) -> None:
+    def test_int_dtype(self, data: _JSONDict) -> None:
         """Tests that integers are given the correct semantic type."""
         types = TableSchema._dtype_based_stype_split(data)
         assert types == {SemanticType.CONTINUOUS: ["A"]}
@@ -906,7 +946,7 @@ class TestBitfountSchema:
 
     @unit_test
     def test_datasource_cannot_be_added_when_frozen(
-        self, datasource: DataSource, schema: BitfountSchema
+        self, datasource: BaseSource, schema: BitfountSchema
     ) -> None:
         """Tests that add_datasource_tables() fails when the schema is frozen."""
         schema.freeze()
@@ -942,7 +982,7 @@ class TestBitfountSchema:
 
     @unit_test
     def test_add_datasource_tables_requires_table_name_for_single_table_datasource(
-        self, schema: BitfountSchema, datasource: DataSource
+        self, datasource: BaseSource, schema: BitfountSchema
     ) -> None:
         """Tests that add_datasource_tables raises a ValueError.
 
@@ -958,15 +998,15 @@ class TestBitfountSchema:
     @integration_test
     def test_add_datasource_tables_with_multi_table_datasource(
         self,
-        schema: BitfountSchema,
-        db_session: sqlalchemy.engine.base.Engine,
         data: pd.DataFrame,
+        db_session: sqlalchemy.engine.base.Engine,
+        schema: BitfountSchema,
     ) -> None:
         """Tests that add_datasource_tables works with a multi-table datasource."""
         db_conn = DatabaseConnection(
             db_session, table_names=["dummy_data", "dummy_data_2"]
         )
-        ds = DataSource(db_conn, seed=420)
+        ds = DatabaseSource(db_conn, seed=420)
         schema.add_datasource_tables(
             ds,
             force_stypes={
@@ -983,24 +1023,28 @@ class TestBitfountSchema:
     @unit_test
     def test_mock_add_datasource_tables_with_multi_table_datasource(
         self,
-        schema: BitfountSchema,
-        mock_engine: Mock,
-        mock_pandas_read_sql_table: None,
         data: pd.DataFrame,
+        mock_engine: Mock,
+        mock_pandas_read_sql_query: None,
         mocker: MockerFixture,
+        schema: BitfountSchema,
     ) -> None:
         """Tests that add_datasource_tables works with a multi-table datasource."""
         db_conn = DatabaseConnection(
             mock_engine, table_names=["dummy_data", "dummy_data_2"]
         )
-        ds = DataSource(db_conn, seed=420)
-        mocker.patch.object(
-            ds.loader, "get_dtypes", return_value=create_dataset().dtypes.to_dict()
-        )
+        ds = DatabaseSource(db_conn, seed=420)
+
+        mock_dtypes = create_dataset().dtypes.to_dict()
+        mocker.patch.object(ds, "get_dtypes", return_value=mock_dtypes)
+        ds._table_hashes.add(_generate_dtypes_hash(mock_dtypes))
+
         mocker.patch.object(
             ds,
             "get_values",
-            side_effect=lambda x, y: {col: create_dataset()[col].unique() for col in x},
+            side_effect=lambda x, table_name: {
+                col: create_dataset()[col].unique() for col in x
+            },
         )
 
         schema.add_datasource_tables(ds)
@@ -1013,37 +1057,41 @@ class TestBitfountSchema:
     @unit_test
     def test_mock_add_datasource_tables_with_multi_table_datasource_table_name_warning(
         self,
-        schema: BitfountSchema,
-        mock_engine: Mock,
-        mock_pandas_read_sql_table: None,
         caplog: LogCaptureFixture,
+        mock_engine: Mock,
+        mock_pandas_read_sql_query: None,
         mocker: MockerFixture,
+        schema: BitfountSchema,
     ) -> None:
         """Tests that add_datasource_tables works with a multi-table datasource."""
         db_conn = DatabaseConnection(
             mock_engine, table_names=["dummy_data", "dummy_data_2"]
         )
-        ds = DataSource(db_conn, seed=420)
-        mocker.patch.object(
-            ds.loader, "get_dtypes", return_value=create_dataset().dtypes.to_dict()
-        )
+        ds = DatabaseSource(db_conn, seed=420)
+
+        mock_dtypes = create_dataset().dtypes.to_dict()
+        mocker.patch.object(ds, "get_dtypes", return_value=mock_dtypes)
+        ds._table_hashes.add(_generate_dtypes_hash(mock_dtypes))
+
         mocker.patch.object(
             ds,
             "get_values",
-            side_effect=lambda x, y: {col: create_dataset()[col].unique() for col in x},
+            side_effect=lambda x, table_name: {
+                col: create_dataset()[col].unique() for col in x
+            },
         )
 
         schema.add_datasource_tables(ds, table_name="this_should_be_ignored")
         assert len(schema.tables) == 2
         assert schema.table_names == ["dummy_data", "dummy_data_2"]
-        assert caplog.records[3].levelname == "WARNING"
-        assert caplog.records[3].message == (
-            "Ignoring table_name argument for multi-table datasource."
+        warning_logs = get_warning_logs(caplog)
+        assert (
+            "Ignoring table_name argument for multi-table datasource." in warning_logs
         )
 
     @unit_test
     def test_add_datasource_tables_adds_to_hash_list(
-        self, datasource: DataSource, mocker: MockerFixture, schema: BitfountSchema
+        self, datasource: BaseSource, mocker: MockerFixture, schema: BitfountSchema
     ) -> None:
         """Checks add_datasource_tables() adds to hash list.
 
@@ -1052,7 +1100,7 @@ class TestBitfountSchema:
         """
         # Mock out datasource.hash
         mock_hash = PropertyMock(return_value="datasource_hash")
-        mocker.patch.object(DataSource, "hash", mock_hash)
+        mocker.patch.object(BaseSource, "hash", mock_hash)
 
         assert schema._datasource_hashes == set()
 
@@ -1111,7 +1159,7 @@ class TestBitfountSchema:
 
     @unit_test
     def test_empty_schema_cannot_be_applied(
-        self, schema: BitfountSchema, data: pd.DataFrame
+        self, data: pd.DataFrame, schema: BitfountSchema
     ) -> None:
         """Tests that applying an empty schema raises an error."""
         with pytest.raises(
@@ -1122,7 +1170,7 @@ class TestBitfountSchema:
 
     @unit_test
     def test_multi_table_schema_cannot_be_applied_to_single_table_data(
-        self, schema_with_data: BitfountSchema, data: pd.DataFrame
+        self, data: pd.DataFrame, schema_with_data: BitfountSchema
     ) -> None:
         """Tests that applying a multi-table schema to single table data errors."""
         schema_with_data.tables.append(schema_with_data.tables[0])
@@ -1132,6 +1180,16 @@ class TestBitfountSchema:
         ):
             schema_with_data.apply(data)
 
+    @unit_test
+    def test_single_table_schema_apply_method_applies_schema(
+        self, data: pd.DataFrame, schema: BitfountSchema
+    ) -> None:
+        """Tests that apply() applies the schema to the data."""
+        mock_table_schema = Mock(spec=TableSchema)
+        schema.tables = [mock_table_schema]
+        schema.apply(data)
+        mock_table_schema.apply.assert_called_once()
+
     @unit_test
     def test__eq__passes_same_item(self, schema: BitfountSchema) -> None:
         """Test __eq__ passes when items are exact same."""
