@@ -2,9 +2,11 @@
 from typing import TYPE_CHECKING
 from unittest.mock import Mock, create_autospec
 
-from marshmallow import Schema as MarshmallowSchema
+import pytest
 from pytest_mock import MockerFixture
 
+from bitfount import BitfountSchema
+from bitfount.backends.pytorch.models.models import PyTorchTabularClassifier
 from bitfount.federated.algorithms.base import (
     _BaseAlgorithm,
     _BaseModellerAlgorithm,
@@ -21,8 +23,11 @@ from bitfount.federated.algorithms.model_algorithms.evaluate import (
     _ModellerSide,
     _WorkerSide,
 )
+from bitfount.federated.utils import _ALGORITHMS
 from bitfount.hub import BitfountHub
-from tests.utils.helper import unit_test
+from bitfount.schemas.utils import bf_dump, bf_load
+from tests.bitfount.backends.pytorch.models.test_models import assert_vars_equal
+from tests.utils.helper import backend_test, create_datastructure, unit_test
 
 
 class TestModelEvaluation:
@@ -56,6 +61,37 @@ class TestModelEvaluation:
         ]:
             assert isinstance(algorithm, type_)
 
+    @unit_test
+    def test_attribute_error_raised_run(self) -> None:
+        """Test non abstract class without run method raises AttributeError."""
+        err_msg = "does not have a `run` method"
+        for type_ in [
+            _BaseWorkerAlgorithm,
+            _BaseWorkerModelAlgorithm,
+        ]:
+            with pytest.raises(AttributeError) as e_info:
+
+                class NoRunMethod(type_):  # type: ignore[valid-type, misc] # reason: test child methods need run method # noqa: B950
+                    def initialise(self) -> None:
+                        pass
+
+            assert err_msg in str(e_info)
+
+    @unit_test
+    def test_attribute_error_raised_initialise(self) -> None:
+        """Test non abstract class without run method raises AttributeError."""
+        err_msg = "does not have a `initialise` method"
+
+        # All model algorithms have the initialise method on the
+        # base class so we only test the `_BaseWorkerAlgorithm`
+        with pytest.raises(AttributeError) as e_info:
+
+            class NoInitialiseMethod(_BaseWorkerAlgorithm):
+                def run(self) -> None:
+                    pass
+
+        assert err_msg in str(e_info)
+
     @unit_test
     def test_worker_run(self, mocker: MockerFixture, model: Mock) -> None:
         """Tests that worker run does metric calculation."""
@@ -72,16 +108,28 @@ class TestModelEvaluation:
     def test_modeller_run(self, model: Mock) -> None:
         """Tests that modeller run returns results."""
         modeller = evaluate._ModellerSide(model=model)
-        results = [{"AUC": 0.5}]
+        results = {"pod1": {"AUC": 0.5}}
         assert results == modeller.run(results=results)
 
-    @unit_test
-    def test_schema(self, model: Mock) -> None:
-        """Tests that schema returns parent class."""
-        schema_cls = ModelEvaluation.get_schema(create_autospec(MarshmallowSchema))
-        schema = schema_cls()
-        factory = schema.recreate_factory(data={"model": model})  # type: ignore[attr-defined]  # Reason: test will fail if wrong type  # noqa: B950
-        assert isinstance(factory, ModelEvaluation)
+
+@backend_test
+@unit_test
+class TestMarshmallowSerialization:
+    """Test Marshmallow Serialization for column average algorithm."""
+
+    def test_serialization(self) -> None:
+        """Test dump and load for ModelEvaluation."""
+        model = PyTorchTabularClassifier(
+            datastructure=create_datastructure(),
+            schema=BitfountSchema(),
+            epochs=2,
+        )
+        algorithm_factory = ModelEvaluation(model=model)
+        dumped = bf_dump(algorithm_factory)
+        loaded = bf_load(dumped, _ALGORITHMS)
+
+        assert algorithm_factory.class_name == loaded.class_name
+        assert_vars_equal(vars(algorithm_factory.model), vars(loaded.model))
 
 
 # Static tests for algorithm-protocol compatibility
@@ -90,8 +138,8 @@ if TYPE_CHECKING:
 
     from bitfount.federated.protocols.results_only import (
         _ResultsOnlyCompatibleModelAlgoFactory,
-        _ResultsOnlyCompatibleModeller,
-        _ResultsOnlyDataIncompatibleWorker,
+        _ResultsOnlyCompatibleModellerAlgorithm,
+        _ResultsOnlyDataIncompatibleWorkerAlgorithm,
     )
     from bitfount.types import (
         DistributedModelProtocol,
@@ -102,9 +150,9 @@ if TYPE_CHECKING:
     _algo_factory: _ResultsOnlyCompatibleModelAlgoFactory = ModelEvaluation(
         model=cast(_DistributedModelTypeOrReference, object())
     )
-    _modeller_side: _ResultsOnlyCompatibleModeller = _ModellerSide(
+    _modeller_side: _ResultsOnlyCompatibleModellerAlgorithm = _ModellerSide(
         model=cast(DistributedModelProtocol, object())
     )
-    _worker_side: _ResultsOnlyDataIncompatibleWorker = _WorkerSide(
+    _worker_side: _ResultsOnlyDataIncompatibleWorkerAlgorithm = _WorkerSide(
         model=cast(DistributedModelProtocol, object())
     )
