@@ -1,7 +1,7 @@
 """Tests for the federated averaging protocol."""
 from pathlib import Path
 import re
-from typing import Tuple
+from typing import Dict, Tuple, cast
 from unittest import mock
 from unittest.mock import ANY, AsyncMock, MagicMock, Mock, create_autospec
 
@@ -10,7 +10,9 @@ import pytest
 from pytest import fixture
 from pytest_mock import MockerFixture
 
-from bitfount.data import DataSource
+from bitfount import BitfountModelReference, BitfountSchema
+from bitfount.backends.pytorch import PyTorchTabularClassifier
+from bitfount.data.datasources.base_source import BaseSource
 import bitfount.federated.aggregators
 import bitfount.federated.aggregators.aggregator
 from bitfount.federated.aggregators.aggregator import Aggregator
@@ -20,13 +22,15 @@ from bitfount.federated.algorithms.model_algorithms import federated_training
 from bitfount.federated.algorithms.model_algorithms.federated_training import (
     FederatedModelTraining,
 )
-from bitfount.federated.protocols import fed_avg
+from bitfount.federated.helper import _create_aggregator
+import bitfount.federated.protocols.base as protocols
 from bitfount.federated.protocols.base import (
     _BaseModellerProtocol,
     _BaseProtocol,
     _BaseWorkerProtocol,
 )
-from bitfount.federated.protocols.fed_avg import (
+from bitfount.federated.protocols.model_protocols import federated_averaging
+from bitfount.federated.protocols.model_protocols.federated_averaging import (
     FederatedAveraging,
     _FederatedAveragingCompatibleAlgoFactory,
 )
@@ -35,15 +39,18 @@ from bitfount.federated.transport.worker_transport import (
     _InterPodWorkerMailbox,
     _WorkerMailbox,
 )
-from tests.bitfount import TEST_SECURITY_FILES
-from tests.utils.helper import unit_test
+from bitfount.schemas.utils import bf_dump, bf_load
+from bitfount.types import _JSONDict
+from tests.bitfount.backends.pytorch.models.test_models import assert_vars_equal
+from tests.utils import PytestRequest
+from tests.utils.helper import backend_test, create_datastructure, unit_test
 
 
 def mocked_modeller_runner_functions(
     mocker: MockerFixture,
 ) -> Tuple[Aggregator, AsyncMock, Mock]:
     """Mock functions used in protocol.modeller.run."""
-    mocker.patch.object(fed_avg._ModellerSide, "perform_iterations_checks")
+    mocker.patch.object(federated_averaging._ModellerSide, "perform_iterations_checks")
 
     mock_algorithm_run = Mock()
     mocker.patch.object(
@@ -53,25 +60,25 @@ def mocked_modeller_runner_functions(
     )
 
     mocker.patch.object(
-        fed_avg._ModellerSide,
+        federated_averaging._ModellerSide,
         "_send_parameters",
     )
 
     mocker.patch.object(
-        fed_avg._ModellerSide,
+        federated_averaging._ModellerSide,
         "_receive_parameter_updates",
         return_value=["some params"],
     )
 
     mocker.patch.object(
-        fed_avg._ModellerSide,
+        federated_averaging._ModellerSide,
         "get_num_federated_iterations",
         return_value=2,
     )
 
     mock_get_training_metrics = AsyncMock()
     mocker.patch.object(
-        fed_avg._ModellerSide,
+        federated_averaging._ModellerSide,
         "_get_training_metrics_updates",
         mock_get_training_metrics,
     )
@@ -85,7 +92,7 @@ def mocked_modeller_runner_functions(
         mock_aggregator_run,
     )
     mock_aggregator_run.return_value = {"some_weight": "some_value"}
-    mocked_aggregator_factory = Aggregator(tensor_shim=AsyncMock())
+    mocked_aggregator_factory = Aggregator()
 
     mock_task_complete = MagicMock()
     mock_task_complete.return_value.set_result(True)
@@ -100,7 +107,7 @@ def mocked_worker_runner_functions(
     mocker: MockerFixture,
 ) -> Tuple[Aggregator, AsyncMock]:
     """Mock functions used in protocol.worker.run."""
-    mocker.patch.object(fed_avg._WorkerSide, "perform_iterations_checks")
+    mocker.patch.object(federated_averaging._WorkerSide, "perform_iterations_checks")
 
     mocker.patch.object(
         federated_training._WorkerSide,
@@ -114,18 +121,18 @@ def mocked_worker_runner_functions(
     )
 
     mocker.patch.object(
-        fed_avg._WorkerSide,
+        federated_averaging._WorkerSide,
         "get_num_federated_iterations",
         return_value=2,
     )
 
     mocker.patch.object(
-        fed_avg._WorkerSide,
+        federated_averaging._WorkerSide,
         "_send_parameter_update",
     )
     mock_send_training_metrics = AsyncMock()
     mocker.patch.object(
-        fed_avg._WorkerSide,
+        federated_averaging._WorkerSide,
         "_send_training_metrics",
         mock_send_training_metrics,
     )
@@ -137,7 +144,7 @@ def mocked_worker_runner_functions(
         "run",
         mock_aggregator_run,
     )
-    mocked_aggregator_factory = Aggregator(tensor_shim=AsyncMock())
+    mocked_aggregator_factory = Aggregator()
 
     mock_task_complete = AsyncMock(return_value=True)
     mocker.patch.object(
@@ -150,10 +157,12 @@ def mocked_worker_runner_functions(
 class TestFederatedAveraging:
     """Test Federated Averaging protocol."""
 
-    @fixture
-    def federated_algorithm(self) -> FederatedModelTraining:
+    @fixture(scope="function", params=[None, "/mock/file"])
+    def federated_algorithm(self, request: PytestRequest) -> FederatedModelTraining:
         """Returns federated algorithm."""
-        return FederatedModelTraining(model=Mock())
+        return FederatedModelTraining(
+            model=Mock(), pretrained_file=request.param, modeller_checkpointing=False
+        )
 
     @fixture
     def mock_federated_algorithm(self) -> Mock:
@@ -222,7 +231,7 @@ class TestFederatedAveraging:
         for type_ in [
             _BaseProtocol,
             _BaseModellerProtocol,
-            fed_avg._ModellerSide,
+            federated_averaging._ModellerSide,
         ]:
             assert isinstance(protocol, type_)
 
@@ -245,7 +254,7 @@ class TestFederatedAveraging:
         for type_ in [
             _BaseProtocol,
             _BaseWorkerProtocol,
-            fed_avg._WorkerSide,
+            federated_averaging._WorkerSide,
         ]:
             assert isinstance(protocol, type_)
 
@@ -260,7 +269,8 @@ class TestFederatedAveraging:
         """Test worker method with different aggregator types."""
         # Mock out WorkerSide constructor
         mock_worker_side_cls = mocker.patch(
-            "bitfount.federated.protocols.fed_avg._WorkerSide", autospec=True
+            "bitfount.federated.protocols.model_protocols.federated_averaging._WorkerSide",
+            autospec=True,
         )
 
         # Test with an instance of AggregatorWorkerFactory
@@ -326,15 +336,12 @@ class TestFederatedAveraging:
             protocol_factory.worker(mailbox=mock_worker_mailbox, hub=mock_hub)
 
     @unit_test
-    def test_helper_run_method_without_model_or_algorithm_raises_value_error(
+    def test_helper_run_method_without_algorithm_raises_type_error(
         self,
     ) -> None:
-        """Tests helper run method without algorithm or model raises ValueError."""
-        with pytest.raises(ValueError):
-            FederatedAveraging.run(
-                pod_identifiers=["bitfount/prosper", "bitfount/prosper2"],
-                private_key_or_file=TEST_SECURITY_FILES / "test_private.testkey",
-            )
+        """Tests helper run method without algorithm raises TypeError."""
+        with pytest.raises(TypeError):
+            FederatedAveraging()  # type: ignore[call-arg] # Reason: this is what we are testing for # noqa: B950
 
     @unit_test
     def test_helper_run_method_with_algorithm(
@@ -342,24 +349,19 @@ class TestFederatedAveraging:
     ) -> None:
         """Tests helper run method with algorithm."""
         federated_algorithm_with_model = FederatedModelTraining(model=Mock())
-        mock_run_protocol = mocker.patch(
-            "bitfount.federated.protocols.fed_avg._run_protocol"
-        )
+        mock_run_protocol = mocker.patch.object(FederatedAveraging, "run")
         mock_create_aggregator = mocker.patch(
-            "bitfount.federated.protocols.fed_avg._create_aggregator"
-        )
-        mock_algorithm_constructor = mocker.patch(
-            "bitfount.federated.protocols.fed_avg.FederatedModelTraining",
+            "bitfount.federated.protocols.model_protocols.federated_averaging._create_aggregator"
         )
 
-        FederatedAveraging.run(
+        protocol = FederatedAveraging(algorithm=federated_algorithm_with_model)
+
+        protocol.run(
             pod_identifiers=["bitfount/fake", "bitfount/fake2"],
-            algorithm=federated_algorithm_with_model,
             hub=mock_hub,
             private_key_or_file=Path("fake.testkey"),
         )
 
-        mock_algorithm_constructor.assert_not_called()
         mock_run_protocol.assert_called_once()
         mock_create_aggregator.assert_called()
 
@@ -369,26 +371,20 @@ class TestFederatedAveraging:
     ) -> None:
         """Tests helper run method with a model."""
         federated_model = Mock()
-
-        mock_run_protocol = mocker.patch(
-            "bitfount.federated.protocols.fed_avg._run_protocol"
-        )
+        mock_run_protocol = mocker.patch.object(FederatedAveraging, "run")
         mock_create_aggregator = mocker.patch(
-            "bitfount.federated.protocols.fed_avg._create_aggregator"
+            "bitfount.federated.protocols.model_protocols.federated_averaging._create_aggregator"
         )
-        mock_algorithm_constructor = mocker.patch(
-            "bitfount.federated.protocols.fed_avg.FederatedModelTraining",
-            return_value=FederatedModelTraining(model=Mock()),
+        protocol = FederatedAveraging(
+            algorithm=FederatedModelTraining(model=federated_model)
         )
 
-        FederatedAveraging.run(
+        protocol.run(
             pod_identifiers=["bitfount/fake", "bitfount/fake2"],
-            model=federated_model,
             hub=mock_hub,
             private_key_or_file=Path("fake.testkey"),
         )
 
-        mock_algorithm_constructor.assert_called()
         mock_run_protocol.assert_called_once()
         mock_create_aggregator.assert_called()
 
@@ -402,28 +398,25 @@ class TestFederatedAveraging:
         regarding the extra model argument.
         """
         model = Mock()
+        mocker.patch.object(
+            model.datastructure, "get_pod_identifiers", return_value=None
+        )
         federated_algorithm_with_model = FederatedModelTraining(model=Mock())
 
-        mock_run_protocol = mocker.patch(
-            "bitfount.federated.protocols.fed_avg._run_protocol"
-        )
+        mock_run_protocol = mocker.patch.object(FederatedAveraging, "run")
         mock_create_aggregator = mocker.patch(
-            "bitfount.federated.protocols.fed_avg._create_aggregator",
-        )
-        mock_algorithm_constructor = mocker.patch(
-            "bitfount.federated.protocols.fed_avg.FederatedModelTraining",
+            "bitfount.federated.protocols.model_protocols.federated_averaging._create_aggregator",
         )
 
-        FederatedAveraging.run(
+        protocol = FederatedAveraging(
+            model=model, algorithm=federated_algorithm_with_model
+        )
+        protocol.run(
             pod_identifiers=["bitfount/fake", "bitfount/fake2"],
-            model=model,
-            algorithm=federated_algorithm_with_model,
             hub=mock_hub,
-            auto_eval=True,
             private_key_or_file=Path("fake.testkey"),
         )
 
-        mock_algorithm_constructor.assert_not_called()
         mock_run_protocol.assert_called_once()
         mock_create_aggregator.assert_called()
 
@@ -456,7 +449,7 @@ class TestFederatedAveraging:
             auto_eval=True,
         )
         protocol = protocol_factory.worker(mailbox=mock_worker_mailbox, hub=mock_hub)
-        await protocol.run(datasource=create_autospec(DataSource, instance=True))
+        await protocol.run(datasource=create_autospec(BaseSource, instance=True))
         mock_send_training_metrics.assert_called_once()
 
     @unit_test
@@ -481,7 +474,7 @@ class TestFederatedAveraging:
             auto_eval=False,
         )
         protocol = protocol_factory.worker(mailbox=mock_worker_mailbox, hub=mock_hub)
-        await protocol.run(datasource=create_autospec(DataSource, instance=True))
+        await protocol.run(datasource=create_autospec(BaseSource, instance=True))
         mock_send_training_metrics.assert_not_called()
 
     @unit_test
@@ -525,19 +518,25 @@ class TestFederatedAveraging:
             aggregator=mocked_aggregator_factory,
             steps_between_parameter_updates=1,
         )
+        mock_modeller_mailbox._task_id = "task_id"
         protocol = protocol_factory.modeller(
             mailbox=mock_modeller_mailbox, hub=mock_hub
         )
+        protocol.mailbox._task_id = "task_id"
         await protocol.run()
         mock_get_training_metrics.assert_called()
         assert protocol.validation_results == [{"AUC": "0.7"}, {"AUC": "0.7"}]
         algorithm_mock_calls = [
             mock.call(update=None),
             mock.call(
-                update={"some_weight": "some_value"}, validation_metrics={"AUC": "0.7"}
+                update={"some_weight": "some_value"},
+                validation_metrics={"AUC": "0.7"},
+                iteration=1,
             ),
             mock.call(
-                update={"some_weight": "some_value"}, validation_metrics={"AUC": "0.7"}
+                update={"some_weight": "some_value"},
+                validation_metrics={"AUC": "0.7"},
+                iteration=2,
             ),
         ]
         assert mock_algorithm_run.mock_calls == algorithm_mock_calls
@@ -566,6 +565,7 @@ class TestFederatedAveraging:
         protocol = protocol_factory.modeller(
             mailbox=mock_modeller_mailbox, hub=mock_hub
         )
+        protocol.mailbox._task_id = "task_id"
         await protocol.run()
         mock_get_training_metrics.assert_not_called()
         assert protocol.validation_results == []
@@ -575,3 +575,129 @@ class TestFederatedAveraging:
             mock.call(update={"some_weight": "some_value"}),
         ]
         assert mock_algorithm_run.mock_calls == algorithm_mock_calls
+
+    @unit_test
+    @pytest.mark.parametrize(
+        "iterations, model_reference",
+        [
+            ({"epochs": 1, "steps": None}, True),
+            ({"epochs": None, "steps": 1}, True),
+            ({"epochs": 1, "steps": None}, False),
+            ({"epochs": None, "steps": 1}, False),
+        ],
+    )
+    async def test_federated_averaging_protocol_no_iter(
+        self, iterations: Dict[str, int], mock_aggregator: Mock, model_reference: bool
+    ) -> None:
+        """Tests FederatedAveraging with no iteration provided.
+
+        Tests that iterations is set to a default value if no value
+        is provided.
+        """
+        if model_reference:
+            model = create_autospec(BitfountModelReference)
+            model.hyperparameters = iterations
+        else:
+            model = Mock(**iterations)
+        algorithm = FederatedModelTraining(model=model)
+
+        protocol = FederatedAveraging(
+            algorithm=algorithm,
+            aggregator=mock_aggregator,
+            steps_between_parameter_updates=None,
+            epochs_between_parameter_updates=None,
+            auto_eval=False,
+        )
+
+        if iterations.get("steps"):
+            assert protocol.steps_between_parameter_updates == 1
+            assert protocol.epochs_between_parameter_updates is None
+        else:
+            assert protocol.epochs_between_parameter_updates == 1
+            assert protocol.steps_between_parameter_updates is None
+
+    @unit_test
+    @pytest.mark.parametrize("secure_aggregation", [True, False])
+    async def test_federated_averaging_protocol_successfully(
+        self, secure_aggregation: bool
+    ) -> None:
+        """Tests that FederatedAveraging works as expected."""
+        model = Mock(epochs=1)
+
+        algorithm = FederatedModelTraining(model=model)
+
+        aggregator = _create_aggregator(secure_aggregation=secure_aggregation)
+        protocol = FederatedAveraging(
+            algorithm=algorithm,
+            aggregator=aggregator,
+            steps_between_parameter_updates=None,
+            epochs_between_parameter_updates=2,
+            auto_eval=False,
+        )
+        assert isinstance(protocol, FederatedAveraging)
+        assert protocol.epochs_between_parameter_updates == 2
+        assert protocol.steps_between_parameter_updates is None
+
+    @unit_test
+    def test_protocol_datastructure_error(self, mock_hub: Mock) -> None:
+        """Tests that protocol raises `DataStructure` error."""
+        model = Mock(epochs=1, datastructure=create_datastructure())
+
+        algorithm = FederatedModelTraining(model=model)
+
+        protocol = FederatedAveraging(
+            algorithm=algorithm,
+            steps_between_parameter_updates=None,
+            epochs_between_parameter_updates=2,
+            auto_eval=False,
+        )
+        with pytest.raises(ValueError):
+            protocol.run(pod_identifiers=["pod1", "pod2"], hub=mock_hub)
+
+
+@backend_test
+@unit_test
+class TestMarshmallowSerialization:
+    """Test Marshmallow Serialization for FederatedAveraging protocol."""
+
+    def test_serialization_model_algorithm(self) -> None:
+        """Test Marshmallow Serialization for FederatedAveraging protocol."""
+        model = PyTorchTabularClassifier(
+            datastructure=create_datastructure(),
+            schema=BitfountSchema(),
+            epochs=2,
+        )
+        algorithm_factory = FederatedModelTraining(model=model)
+        fed_avg = FederatedAveraging(
+            algorithm=algorithm_factory, aggregator=Aggregator()
+        )
+
+        dumped = bf_dump(fed_avg)
+        loaded = bf_load(dumped, protocols.registry)
+
+        assert vars(fed_avg).keys() == vars(loaded).keys()
+        assert fed_avg.algorithm.class_name == loaded.algorithm.class_name
+        assert fed_avg.aggregator.class_name == loaded.aggregator.class_name
+
+        assert_vars_equal(vars(fed_avg.algorithm.model), vars(loaded.algorithm.model))
+
+    def test_serialization_model_algorithm_from_protocol(self) -> None:
+        """Test Marshmallow Serialization for FederatedAveraging protocol."""
+        model = PyTorchTabularClassifier(
+            datastructure=create_datastructure(),
+            schema=BitfountSchema(),
+            epochs=2,
+        )
+        algorithm_factory = FederatedModelTraining(model=model)
+        fed_avg = FederatedAveraging(
+            algorithm=algorithm_factory, aggregator=Aggregator()
+        )
+
+        dumped = fed_avg.dump()
+        loaded = bf_load(cast(_JSONDict, dumped), protocols.registry)
+
+        assert vars(fed_avg).keys() == vars(loaded).keys()
+        assert fed_avg.algorithm.class_name == loaded.algorithm.class_name
+        assert fed_avg.aggregator.class_name == loaded.aggregator.class_name
+
+        assert_vars_equal(vars(fed_avg.algorithm.model), vars(loaded.algorithm.model))
