@@ -4,7 +4,7 @@ from pathlib import Path
 import re
 import sys
 from typing import Dict, List, Optional, cast
-from unittest.mock import Mock, create_autospec
+from unittest.mock import AsyncMock, MagicMock, Mock, create_autospec
 
 from _pytest.logging import LogCaptureFixture
 from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey
@@ -12,8 +12,9 @@ import pytest
 from pytest import fixture
 from pytest_mock import MockerFixture
 
-from bitfount.federated.aggregators.aggregator import Aggregator
-from bitfount.federated.algorithms.base import _BaseAlgorithmFactory
+from bitfount.backends.pytorch import PyTorchTabularClassifier
+from bitfount.data.schema import BitfountSchema
+from bitfount.federated.aggregators.secure import SecureAggregator
 from bitfount.federated.algorithms.model_algorithms.base import (
     _BaseModelAlgorithmFactory,
 )
@@ -26,20 +27,36 @@ from bitfount.federated.authorisation_checkers import (
     _SignatureBasedAuthorisation,
 )
 from bitfount.federated.encryption import _RSAEncryption
+from bitfount.federated.exceptions import PodResponseError
 from bitfount.federated.model_reference import BitfountModelReference
-from bitfount.federated.modeller import Modeller
+from bitfount.federated.modeller import _Modeller
 from bitfount.federated.protocols.base import _BaseProtocolFactory
-from bitfount.federated.task_requests import _ProtocolDetails
-from bitfount.federated.transport.base_transport import _WorkerMailboxDetails
+from bitfount.federated.protocols.model_protocols.federated_averaging import (
+    FederatedAveraging,
+)
 from bitfount.federated.transport.identity_verification.types import _ResponseHandler
-from bitfount.federated.transport.message_service import _MessageService
-from bitfount.federated.transport.modeller_transport import _ModellerMailbox
-from bitfount.federated.types import _TaskRequestMessageGenerator
+from bitfount.federated.transport.message_service import (
+    _BitfountMessageType,
+    _MessageService,
+)
+from bitfount.federated.transport.modeller_transport import (
+    _ModellerMailbox,
+    _WorkerMailboxDetails,
+)
+from bitfount.federated.types import (
+    SerializedAggregator,
+    SerializedAlgorithm,
+    SerializedModel,
+    SerializedProtocol,
+    _TaskRequestMessageGenerator,
+)
 from bitfount.hub.api import BitfountHub
 from bitfount.hub.authentication_flow import _AuthEnv
+from bitfount.models.base_models import _BaseModel
 from tests.bitfount import TEST_SECURITY_FILES
 from tests.utils import PytestRequest
 from tests.utils.helper import (
+    create_datastructure,
     get_debug_logs,
     get_error_logs,
     get_warning_logs,
@@ -103,32 +120,28 @@ class TestModeller:
         mock_protocol_factory: Mock = create_autospec(
             _BaseProtocolFactory, instance=True
         )
-        mock_protocol_factory.name = protocol_factory_name
-        mock_algorithm_factory = create_autospec(_BaseAlgorithmFactory, instance=True)
-        mock_algorithm_factory.name = algorithm_factory_name
-        mock_protocol_factory.algorithm = mock_algorithm_factory
+        mock_protocol_factory.algorithm = Mock()
+        dump_return_value = {
+            "class_name": protocol_factory_name,
+            "algorithm": {"class_name": algorithm_factory_name},
+        }
+
         # If we are testing with a protocol+aggregator, mock it out.
         if mock_protocol_factory_param != "protocol_without_aggregator":
-            mock_aggregator_factory = create_autospec(Aggregator, instance=True)
-            mock_protocol_factory.aggregator = mock_aggregator_factory
-            mock_aggregator_factory.name = aggregator_factory_name
+            dump_return_value["aggregator"] = {"class_name": aggregator_factory_name}
 
             if mock_protocol_factory_param == "protocol_with_aggregator":
-                mock_algorithm_factory = FederatedModelTraining(model=Mock())
-                mock_algorithm_factory.model.name = model_name
-                mock_protocol_factory.algorithm = mock_algorithm_factory
+                dump_return_value["algorithm"] = {
+                    "class_name": "FederatedModelTraining",
+                    "model": {"class_name": model_name},
+                }
             else:
-                mock_algorithm_factory = FederatedModelTraining(
-                    model=BitfountModelReference(
-                        model_ref="MyModel",
-                        username="username",
-                        datastructure=Mock(),
-                        schema=Mock(),
-                        hub=Mock(),
-                    )
-                )
-                mock_protocol_factory.algorithm = mock_algorithm_factory
+                dump_return_value["algorithm"] = {
+                    "class_name": "FederatedModelTraining",
+                    "model": {"class_name": "bitfount.BitfountModelReference"},
+                }
 
+        mock_protocol_factory.dump.return_value = dump_return_value
         return mock_protocol_factory
 
     @staticmethod
@@ -202,12 +215,12 @@ class TestModeller:
         mock_message_service: Mock,
         mock_private_key: Mock,
         mock_protocol_factory: Mock,
-    ) -> Modeller:
+    ) -> _Modeller:
         """Creates a modeller instance with most aspects mocked out.
 
         Optional args are not mocked, but are left as None.
         """
-        modeller = Modeller(
+        modeller = _Modeller(
             protocol=mock_protocol_factory,
             message_service=mock_message_service,
             bitfounthub=mock_bitfount_hub,
@@ -216,10 +229,15 @@ class TestModeller:
         return modeller
 
     @fixture
-    def mock_modeller_mailbox(self) -> Mock:
+    def mailbox_id(self) -> str:
+        """Fake mailbox id for modeller mailbox."""
+        return "modeller-mailbox-id"
+
+    @fixture
+    def mock_modeller_mailbox(self, mailbox_id: str) -> Mock:
         """Returns mock mailbox."""
-        # Returning "Any" since that is the given return type for create_autospec
         mock_mailbox: Mock = create_autospec(_ModellerMailbox, instance=True)
+        mock_mailbox.mailbox_id = mailbox_id
         return mock_mailbox
 
     async def test__send_task_requests(
@@ -227,16 +245,13 @@ class TestModeller:
         aggregator_factory_name: str,
         algorithm_factory_name: str,
         default_task_request_msg_gen: _TaskRequestMessageGenerator,
-        mock_aggregator_type_enum: Mock,
-        mock_algorithm_type_enum: Mock,
         mock_message_service: Mock,
         mock_modeller_mailbox: Mock,
         mock_private_key: Mock,
         mock_protocol_factory_param: str,
-        mock_protocol_type_enum: Mock,
         mocker: MockerFixture,
         model_name: str,
-        modeller: Modeller,
+        modeller: _Modeller,
         pod_identifiers: List[str],
         protocol_factory_name: str,
     ) -> None:
@@ -254,26 +269,30 @@ class TestModeller:
             _ModellerMailbox, "send_task_requests", autospec=True
         )
         mock_mailbox_send_task_requests.return_value = mock_modeller_mailbox
-
+        # mock_protocol_factory_param.class_name ="FederatedModelTraining"
         # Recreate the expected task_request_body
         if mock_protocol_factory_param == "protocol_with_aggregator":
-            expected_task_request_body = _ProtocolDetails(
-                protocol_factory_name,
-                "FederatedModelTraining",  # must be an instance of BaseModelAlgorithmFactory # noqa: B950
-                f"bitfount.{model_name}",
-                aggregator_factory_name,
+            expected_task_request_body = SerializedProtocol(
+                class_name=protocol_factory_name,
+                algorithm=SerializedAlgorithm(
+                    class_name="FederatedModelTraining",  # must be an instance of BaseModelAlgorithmFactory # noqa: B950
+                    model=SerializedModel(class_name=f"{model_name}"),
+                ),
+                aggregator=SerializedAggregator(class_name=aggregator_factory_name),
             )
         elif mock_protocol_factory_param == "protocol_with_aggregator_custom_model":
-            expected_task_request_body = _ProtocolDetails(
-                protocol_factory_name,
-                "FederatedModelTraining",  # must be an instance of BaseModelAlgorithmFactory # noqa: B950
-                "username.MyModel",
-                aggregator_factory_name,
+            expected_task_request_body = SerializedProtocol(
+                class_name=protocol_factory_name,
+                algorithm=SerializedAlgorithm(
+                    class_name="FederatedModelTraining",  # must be an instance of BaseModelAlgorithmFactory # noqa: B950
+                    model=SerializedModel(class_name="bitfount.BitfountModelReference"),
+                ),
+                aggregator=SerializedAggregator(class_name=aggregator_factory_name),
             )
         else:
-            expected_task_request_body = _ProtocolDetails(
-                protocol_factory_name,
-                algorithm_factory_name,
+            expected_task_request_body = SerializedProtocol(
+                class_name=protocol_factory_name,
+                algorithm=SerializedAlgorithm(class_name=algorithm_factory_name),
             )
 
         modeller_mailbox = await modeller._send_task_requests(pod_identifiers)
@@ -283,7 +302,7 @@ class TestModeller:
         assert modeller_mailbox == mock_modeller_mailbox
         # The class method should have been called with correctly constructed args.
         mock_mailbox_send_task_requests.assert_called_once_with(
-            protocol_details=expected_task_request_body,
+            serialized_protocol=expected_task_request_body,
             pod_public_keys=mock_get_public_keys(),
             message_service=mock_message_service,
             task_request_msg_gen=default_task_request_msg_gen,
@@ -302,14 +321,13 @@ class TestModeller:
             ValueError,
             match=re.escape(f"Unsupported identity verification method: {fake_method}"),
         ):
-            Modeller(
+            _Modeller(
                 protocol=mock_protocol_factory,
                 message_service=mock_message_service,
                 bitfounthub=mock_bitfount_hub,
                 identity_verification_method=fake_method,
             )
 
-    @pytest.mark.skip("Signature-based access not working until [BIT-1291] implemented")
     def test_modeller_init_fails_key_based_verification_but_no_key(
         self,
         mock_bitfount_hub: Mock,
@@ -323,11 +341,11 @@ class TestModeller:
                 "Key-based identity verification selected but no private key provided."
             ),
         ):
-            Modeller(
+            _Modeller(
                 protocol=mock_protocol_factory,
                 message_service=mock_message_service,
                 bitfounthub=mock_bitfount_hub,
-                identity_verification_method=IdentityVerificationMethod.KEYS,  # type: ignore[attr-defined] # Reason: Disabled until [BIT-1291] resolved # noqa: B950
+                identity_verification_method=IdentityVerificationMethod.KEYS,
                 private_key=None,
             )
 
@@ -340,7 +358,7 @@ class TestModeller:
         mock_protocol_factory: Mock,
     ) -> None:
         """Tests Modeller.__init__ warns if key provided for non-key verification."""
-        Modeller(
+        _Modeller(
             protocol=mock_protocol_factory,
             message_service=mock_message_service,
             bitfounthub=mock_bitfount_hub,
@@ -355,7 +373,6 @@ class TestModeller:
             f"will be ignored." in warning_logs
         )
 
-    @pytest.mark.skip("Signature-based access not working until [BIT-1291] implemented")
     def test_modeller_init_loads_key_if_path(
         self,
         mock_bitfount_hub: Mock,
@@ -372,11 +389,11 @@ class TestModeller:
         # Fake key path
         fake_path = create_autospec(Path, instance=True)
 
-        m = Modeller(
+        m = _Modeller(
             protocol=mock_protocol_factory,
             message_service=mock_message_service,
             bitfounthub=mock_bitfount_hub,
-            identity_verification_method=IdentityVerificationMethod.KEYS,  # type: ignore[attr-defined] # Reason: Disabled until [BIT-1291] resolved # noqa: B950
+            identity_verification_method=IdentityVerificationMethod.KEYS,
             private_key=fake_path,
         )
 
@@ -384,7 +401,6 @@ class TestModeller:
         mock_load_private_key.assert_called_once_with(fake_path)
         assert m._private_key == mock_load_private_key.return_value
 
-    @pytest.mark.skip("Signature-based access not working until [BIT-1291] implemented")
     def test_modeller_init_loads_key_if_key(
         self,
         mock_bitfount_hub: Mock,
@@ -393,11 +409,11 @@ class TestModeller:
         mock_protocol_factory: Mock,
     ) -> None:
         """Tests Modeller.__init__ uses provided key."""
-        m = Modeller(
+        m = _Modeller(
             protocol=mock_protocol_factory,
             message_service=mock_message_service,
             bitfounthub=mock_bitfount_hub,
-            identity_verification_method=IdentityVerificationMethod.KEYS,  # type: ignore[attr-defined] # Reason: Disabled until [BIT-1291] resolved # noqa: B950
+            identity_verification_method=IdentityVerificationMethod.KEYS,
             private_key=mock_private_key,
         )
 
@@ -408,7 +424,7 @@ class TestModeller:
         self,
         mock_modeller_mailbox: Mock,
         mocker: MockerFixture,
-        modeller: Modeller,
+        modeller: _Modeller,
         pod_identifiers: List[str],
     ) -> None:
         """Tests _modeller_run performs expected operations."""
@@ -418,25 +434,227 @@ class TestModeller:
             for pod_identifier in pod_identifiers
         }
         mock_modeller_mailbox.accepted_worker_mailboxes = mock_worker_mailboxes
+        # Mock out next call in chain
+        mock__run_modeller_protocol = mocker.patch.object(
+            modeller, "_run_modeller_protocol", autospec=True
+        )
+
+        await modeller._modeller_run(
+            modeller_mailbox=mock_modeller_mailbox, pod_identifiers=pod_identifiers
+        )
+
+        # Check task start message was sent
+        mock_modeller_mailbox.send_task_start_message.assert_awaited_once()
+        # Check task response processing was called
+        mock_modeller_mailbox.process_task_request_responses.assert_awaited_once()
+        # Check next stage called
+        mock__run_modeller_protocol.assert_awaited_once()
+
+    def test_modeller_run_secure_aggregation_params(
+        self,
+        caplog: LogCaptureFixture,
+        mock_bitfount_hub: Mock,
+        mock_message_service: Mock,
+        mock_private_key: Mock,
+        mocker: MockerFixture,
+        pod_identifiers: List[str],
+    ) -> None:
+        """Tests parameters for clipping are set for secure aggregation."""
+        model = PyTorchTabularClassifier(
+            datastructure=create_datastructure(),
+            schema=BitfountSchema(),
+            epochs=2,
+        )
+        algorithm_factory = FederatedModelTraining(model=model)
+        fed_avg = FederatedAveraging(
+            algorithm=algorithm_factory, aggregator=SecureAggregator()
+        )
+        modeller = _Modeller(
+            protocol=fed_avg,
+            message_service=mock_message_service,
+            bitfounthub=mock_bitfount_hub,
+            private_key=mock_private_key,
+        )
+
+        mocker.patch.object(
+            _Modeller, "run_async", new_callable=AsyncMock, return_value=False
+        )
+
+        modeller.run(pod_identifiers=pod_identifiers)
+        assert (
+            "SecureAggregation in use. "
+            "We recommend normalization of continuous features prior to training."
+            in get_warning_logs(caplog)
+        )
+        assert (
+            modeller.protocol.algorithm.model.param_clipping  # type: ignore[attr-defined] # Reason: algorithm defined with model # noqa: B950
+            == {
+                "prime_q": 2**61 - 1,
+                "precision": 10**10,
+                "num_workers": 2,
+            }
+        )
+
+    def test_modeller_run_secure_aggregation_params_steps(
+        self,
+        caplog: LogCaptureFixture,
+        mock_bitfount_hub: Mock,
+        mock_message_service: Mock,
+        mock_private_key: Mock,
+        mocker: MockerFixture,
+        pod_identifiers: List[str],
+    ) -> None:
+        """Tests parameters for clipping are not set for secure aggregation.
+
+        Tests the case when steps_between_parameter_updates=1 in
+        the protocol definition.
+        """
+        model = PyTorchTabularClassifier(
+            datastructure=create_datastructure(),
+            schema=BitfountSchema(),
+            steps=2,
+        )
+        algorithm_factory = FederatedModelTraining(model=model)
+        fed_avg = FederatedAveraging(
+            algorithm=algorithm_factory,
+            aggregator=SecureAggregator(),
+            steps_between_parameter_updates=1,
+        )
+        modeller = _Modeller(
+            protocol=fed_avg,
+            message_service=mock_message_service,
+            bitfounthub=mock_bitfount_hub,
+            private_key=mock_private_key,
+        )
+
+        mocker.patch.object(
+            _Modeller, "run_async", new_callable=AsyncMock, return_value=False
+        )
+
+        modeller.run(pod_identifiers=pod_identifiers)
+        assert (
+            caplog.records[-1].msg != "SecureAggregation in use. "
+            "We recommend normalization of continuous features prior to training."
+        )
+        assert modeller.protocol.algorithm.model.param_clipping is None  # type: ignore[attr-defined] # Reason: algorithm defined with model # noqa: B950
+
+    def test_modeller_run_secure_aggregation_params_bfmodel(
+        self,
+        bitfount_model_correct_structure: str,
+        caplog: LogCaptureFixture,
+        mock_bitfount_hub: Mock,
+        mock_message_service: Mock,
+        mock_private_key: Mock,
+        mocker: MockerFixture,
+        pod_identifiers: List[str],
+        tmp_path: Path,
+    ) -> None:
+        """Tests parameters for clipping are set for secure aggregation.
+
+        Tests that an extra warning is raised in the case
+        of a custom model in the protocol definition.
+        """
+        model_file = tmp_path / "MyModel.py"
+        model_file.touch()
+        model_file.write_text(bitfount_model_correct_structure)
+        hub_mock = MagicMock()
+        hub_mock.send_model.return_value = True
+        hub_mock.username = "test_username"
+        model_ref = BitfountModelReference(
+            username="test",
+            datastructure=create_datastructure(),
+            schema=BitfountSchema(),
+            model_ref=model_file,
+            hub=hub_mock,
+            hyperparameters={"param1": 1, "param2": 2},
+        )
+        algorithm_factory = FederatedModelTraining(model=model_ref)
+        fed_avg = FederatedAveraging(
+            algorithm=algorithm_factory,
+            aggregator=SecureAggregator(),
+            steps_between_parameter_updates=2,
+        )
+        modeller = _Modeller(
+            protocol=fed_avg,
+            message_service=mock_message_service,
+            bitfounthub=mock_bitfount_hub,
+            private_key=mock_private_key,
+        )
+
+        mocker.patch.object(
+            _Modeller, "run_async", new_callable=AsyncMock, return_value=False
+        )
+
+        modeller.run(pod_identifiers=pod_identifiers)
+
+        # Check logs
+        warning_logs = get_warning_logs(caplog)
+        assert (
+            "SecureAggregation in use. "
+            "We recommend normalization of continuous features prior to training."
+            in warning_logs
+        )
+        assert (
+            "You are using a custom model with Secure Aggregation."
+            "We recommend clipping the model parameters." in warning_logs
+        )
+
+    @pytest.mark.parametrize("require_all_pods", [True, False])
+    async def test__modeller_run_offline_pods(
+        self,
+        caplog: LogCaptureFixture,
+        mock_modeller_mailbox: Mock,
+        mocker: MockerFixture,
+        modeller: _Modeller,
+        pod_identifiers: List[str],
+        require_all_pods: bool,
+    ) -> None:
+        """Tests _modeller_run performs errors/warns about offline pods."""
+        # Set "accepted" pods
+        mock_worker_mailboxes = {
+            pod_identifiers[0]: create_autospec(_WorkerMailboxDetails, instance=True)
+        }
+        mock_modeller_mailbox.accepted_worker_mailboxes = mock_worker_mailboxes
 
         # Mock out next call in chain
         mock__run_modeller_protocol = mocker.patch.object(
             modeller, "_run_modeller_protocol", autospec=True
         )
 
-        await modeller._modeller_run(modeller_mailbox=mock_modeller_mailbox)
+        err_msg = "Pods user2/pod2 rejected task request or failed to respond. "
+
+        if require_all_pods:
+            with pytest.raises(
+                PodResponseError,
+                match=err_msg + "Task requires all pods accept the task request.",
+            ):
+                await modeller._modeller_run(
+                    modeller_mailbox=mock_modeller_mailbox,
+                    pod_identifiers=pod_identifiers,
+                    require_all_pods=require_all_pods,
+                )
+                # Check next stage not called
+                mock__run_modeller_protocol.assert_not_called()
+        else:
+            await modeller._modeller_run(
+                modeller_mailbox=mock_modeller_mailbox,
+                pod_identifiers=pod_identifiers,
+                require_all_pods=require_all_pods,
+            )
+            # Check next stage called
+            mock__run_modeller_protocol.assert_awaited_once()
+            warning_logs = get_warning_logs(caplog)
+            assert warning_logs == err_msg + "Continuing task without these pods ..."
 
         # Check task response processing was called
         mock_modeller_mailbox.process_task_request_responses.assert_awaited_once()
-        # Check next stage called
-        mock__run_modeller_protocol.assert_awaited_once()
 
     async def test__modeller_run_fails_fast_if_no_accepted_pods(
         self,
         caplog: LogCaptureFixture,
         mock_modeller_mailbox: Mock,
         mocker: MockerFixture,
-        modeller: Modeller,
+        modeller: _Modeller,
     ) -> None:
         """Tests _modeller_run method fails fast if no pods accept."""
         # Set no "accepted" pods
@@ -447,7 +665,9 @@ class TestModeller:
             modeller, "_run_modeller_protocol", autospec=True
         )
 
-        result = await modeller._modeller_run(modeller_mailbox=mock_modeller_mailbox)
+        result = await modeller._modeller_run(
+            modeller_mailbox=mock_modeller_mailbox, pod_identifiers=["pod_id"]
+        )
 
         # Check false return value
         assert result is False
@@ -461,7 +681,7 @@ class TestModeller:
         self,
         mock_modeller_mailbox: Mock,
         mocker: MockerFixture,
-        modeller: Modeller,
+        modeller: _Modeller,
         pod_identifiers: List[str],
     ) -> None:
         """Tests _modeller_run method when provided with a ResponseHandler."""
@@ -483,6 +703,7 @@ class TestModeller:
         await modeller._modeller_run(
             modeller_mailbox=mock_modeller_mailbox,
             response_handler=mock_response_handler,
+            pod_identifiers=pod_identifiers,
         )
 
         # Check task response pre-processing was called
@@ -496,7 +717,7 @@ class TestModeller:
         self,
         mock_modeller_mailbox: Mock,
         mocker: MockerFixture,
-        modeller: Modeller,
+        modeller: _Modeller,
         pod_identifiers: List[str],
     ) -> None:
         """Tests run_async method performs expected operations."""
@@ -512,14 +733,16 @@ class TestModeller:
             modeller, "_modeller_run", autospec=True
         )
 
-        result = await modeller.run_async(pod_identifiers)
+        result = await modeller.run_async(pod_identifiers, require_all_pods=False)
 
         # Assert task requests sent
         mock__send_task_requests.assert_awaited_once()
         # Assert next stage called
         mock__modeller_run.assert_awaited_once()
         # Check log message handler removed
-        mock_modeller_mailbox.delete_handler.assert_called_once()
+        mock_modeller_mailbox.delete_all_handlers.assert_called_once_with(
+            _BitfountMessageType.LOG_MESSAGE
+        )
         # Check return value
         assert result == mock__modeller_run.return_value
 
@@ -527,7 +750,7 @@ class TestModeller:
         self,
         mock_modeller_mailbox: Mock,
         mocker: MockerFixture,
-        modeller: Modeller,
+        modeller: _Modeller,
         pod_identifiers: List[str],
     ) -> None:
         """Test run_async correctly starts SAML challenge handler."""
@@ -552,7 +775,7 @@ class TestModeller:
         )
         mock_saml_handler = mock_saml_handler_cls.return_value
 
-        result = await modeller.run_async(pod_identifiers)
+        result = await modeller.run_async(pod_identifiers, require_all_pods=False)
 
         # Assert SAML server running
         mock_saml_handler.start_server.assert_called_once()
@@ -562,7 +785,9 @@ class TestModeller:
         # Assert next stage called
         mock__modeller_run.assert_awaited_once()
         # Check log message handler removed
-        mock_modeller_mailbox.delete_handler.assert_called_once()
+        mock_modeller_mailbox.delete_all_handlers.assert_called_once_with(
+            _BitfountMessageType.LOG_MESSAGE
+        )
         # Check return value
         assert result == mock__modeller_run.return_value
 
@@ -571,7 +796,7 @@ class TestModeller:
         caplog: LogCaptureFixture,
         mock_modeller_mailbox: Mock,
         mocker: MockerFixture,
-        modeller: Modeller,
+        modeller: _Modeller,
         pod_identifiers: List[str],
     ) -> None:
         """Test run_async correctly starts OIDC auth code challenge handler."""
@@ -611,7 +836,7 @@ class TestModeller:
         mock_oidc_handler = mock_oidc_handler_cls.return_value
 
         with caplog.at_level(logging.DEBUG):
-            result = await modeller.run_async(pod_identifiers)
+            result = await modeller.run_async(pod_identifiers, require_all_pods=False)
 
         # Assert OIDC server running
         mock_oidc_handler.start_server.assert_called_once()
@@ -631,7 +856,9 @@ class TestModeller:
         # Assert next stage called
         mock__modeller_run.assert_awaited_once()
         # Check log message handler removed
-        mock_modeller_mailbox.delete_handler.assert_called_once()
+        mock_modeller_mailbox.delete_all_handlers.assert_called_once_with(
+            _BitfountMessageType.LOG_MESSAGE
+        )
         # Check return value
         assert result == mock__modeller_run.return_value
 
@@ -639,7 +866,7 @@ class TestModeller:
         self,
         mock_modeller_mailbox: Mock,
         mocker: MockerFixture,
-        modeller: Modeller,
+        modeller: _Modeller,
         pod_identifiers: List[str],
     ) -> None:
         """Test run_async correctly starts OIDC device code challenge handler."""
@@ -677,7 +904,7 @@ class TestModeller:
             "bitfount.federated.modeller._OIDCDeviceCodeHandler", autospec=True
         )
 
-        result = await modeller.run_async(pod_identifiers)
+        result = await modeller.run_async(pod_identifiers, require_all_pods=False)
 
         # Check OIDC handler constructed correctly
         mock_oidc_handler_cls.assert_called_once_with(
@@ -689,7 +916,9 @@ class TestModeller:
         # Assert next stage called
         mock__modeller_run.assert_awaited_once()
         # Check log message handler removed
-        mock_modeller_mailbox.delete_handler.assert_called_once()
+        mock_modeller_mailbox.delete_all_handlers.assert_called_once_with(
+            _BitfountMessageType.LOG_MESSAGE
+        )
         # Check return value
         assert result == mock__modeller_run.return_value
 
@@ -697,14 +926,14 @@ class TestModeller:
         self,
         caplog: LogCaptureFixture,
         mocker: MockerFixture,
-        modeller: Modeller,
+        modeller: _Modeller,
         pod_identifiers: List[str],
     ) -> None:
         """Test run_async finally block handles non-existent handler variable."""
         mocker.patch.object(modeller, "_get_response_handler", side_effect=Exception)
 
         with pytest.raises(Exception):
-            await modeller.run_async(pod_identifiers)
+            await modeller.run_async(pod_identifiers, require_all_pods=False)
 
         warning_logs = get_warning_logs(caplog)
         assert "Tried to shutdown non-existent response handler" in warning_logs
@@ -712,10 +941,11 @@ class TestModeller:
     async def test_run_async_finally_works_with_non_server_response_handler(
         self,
         mocker: MockerFixture,
-        modeller: Modeller,
+        modeller: _Modeller,
         pod_identifiers: List[str],
     ) -> None:
         """Test run_async finally block handles non-server handler variable."""
+        require_all_pods = False
         # Mock out response_handler. We don't use autospec so we can guarantee
         # that stop_server isn't present.
         mock_get_response_handler = mocker.patch.object(
@@ -735,16 +965,56 @@ class TestModeller:
         )
 
         with pytest.raises(Exception, match="specific exception"):
-            await modeller.run_async(pod_identifiers)
+            await modeller.run_async(pod_identifiers, require_all_pods)
+
+    async def test_modeller_aborts_task_correctly_if_pod_does_not_exist(
+        self,
+        caplog: LogCaptureFixture,
+        mocker: MockerFixture,
+        modeller: _Modeller,
+    ) -> None:
+        """Test modeller aborts the task as expected if there is a non-existent pod.
+
+        This includes raising an informative error message.
+        """
+        pod_identifier = "non-existent-pod"
+        mocker.patch.object(modeller, "_get_response_handler", return_value=AsyncMock())
+        mock_get_pod_key = mocker.patch.object(
+            modeller._hub, "get_pod_key", return_value=None
+        )
+        result = await modeller.run_async(
+            pod_identifiers=[pod_identifier], require_all_pods=False
+        )
+        assert result is False
+        final_log_message = caplog.records[-1]
+        penultimate_log_message = caplog.records[-2]
+
+        assert final_log_message.levelname == "ERROR"
+        assert final_log_message.message == "Aborted task request."
+        assert penultimate_log_message.levelname == "ERROR"
+        assert (
+            penultimate_log_message.message
+            == f"No public key found for pod: {pod_identifier}"
+        )
+        mock_get_pod_key.assert_called_once_with(pod_identifier=pod_identifier)
 
-    @pytest.mark.skip("Signature-based access not working until [BIT-1291] implemented")
     def test__get_task_request_msg_gen_key_based(
-        self, mock_private_key: Mock, mocker: MockerFixture, modeller: Modeller
+        self,
+        mock_bitfount_hub: Mock,
+        mock_message_service: Mock,
+        mock_private_key: Mock,
+        mock_protocol_factory: Mock,
+        mocker: MockerFixture,
     ) -> None:
         """Test request message generator creation for key-based authentication."""
         # Guarantee private key and key-based on modeller
-        modeller._private_key = mock_private_key
-        modeller._identity_verification_method = IdentityVerificationMethod.KEYS  # type: ignore[attr-defined] # Reason: Disabled until [BIT-1291] resolved # noqa: B950
+        modeller = _Modeller(
+            protocol=mock_protocol_factory,
+            message_service=mock_message_service,
+            bitfounthub=mock_bitfount_hub,
+            private_key=mock_private_key,
+            identity_verification_method=IdentityVerificationMethod.KEYS,
+        )
 
         # Wrap mock around authorisation checker class
         wrapped_creator = mocker.patch.object(
@@ -760,31 +1030,35 @@ class TestModeller:
         # Check returned gen
         assert isinstance(msg_gen, _TaskRequestMessageGenerator)
 
-    @pytest.mark.skip("Signature-based access not working until [BIT-1291] implemented")
     def test__get_task_request_msg_gen_key_based_fails_no_key(
         self,
-        modeller: Modeller,
+        mock_bitfount_hub: Mock,
+        mock_message_service: Mock,
+        mock_protocol_factory: Mock,
     ) -> None:
         """Test message generator creation failure for key-based authentication.
 
         Will fail if no key provided.
         """
-        # Guarantee NO private key and key-based on modeller
-        modeller._private_key = None
-        modeller._identity_verification_method = IdentityVerificationMethod.KEYS  # type: ignore[attr-defined] # Reason: Disabled until [BIT-1291] resolved # noqa: B950
-
         with pytest.raises(
             ValueError,
             match=re.escape(
-                "Signature-based identification selected but no private key provided."
+                "Key-based identity verification selected but no private key provided."
             ),
         ):
-            modeller._get_task_request_msg_gen()
+            # Guarantee NO private key and key-based on modeller
+            _Modeller(
+                protocol=mock_protocol_factory,
+                message_service=mock_message_service,
+                bitfounthub=mock_bitfount_hub,
+                private_key=None,
+                identity_verification_method=IdentityVerificationMethod.KEYS,
+            )
 
     def test__get_task_request_msg_gen_saml_based(
         self,
         mocker: MockerFixture,
-        modeller: Modeller,
+        modeller: _Modeller,
     ) -> None:
         """Test request message generator creation for key-based authentication."""
         # Guarantee SAML-based on modeller
@@ -806,7 +1080,7 @@ class TestModeller:
 
     @pytest.mark.parametrize("model_out", (None, Path("out_file.pt")))
     def test_run(
-        self, modeller: Modeller, mocker: MockerFixture, model_out: Optional[Path]
+        self, mocker: MockerFixture, model_out: Optional[Path], modeller: _Modeller
     ) -> None:
         """Test modeller run method."""
         pod_id = "pod_id"
@@ -815,7 +1089,7 @@ class TestModeller:
         )
         mock_run = mocker.patch("asyncio.run")
         mock_serialize = mocker.patch.object(modeller, "_serialize")
-        modeller.run(pod_id, model_out)
+        modeller.run(pod_id, model_out=model_out)
         mock_check_and_update_pod_ids.assert_called_once()
         if model_out:
             mock_serialize.assert_called_once()
@@ -826,23 +1100,33 @@ class TestModeller:
     @pytest.mark.parametrize("protocol", ("ResultsOnly", "NotResultsOnly"))
     def test__serialize(
         self,
-        modeller: Modeller,
-        mocker: MockerFixture,
-        protocol: str,
+        algorithm_factory_name: str,
         mock_protocol_factory_param: str,
+        mocker: MockerFixture,
+        model_name: str,
+        modeller: _Modeller,
+        protocol: str,
     ) -> None:
         """Test serializing of model."""
         if mock_protocol_factory_param == "protocol_with_aggregator":
             file_name = Path("file_name")
-            modeller.protocol.name = protocol
+
+            modeller.protocol.class_name = protocol
+            mock_algorithm_factory = create_autospec(
+                _BaseModelAlgorithmFactory, instance=True
+            )
+            mock_algorithm_factory.class_name = algorithm_factory_name
+            mock_algorithm_factory.model = Mock(spec=_BaseModel)
+            mock_algorithm_factory.model.class_name = model_name
+            modeller.protocol.algorithm = mock_algorithm_factory
+
             assert isinstance(  # nosec
                 modeller.protocol.algorithm, _BaseModelAlgorithmFactory
             )
-            mock_serialize = mocker.patch.object(
-                modeller.protocol.algorithm.model, "serialize"
-            )
+
             modeller._serialize(file_name)
+            # Mypy can't detect that the model is actually a Mock
             if protocol == "ResultsOnly":
-                mock_serialize.assert_not_called()
+                modeller.protocol.algorithm.model.serialize.assert_not_called()  # type: ignore[union-attr] # Reason: see above # noqa: B950
             else:
-                mock_serialize.assert_called_once()
+                modeller.protocol.algorithm.model.serialize.assert_called_once()  # type: ignore[union-attr] # Reason: see above # noqa: B950
