@@ -1,10 +1,12 @@
 """Test worker can communicate with modeller."""
+from __future__ import annotations
+
 import asyncio
 from collections import namedtuple
 from datetime import datetime, timezone
 import logging
 import re
-from typing import Callable, Dict, Iterable, List
+from typing import Callable, Dict, Iterable, List, Optional
 from unittest.mock import ANY, Mock, NonCallableMagicMock, call, create_autospec
 
 from _pytest.logging import LogCaptureFixture
@@ -27,15 +29,24 @@ from bitfount.federated.transport.message_service import (
 )
 from bitfount.federated.transport.protos.messages_pb2 import SuccessResponse
 from bitfount.federated.transport.worker_transport import (
+    _DEFAULT_AUTHENTICATION_MODELLER_RESPONSE_TIMEOUT,
+    _SOFT_LIMIT_MESSAGE_TIMEOUT,
+    _get_psi_dataset,
     _get_worker_secure_shares,
     _InterPodWorkerMailbox,
     _send_secure_shares_to_others,
     _WorkerMailbox,
 )
 from bitfount.federated.types import _PodResponseType
-from tests.utils.fixtures import MessageOrException
-from tests.utils.helper import get_debug_logs, get_error_logs, unit_test
-from tests.utils.mocks import AsyncIteratorMock
+from tests.utils import PytestRequest
+from tests.utils.fixtures import PollForMessagesPatcher
+from tests.utils.helper import (
+    get_debug_logs,
+    get_error_logs,
+    get_info_logs,
+    get_warning_logs,
+    unit_test,
+)
 
 WorkerDetails = namedtuple("WorkerDetails", ["pod_identifier", "mailbox_id"])
 
@@ -101,12 +112,29 @@ def pod_mailbox_ids(
     }
 
 
+@fixture(params=(True, False), ids=("task_id_incl", "no_task_id"))
+def opt_task_id(request: PytestRequest) -> Optional[str]:
+    """Returns a task ID or None, to cover both cases."""
+    incl_task_id: bool = request.param
+    if incl_task_id:
+        return "this-is-a-task-id"
+    else:
+        return None
+
+
+@fixture
+def online_check_uuid() -> str:
+    """UUID for online checking."""
+    return "auuidforonlinecheck"
+
+
 @fixture
 def worker_mailbox(
     aes_key: bytes,
     mock_message_service: Mock,
     modeller_mailbox_id: str,
     modeller_name: str,
+    opt_task_id: Optional[str],
     pod_identifier: str,
     pod_mailbox_ids: Dict[str, str],
 ) -> _WorkerMailbox:
@@ -118,6 +146,7 @@ def worker_mailbox(
         aes_encryption_key=aes_key,
         message_service=mock_message_service,
         pod_mailbox_ids=pod_mailbox_ids,
+        task_id=opt_task_id,
     )
 
 
@@ -145,20 +174,22 @@ def interpod_worker_mailbox(
     mock_private_key: Mock,
     modeller_mailbox_id: str,
     modeller_name: str,
+    opt_task_id: Optional[str],
     other_pod_details: List[WorkerDetails],
     pod_identifier: str,
     pod_mailbox_ids: Dict[str, str],
 ) -> _InterPodWorkerMailbox:
     """An InterPodWorkerMailbox instance with components mocked out."""
     return _InterPodWorkerMailbox(
+        pod_public_keys=mock_pod_public_keys,
+        private_key=mock_private_key,
         pod_identifier=pod_identifier,
         modeller_mailbox_id=modeller_mailbox_id,
         modeller_name=modeller_name,
         aes_encryption_key=aes_key,
         message_service=mock_message_service,
         pod_mailbox_ids=pod_mailbox_ids,
-        pod_public_keys=mock_pod_public_keys,
-        private_key=mock_private_key,
+        task_id=opt_task_id,
     )
 
 
@@ -169,6 +200,46 @@ def mock_message_type() -> Mock:
     return mock_message_type
 
 
+@fixture
+def dummy_message_awaitable() -> Callable[
+    [Optional[_BitfountMessage], Optional[bool]], object
+]:
+    """A dummy awaitable for testing."""
+
+    class MockMessageAwaitable:
+        """Dummy awaitable in place of _AsyncCallback.
+
+        The `timeout` indicates whether the first time the awaitable is called,
+        it should return a message or raise an `asyncio.TimeoutError`.
+        """
+
+        def __init__(
+            self,
+            message: Optional[_BitfountMessage] = None,
+            timeout: Optional[bool] = False,
+        ) -> None:
+            self.message = message
+            self.timeout = timeout
+            self.timeout_handler = Mock()
+
+        async def result(
+            self, timeout: Optional[float] = None
+        ) -> Optional[_BitfountMessage]:
+            """Returns the result of the awaitable.
+
+            In this dummy implementation, the result is just the original message. If a
+            timeout is specified, this will be recorded using `self.timeout_handler`.
+            """
+            if self.timeout:
+                self.timeout = None
+                raise asyncio.TimeoutError()
+
+            self.timeout_handler(timeout)
+            return self.message
+
+    return MockMessageAwaitable
+
+
 @unit_test
 class TestWorkerMailbox:
     """Test WorkerMailbox class."""
@@ -181,6 +252,7 @@ class TestWorkerMailbox:
         mocker: MockerFixture,
         modeller_mailbox_id: str,
         modeller_name: str,
+        opt_task_id: Optional[str],
         pod_identifier: str,
         worker_mailbox: _WorkerMailbox,
         worker_mailbox_id: str,
@@ -210,6 +282,7 @@ class TestWorkerMailbox:
                 sender=pod_identifier,
                 sender_mailbox_id=worker_mailbox_id,
                 timestamp=fake_timestamps[0],
+                task_id=opt_task_id,
             ),
             already_packed=True,
         )
@@ -222,7 +295,9 @@ class TestWorkerMailbox:
         mocker: MockerFixture,
         modeller_mailbox_id: str,
         modeller_name: str,
+        opt_task_id: Optional[str],
         pod_identifier: str,
+        rpc_error: RpcError,
         worker_mailbox: _WorkerMailbox,
         worker_mailbox_id: str,
     ) -> None:
@@ -237,7 +312,7 @@ class TestWorkerMailbox:
         message = "some message"
         dumped_message = msgpack.dumps(message)
 
-        mock_message_service.send_message.side_effect = RpcError()
+        mock_message_service.send_message.side_effect = rpc_error
 
         with raises(RpcError):
             await worker_mailbox._send_aes_encrypted_message(
@@ -254,19 +329,195 @@ class TestWorkerMailbox:
                 sender=pod_identifier,
                 sender_mailbox_id=worker_mailbox_id,
                 timestamp=fake_timestamps[0],
+                task_id=opt_task_id,
             ),
             already_packed=True,
         )
 
-    async def test__get_message_and_decrypt_successful(
+    async def test__get_message_returns_message_without_breaching_timeout(
         self,
-        aes_key: bytes,
-        mock_poll_for_messages: Callable[
-            [Iterable[MessageOrException]], AsyncIteratorMock
+        caplog: LogCaptureFixture,
+        modeller_mailbox_id: str,
+        modeller_name: str,
+        opt_task_id: Optional[str],
+        patch_poll_for_messages: PollForMessagesPatcher,
+        pod_identifier: str,
+        worker_mailbox: _WorkerMailbox,
+        worker_mailbox_id: str,
+    ) -> None:
+        """Message is returned as expected without breaching timeout."""
+        # Set up the mocked messages to be received
+        patch_poll_for_messages(
+            [
+                _BitfountMessage(
+                    message_type=_BitfountMessageType.TRAINING_UPDATE,
+                    body=b"123",
+                    recipient=modeller_name,
+                    recipient_mailbox_id=modeller_mailbox_id,
+                    sender=pod_identifier,
+                    sender_mailbox_id=worker_mailbox_id,
+                    task_id=opt_task_id,
+                )
+            ]
+        )
+
+        # Start listening and processing messages
+        mock_soft_limit_timeout = 30
+        message = await _run_func_and_listen_to_mailbox(
+            worker_mailbox._get_message(
+                _BitfountMessageType.TRAINING_UPDATE, timeout=mock_soft_limit_timeout
+            ),
+            worker_mailbox,
+        )
+        assert message.body == b"123"
+
+        # Check no logs due to timeout not being reached
+        logs = get_info_logs(caplog, and_higher=True)
+        assert "Checking if Modeller is still online..." not in logs
+
+    async def test__get_message_breaches_soft_limit_timeout(
+        self,
+        caplog: LogCaptureFixture,
+        dummy_message_awaitable: Callable[
+            [Optional[_BitfountMessage], Optional[bool]], object
         ],
         mocker: MockerFixture,
         modeller_mailbox_id: str,
         modeller_name: str,
+        online_check_uuid: str,
+        opt_task_id: Optional[str],
+        pod_identifier: str,
+        worker_mailbox: _WorkerMailbox,
+        worker_mailbox_id: str,
+    ) -> None:
+        """Message is returned as expected after breaching soft timeout."""
+        caplog.set_level("INFO")
+        original_message_awaitable = dummy_message_awaitable(
+            _BitfountMessage(
+                message_type=_BitfountMessageType.TRAINING_UPDATE,
+                body=b"123",
+                recipient=modeller_name,
+                recipient_mailbox_id=modeller_mailbox_id,
+                sender=pod_identifier,
+                sender_mailbox_id=worker_mailbox_id,
+                task_id=opt_task_id,
+            ),
+            True,
+        )
+
+        # Patch out the waiting for response
+        mocker.patch.object(
+            worker_mailbox._online_response_handler,
+            "wait_for_response",
+            autospec=True,
+            return_value=None,
+        )
+
+        # Mock the bitfount messages retrieved from the modeller
+        mocker.patch(
+            "bitfount.federated.transport.worker_transport._get_message_awaitable",
+            return_value=original_message_awaitable,
+        )
+
+        # Mock checking the modeller is online
+        mock_check_modeller_online = mocker.patch.object(
+            worker_mailbox, "check_modeller_online"
+        )
+
+        # Mock hard limit timeout
+        mocker.patch(
+            "bitfount.federated.transport.worker_transport._HARD_LIMIT_MESSAGE_TIMEOUT",
+            5,
+        )
+
+        # Start listening and processing messages
+        mock_soft_limit_timeout = 3
+        message = await worker_mailbox._get_message(
+            _BitfountMessageType.TRAINING_UPDATE, timeout=mock_soft_limit_timeout
+        )
+
+        assert message.body == b"123"
+
+        # Check logs
+        info_logs = get_info_logs(caplog)
+        assert "Checking if Modeller is still online..." in info_logs
+        assert "Modeller is online, responded with expected message." in info_logs
+
+        mock_check_modeller_online.assert_called_once()
+
+        # Should be one call to timeout_handler; the first call to result() raises
+        # a TimeoutError, the second call passes the timeout through to
+        # timeout_handler. This second call is made whilst we are waiting for the
+        # online response check event to be set and so has no timeout.
+        # mypy reason: Mypy can't detect the actual type of this mocked awaitable.
+        original_message_awaitable.timeout_handler.assert_called_once()  # type: ignore[attr-defined] # Reason: see above # noqa: B950
+
+    async def test__get_message_breaches_hard_limit_timeout(
+        self,
+        caplog: LogCaptureFixture,
+        dummy_message_awaitable: Callable[
+            [Optional[_BitfountMessage], Optional[bool]], object
+        ],
+        mocker: MockerFixture,
+        worker_mailbox: _WorkerMailbox,
+    ) -> None:
+        """Test worker gives up on waiting for message after breaching hard limit.
+
+        This should raise an `asyncio.TimeoutError` exception.
+        """
+        caplog.set_level("INFO")
+
+        # No messages retrieved from the Modeller, just `TimeOutError`s
+        original_message_awaitable = dummy_message_awaitable(None, True)
+
+        # Mock the bitfount messages retrieved from the modeller
+        mocker.patch(
+            "bitfount.federated.transport.worker_transport._get_message_awaitable",
+            return_value=original_message_awaitable,
+        )
+
+        # Mock checking the modeller is online
+        mock_check_modeller_online = mocker.patch.object(
+            worker_mailbox, "check_modeller_online"
+        )
+
+        # Mock sending of task abort message
+        mock_send_task_abort_message = mocker.patch.object(
+            worker_mailbox, "send_task_abort_message"
+        )
+
+        # Mock hard limit timeout
+        mocker.patch(
+            "bitfount.federated.transport.worker_transport._HARD_LIMIT_MESSAGE_TIMEOUT",
+            5,
+        )
+        mock_wait_for = mocker.patch.object(
+            asyncio, "wait_for", autospec=True, side_effect=asyncio.TimeoutError()
+        )
+
+        # Start listening and processing messages
+        mock_soft_limit_timeout = 3
+        with pytest.raises(asyncio.TimeoutError):
+            await worker_mailbox._get_message(
+                _BitfountMessageType.TRAINING_UPDATE, timeout=mock_soft_limit_timeout
+            )
+
+        # Check logs
+        assert "Checking if Modeller is still online..." in get_info_logs(caplog)
+        assert "Modeller is offline. Aborting task." in get_warning_logs(caplog)
+
+        mock_check_modeller_online.assert_called_once()
+        mock_wait_for.assert_awaited_once_with(ANY, timeout=5)
+        mock_send_task_abort_message.assert_called_once()
+
+    async def test__get_message_and_decrypt_successful(
+        self,
+        aes_key: bytes,
+        mocker: MockerFixture,
+        modeller_mailbox_id: str,
+        modeller_name: str,
+        opt_task_id: Optional[str],
+        patch_poll_for_messages: PollForMessagesPatcher,
         pod_identifier: str,
         worker_mailbox: _WorkerMailbox,
         worker_mailbox_id: str,
@@ -283,7 +534,7 @@ class TestWorkerMailbox:
         decrypt.return_value = dumped_message
 
         # Set up the mocked messages to be received
-        mock_poll_for_messages(
+        patch_poll_for_messages(
             [
                 _BitfountMessage(
                     message_type=_BitfountMessageType.TRAINING_UPDATE,
@@ -292,6 +543,7 @@ class TestWorkerMailbox:
                     recipient_mailbox_id=modeller_mailbox_id,
                     sender=pod_identifier,
                     sender_mailbox_id=worker_mailbox_id,
+                    task_id=opt_task_id,
                 )
             ]
         )
@@ -307,39 +559,14 @@ class TestWorkerMailbox:
         assert message == expected_message
         decrypt.assert_called_once_with(mock_encrypted_message, aes_key)
 
-    async def test__get_message_and_decrypt_error_never_receives_message(
-        self,
-        mock_poll_for_messages: Callable[
-            [Iterable[MessageOrException]], AsyncIteratorMock
-        ],
-        worker_mailbox: _WorkerMailbox,
-    ) -> None:
-        """Never receives message using provided AES key."""
-        # Use empty list to indicate no message scenario
-        mock_poll_for_messages([])
-
-        with raises(asyncio.TimeoutError):
-            # Start listening and processing (non-existent) messages. We will
-            # timeout on the responses to simulate what happens when no message
-            # is received.
-            await _run_func_and_listen_to_mailbox(
-                worker_mailbox._get_message_and_decrypt(
-                    _BitfountMessageType.TRAINING_UPDATE,  # this is arbitrary
-                    timeout=1,
-                ),
-                worker_mailbox,
-            )
-
     async def test__get_message_and_decrypt_error(
         self,
-        mock_poll_for_messages: Callable[
-            [Iterable[MessageOrException]], AsyncIteratorMock
-        ],
+        patch_poll_for_messages: PollForMessagesPatcher,
         rpc_error: RpcError,
         worker_mailbox: _WorkerMailbox,
     ) -> None:
         """Receiving message throws RpcError."""
-        mock_poll_for_messages([rpc_error])
+        patch_poll_for_messages([rpc_error])
 
         with raises(
             MessageRetrievalError,
@@ -376,11 +603,14 @@ class TestWorkerMailbox:
     async def test_accept_job_unsuccessful(
         self,
         mocker: MockerFixture,
+        rpc_error: RpcError,
         worker_mailbox: _WorkerMailbox,
     ) -> None:
         """Sending training job acceptance fails."""
         mock_message_send = mocker.patch.object(
-            worker_mailbox, "_send_aes_encrypted_message", side_effect=RpcError()
+            worker_mailbox,
+            "_send_aes_encrypted_message",
+            side_effect=rpc_error,
         )
 
         with pytest.raises(RpcError):
@@ -416,12 +646,15 @@ class TestWorkerMailbox:
     async def test_reject_job_unsuccessful(
         self,
         mocker: MockerFixture,
+        rpc_error: RpcError,
         worker_mailbox: _WorkerMailbox,
     ) -> None:
         """Sending training job rejection fails."""
         expected_error_messages = {"error": "messages"}
         mock_message_send = mocker.patch.object(
-            worker_mailbox, "_send_aes_encrypted_message", side_effect=RpcError()
+            worker_mailbox,
+            "_send_aes_encrypted_message",
+            side_effect=rpc_error,
         )
 
         with pytest.raises(RpcError):
@@ -459,12 +692,15 @@ class TestWorkerMailbox:
     async def test_issue_saml_challenge_unsuccessful(
         self,
         mocker: MockerFixture,
+        rpc_error: RpcError,
         worker_mailbox: _WorkerMailbox,
     ) -> None:
         """Sending SAML challenge fails."""
         expected_saml_request = "some saml request"
         mock_message_send = mocker.patch.object(
-            worker_mailbox, "_send_aes_encrypted_message", side_effect=RpcError()
+            worker_mailbox,
+            "_send_aes_encrypted_message",
+            side_effect=rpc_error,
         )
 
         with pytest.raises(RpcError):
@@ -494,15 +730,33 @@ class TestWorkerMailbox:
 
         assert response == expected_saml_response
         mock_message_decrypt.assert_awaited_once_with(
-            _BitfountMessageType.SAML_RESPONSE, None
+            _BitfountMessageType.SAML_RESPONSE,
+            _DEFAULT_AUTHENTICATION_MODELLER_RESPONSE_TIMEOUT,
         )
 
-    async def test_get_task_complete_response(
+    async def test_get_task_start_update(
         self,
         mocker: MockerFixture,
         worker_mailbox: _WorkerMailbox,
     ) -> None:
-        """Tests worker gets task complete empty message ."""
+        """Tests worker gets task start empty message."""
+        mock_message_decrypt = mocker.patch.object(
+            worker_mailbox,
+            "_get_message_and_decrypt",
+            return_value=None,
+        )
+
+        await worker_mailbox.get_task_start_update()
+        mock_message_decrypt.assert_awaited_once_with(
+            _BitfountMessageType.TASK_START, timeout=_SOFT_LIMIT_MESSAGE_TIMEOUT
+        )
+
+    async def test_get_task_complete_update(
+        self,
+        mocker: MockerFixture,
+        worker_mailbox: _WorkerMailbox,
+    ) -> None:
+        """Tests worker gets task complete empty message."""
         mock_message_decrypt = mocker.patch.object(
             worker_mailbox,
             "_get_message_and_decrypt",
@@ -511,7 +765,7 @@ class TestWorkerMailbox:
 
         await worker_mailbox.get_task_complete_update()
         mock_message_decrypt.assert_awaited_once_with(
-            _BitfountMessageType.TASK_COMPLETE, timeout=None
+            _BitfountMessageType.TASK_COMPLETE, timeout=_SOFT_LIMIT_MESSAGE_TIMEOUT
         )
 
     async def test_get_training_complete_update(
@@ -519,7 +773,7 @@ class TestWorkerMailbox:
         mocker: MockerFixture,
         worker_mailbox: _WorkerMailbox,
     ) -> None:
-        """Tests worker gets task complete empty message ."""
+        """Tests worker gets training complete empty message."""
         expected_message = True
         mock_message_decrypt = mocker.patch.object(
             worker_mailbox,
@@ -530,7 +784,7 @@ class TestWorkerMailbox:
         response = await worker_mailbox.get_training_iteration_complete_update()
         assert response == expected_message
         mock_message_decrypt.assert_awaited_once_with(
-            _BitfountMessageType.TRAINING_COMPLETE, timeout=None
+            _BitfountMessageType.TRAINING_COMPLETE, timeout=_SOFT_LIMIT_MESSAGE_TIMEOUT
         )
 
     async def test_send_oidc_client_id(
@@ -751,6 +1005,35 @@ class TestWorkerMailbox:
             _BitfountMessageType.OIDC_DEVICE_CODE_RESPONSE, ANY
         )
 
+    async def test__get_psi_dataset(
+        self,
+        mocker: MockerFixture,
+        worker_mailbox: _WorkerMailbox,
+    ) -> None:
+        """Test _get_psi_dataset method."""
+        # Patch out _WorkerMailbox._get_message_and_decrypt()
+        expected_message = ["some message"]
+        mock_message_decrypt = mocker.patch.object(
+            worker_mailbox,
+            "_get_message_and_decrypt",
+            return_value=expected_message,
+        )
+        await _get_psi_dataset(worker_mailbox)
+        # Check calls/returns
+        mock_message_decrypt.assert_awaited_once_with(
+            _BitfountMessageType.PSI_DATASET, timeout=_SOFT_LIMIT_MESSAGE_TIMEOUT
+        )
+
+    async def test_log(
+        self, mocker: MockerFixture, worker_mailbox: _WorkerMailbox
+    ) -> None:
+        """Test that log message is sent appropriately."""
+        mock_log = mocker.patch.object(worker_mailbox, "_send_aes_encrypted_message")
+        await worker_mailbox.log({"msg": "message"})
+        mock_log.assert_awaited_once_with(
+            {"msg": "message"}, _BitfountMessageType.LOG_MESSAGE
+        )
+
 
 @unit_test
 class TestInterPodWorkerMailbox:
@@ -782,14 +1065,14 @@ class TestInterPodWorkerMailbox:
             ),
         ):
             _InterPodWorkerMailbox(
+                pod_public_keys=mock_pod_public_keys,
+                private_key=mock_private_key,
                 pod_identifier=pod_identifier,
                 modeller_mailbox_id=modeller_mailbox_id,
                 modeller_name=modeller_name,
                 aes_encryption_key=aes_key,
                 message_service=mock_message_service,
                 pod_mailbox_ids=pod_mailbox_ids,
-                pod_public_keys=mock_pod_public_keys,
-                private_key=mock_private_key,
             )
 
     async def test__send_pod_to_pod_message(
@@ -800,6 +1083,7 @@ class TestInterPodWorkerMailbox:
         mock_message_type: Mock,
         mock_pod_public_keys: Dict[str, Mock],
         mock_rsa_encryption: Mock,
+        opt_task_id: Optional[str],
         other_pod_details: List[WorkerDetails],
         pod_identifier: str,
         worker_mailbox_id: str,
@@ -832,6 +1116,7 @@ class TestInterPodWorkerMailbox:
             sender=pod_identifier,
             sender_mailbox_id=worker_mailbox_id,
             timestamp=fake_timestamps[0],
+            task_id=opt_task_id,
         )
         mock_message_service.send_message.assert_awaited_once_with(
             expected_message, already_packed=True
@@ -845,7 +1130,7 @@ class TestInterPodWorkerMailbox:
         mock_message_type: Mock,
     ) -> None:
         """Test _send_pod_to_pod_message logs error if no public key for recipient."""
-        result = await interpod_worker_mailbox._send_pod_to_pod_message(
+        await interpod_worker_mailbox._send_pod_to_pod_message(
             recipient="not_a_real_recipient/pod",
             recipient_mailbox_id="not_a_real_mailbox_id",
             object_to_send=Mock(),
@@ -860,8 +1145,6 @@ class TestInterPodWorkerMailbox:
         )
         # Check didn't try to send
         mock_message_service.send_message.assert_not_called()
-        # Check returned None
-        assert result is None
 
     def test__pod_to_pod_message_handler(
         self,
@@ -877,6 +1160,42 @@ class TestInterPodWorkerMailbox:
         mock_message.decrypt_rsa.assert_called_once_with(mock_private_key)
         assert decrypted == mock_message.decrypt_rsa.return_value
 
+    async def test_check_modeller_online(
+        self,
+        mocker: MockerFixture,
+        online_check_uuid: str,
+        worker_mailbox: _WorkerMailbox,
+    ) -> None:
+        """Tests worker `check_modeller_online` method."""
+        mock_message_send = mocker.patch.object(
+            worker_mailbox,
+            "_send_aes_encrypted_message",
+            return_value=None,
+        )
+
+        await worker_mailbox.check_modeller_online(online_check_uuid)
+        mock_message_send.assert_awaited_once_with(
+            online_check_uuid,
+            _BitfountMessageType.ONLINE_CHECK,
+        )
+
+    async def test_send_task_abort_message(
+        self,
+        mocker: MockerFixture,
+        worker_mailbox: _WorkerMailbox,
+    ) -> None:
+        """Tests worker `send_task_abort_message` method."""
+        mock_message_send = mocker.patch.object(
+            worker_mailbox,
+            "_send_aes_encrypted_message",
+            return_value=None,
+        )
+
+        await worker_mailbox.send_task_abort_message()
+        mock_message_send.assert_awaited_once_with(
+            None, _BitfountMessageType.TASK_ABORT
+        )
+
 
 @unit_test
 class TestWorkerTransportFunctions:
@@ -888,6 +1207,7 @@ class TestWorkerTransportFunctions:
         mock_message_service: Mock,
         mock_message_timestamps: Callable[[Iterable[str]], Mock],
         mock_rsa_encryption: Mock,
+        opt_task_id: Optional[str],
         other_pod_details: List[WorkerDetails],
         pod_identifier: str,
         worker_mailbox_id: str,
@@ -930,6 +1250,7 @@ class TestWorkerTransportFunctions:
                         sender=pod_identifier,
                         sender_mailbox_id=worker_mailbox_id,
                         timestamp=fake_timestamps[0],
+                        task_id=opt_task_id,
                     ),
                     already_packed=True,
                 ),
@@ -942,6 +1263,7 @@ class TestWorkerTransportFunctions:
                         sender=pod_identifier,
                         sender_mailbox_id=worker_mailbox_id,
                         timestamp=fake_timestamps[1],
+                        task_id=opt_task_id,
                     ),
                     already_packed=True,
                 ),
@@ -952,11 +1274,9 @@ class TestWorkerTransportFunctions:
         self,
         caplog: LogCaptureFixture,
         interpod_worker_mailbox: _InterPodWorkerMailbox,
-        mock_poll_for_messages: Callable[
-            [Iterable[MessageOrException]], AsyncIteratorMock
-        ],
         mock_private_key: Mock,
         other_pod_details: List[WorkerDetails],
+        patch_poll_for_messages: PollForMessagesPatcher,
     ) -> None:
         """Tests the _get_worker_secure_shares function works."""
         # Create mock messages to process
@@ -971,7 +1291,7 @@ class TestWorkerTransportFunctions:
             )
             for i, details in enumerate(other_pod_details)
         ]
-        mock_poll_for_messages(mock_messages)
+        patch_poll_for_messages(mock_messages)
 
         with caplog.at_level(logging.DEBUG):
             shares = await _run_func_and_listen_to_mailbox(
@@ -980,7 +1300,8 @@ class TestWorkerTransportFunctions:
             )
 
         # Check decoded shares are what we expect and that each message was decoded
-        assert shares == [i for i in range(len(other_pod_details))]
+        # Need to sort for comparison as order is not guaranteed!
+        assert sorted(shares) == [i for i in range(len(other_pod_details))]
         for mock_message in mock_messages:
             mock_message.decrypt_rsa.assert_called_once_with(mock_private_key)
 
