@@ -13,8 +13,10 @@ from _pytest.monkeypatch import MonkeyPatch
 import pytest
 from pytest_mock import MockerFixture
 
-from bitfount import BitfountSchema, DataSource, DataStructure
+from bitfount import BitfountSchema, DataStructure
 from bitfount.backends.pytorch.models.models import PyTorchTabularClassifier
+from bitfount.data.datasources.base_source import BaseSource
+from bitfount.data.datasources.csv_source import CSVSource
 from bitfount.runners.config_schemas import ModellerConfig, PodConfig
 from bitfount.runners.modeller_runner import setup_modeller_from_config
 from bitfount.runners.pod_runner import setup_pod_from_config
@@ -50,8 +52,8 @@ MODELLER_RUN_TIMEOUT: int = int(
 
 @backend_test
 @end_to_end_test
-class TestAdultEndToEnd:
-    """End-to-end tests using the Adult dataset."""
+class TestCensusIncomeEndToEnd:
+    """End-to-end tests using the Census Income dataset."""
 
     @pytest.fixture(autouse=True)
     def bitfount_env(self, monkeypatch: MonkeyPatch) -> None:
@@ -69,9 +71,11 @@ class TestAdultEndToEnd:
         return load_modeller_config(CONFIG_DIR / "modeller.yaml")
 
     @pytest.fixture
-    def adult_modeller_config(self) -> ModellerConfig:
-        """Adult modeller config."""
-        return load_modeller_config(CONFIG_DIR / "modeller_adult_result_only.yaml")
+    def census_income_modeller_config(self) -> ModellerConfig:
+        """Census income modeller config."""
+        return load_modeller_config(
+            CONFIG_DIR / "modeller_census_income_result_only.yaml"
+        )
 
     @pytest.fixture
     def modeller_pswd(self, modeller_config: ModellerConfig) -> str:
@@ -79,14 +83,16 @@ class TestAdultEndToEnd:
         return password_from_modeller_config(modeller_config)
 
     @pytest.fixture
-    def pod_1_config(self, adult_data: Path) -> PodConfig:
+    def pod_1_config(self, census_income_data: Path) -> PodConfig:
         """Pod 1 config."""
-        return load_pod_config(CONFIG_DIR / "pod_1.yaml", adult_data)
+        return load_pod_config(CONFIG_DIR / "pod_1.yaml", census_income_data)
 
     @pytest.fixture
-    def adult_pod_config(self, adult_data: Path) -> PodConfig:
-        """Adult pod config."""
-        return load_pod_config(CONFIG_DIR / "pod_adult.yaml", adult_data)
+    def census_income_pod_config(self, census_income_data: Path) -> PodConfig:
+        """Census income pod config."""
+        return load_pod_config(
+            CONFIG_DIR / "pod_census_income.yaml", census_income_data
+        )
 
     @pytest.fixture
     def pod_1_pswd(self, pod_1_config: PodConfig) -> str:
@@ -94,9 +100,9 @@ class TestAdultEndToEnd:
         return password_from_pod_config(pod_1_config)
 
     @pytest.fixture
-    def pod_2_config(self, adult_data: Path) -> PodConfig:
+    def pod_2_config(self, census_income_data: Path) -> PodConfig:
         """Pod 2 config."""
-        return load_pod_config(CONFIG_DIR / "pod_2.yaml", adult_data)
+        return load_pod_config(CONFIG_DIR / "pod_2.yaml", census_income_data)
 
     @pytest.fixture
     def pod_2_pswd(self, pod_2_config: PodConfig) -> str:
@@ -104,9 +110,9 @@ class TestAdultEndToEnd:
         return password_from_pod_config(pod_2_config)
 
     @pytest.fixture
-    def datasource(self, adult_data: Path) -> DataSource:
-        """Adult datasource."""
-        return DataSource(data_ref=adult_data, ignore_cols=["fnlwgt"], seed=100)
+    def datasource(self, census_income_data: Path) -> BaseSource:
+        """Census Income datasource."""
+        return CSVSource(census_income_data, seed=100, ignore_cols=["fnlwgt"])
 
     @pytest.fixture
     def mock_modeller_oidc(self, mocker: MockerFixture) -> Any:
@@ -120,8 +126,8 @@ class TestAdultEndToEnd:
         return oidc_patcher
 
     @pytest.fixture
-    def schema(self, datasource: DataSource) -> BitfountSchema:
-        """Adult schema."""
+    def schema(self, datasource: BaseSource) -> BitfountSchema:
+        """Census income schema."""
         return BitfountSchema(
             datasource,
             table_name=TABLE_NAME,
@@ -142,23 +148,25 @@ class TestAdultEndToEnd:
             },
         )
 
-    @pytest.mark.skip("[BIT-1466] Investigate why local results differ from remote")
+    @pytest.mark.skip("[BIT-1669] Fix e2e tests")
     async def test_local_vs_pod(
         self,
-        adult_pod_config: PodConfig,
-        pod_1_pswd: str,
-        token_dir: Path,
-        schema: BitfountSchema,
-        datasource: DataSource,
+        census_income_modeller_config: ModellerConfig,
+        census_income_pod_config: PodConfig,
+        datasource: BaseSource,
         mock_bitfount_session: Mock,
-        adult_modeller_config: ModellerConfig,
-        modeller_pswd: str,
         mock_modeller_oidc: Mock,
+        modeller_pswd: str,
+        pod_1_pswd: str,
+        schema: BitfountSchema,
+        token_dir: Path,
     ) -> None:
         """Tests local vs remote results with oidc device code auth."""
         try:
             # Load configs and tie together
-            adult_modeller_config.pods.identifiers = [adult_pod_config.pod_id]
+            census_income_modeller_config.pods.identifiers = [
+                census_income_pod_config.pod_id
+            ]
             # Mock BitfountSession
             # Override default return_value with side_effect
             # get_bitfount_session() one. side_effect takes precedence
@@ -167,22 +175,26 @@ class TestAdultEndToEnd:
             # then be sure to change the order of these...
             mock_bitfount_session.side_effect = [
                 # pod 1 bitfount session
-                get_bitfount_session(adult_pod_config.username, pod_1_pswd, token_dir),
+                get_bitfount_session(
+                    census_income_pod_config.username, pod_1_pswd, token_dir
+                ),
                 # modeller bitfount session
                 get_bitfount_session(
-                    adult_modeller_config.modeller.username, modeller_pswd, token_dir
+                    census_income_modeller_config.modeller.username,
+                    modeller_pswd,
+                    token_dir,
                 ),
             ]
 
             mock_modeller_oidc.side_effect = [
                 partial(
                     oidc_flow,
-                    username=adult_modeller_config.modeller.username,
+                    username=census_income_modeller_config.modeller.username,
                     password=modeller_pswd,
                 )
             ]
             # Load pods
-            pod = setup_pod_from_config(adult_pod_config)
+            pod = setup_pod_from_config(census_income_pod_config)
 
             # Start pods
             logger.info("Spinning up pods... ")
@@ -199,11 +211,11 @@ class TestAdultEndToEnd:
 
             # Sort proactive access to each pod
             logger.info(
-                f"Granting proactive access to pod {adult_pod_config.pod_id}..."
+                f"Granting proactive access to pod {census_income_pod_config.pod_id}..."
             )
             grant_proactive_access(
-                modeller_username=adult_modeller_config.modeller.username,
-                pod_id=adult_pod_config.pod_id,
+                modeller_username=census_income_modeller_config.modeller.username,
+                pod_id=census_income_pod_config.pod_id,
                 role="general_modeller",
                 # This is a WebdriverBitfountSession as we mocked it out
                 pod_session=cast(WebdriverBitfountSession, pod._session),
@@ -213,7 +225,7 @@ class TestAdultEndToEnd:
             # We do this after the pods have started, as we need the pods to have
             # published their public keys to the Hub.
             modeller, pod_identifiers = setup_modeller_from_config(
-                adult_modeller_config
+                census_income_modeller_config
             )
 
             queue: Queue = Queue()
@@ -243,10 +255,10 @@ class TestAdultEndToEnd:
             assert (
                 modeller_process.exitcode is not None and modeller_process.exitcode <= 0
             )
+
             pod_results = queue.get()[0]
             # Check pod_results exists
             assert pod_results
-
             # Ensure pods are still running
             assert pod_process.is_alive()
 
@@ -259,8 +271,8 @@ class TestAdultEndToEnd:
 
             schema.unfreeze()
             datastructure = DataStructure(table=pod.name, target="TARGET")
-            hyperparameters = adult_modeller_config.task.model.hyperparameters
-            model_name = adult_modeller_config.task.model.name
+            hyperparameters = census_income_modeller_config.task.model.hyperparameters
+            model_name = census_income_modeller_config.task.model.name
             if model_name:
                 assert PyTorchTabularClassifier == eval(model_name)
                 local_model = PyTorchTabularClassifier(
@@ -271,8 +283,6 @@ class TestAdultEndToEnd:
 
                 local_results = local_model.fit(data=datasource)
                 if local_results:
-
-                    # TODO: [BIT-1466] verify why these aren't the same
                     # Check results are the same
                     for k, pod_value in pod_results:
                         assert pod_value == local_results.get(k)
@@ -286,32 +296,30 @@ class TestAdultEndToEnd:
                 modeller_process.terminate()
             except Exception:
                 pass
+
             # noinspection PyBroadException
             try:
                 # noinspection PyUnboundLocalVariable
                 pod_process.terminate()
             except Exception:
                 pass
-            # noinspection PyBroadException
-            except Exception:
-                pass
 
-    @pytest.mark.skip("Signature-based access not working until [BIT-1291] implemented")
-    async def test_adult_e2e_results_only(
+    @pytest.mark.skip("[BIT-1669] Fix e2e tests")
+    async def test_census_income_e2e_results_only(
         self,
         mock_bitfount_session: Mock,
-        tmp_path: Path,
-        token_dir: Path,
         modeller_config: ModellerConfig,
         modeller_pswd: str,
         pod_1_config: PodConfig,
         pod_1_pswd: str,
         pod_2_config: PodConfig,
         pod_2_pswd: str,
+        tmp_path: Path,
+        token_dir: Path,
     ) -> None:
         """Tests a full end-to-end run.
 
-        End to end run using Adult dataset with key-based id verification.
+        End to end run using Census income dataset with key-based id verification.
         """
         try:
             # Load configs and tie together
