@@ -4,12 +4,12 @@ import copy
 import logging
 import multiprocessing
 from multiprocessing import current_process
-from multiprocessing.managers import SyncManager
+from multiprocessing.managers import DictProxy, SyncManager
 import os
 from pathlib import Path
 from queue import Queue
 import time
-from typing import Any, Callable, Dict, Generator, Iterable, Union, cast
+from typing import Any, Callable, Dict, Generator, Iterable, Type, Union, cast
 from unittest.mock import Mock
 
 from _pytest.monkeypatch import MonkeyPatch
@@ -21,8 +21,9 @@ from pytest_flask.live_server import LiveServer
 from pytest_mock import MockerFixture
 from pytorch_lightning.loggers import TensorBoardLogger
 
-from bitfount.federated.exceptions import PodRegistrationError
-from bitfount.federated.modeller import Modeller
+import bitfount
+from bitfount import PodRegistrationError
+from bitfount.federated.modeller import _Modeller
 from bitfount.federated.transport.base_transport import _run_func_and_listen_to_mailbox
 from bitfount.federated.transport.message_service import _LargeObjectRequestHandler
 from bitfount.hub.authentication_flow import BitfountSession
@@ -60,8 +61,7 @@ class MockBitfountSession(BitfountSession):
 
     def __init__(self, access_token: str, username: str, **kwargs: Any):
         self._faked_access_token = access_token
-        self._username = username
-        super().__init__(**kwargs)
+        super().__init__(username=username, **kwargs)
 
     @property
     def access_token(self) -> str:
@@ -110,10 +110,10 @@ class TestEndToEnd:
         )
 
     @fixture
-    def pod_config(self, adult_data: Path) -> PodConfig:
+    def pod_config(self, census_income_data: Path) -> PodConfig:
         """Pod config to use for tests."""
-        pod_config = load_pod_config(CONFIG_DIR / "pod_1.yaml", adult_data)
-        pod_config.other_pods = [f"{pod_config.pod_id}2"]
+        pod_config = load_pod_config(CONFIG_DIR / "pod_1.yaml", census_income_data)
+        pod_config.approved_pods = [f"{pod_config.pod_id}2"]
         pod_config.hub.url = url_for("index", _external=True)
         return pod_config
 
@@ -158,9 +158,10 @@ class TestEndToEnd:
         def _create_mock_bitfount_session(
             username: str, access_token: str
         ) -> MockBitfountSession:
-            return MockBitfountSession(
-                access_token=access_token, username=username, user_storage_path=tmp_path
-            )
+            session = MockBitfountSession(access_token=access_token, username=username)
+            session.user_storage_path = tmp_path / username
+            session.token_file = session.user_storage_path / ".token"
+            return session
 
         return _create_mock_bitfount_session
 
@@ -181,7 +182,7 @@ class TestEndToEnd:
             LargeObjectRequestHandler.get_large_object_from_url() to use local storage.
             """
             # Mock out message service functions
-            large_object_storage: Dict[
+            large_object_storage: DictProxy[
                 Union[_S3PresignedURL, _S3PresignedPOSTURL], bytes
             ] = manager.dict()
 
@@ -233,7 +234,7 @@ class TestEndToEnd:
 
     @staticmethod
     def _modeller_send_task_requests(
-        modeller: Modeller,
+        modeller: _Modeller,
         pod_identifiers: Iterable[str],
     ) -> None:
         """Helper for sending Modeller task requests in a separate process.
@@ -256,22 +257,32 @@ class TestEndToEnd:
             print(f"Caught exception in {current_process().name} process: {e}")
             raise e
 
+    @pytest.mark.parametrize(
+        argnames=("cli_mode", "expected_exc"),
+        argvalues=((True, SystemExit), (False, PodRegistrationError)),
+    )
     @pytest.mark.parametrize("app", [(True, False, False)], indirect=True)
     @pytest.mark.timeout(STD_TIMEOUT)
     def test_pod_setup_raises_pod_registration_error(
         self,
         app: Flask,
         bitfount_session_factory: Callable[[str, str], MockBitfountSession],
+        cli_mode: bool,
+        expected_exc: Type[BaseException],
         live_server: LiveServer,
         mock_bitfount_session: Mock,
+        monkeypatch: MonkeyPatch,
         pod_config: PodConfig,
     ) -> None:
         """Tests handling when pod raises registration error."""
+        # Set _BITFOUNT_CLI_MODE config variable to elicit different error handling
+        monkeypatch.setattr(bitfount.config, "_BITFOUNT_CLI_MODE", cli_mode)
+
         username = pod_config.username
         # Override default return_value with bitfount_session_factory() one
         mock_bitfount_session.return_value = bitfount_session_factory(username, "pod")
 
-        with pytest.raises(PodRegistrationError):
+        with pytest.raises(expected_exc):
             setup_pod_from_config(pod_config)
 
     @pytest.mark.timeout(STD_TIMEOUT)
@@ -292,9 +303,6 @@ class TestEndToEnd:
         (Otherwise this test will hang indefinitely)
         """
         with multiprocessing.Manager() as manager:
-            # multiprocessing.Manager() actually returns a SyncManager
-            manager = cast(SyncManager, manager)
-
             pod_token = "pod"
             username = pod_config.username
             mock_message_service_stub.return_value = GRPCStubMock(
@@ -322,7 +330,7 @@ class TestEndToEnd:
 
             assert p.exitcode is not None and p.exitcode <= 0
 
-    @pytest.mark.skip("Signature-based access not working until [BIT-1291] implemented")
+    @pytest.mark.skip("[BIT-1669] Fix e2e tests")
     @backend_test
     @pytest.mark.parametrize("app", [(False, True, False)], indirect=True)
     @pytest.mark.timeout(STD_TIMEOUT)
@@ -343,9 +351,6 @@ class TestEndToEnd:
         """Tests Modeller handling if task requests are rejected by pods."""
         try:
             with multiprocessing.Manager() as manager:
-                # multiprocessing.Manager() actually returns a SyncManager
-                manager = cast(SyncManager, manager)
-
                 modeller_token = "modeller_token"
                 pod_token = "pod_token"
 
@@ -470,7 +475,7 @@ class TestEndToEnd:
             side_effect=_download,
         )
 
-    @pytest.mark.skip("Signature-based access not working until [BIT-1291] implemented")
+    @pytest.mark.skip("[BIT-1669] Fix e2e tests")
     @backend_test
     @pytest.mark.timeout(EXT_TIMEOUT)
     @pytest.mark.parametrize("use_local_storage", [True, False])
@@ -484,11 +489,11 @@ class TestEndToEnd:
         mock_grpc_insecure_channel: Mock,
         mock_message_service_stub: Mock,
         mock_schema_upload_download: None,
+        mocker: MockerFixture,
         modeller_config: ModellerConfig,
         pod_config: PodConfig,
         tmp_path: Path,
         use_local_storage: bool,
-        mocker: MockerFixture,
     ) -> None:
         """Tests that training works end to end.
 
@@ -497,9 +502,6 @@ class TestEndToEnd:
         """
         try:
             with multiprocessing.Manager() as manager:
-                # multiprocessing.Manager() actually returns a SyncManager
-                manager = cast(SyncManager, manager)
-
                 pod1_config = pod_config
 
                 # Determine if we are using file paths for messages and set it up
@@ -510,7 +512,7 @@ class TestEndToEnd:
                 # Create second pod config from first
                 second_pod_owner = "e2e_provider_2"
                 pod2_config = copy.deepcopy(pod1_config)
-                pod2_config.pod_name += "2"
+                pod2_config.name += "2"
                 pod2_config.username = second_pod_owner
 
                 # Patch out GRPC interactions
@@ -532,8 +534,8 @@ class TestEndToEnd:
                 apply_mock_large_object_interactions(manager)
 
                 # Tie pod configs together
-                pod1_config.other_pods = [pod2_config.pod_id]
-                pod2_config.other_pods = [pod1_config.pod_id]
+                pod1_config.approved_pods = [pod2_config.pod_id]
+                pod2_config.approved_pods = [pod1_config.pod_id]
                 modeller_config.pods.identifiers = [
                     pod1_config.pod_id,
                     pod2_config.pod_id,
