@@ -3,31 +3,21 @@ import datetime
 import inspect
 from io import BytesIO
 import logging
+from pathlib import Path
 import random
 import string
-from typing import (
-    Any,
-    Dict,
-    List,
-    Literal,
-    Optional,
-    Protocol,
-    Sequence,
-    Tuple,
-    Union,
-    cast,
-)
+from typing import Any, List, Literal, Optional, Protocol, Sequence, Tuple, Union, cast
 from unittest.mock import create_autospec
 
 from PIL import Image, ImageOps
 from _pytest.logging import LogCaptureFixture
 from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey, RSAPublicKey
-import databricks.koalas as ks
 import numpy as np
 import pandas as pd
 import pytest
 
-from bitfount.data.datasource import DataSource
+from bitfount.data.datasources.base_source import BaseSource
+from bitfount.data.datasources.dataframe_source import DataFrameSource
 from bitfount.data.datastructure import DataStructure
 from bitfount.data.schema import BitfountSchema
 from bitfount.data.types import SchemaOverrideMapping
@@ -39,15 +29,14 @@ from bitfount.federated.authorisation_checkers import (
     _LocalAuthorisation,
 )
 from bitfount.federated.early_stopping import FederatedEarlyStopping
-from bitfount.federated.modeller import Modeller
+from bitfount.federated.modeller import _Modeller
 from bitfount.federated.pod_response_message import _PodResponseMessage
 from bitfount.federated.secure import SecureShare
-from bitfount.federated.task_requests import _ProtocolDetails
 from bitfount.federated.transport.worker_transport import (
     _InterPodWorkerMailbox,
     _WorkerMailbox,
 )
-from bitfount.federated.types import AggregatorType, AlgorithmType, ProtocolType
+from bitfount.federated.types import AlgorithmType, ProtocolType
 from bitfount.federated.utils import _ALGORITHMS, _MODELS, _PROTOCOLS
 from bitfount.federated.worker import _Worker
 from bitfount.hub.api import BitfountHub
@@ -58,8 +47,8 @@ from bitfount.metrics import (
     MetricCollection,
 )
 from bitfount.models.base_models import ClassifierMixIn, Optimizer, _BaseModel
-from bitfount.types import DistributedModelProtocol, _DataFrameType, _SeriesType
-from bitfount.utils import _add_this_to_list, seed_all
+from bitfount.types import DistributedModelProtocol, _StrAnyDict
+from bitfount.utils import ExampleSegmentationData, _add_this_to_list, seed_all
 from tests.utils.mocks import LocalMessageService, LocalMessageServiceSharedQueues
 
 # PYTEST MARKS
@@ -68,8 +57,6 @@ backend_test = pytest.mark.backend_test
 
 # tests that run the full end-to-end system in the form of tutorials
 tutorial_test = pytest.mark.tutorial_test
-# tests that run the full end-to-end system in the form of tutorials with YAML config
-tutorial_yaml_test = pytest.mark.tutorial_yaml_test
 # tests that run the full end-to-end system, but mock out web calls
 end_to_end_mocks_test = pytest.mark.end_to_end_mocks_test
 # tests that run the full end-to-end system, including web calls
@@ -78,6 +65,18 @@ end_to_end_test = pytest.mark.end_to_end_test
 integration_test = pytest.mark.integration_test
 # faster, unit-test-like tests
 unit_test = pytest.mark.unit_test
+# Individual test markers for tutorial tests
+tutorial_1_test = pytest.mark.tutorial_1_test
+tutorial_2_test = pytest.mark.tutorial_2_test
+tutorial_3_test = pytest.mark.tutorial_3_test
+tutorial_4_test = pytest.mark.tutorial_4_test
+tutorial_5_test = pytest.mark.tutorial_5_test
+tutorial_6_test = pytest.mark.tutorial_6_test
+tutorial_7_test = pytest.mark.tutorial_7_test
+tutorial_8_test = pytest.mark.tutorial_8_test
+tutorial_9_test = pytest.mark.tutorial_9_test
+tutorial_10_test = pytest.mark.tutorial_10_test
+tutorial_11_test = pytest.mark.tutorial_11_test
 
 DATASET_ROW_COUNT = 4000
 DIMS = (DATASET_ROW_COUNT, 4)
@@ -199,12 +198,12 @@ TABLE_NAME: str = "test_table"
 def create_dataset(
     classification: bool = True,
     seed: int = 420,
-    koalas: bool = False,
     image: bool = False,
     multiimage: bool = False,
     multihead: bool = False,
     img_size: int = 50,
-) -> _DataFrameType:
+    grayscale_image: bool = False,
+) -> pd.DataFrame:
     """Creates a random (seeded) dataset for testing."""
     seed_all(seed)
 
@@ -235,14 +234,14 @@ def create_dataset(
     size = img_size if img_size is not None else 50
     if image or multiimage:
 
-        def create_image(x: _DataFrameType, size: int = size) -> BytesIO:
-            def sigmoid(n: _SeriesType) -> int:
+        def create_image(x: pd.DataFrame, size: int = size) -> BytesIO:
+            def sigmoid(n: pd.Series) -> int:
                 return int(255 / (1 + np.exp(-n)))
 
             # Use column "M" as defining whether the image is grayscale
             R, G, B = sigmoid(x["A"] / 100), sigmoid(x["F"]), sigmoid(x["D"] / 100)
             image = Image.new("RGB", size=(size, size), color=(R, G, B))
-            if x["M"]:
+            if grayscale_image:
                 image = ImageOps.grayscale(image)
             temp = BytesIO()
             image.save(temp, format="png")
@@ -255,7 +254,66 @@ def create_dataset(
             data["image2"] = data.apply(create_image, args=[size], axis=1)
     if multihead:
         data = data.assign(category=list(np.random.choice(["A", "B"], size=len(data))))
-    return ks.from_pandas(data) if koalas else data
+    return data
+
+
+def create_dataset_pii(
+    seed: int = 420,
+) -> pd.DataFrame:
+    """Creates a PII dataset with (seeded) random values for testing."""
+    seed_all(seed)
+
+    # Create a list of private identifiers
+    df_names = pd.DataFrame(
+        np.asarray(range(1, 5001)),
+        columns=["name"],
+    )
+    df_int = pd.DataFrame(
+        np.random.randint(1, 1000, size=(5000, 1)), columns=["int_field"]
+    ).astype("int")
+    df_str = pd.DataFrame(
+        np.random.choice(list(string.ascii_lowercase), size=(5000, 1)).tolist(),
+        columns=["str_field"],
+    ).astype("str")
+    df_obj = pd.DataFrame(
+        np.random.choice(list(string.ascii_lowercase), size=(5000, 1)).tolist(),
+        columns=["obj_field"],
+    ).astype("object")
+
+    df_age = pd.DataFrame(np.random.randint(1, 110, size=(5000, 1)), columns=["age"])
+    df_height = pd.DataFrame(
+        np.random.triangular(left=25, mode=168, right=230, size=(5000, 1)),
+        columns=["height"],
+    )
+    df_weight = pd.DataFrame(
+        np.random.triangular(left=1.8, mode=70, right=230, size=(5000, 1)),
+        columns=["weight"],
+    )
+
+    exercise_choices = [
+        "none",
+        "run",
+        "cycle",
+        "swim",
+        "row",
+        "weights",
+        "elliptical",
+        "yoga",
+        "pilates",
+        "sport",
+        "other",
+    ]
+    exercise_priors = [0.2, 0.08, 0.08, 0.08, 0.08, 0.08, 0.08, 0.08, 0.08, 0.08, 0.08]
+    df_exercise = pd.DataFrame(
+        np.random.choice(exercise_choices, size=(5000, 1), p=exercise_priors),
+        columns=["exercise"],
+    )
+    data: pd.DataFrame = pd.concat(
+        [df_names, df_age, df_weight, df_height, df_exercise, df_int, df_str, df_obj],
+        axis=1,
+    )
+
+    return data
 
 
 def create_datastructure(
@@ -303,7 +361,7 @@ def create_query_datastructure(
                 dd1."N" as n1, dd1."TARGET" as target,
                 dd2."A" as a2, dd2."D" as d2, dd2."M" as m2, dd2."N" as n2
                 FROM dummy_data dd1, dummy_data_2 dd2
-                WHERE dd1.index = dd2.index
+                WHERE dd1."Date" = dd2."Date"
     """
     feature_override: SchemaOverrideMapping = {
         "continuous": ["a1", "d1", "a2", "d2"],
@@ -338,33 +396,24 @@ def create_datasource(
     classification: bool,
     multilabel: bool = False,
     multihead: bool = False,
-    koalas: bool = False,
     loss_weights: bool = False,
     image: bool = False,
     multiimage: bool = False,
-) -> DataSource:
-    """This method creates and returns a DataSource object."""
+) -> BaseSource:
+    """This method creates and returns a BaseSource object."""
     image_col: Optional[List[str]]
     if image:
         data = create_dataset(
-            classification=classification,
-            koalas=koalas,
-            image=image,
-            multihead=multihead,
+            classification=classification, image=image, multihead=multihead
         )
         image_col = ["image"]
     elif multiimage:
         data = create_dataset(
-            classification=classification,
-            koalas=koalas,
-            multiimage=multiimage,
-            multihead=multihead,
+            classification=classification, multiimage=multiimage, multihead=multihead
         )
         image_col = ["image1", "image2"]
     else:
-        data = create_dataset(
-            classification=classification, koalas=koalas, multihead=multihead
-        )
+        data = create_dataset(classification=classification, multihead=multihead)
         image_col = None
 
     if loss_weights:
@@ -377,7 +426,16 @@ def create_datasource(
         data = data.assign(TARGET_2=np.zeros(len(data)))
         data.loc[(data.A < 700) & (data.F < 0.5) & (data.D % 2 == 1), "TARGET_2"] = 1
 
-    dataset = DataSource(data, seed=420, koalas=koalas, image_col=image_col)
+    dataset = DataFrameSource(data, seed=420, image_col=image_col)
+
+    return dataset
+
+
+def create_datasource_pii() -> BaseSource:
+    """This method creates and returns a PII DataFrameSource dataset."""
+    data = create_dataset_pii()
+
+    dataset = DataFrameSource(data, seed=420, image_col=None)
 
     return dataset
 
@@ -387,17 +445,10 @@ def create_schema(
     multilabel: bool = False,
     multihead: bool = False,
     loss_weights: bool = False,
-    koalas: bool = False,
 ) -> BitfountSchema:
     """Helper function for creating datastructure and datasource."""
     datastructure = create_datastructure(multilabel, multihead, loss_weights)
-    datasource = create_datasource(
-        classification,
-        multilabel,
-        multihead,
-        koalas,
-        loss_weights,
-    )
+    datasource = create_datasource(classification, multilabel, multihead, loss_weights)
     schema = BitfountSchema()
     force_stypes = {}
     force_categorical: List[Optional[str]] = []
@@ -420,28 +471,68 @@ def create_schema(
     return schema
 
 
+def create_schema_pii() -> BitfountSchema:
+    """Helper function for creating datastructure and datasource for a PII dataset."""
+    datasource = create_datasource_pii()
+    schema = BitfountSchema()
+
+    datasource.load_data()
+    schema.add_datasource_tables(
+        datasource=datasource,
+        table_name=TABLE_NAME,
+        force_stypes=None,
+    )
+
+    return schema
+
+
 def get_datastructure_and_datasource(
     classification: bool,
     multilabel: bool = False,
     multihead: bool = False,
     multihead_size: int = 1,
     loss_weights: bool = False,
-    koalas: bool = False,
-) -> Tuple[DataStructure, DataSource]:
+) -> Tuple[DataStructure, BaseSource]:
     """Helper function for creating datastructure and datasource."""
     datastructure = create_datastructure(
         multilabel, multihead, multihead_size, loss_weights
     )
-    datasource = create_datasource(
-        classification,
-        multilabel,
-        multihead,
-        koalas,
-        loss_weights,
-    )
+    datasource = create_datasource(classification, multilabel, multihead, loss_weights)
     return datastructure, datasource
 
 
+def create_segmentation_dataset(
+    seg_dir: Path, height: int = 100, width: int = 100, count: int = 25
+) -> pd.DataFrame:
+    """Create a segmentation dataset."""
+    seg = ExampleSegmentationData()
+    input_images, target_masks = seg.generate_data(height, width, count=count)
+    input_images_rgb = [x.astype(np.uint8) for x in input_images]
+    target_masks_rgb = [seg.masks_to_colorimg(x.astype(np.uint8)) for x in target_masks]
+    img_names_list = []
+    masks_names_list = []
+    # Save images
+    for i in range(count):
+        im2 = Image.fromarray((input_images_rgb[i]).astype(np.uint8))
+        im2.save(f"{seg_dir}/img_{i}.png")
+        img_names_list.append(f"img_{i}.png")
+    # Save masks
+    for i in range(count):
+        im2 = Image.fromarray((target_masks_rgb[i]).astype(np.uint8))
+        im2.save(f"{seg_dir}/masks_{i}.png")
+        masks_names_list.append(f"masks_{i}.png")
+
+    # Create dataframe with image and masks locations
+    df = pd.DataFrame(
+        {
+            "img": [str(seg_dir) + "/" + img_name for img_name in img_names_list],
+            "masks": [str(seg_dir) + "/" + mask_name for mask_name in masks_names_list],
+        },
+        columns=["img", "masks"],
+    )
+    return df
+
+
 def create_local_modeller_and_workers(
     model_name: str,
     protocol_name: str,
@@ -449,7 +540,7 @@ def create_local_modeller_and_workers(
     secure_aggregation: bool = False,
     early_stopping: bool = False,
     auto_eval: bool = True,
-) -> Tuple[Modeller, List[_Worker]]:
+) -> Tuple[_Modeller, List[_Worker]]:
     """Creates a modeller and workers to run in tests."""
     pod_ids = ["Alice/AlicePod", "Bob/BobPod"]
     pod_public_keys = {
@@ -461,7 +552,7 @@ def create_local_modeller_and_workers(
     seed_all(43)
 
     # Create default hyperparams
-    hyperparams: Dict[str, Any] = {
+    hyperparams: _StrAnyDict = {
         "epochs": 1,
         "batch_size": 32,
         "optimizer": Optimizer("RAdam", {"lr": 0.001}),
@@ -487,7 +578,9 @@ def create_local_modeller_and_workers(
 
     # Create datasets for each pod and combine them in the schema
     for _ in pod_ids:
-        dataset = DataSource(data.sample(n=1000, random_state=random.randint(1, 101)))
+        dataset = DataFrameSource(
+            cast(pd.DataFrame, data.sample(n=1000, random_state=random.randint(1, 101)))
+        )
         pod_datasets.append(dataset)
 
         # to make sure modeller schema corresponds to the worker schema
@@ -508,71 +601,6 @@ def create_local_modeller_and_workers(
     # Just use the same key for each as we're running local tests
     aes_encryption_key = b"aes_encryption_key"
 
-    if secure_aggregation:
-        task_request = _ProtocolDetails(
-            ProtocolType[protocol_name].value,
-            AlgorithmType[algorithm_name].value,
-            aggregator=AggregatorType["SecureAggregator"].value,
-            model=f"bitfount.{model_name}",
-        )
-    elif protocol_name == "FederatedAveraging":
-        task_request = _ProtocolDetails(
-            ProtocolType[protocol_name].value,
-            AlgorithmType[algorithm_name].value,
-            aggregator=AggregatorType["Aggregator"].value,
-            model=f"bitfount.{model_name}",
-        )
-    else:
-        task_request = _ProtocolDetails(
-            ProtocolType[protocol_name].value,
-            AlgorithmType[algorithm_name].value,
-            model=f"bitfount.{model_name}",
-        )
-
-    # Create workers directly
-    workers = []
-    for pod_identifier, dataset in zip(pod_ids, pod_datasets):
-        message_service = LocalMessageService(
-            username=pod_identifier.split("/")[0],
-            shared_queues=mailbox_queues,
-            modeller_mailbox_id=modeller_mailbox_id,
-            pod_ids=pod_ids,
-        )
-
-        worker_mailbox: _WorkerMailbox
-        if secure_aggregation:
-            worker_mailbox = _InterPodWorkerMailbox(
-                pod_identifier=pod_identifier,
-                modeller_mailbox_id=modeller_mailbox_id,
-                modeller_name=modeller_name,
-                aes_encryption_key=aes_encryption_key,
-                message_service=message_service,
-                pod_mailbox_ids=message_service.worker_mailbox_ids,
-                pod_public_keys=pod_public_keys,
-                private_key=create_autospec(RSAPrivateKey, instance=True),
-            )
-        else:
-            worker_mailbox = _WorkerMailbox(
-                pod_identifier=pod_identifier,
-                modeller_mailbox_id=modeller_mailbox_id,
-                modeller_name=modeller_name,
-                aes_encryption_key=aes_encryption_key,
-                message_service=message_service,
-                pod_mailbox_ids=message_service.worker_mailbox_ids,
-            )
-
-        worker = _Worker(
-            datasource=dataset,
-            mailbox=worker_mailbox,
-            bitfounthub=create_autospec(BitfountHub, instance=True),
-            authorisation=_LocalAuthorisation(
-                _PodResponseMessage(modeller_name, pod_identifier),
-                task_request,
-            ),
-            pod_identifier=pod_identifier,
-        )
-        workers.append(worker)
-
     # Handle early stopping
     early_stopping_ = None
     if early_stopping:
@@ -588,16 +616,16 @@ def create_local_modeller_and_workers(
     )
 
     # Create protocol
-    protocol_kwargs: Dict[str, Any] = {}
+    protocol_kwargs: _StrAnyDict = {}
     if protocol_name == "FederatedAveraging":
         assert isinstance(model, DistributedModelProtocol)
         if secure_aggregation:
-            sec_share = SecureShare(tensor_shim=model.backend_tensor_shim())
+            sec_share = SecureShare()
             aggregator: _BaseAggregatorFactory = SecureAggregator(
-                secure_share=sec_share, tensor_shim=model.backend_tensor_shim()
+                secure_share=sec_share
             )
         else:
-            aggregator = Aggregator(tensor_shim=model.backend_tensor_shim())
+            aggregator = Aggregator()
 
         protocol_kwargs = {
             "epochs_between_parameter_updates": 1,
@@ -608,8 +636,56 @@ def create_local_modeller_and_workers(
         early_stopping=early_stopping_, algorithm=algorithm, **protocol_kwargs
     )
 
+    # Create workers directly
+    workers = []
+    serialized_protocol = protocol.dump()
+    for pod_identifier, dataset in zip(pod_ids, pod_datasets):
+        message_service = LocalMessageService(
+            username=pod_identifier.split("/")[0],
+            shared_queues=mailbox_queues,
+            modeller_mailbox_id=modeller_mailbox_id,
+            pod_ids=pod_ids,
+        )
+
+        worker_mailbox: _WorkerMailbox
+        if secure_aggregation:
+            worker_mailbox = _InterPodWorkerMailbox(
+                pod_public_keys=pod_public_keys,
+                private_key=create_autospec(RSAPrivateKey, instance=True),
+                pod_identifier=pod_identifier,
+                modeller_mailbox_id=modeller_mailbox_id,
+                modeller_name=modeller_name,
+                aes_encryption_key=aes_encryption_key,
+                message_service=message_service,
+                pod_mailbox_ids=message_service.worker_mailbox_ids,
+                task_id="this-is-a-task-id",
+            )
+        else:
+            worker_mailbox = _WorkerMailbox(
+                pod_identifier=pod_identifier,
+                modeller_mailbox_id=modeller_mailbox_id,
+                modeller_name=modeller_name,
+                aes_encryption_key=aes_encryption_key,
+                message_service=message_service,
+                pod_mailbox_ids=message_service.worker_mailbox_ids,
+                task_id="this-is-a-task-id",
+            )
+
+        worker = _Worker(
+            datasource=dataset,
+            mailbox=worker_mailbox,
+            bitfounthub=create_autospec(BitfountHub, instance=True),
+            authorisation=_LocalAuthorisation(
+                _PodResponseMessage(modeller_name, pod_identifier),
+                serialized_protocol,
+            ),
+            pod_identifier=pod_identifier,
+            serialized_protocol=serialized_protocol,
+        )
+        workers.append(worker)
+
     # Create modeller
-    modeller = Modeller(
+    modeller = _Modeller(
         protocol=protocol,
         message_service=LocalMessageService(
             username=modeller_name,
@@ -618,7 +694,7 @@ def create_local_modeller_and_workers(
             pod_ids=pod_ids,
         ),
         bitfounthub=create_autospec(BitfountHub, instance=True),
-        identity_verification_method=IdentityVerificationMethod.KEYS,  # type: ignore[attr-defined] # Reason: Disabled until [BIT-1291] resolved # noqa: B950
+        identity_verification_method=IdentityVerificationMethod.KEYS,
         private_key=create_autospec(RSAPrivateKey, instance=True),
     )
 
@@ -636,15 +712,15 @@ def assert_results(
     on the output.
     """
     if test_preds is None or test_target is None:
-        preds, targs = model.evaluate()
+        preds, target = model.evaluate()
         # TODO: [BIT-1604] Remove these assert statements once they become superfluous.
         assert isinstance(preds, np.ndarray)
-        assert isinstance(targs, np.ndarray)
+        assert isinstance(target, np.ndarray)
     else:
         preds = test_preds
-        targs = np.asarray(test_target)
+        target = np.asarray(test_target)
     metrics = MetricCollection.create_from_model(model)
-    results = metrics.compute(targs, preds)
+    results = metrics.compute(target, preds)
     assert results is not None
     assert isinstance(results, dict)
     if isinstance(model, ClassifierMixIn):
@@ -664,7 +740,7 @@ class CallArgsLike(Protocol):
     """A protocol that defines a unittest.mock.call_args-like interface."""
 
     args: List[Any]
-    kwargs: Dict[str, Any]
+    kwargs: _StrAnyDict
 
 
 def get_arg_from_args_or_kwargs(
@@ -713,6 +789,7 @@ def get_logs_at_level(
     caplog_fixture: LogCaptureFixture,
     levelname: _LOG_LEVEL_NAMES,
     full_details: bool = False,
+    and_higher: bool = False,
 ) -> str:
     """Returns all logs at a given level as a single string.
 
@@ -720,15 +797,26 @@ def get_logs_at_level(
         caplog_fixture: The caplog fixture to extract from.
         levelname: The name of the log level to retrieve messages of.
         full_details: Whether to extract only the log record message or the full details
-                  (timestamp, exception info, etc.). Uses the default formatter
-                  to get the full log record output.
+                      (timestamp, exception info, etc.). Uses the default formatter
+                      to get the full log record output.
+        and_higher: Whether to get logs only at target level or at any level greater
+                    than or equal to the target level.
 
     Returns:
-         All the log messages at the target level as a newline-joined string.
+        All the log messages at the target level (or higher) as a newline-joined
+        string.
     """
-    level_records = [
-        record for record in caplog_fixture.records if record.levelname == levelname
-    ]
+    if not and_higher:
+        level_records = [
+            record for record in caplog_fixture.records if record.levelname == levelname
+        ]
+    else:
+        level_records = [
+            record
+            for record in caplog_fixture.records
+            if record.levelno >= logging.getLevelName(levelname)
+        ]
+
     # Need to interpolate/format the messages
     if full_details:
         formatter = logging.Formatter()
@@ -739,7 +827,9 @@ def get_logs_at_level(
 
 
 def get_debug_logs(
-    caplog_fixture: LogCaptureFixture, full_details: bool = False
+    caplog_fixture: LogCaptureFixture,
+    full_details: bool = False,
+    and_higher: bool = False,
 ) -> str:
     """Returns all logs at debug level as a single string.
 
@@ -747,21 +837,27 @@ def get_debug_logs(
 
     See `get_logs_at_level` for more information.
     """
-    return get_logs_at_level(caplog_fixture, "DEBUG", full_details)
+    return get_logs_at_level(caplog_fixture, "DEBUG", full_details, and_higher)
 
 
-def get_info_logs(caplog_fixture: LogCaptureFixture, full_details: bool = False) -> str:
+def get_info_logs(
+    caplog_fixture: LogCaptureFixture,
+    full_details: bool = False,
+    and_higher: bool = False,
+) -> str:
     """Returns all logs at info level as a single string.
 
     All the log messages are combined as a newline-joined string.
 
     See `get_logs_at_level` for more information.
     """
-    return get_logs_at_level(caplog_fixture, "INFO", full_details)
+    return get_logs_at_level(caplog_fixture, "INFO", full_details, and_higher)
 
 
 def get_warning_logs(
-    caplog_fixture: LogCaptureFixture, full_details: bool = False
+    caplog_fixture: LogCaptureFixture,
+    full_details: bool = False,
+    and_higher: bool = False,
 ) -> str:
     """Returns all logs at warning level as a single string.
 
@@ -769,11 +865,13 @@ def get_warning_logs(
 
     See `get_logs_at_level` for more information.
     """
-    return get_logs_at_level(caplog_fixture, "WARNING", full_details)
+    return get_logs_at_level(caplog_fixture, "WARNING", full_details, and_higher)
 
 
 def get_error_logs(
-    caplog_fixture: LogCaptureFixture, full_details: bool = False
+    caplog_fixture: LogCaptureFixture,
+    full_details: bool = False,
+    and_higher: bool = False,
 ) -> str:
     """Returns all logs at error level as a single string.
 
@@ -781,11 +879,13 @@ def get_error_logs(
 
     See `get_logs_at_level` for more information.
     """
-    return get_logs_at_level(caplog_fixture, "ERROR", full_details)
+    return get_logs_at_level(caplog_fixture, "ERROR", full_details, and_higher)
 
 
 def get_critical_logs(
-    caplog_fixture: LogCaptureFixture, full_details: bool = False
+    caplog_fixture: LogCaptureFixture,
+    full_details: bool = False,
+    and_higher: bool = False,
 ) -> str:
     """Returns all logs at critical level as a single string.
 
@@ -793,4 +893,4 @@ def get_critical_logs(
 
     See `get_logs_at_level` for more information.
     """
-    return get_logs_at_level(caplog_fixture, "CRITICAL", full_details)
+    return get_logs_at_level(caplog_fixture, "CRITICAL", full_details, and_higher)
